unit CardDeckForm;
//Selects index for a string of available CardDecks.
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls;

type
  TCardDeckForm = class(TForm)
    ListBox1: TListBox;
    Button1: TButton;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ListBox1Click(Sender: TObject);
    procedure ListBox1KeyPress(Sender: TObject; var Key: Char);
  private
    { Private declarations }
    procedure ShowCards;
    procedure GetDirList(SearchDir: TFileName; var Files: TStringList);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent; const Dir: String; const Current:
      String = 'Standard'); reintroduce;
  end;

implementation

{$R *.dfm}

uses
  CardTable;

var
  SearchDir, CurrentDeck: String;
  CardTable: TCardTable;

constructor TCardDeckForm.Create(AOwner: TComponent; const Dir: String; const
  Current: String = 'Standard');
begin
  inherited Create(AOwner);
  SearchDir := Dir;
  CurrentDeck := Current;
end;

procedure TCardDeckForm.FormCreate(Sender: TObject);
var
  CardSetList: TStringList;
  i: Integer;
  S: string;
begin
  CardSetList := TStringList.Create;
  try
    GetDirList(SearchDir, CardSetList);
    ListBox1.Items.Assign(CardSetList);
  finally
    CardSetList.Free;
  end;
  ListBox1.Selected[ListBox1.Items.IndexOf(CurrentDeck)] := True;
//Creating a cardtable. We know the location of the cardsets. But there is no
//way to inform cardtable of this & we could be in any directory! But if the
//current dir is set to the dir that contains the cardsets all will be ok.
  for i := length(SearchDir) downto 1 do
    if IsPathDelimiter(SearchDir, i) then break;
  SetLength(SearchDir, i-1);
  SetLength(S, MAX_PATH);
  SetLength(S, GetCurrentDirectory(MAX_PATH, PChar(S)));
  if not SameText(S, SearchDir) then
    SetCurrentDirectory(PChar(SearchDir));
  CardTable := TCardTable.Create(Self);
  CardTable.Left := 300;
  ShowCards;
  if not SameText(S, SearchDir) then//Switch back to original dir.
    SetCurrentDirectory(PChar(S));
end;

procedure TCardDeckForm.Button1Click(Sender: TObject);
begin
  ModalResult := mrOk;
end;

procedure TCardDeckForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  CardTable.Free;
end;

procedure TCardDeckForm.ShowCards;
var
  i: Integer;
  ACard: TCard;
begin
  if CardTable.CardDeck.DeckName <> ListBox1.Items[ListBox1.ItemIndex] then
    CardTable.CardDeck.DeckName := ListBox1.Items[ListBox1.ItemIndex];
  CardTable.PickUpAllCardsAndReset;//Picks up cards AND resets the deck.
  CardTable.Width := CardTable.CardDeck.CardWidth * 2 + 10;
  CardTable.Height := CardTable.CardDeck.CardHeight + 24;
  Width := 320 + CardTable.Width;
  for i := 1 to 24 do
    ACard := CardTable.CardDeck.Draw;
  CardTable.TurnOverCard(ACard);
  CardTable.PlaceCard(ACard, 8, 8);
  for i := 1 to 28 do
    ACard := CardTable.CardDeck.Draw;
  CardTable.TurnOverCard(ACard);
  CardTable.PlaceCard(ACard, 8+CardTable.CardDeck.CardWidth, 8);
end;

procedure TCardDeckForm.ListBox1Click(Sender: TObject);
begin
  ShowCards;
end;

procedure TCardDeckForm.ListBox1KeyPress(Sender: TObject; var Key: Char);
begin
  if Ord(Key) = 13 then
    ModalResult := mrOk;
end;

procedure TCardDeckForm.GetDirList(SearchDir: TFileName; var Files:
  TStringList);
var
  SearchRec: TSearchRec;
  DosError: Integer;
begin
 	Files.Clear;
  Files.CaseSensitive := False;
  Files.Sorted := True;
	SearchDir := SearchDir + '\*.*';
	try
		DosError := FindFirst(SearchDir, faDirectory, SearchRec);
		while DosError = 0 do
		begin
			if (SearchRec.Attr and faDirectory = faDirectory) and
        (SearchRec.Name[1] <> '.') then Files.Add(SearchRec.Name);
			DosError := FindNext(SearchRec);
		end;
	finally
		SysUtils.FindClose(SearchRec);
	end;
end;


end.
