{

CardTable v1.02.
Copyright  2003 by David Mayne.

This is FreeWare. You may use this code in your own programs freely, with no
restrictions and with no charge.

}
unit CardTable;
{$R-,T-,H+,X+,O+}
interface

uses
  Windows, Types, Classes, Graphics, Controls, SysUtils, Jpeg;

const
  MaxCards = 104;//Max no of cards that can be displayed on the table.
  NoOfCards = 52;//The no of cards in the deck.

type
  //Enumerated types.
  TCardSuit = (csClub, csDiamond, csHeart, csSpade);
  TCardValue = (cvAce, cvTwo, cvThree, cvFour, cvFive, cvSix, cvSeven,
    cvEight, cvNine, cvTen, cvJack, cvQueen, cvKing);//Nb aces are low.
  TMode = (mUp, mDown, mHorizontal, mAuto);//Only used by CardTable.SortHand.
  TCardMarker = (cmOutline, cmMark, cmClub, cmDiamond, cmHeart, cmSpade,
    cmDragShade);//Used in drawing outlines & shades of cards on the table.

  //Forward declarations.
  TCard = class;
  TCardDeck = class;
  TCardTable = class;

  //Exceptions and Events.
  ECardTableError = class(Exception);
  TCardClickEvent = procedure(ACard: TCard; Button: TMouseButton) of object;
  TCardDropEvent = procedure(ACard: TCard; X, Y, RelX, RelY: Integer) of object;

  //Records.
  TCardMark = record//Used to draw a card outline or shade on the table.
    Mark: TBitMap;
    Under: TBitMap;
    Displayed: Boolean;
    Position: TPoint;
  end;
  PCardMark = ^TCardMark;

  //Classes.
  TCard = class(TObject)
  private
    FFaceUp: Boolean;
    FSuit: TCardSuit;
    FValue: TCardValue;
    FUnder: TBitMap;//A copy of whats below the card.
    Fx: Integer;//X,Y pos of card on table.
    Fy: Integer;
    FStatus: Integer;//Status of card - 0 in deck, 1 in play, -1 discarded.
    FDisplayed: Boolean;//Is the card on the table?
    FFace: TBitMap;
    FOwner: TCardDeck;
    procedure LoadCardFace(const Path, DeckName: string; const ASuit: TCardSuit;
      const AValue: TCardValue);
    procedure SetStatus(const Status: Integer);
  public
    constructor Create(AOwner: TCardDeck);
    destructor Destroy; override;
    function NumericValue: Integer;
    procedure Assign(Source: TCard); virtual;
    property Owner: TCardDeck read FOwner;
    property Suit: TCardSuit read FSuit write FSuit;
    property Value: TCardValue read FValue write FValue;
    //If on table its best to use CardTable.TurnOverCard & not this property.
    property FaceUp: Boolean read FFaceUp write FFaceUp;
    property Status: Integer read FStatus write SetStatus;
    property Displayed: Boolean read FDisplayed write FDisplayed;
    property X: Integer read Fx write Fx;
    property Y: Integer read Fy write Fx;
  end;

  TCardDeck = class(TPersistent)
  private
    FCardArray: Array[1..NoOfCards] of TCard;
    FTop: Integer;//Cards dealt 0 - NoOfCards
    FCardDeckName: string;
    FNoOfCardBacks: Integer;
    FCardBack: Integer;
    FBack: TBitMap;
    FCardMask: TBitMap;
    FCardHeight: Integer;
    FCardWidth: Integer;
    FOwner: TComponent;
    FCardMarks: Array[cmOutline..cmDragShade] of TCardMark;//Card placeholders.
    FDeckDirectory: string;
    FShadow00: TBitMap;//Now various fields used for the decks shadow.
    FShadow01: TBitMap;
    FShadow00Under: TBitMap;
    FShadow01Under: TBitMap;
    FShadow00Point: TPoint;
    FShadow01Point: TPoint;
    procedure LoadBitMap(var BitMap: TBitMap; const FileName: string);
    function GetNoOfBacks: Integer;
    procedure LoadCardBack(const Back: Integer);
    procedure LoadDeck;
    procedure SetDeck(const DeckType: string);
    procedure SetNewBitmap(var BitMap, NewBitMap: TBitMap);
  public
    //2nd Overloaded constructor should be used when you know the CardSets dir.
    constructor Create(AOwner: TComponent); overload;
    constructor Create(AOwner: TComponent; const DeckDirectory: string);
      overload;
    destructor Destroy; override;
    function NoOfCardsInDeck: Integer;//Returns no of unused cards.
    procedure Shuffle;
    function Draw: TCard;
    procedure SelectBack;//Select back graphic via a form.
    procedure SelectDeck;//Select deck used via a form.
    procedure Assign(Source: TPersistent); override;
    procedure ResetDeck;//Resets card & deck fields to initial full deck state.
    procedure AddCardsToDeck(var Cards: Array of TCard);
    property Owner: TComponent read FOwner;
    property CardHeight: Integer read FCardHeight;
    property CardWidth: Integer read FCardWidth;
    property NoOfCardBacks: Integer read FNoOfCardBacks;
    property DeckDirectory: string read FDeckDirectory;
  published
    property DeckName: string read FCardDeckName write SetDeck;
    property CardBack: Integer read FCardBack write LoadCardBack;
  end;

  TCardTable = class(TGraphicControl)
  private
    FBitMap: TBitMap;//Table buffer.
    FDirectory: string;//Program Startup Directory;
    FJpeg: TJpegImage;//Holds jpeg picture if used.
    FJpegPerformance: TJPEGPerformance;//Display speed of the Jpeg.
    FCardDeck: TCardDeck;//Owned & freed by this object.
    FExternalDeck: TCardDeck;//Pointer to an external deck that can be clicked.
    FDiscardsAt: TPoint;//A pile of unusable discards.
    FNoOfDiscards: Integer;
    FDiscardArray: Array[1..MaxCards] of TCard;
    FOnCardClickEvent: TCardClickEvent;
    FOnCardDropEvent: TCardDropEvent;
    FColor: TColor;//If no background bitmap then this is table color.
    FFileName: TFileName;//Name of background bitmap if used or 'None' if not.
    FStretchBackground: Boolean;//Is the bimap stretched to fit the table.
    FCardSpeed: Integer;
    FCardsOnTable: Array[1..MaxCards] of TCard;//A stack of the cards on table.
    FNoOfCardsOnTable: Integer;//The no of cards on the table.
    FBufferFix: Boolean;//If True buffer errors are fixed in Paint.
    FResizedByPicture: Boolean;//True if loading a picture has resized control.
    FCopyToScreen: Boolean;//Set to false during resizing to reduce flicker.
    FPlaceDeckOffset: Integer;//Allows a pseudo 3d look to the deck.
    FDragAndDrop: Boolean;//Is drag & drop enabled.
    FDragFromDeck: Boolean;//If DragAndDrop can we drag a card from the deck.
    FCardDraggedFromDeck : Boolean;//Has card been dragged from the deck.
    FDragFaceDown: Boolean;//Can we drag face down cards.
    FCardDragging: Boolean;//Is a card currently being dragged.
    FDraggingCard: TCard;//The dragged card.
    FDragOrigin: TPoint;//The Dragged cards original position on screen.
    FCOTPosition: Integer;//Drag card pos in CardsOnTable array, 0 if in deck.
    FMouseRelX: Integer;//Mouse x pos relative to dragged card.
    FMouseRelY: Integer;//Mouse y pos relative to dragged card.
    FPreDragCursor: TCursor;//Cursor used prior to the drag.
    FShadePlaced: Boolean;//Has a shade been placed whilst card is dragged.
    FSaveTable: Boolean;//Flags to DefineProperties to save cards as well.
    FRefreshLabels: Boolean;//Should we repaint any labels on the table.
  protected
    procedure PutDown(const ACard: TCard; const X, Y: Integer;
      const Deck: TCardDeck; const SaveBack: boolean = true);//To buffer.
    procedure PutDownShadow(const X, Y: Integer; const Deck: TCardDeck);
    procedure PickUp(const ACard: TCard; const Deck: TCardDeck);//From buffer.
    procedure PickUpShadow(const Deck: TCardDeck);
    procedure CopyBuffer(const X, Y, Width, Height: Integer);
    procedure SetColor(const Color: TColor);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer);
      override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
      override;
    function MouseButtonPressed(const X, Y: Integer): TCard;
    procedure LoadTableTop(const FileName: TFileName);
    procedure Resize; override;
    procedure SetSpeed(const Speed: Integer);
    function GetCanvas: TCanvas;
    function PickUpCoveringCards(const ACard: TCard; const Deck: TCardDeck;
      var CoveringNo: Integer; const BeginingMove: Boolean = True): Integer;
    procedure ReplaceCoveringCards(const ArrayPos: Integer);
    function StillUnder: Boolean;
    function GetCardOnTable(I: Integer): TCard;
    procedure SetBufferFix(const Value: Boolean);
    procedure SetDeckOffset(const Offset: Integer);
    procedure SetDragAndDrop(const Mode: Boolean);
    procedure SetDragFromDeck(const Mode: Boolean);
    procedure SetDragFaceDown(const Mode: Boolean);
    procedure SetCardDeck(const Deck: TCardDeck);
    procedure SetJpegPerformance(const Performance: TJPEGPerformance);
    procedure RedrawBuffer;
    procedure PlaceCardMarkers;
    //Methods to save/load table + cards via streaming system.
    procedure DefineProperties(Filer: TFiler); override;
    procedure WriteCardValue(Writer: TWriter);
    procedure ReadCardValue(Reader: TReader);
    procedure WriteCardSuit(Writer: TWriter);
    procedure ReadCardSuit(Reader: TReader);
    procedure WriteFaceUp(Writer: TWriter);
    procedure ReadFaceUp(Reader: TReader);
    procedure ReadStatus(Reader: TReader);
    procedure WriteStatus(Writer: TWriter);
    procedure WriteDisplayed(Writer: TWriter);
    procedure ReadDisplayed(Reader: TReader);
    procedure ReadX(Reader: TReader);
    procedure WriteX(Writer: TWriter);
    procedure ReadY(Reader: TReader);
    procedure WriteY(Writer: TWriter);
    procedure ReadFT(Reader: TReader);
    procedure WriteFT(Writer: TWriter);
    procedure ReadNoOfCOT(Reader: TReader);
    procedure WriteNoOfCOT(Writer: TWriter);
    procedure ReadCOT(Reader: TReader);
    procedure WriteCOT(Writer: TWriter);
    procedure ReadCardMarks(Reader: TReader);
    procedure WriteCardMarks(Writer: TWriter);
    procedure ReadD2Name(Reader: TReader);
    procedure WriteD2Name(Writer: TWriter);
    procedure ReadD2Back(Reader: TReader);
    procedure WriteD2Back(Writer: TWriter);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure PlaceDeck(const X, Y: Integer; Deck: TCardDeck = nil);
    procedure SetDrawnCard(const ACard: TCard);//Sets the drawn card.
    //DrawCardFromDeck is quick way to draw from a displayed deck.
    function DrawCardFromDeck(Deck: TCardDeck = nil): TCard;
    procedure PlaceCard(const ACard: TCard; const X, Y: Integer);
    procedure PickUpCard(const ACard: TCard);
    procedure MoveTo(const ACard: TCard; const X, Y: Integer);
    procedure TurnOverCard(const ACard: TCard);
    procedure Discard(const ACard: TCard);
    procedure PickUpDeck(Deck: TCardDeck = nil);
    procedure PickUpDiscards;
    procedure PickUpCardsOnTable;
    procedure PickUpAllCardsAndReset(Deck: TCardDeck = nil);
    procedure Paint; override;
    procedure Show;
    procedure SortHand(var Hand: array of TCard; const AcesLow: Boolean;
      const SuitSort: Boolean; Mode: TMode = mAuto);
    procedure DropCard(X, Y: Integer);//Drop dragged card.
    procedure RestoreCard;//Restore dragged card to its origin.
    procedure PlaceCardMarker(const CardMarker: TCardMarker; const X, Y:
      Integer; Deck: TCardDeck = nil);//Places a CardMarker on the table.
    procedure PickUpMarker(const CardMarker: TCardMarker;
      Deck: TCardDeck = nil);
    procedure PickUpAllMarkers(Deck: TCardDeck = nil);
    procedure DropCardOnShade(const CardMarker: TCardMarker;
      Deck: TCardDeck = nil);
    procedure ClearTable;//Clears the table.
    procedure SelectBackgroundPicture;//Select bitmap via dialog.
    procedure SaveStatus;//Saves changes made to table properties.
    procedure LoadStatus;//Load saved properties.
    function DragCardWithinBounds(const Bounds: TRect): Boolean;
    procedure ResizeTable(const Left, Top, Width, Height: integer);
    procedure DeckOrBackChanged;//Redraws table after deck/back change.
    procedure SaveTable(const FileName: TFileName);//Saves table & cards.
    procedure LoadTable(const FileName: TFileName);//Loads table & cards.
    procedure RefreshLabels;
    property TableCanvas: TCanvas read GetCanvas;//Surfacing the buffer.
    property CardDragging: Boolean read FCardDragging;
    property DraggingCard: TCard read FDraggingCard;
    property CardDraggedFromDeck: Boolean read FCardDraggedFromDeck;
    //External deck can be set to point to an external deck object. This deck
    //can then be selected with the mouse & will be redrawn when necessary. This
    //means a limit of 2 visible decks that can be selected & redrawn
    //automatically when needed on the table.
    property ExternalDeck: TCardDeck read FExternalDeck write FExternalDeck;
    property ShadePlaced: Boolean read FShadePlaced;
    property NoOfCardsOnTable: Integer read FNoOfCardsOnTable;
    property CardsOnTable[I: Integer]: TCard read GetCardOnTable;
    property DiscardsPoint: TPoint read FDiscardsAt write FDiscardsAt;
    property Directory: string read FDirectory;//Program directory.
  published
  //Note no default values as streaming system is used by save/load status/table
  //and default values are not written to the stream. It`s possible to loop thru
  //the properties and obtain the default values at runtime. You could then set
  //the properties before loading the resource files. You could then use default
  //values and save/load status/table would still work ok. But help system is
  //not at all helpfull with regard to this and it would require a fair bit of
  //effort for not much reward so I can`t be bothered as there aren`t that many
  //published properties anyway.
    property OnClick;
    property OnCardClickEvent: TCardClickEvent read FOnCardClickEvent
      write FOnCardClickEvent;
    property OnCardDropEvent: TCardDropEvent read FOnCardDropEvent
      write FOnCardDropEvent;
    property OnMouseMove;
    property CardDeck: TCardDeck read FCardDeck write SetCardDeck;
    property Visible;
    property StretchBackground: Boolean read FStretchBackground write
      FStretchBackground;
    property PlaceDeckOffset: Integer read FPlaceDeckOffset write SetDeckOffset;
    property Cursor;
    //The color of table if no bitmap is used.
    property Color: TColor read FColor write SetColor ;
    property JpegPerformance: TJPEGPerformance read FJpegPerformance write
      SetJpegPerformance;//Display speed + quality of Jpeg.
    //Nb for BackgroundPicture filename must be fully qualified. So if in
    //program directory use cardtable.directory + '\pic.***' not just 'pic.***'.
    property Width;//When resizing use the ResizeTable function not Width or
    property Height;//Height.
    property Left;//Left & Top are safe to use if not resizing as well.
    property Top;
    property BackgroundPicture: TFileName read FFileName write LoadTableTop;
    property CardSpeed: Integer read FCardSpeed write SetSpeed;
    //BufferFix is by default True. It can be set to False if you are sure you
    //will not introduce buffer errors. Used by the Card Back Selector form to
    //enable CardTable to show different card backs of the same deck on the same
    //table.
    property BufferFix: Boolean read FBufferFix write SetBufferFix;
    property DragAndDrop: Boolean read FDragAndDrop write SetDragAndDrop;
    property DragFromDeck: Boolean read FDragFromDeck write SetDragFromDeck;
    property DragFaceDown: Boolean read FDragFaceDown write SetDragFaceDown;
    property RepaintLabels: Boolean read FRefreshLabels write FRefreshLabels;
  end;

implementation

uses
  Forms, Math, CardBackForm, CardDeckForm, DeckPathForm, ExtDlgs, TypInfo;



//............................................................................//
//............................ T C A R D .....................................//
//............................................................................//



constructor TCard.Create(AOwner: TCardDeck);
//A card is part of a deck, the deck determines its size & look.
begin
  inherited Create;
  FFaceUp := False;
  FUnder := TBitMap.Create;
  FDisplayed := false;
  FFace := TBitMap.Create;
  FOwner := AOwner;
end;

destructor TCard.Destroy;
//Call free not this method.
begin
  FUnder.Free;
  FFace.Free;
  inherited Destroy;
end;

function TCard.NumericValue: Integer;
//Returns a card's numeric value.
begin
  Result := Ord(FValue) + 1;
  if Result > 10 then Result := 10;
end;

procedure TCard.LoadCardFace(const Path, DeckName: string;
  const ASuit: TCardSuit; const AValue: TCardValue);
//Private method that loads a face bitmap.
var
  FileName: string;
begin
  if Ord(AValue) < 9 then FileName := '\0' else FileName := '\';
  FileName := FileName + IntToStr(Ord(AValue) + 1);
  case ASuit of
    csHeart: FileName := FileName + 'h.bmp';
    csSpade: FileName := FileName + 's.bmp';
    csClub: FileName := FileName + 'c.bmp';
    csDiamond: FileName := FileName + 'd.bmp';
  end;
  FFace.LoadFromFile(Path + '\' + DeckName + FileName);
  FFace.Dormant;
  FFace.FreeImage;
  FUnder.Height := FFace.Height;
  FUnder.Width := FFace.Width;
end;

procedure TCard.SetStatus(const Status: Integer);
//Private method setting a public property that sets FStatus.
begin
  if (Status >= -1) and (Status <= 1) then
    FStatus := Status;
end;

procedure TCard.Assign(Source: TCard);
//Copying all but FOwner.
begin
  FFaceUp := Source.FFaceUp;
  FSuit := Source.FSuit;
  FValue := Source.FValue;
  FUnder.Free;
  FUnder := TBitMap.Create;
  FUnder.Assign(Source.FUnder);
  Fx := Source.Fx;
  Fy := Source.Fy;
  FStatus := Source.FStatus;
  FDisplayed := Source.FDisplayed;
  FFace.Free;
  FFace := TBitMap.Create;
  FFace.Assign(Source.FFace);
end;



//............................................................................//
//......................... T C A R D D E C K ................................//
//............................................................................//



constructor TCardDeck.Create(AOwner: TComponent);
//A deck of cards. The 2nd overloaded constructor should be used when you know
//the CardSets dir as for instance when a CardTable has been instantiated:
//Property CardTable.CardDeck.DeckDirectory.
var
  Result: Integer;
  OptionList: TstringList;
  DeckPathForm: TDeckPathForm;
  S: string;
begin
  inherited Create;
  FOwner := AOwner;
  SetLength(S, MAX_PATH);
  SetLength(S, GetCurrentDirectory(MAX_PATH, PChar(S)));
  if DirectoryExists('Cardsets\Standard') then
    FDeckDirectory := S + '\Cardsets'
  else
    if FileExists(S + '\Config.txt') then
    begin
      OptionList := TstringList.Create;
      try
        OptionList.LoadFromFile(S + '\Config.txt');
        FDeckDirectory := OptionList[0];
      finally
        OptionList.Free;
      end;
      if not DirectoryExists(FDeckDirectory+'\Standard') then
        FDeckDirectory := '';
    end;
  if FDeckDirectory = '' then
    if not (csDesigning in AOwner.ComponentState) then
    begin
      DeckPathForm := TDeckPathForm.Create(FOwner);
      try
        Result := DeckPathForm.ShowModal;
        FDeckDirectory := DeckPathForm.Edit1.Text;
      finally
        DeckPathForm.Release;
      end;
      if (Result = mrOk) and DirectoryExists(FDeckDirectory+'\Standard') then
      begin
        OptionList := TstringList.Create;
        try
          Optionlist.Add(FDeckDirectory);
          OptionList.SaveToFile(S + '\Config.txt');
        finally
          OptionList.Free;
        end;
      end
      else
        raise ECardTableError.Create('Cardsets directory not found.');
    end
    else
      raise ECardTableError.Create('Cardsets directory not found.');
  FCardDeckName := 'Standard';
  FNoOfCardBacks := GetNoOfBacks;
  LoadDeck;
  FTop := 0;
end;

constructor TCardDeck.Create(AOwner: TComponent; const DeckDirectory: string);
//A deck of cards. Use this constructor when you know the CardSets dir.
begin
  inherited Create;
  FOwner := AOwner;
  FDeckDirectory := DeckDirectory;
  FCardDeckName := 'Standard';
  FNoOfCardBacks := GetNoOfBacks;
  LoadDeck;
  FTop := 0;
end;

destructor TCardDeck.Destroy;
//Call free not this method.
var
  i: Integer;
  j: TCardMarker;
begin
  for i := 1 to NoOfCards do
    FCardArray[i].Free;
  FBack.Free;
  FCardMask.Free;
  FShadow00.Free;
  FShadow01.Free;
  FShadow00Under.Free;
  FShadow01Under.Free;
  for j := cmOutline to cmDragShade do
  begin
    FCardMarks[j].Mark.Free;
    FCardMarks[j].Under.Free;
  end;
  inherited Destroy;
end;

procedure TCardDeck.Shuffle;
//Shuffles the cards in the deck does not have to be a full deck. It must NOT be
//displayed at the time.
var
  i, RandNum: Integer;
  RandCard: TCard;
begin
  for i := 1 to NoOfCards - FTop do
  begin
    RandNum := Random(NoOfCards-1) + 1;
    RandCard := FCardArray[RandNum];//Swap next card with
    FCardArray[RandNum] := FCardArray[i];// random card in deck.
    FCardArray[i] := RandCard;
    RandCard.FStatus :=0;
    RandCard.FFaceUp := False;
    RandCard.FDisplayed := False;
    FCardArray[i].FStatus := 0;
    FCardArray[i].FFaceUp := False;
    FCardArray[i].FDisplayed := False;
  end;
end;

function TCardDeck.Draw: TCard;
//Returns TCard increasing FTop. Note that if the deck is displayed on the
//table you must next call CardTable.SetDrawnCard(ACard).
begin
  inc(FTop);
  if FTop = NoOfCards+1 then
  begin
    FTop := NoOfCards;
    Result := nil;
  end else
  begin
    Result := FCardArray[FTop];
    Result.FStatus := 1;
  end;
end;

function TCardDeck.NoOfCardsInDeck: Integer;
//Returns the no of cards left in the deck. You probably guessed that!
begin
  Result := NoOfCards - FTop;
end;

procedure TCardDeck.LoadCardBack(const Back: Integer);
//Private method loads deck back.
var
  FileName: string;
begin
  if Back <= NoOfCardBacks then
  begin
    FileName := FDeckDirectory + '\' + FCardDeckName + '\Back' +
      IntToStr(Back) + '.bmp';
    FBack.LoadFromFile(FileName);
    FBack.Dormant;
    FBack.FreeImage;
    FCardBack := Back;
  end;
end;

procedure TCardDeck.LoadBitMap(var BitMap: TBitMap; const FileName: string);
//Private method loads a bitmap from a file to a TBitMap.
begin
  BitMap.LoadFromFile(FileName);
  BitMap.Dormant;
  BitMap.FreeImage;
end;

function TCardDeck.GetNoOfBacks: Integer;
//Private method returns the no of backs the deck has 1 - 9.
var
  i: Integer;
  FileName: string;
begin
  i := 1;
  repeat
    FileName := FDeckDirectory + '\' + FCardDeckName + '\Back' + IntToStr(i) +
      '.bmp';
    if FileExists(FileName) then Inc(i)
    else break;
  until (i = 10);
  Dec(i);
  if i <>0 then Result := i
  else raise ECardTableError.Create('No CardBacks Found.');
end;

procedure TCardDeck.LoadDeck;
//Private method that loads card bitmaps and assign values and suits if not
//already assigned.
var
  i: Integer;
  ASuit: TCardSuit;
  AValue: TCardValue;
  FileName: TFileName;
  j: TCardMarker;
begin
  ASuit := csClub;
  AValue := cvAce;
  for i := 1 to NoOfCards do
  begin
    if not Assigned(FCardArray[i]) then
    begin
      FCardArray[i] := TCard.Create(self);
      FCardArray[i].FValue := AValue;
      FCardArray[i].FSuit := ASuit;
    end;//Else keep already assigned suit & value.
    FCardArray[i].LoadCardFace(FDeckDirectory, FCardDeckName,
      FCardArray[i].FSuit, FCardArray[i].FValue);
    if (i mod 4 = 0) and (i <> NoOfCards) then inc(AValue);
    if ASuit <> High(TCardSuit) then inc(ASuit) else ASuit := Low(TCardSuit);
  end;
  if not Assigned(FBack) then FBack := TBitMap.Create;
  if not Assigned(FCardMask) then FCardMask := TBitMap.Create;
  if not Assigned(FShadow00) then FShadow00 := TBitMap.Create;
  if not Assigned(FShadow01) then FShadow01 := TBitMap.Create;
  if not Assigned(FShadow00Under) then FShadow00Under := TBitMap.Create;
  if not Assigned(Fshadow01Under) then FShadow01Under := TBitMap.Create;
  for j := cmOutline to cmDragShade do
  begin
    if not Assigned(FCardMarks[j].Mark) then
    begin
      FCardMarks[j].Mark := TBitMap.Create;
      FCardMarks[j].Under := TBitMap.Create;
    end;
    case j of
      cmOutline..cmMark: FileName := '/Bottom0' + IntToStr(Ord(j)+1);
      cmClub..cmSpade: FileName := '/Bottom0' + IntToStr(Ord(j)+2);
      cmDragShade: FileName := '/Shade';
    end;
    FileName := FileName + '.bmp';
    LoadBitmap(FCardMarks[j].Mark, FDeckDirectory + '\' + FCardDeckName +
      FileName);
    FCardMarks[j].Under.Width := FCardMarks[j].Mark.Width;
    FCardMarks[j].Under.Height := FCardMarks[j].Mark.Height;
  end;
  LoadCardBack(1);
  LoadBitMap(FCardMask, FDeckDirectory + '\' + FCardDeckName + '\CardMask.bmp');
  LoadBitMap(FShadow00, FDeckDirectory + '\' + FCardDeckName + '\Shadow00.bmp');
  LoadBitMap(FShadow01, FDeckDirectory + '\' + FCardDeckName + '\Shadow01.bmp');
  FShadow00Under.Width := FShadow00.Width;
  FShadow00Under.Height := FShadow00.Height;
  FShadow01Under.Width := FShadow01.Width;
  FShadow01Under.Height := FShadow01.Height;
  FCardWidth := FBack.Width;
  FCardHeight := FBack.Height;
end;

procedure TCardDeck.SetDeck(const DeckType: string);
//Property method that sets the deck to the given decktype.
begin
  if DirectoryExists(FDeckDirectory + '\' + DeckType) then
  begin
    FCardDeckName := DeckType;
    FNoOfCardBacks := GetNoOfBacks;
    LoadDeck;
  end;
end;

procedure TCardDeck.SelectBack;
//Select back via a form.
var
  CardBackForm: TCardBackForm;
begin
  CardBackForm := TCardBackForm.Create(Application, FDeckDirectory);
  try
    CardBackForm.DrawBacks(Application, FCardDeckName);
    CardBack := CardBackForm.ShowModal;
  finally
    CardBackForm.Release;
  end;
end;

procedure TCardDeck.SelectDeck;
//Select deck used via a form.
var
  CardDeckForm: TCardDeckForm;
  DeckName: string;
begin
  CardDeckForm := TCardDeckForm.Create(Application, FDeckDirectory,
    FCardDeckName);
  try
    CardDeckForm.ShowModal;
    DeckName := CardDeckForm.ListBox1.Items[CardDeckForm.ListBox1.itemindex];
  finally
    CardDeckForm.Release;
    //Ensure form is painted before setting the deck to fix a messy delay.
    Application.ProcessMessages;
    Self.DeckName := DeckName;
  end;
end;

procedure TCardDeck.Assign(Source: TPersistent);
//Copying all but FOwner.
var
  i: Integer;
  j: TCardMarker;
begin
  if Source is TCardDeck then
  begin
    for i := 1 to NoOfCards do
      FCardArray[i].Assign((Source as TCardDeck).FCardArray[i]);
    FTop := (Source as TCardDeck).FTop;
    FCardDeckName := (Source as TCardDeck).FCardDeckName;
    FNoOfCardBacks := (Source as TCardDeck).FNoOfCardBacks;
    FCardBack := (Source as TCardDeck).FCardBack;
    SetNewBitMap(FBack, (Source as TCardDeck).FBack);
    SetNewBitMap(FCardMask, (Source as TCardDeck).FCardMask);
    FCardHeight := (Source as TCardDeck).FCardHeight;
    FCardWidth := (Source as TCardDeck).FCardWidth;
    FDeckDirectory := (Source as TCardDeck).FDeckDirectory;
    SetNewBitMap(FShadow00, (Source as TCardDeck).FShadow00);
    SetNewBitMap(FShadow01, (Source as TCardDeck).FShadow01);
    SetNewBitMap(FShadow00Under, (Source as TCardDeck).FShadow00Under);
    SetNewBitMap(FShadow01Under, (Source as TCardDeck).FShadow01Under);
    FShadow00Point := (Source as TCardDeck).FShadow00Point;
    FShadow01Point := (Source as TCardDeck).FShadow01Point;
    for j := cmOutline to cmDragShade do
    begin
      SetNewBitMap(FCardMarks[j].Mark,
        (Source as TCardDeck).FCardMarks[j].Mark);
      SetNewBitMap(FCardMarks[j].Under,
        (Source as TCardDeck).FCardMarks[j].Under);
    end;
  end
  else
    inherited Assign(Source);
end;

procedure TCardDeck.SetNewBitmap(var BitMap, NewBitMap: TBitMap);
//Private method that assigns one bitmap to another.
begin
  BitMap.Free;
  BitMap := TBitMap.Create;
  BitMap.Assign(NewBitMap);
end;

procedure TCardDeck.ResetDeck;
//Reseting all card & deck fields to full deck state.
var
  i: Integer;
  ACard: TCard;
begin
  FTop := 0;
  for i := 1 to NoOfCards do
  begin
    ACard := FCardArray[i];
    ACard.FFaceUp := False;
    ACard.FStatus := 0;
    ACard.FDisplayed := False;
  end;
end;

procedure TCardDeck.AddCardsToDeck(var Cards: Array of TCard);
//Cards dealt are removed from the deck & moved to the table. To add to the
//bottom of the deck, some cards from the table, pass them to this method. The
//Deck MUST own the Cards and the deck AND the cards from the table also MUST
//NOT be displayed on the table when this method is called.
var
  i, j, k: Integer;
begin
  for i := 0 to High(Cards) do
  begin
    if not Assigned(Cards[i]) then break;
    if Cards[i].FStatus <> 0 then
    begin
      for j := FTop downto 1 do
        if FCardArray[j] = Cards[i] then break;
      if FCardArray[j] = Cards[i] then
      begin
        for k := j to NoOfCards-1 do
          FCardArray[k] := FCardArray[K+1];
        FCardArray[NoOfCards] := Cards[i];
        Cards[i].FStatus := 0;
        Cards[i].FFaceUp := False;
        dec(FTop);
      end;
    end;
  end;
end;



//............................................................................//
//........................ T C A R D T A B L E ...............................//
//............................................................................//



constructor TCardTable.Create(AOwner: TComponent);
//Displaying & moving cards.
begin
  inherited Create(AOwner);
  Parent := AOwner as TWinControl;
  FCardDeck := TCardDeck.Create(self);//1 deck is created & freed by the table.
  FBitMap := TBitMap.Create;
  FJpeg := TJpegImage.Create;
  ControlStyle := [csOpaque, csCaptureMouse, csClickEvents, csSetCaption];
  SetLength(FDirectory, MAX_PATH);
  SetLength(FDirectory, GetCurrentDirectory(MAX_PATH, PChar(FDirectory)));
  FSaveTable := False;
  FColor := clSilver;
  FCardSpeed := 10;
  Height := 350;
  Width := 450;
  FJpegPerformance := jpBestQuality;
  FJpeg.Performance := jpBestQuality;
  FStretchBackground := False;
  BackgroundPicture := 'None';
  FNoOfCardsOnTable := 0;
  FNoOfDiscards := 0;
  FBufferFix := True;
  FPlaceDeckOffset := 53;//7 gives a reasonable 3d look.
  FResizedByPicture := False;
  FDragAndDrop := False;
  FDragFromDeck := False;
  FCardDragging := False;
  FDragFaceDown := False;
  FShadePlaced := False;
  FCopyToScreen := True;
  FRefreshLabels := False;
  SendToBack;
end;

destructor TcardTable.Destroy;
//Call free not this method.
begin
  FBitMap.Free;
  FJpeg.Free;
  FCardDeck.Free;
  inherited Destroy;
end;

procedure TCardTable.SetDrawnCard(const ACard: TCard);
//Needs to be called if Deck is displayed on the table and you draw a card.
//Because although the card is on the table it is not in the CardsOnTable array.
begin
  if ACard.FDisplayed then
  begin
    inc(FNoOfCardsOnTable);
    FCardsOnTable[FNoOfCardsOnTable] := ACard;
  end;
end;

function TCardTable.DrawCardFromDeck(Deck: TCardDeck = nil): TCard;
//Shortcut function - use instead of CardDeck.Draw & SetDrawnCard which you need
//to do otherwise if deck is displayed on screen.
var
  ACard: TCard;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  ACard := Deck.Draw;
  if ACard <> nil then
    SetDrawnCard(ACard);
  Result := ACard;
end;

procedure TCardTable.PlaceDeck(const X, Y: Integer; Deck: TCardDeck = nil);
//Places the deck on table. Shuffling MUST be done prior to this. Cards may be
//drawn before placing the deck. Does not add cards to CardsOnTable array.
//Setting PlaceDeckOffset to for example 7 will give a 3d look to the displayed
//deck & must obviously be done prior to calling this function. To draw a card
//from a placed deck either do CardDeck.Draw & CardTable.SetDrawnCard(ACard) or
//call CardTable.DrawCardFromDeck.
var
  i, Offset: Integer;
begin
  If not Assigned(Deck) then Deck := FCardDeck;
  Offset := 0;
  for i := NoOfCards downto NoOfCards+1 - Deck.NoOfCardsInDeck do
  begin
    PutDown(Deck.FCardArray[i], X+Offset, Y+Offset, Deck, True);
    CopyBuffer(X+Offset, Y+Offset, Deck.FCardWidth, Deck.FCardHeight);
    Deck.FCardArray[i].FDisplayed := true;
    if (i mod FPlaceDeckOffset = 0) then dec(Offset);
  end;
end;

procedure TCardTable.CopyBuffer(const X, Y, Width, Height: Integer);
//Protected method that copies section of FBitMap to Canvas (of owner form if
//visible) X, Y is origin position of invalidated area on FBitMap.
begin
  if Visible and FCopyToScreen then
    BitBlt(Canvas.Handle, X, Y, Width, Height,
      FBitMap.Canvas.Handle, X, Y, SRCCOPY);
end;

procedure TCardTable.PutDown(const ACard: TCard; const X, Y: Integer;
  const Deck: TCardDeck; const SaveBack: Boolean = True);
//Protected method that puts down a card to the buffer.
begin
  if ACard.FDisplayed and SaveBack then exit;//Already on the table.
  ACard.Fx := X;
  ACard.Fy := Y;
  //First store what will be under the card.
  if SaveBack then
    BitBlt(ACard.FUnder.Canvas.Handle, 0,0, Deck.FCardWidth, Deck.FCardHeight,
      FBitMap.Canvas.Handle, X, Y, SRCCOPY);
  if ACard.FFaceUp then
  begin
    BitBlt(FBitMap.Canvas.Handle, X, Y, Deck.FCardWidth, Deck.FCardHeight,
      Deck.FCardMask.Canvas.Handle, 0, 0, SrcAnd);
    BitBlt(FBitMap.Canvas.Handle, X, Y, Deck.FCardWidth, Deck.FCardHeight,
      ACard.FFace.Canvas.Handle, 0, 0, SrcPaint);
  end
  else begin
    BitBlt(FBitMap.Canvas.Handle, X, Y, Deck.FCardWidth, Deck.FCardHeight,
      Deck.FCardMask.Canvas.Handle, 0, 0, SrcAnd);
    BitBlt(FBitMap.Canvas.Handle, X, Y, Deck.FCardWidth, Deck.FCardHeight,
      Deck.FBack.Canvas.Handle, 0, 0, SrcPaint);
  end;
end;

procedure TCardTable.PlaceCard(const ACard: TCard; const X, Y: Integer);
//Places card on the table.
var
  Deck: TCardDeck;
begin
  if not ACard.FDisplayed then
  begin
    Deck := ACard.FOwner;
    PutDown(ACard, X, Y, Deck);
    CopyBuffer(X, Y, Deck.FCardWidth, Deck.FCardHeight);
    ACard.FDisplayed := true;
    inc(FNoOfCardsOnTable);
    FCardsOnTable[FNoOfCardsOnTable] := ACard;
  end;
end;

procedure TCardTable.PickUp(const ACard: TCard; const Deck: TCardDeck);
//Protected method that picks up a card from the buffer.
begin
  BitBlt(FBitMap.Canvas.Handle, ACard.Fx, ACard.Fy, Deck.FCardWidth,
    Deck.FCardHeight, ACard.FUnder.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TCardTable.PickUpCard(const ACard: TCard);
//Picks up card from the table.
var
  ArrayPos, i, CoveringNo: Integer;
  Deck: TCardDeck;
begin
  Deck := ACard.FOwner;
  if ACard.FDisplayed then
  begin
    CoveringNo := 0;
    ArrayPos := PickUpCoveringCards(ACard, Deck, CoveringNo);
    PickUp(ACard, Deck);
    ACard.FDisplayed := False;
    //Delete entry in CardsOnTableArray.
    if ArrayPos <> FNoOfCardsOnTable then
      for i := ArrayPos to FNoOfCardsOnTable-1 do
        FCardsOnTable[i] := FCardsOnTable[i+1];
    dec(FNoOfCardsOnTable);
    if CoveringNo <> 0 then
    begin
      ReplaceCoveringCards(ArrayPos);
      StillUnder;
    end;
    CopyBuffer(Acard.Fx, ACard.Fy, Deck.FCardWidth, Deck.FCardHeight);
  end;
end;

function TCardTable.PickUpCoveringCards(const ACard: TCard;
  const Deck: TCardDeck; var CoveringNo: Integer; const BeginingMove: Boolean)
    : Integer;
//Protected method that picks up cards covering ACard & returns index of ACard
//in the CardsOnTable array. BeginingMove flags the function to mark the
//covering cards. This temporarily uses a Cards FStatus. The following calls to
//this function within the MoveTo procedure will only pick up the original
//covering cards and not any others later in the Move path. This means the card
//will move out only from directly covering cards and onto any others. The var
//parameter CoveringNo has the value of the no of covering cards 0..n OR -1
//which signifies that although 0 cards cover some cards still need to be picked
//up. This function is rather messy.
var
  i, MaxWidth, MaxHeight: Integer;
  PickedUp: Boolean;
  Deck2: TCardDeck;
begin
  Result := 0;
  PickedUp := False;
  for i := FNoOfCardsOnTable downto 1 do
  begin
    if FCardsOnTable[i] = ACard then
    begin
      Result := i;
      break;
    end
    else
    begin
      Deck2 := FCardsOnTable[i].FOwner;
      MaxWidth := Max(Deck.FCardWidth, Deck2.FCardWidth);
      MaxHeight := Max(Deck.FCardHeight, Deck2.FCardHeight);
    //Is this the first stage of a Move Loop or an active marked card
      if (FCardsOnTable[i].FStatus = 2) or BeginingMove then
    //This card might be covering ACard
        if (abs(ACard.Fx - FCardsOnTable[i].Fx) < MaxWidth) and
          (abs(ACard.Fy - FCardsOnTable[i].Fy) < MaxHeight) then
        begin
    //Yes it is so pick it up
          PickUp(FCardsOnTable[i], Deck2);
          FCardsOnTable[i].FDisplayed := False;
    //Mark this card as a covering one
          PickedUp := True;
          if BeginingMove then
          begin
            inc(CoveringNo);
            FCardsOnTable[i].FStatus := 2;
          end;
        end
        else begin
          if not BeginingMove then
          begin
    //A covering card but no longer covering. Still might have to be picked up
    //as it might cover a coverer!
            FCardsOnTable[i].FStatus := 3;
            dec(CoveringNo);
            if CoveringNo <> 0 then
    //Covering the covering?
            begin
              PickUp(FCardsOnTable[i], Deck2);
              FCardsOnTable[i].FDisplayed := False;
              PickedUp := True;
            end
            else
              if PickedUp then
    //0 cards cover but have we already picked a card that needs replacing?
    //So we have to pick this up as well so as to not mess up the display.
              begin
                PickUp(FCardsOnTable[i], Deck2);
                FCardsOnTable[i].FDisplayed := False;
                CoveringNo := -1;
              end;
          end;
        end
      else
        if FCardsOnTable[i].FStatus = 3 then
          if CoveringNo <> 0 then
    //Covering the covering?
          begin
            PickUp(FCardsOnTable[i], Deck2);
            FCardsOnTable[i].FDisplayed := False;
            PickedUp := True;
          end;
    end;
  end;
end;

procedure TCardTable.ReplaceCoveringCards(const ArrayPos: Integer);
//Protected method that replaces cards from ArrayPos to NoOfCardsOnTable. If
//this function needs to be called then so does StillUnder.
var
  i: Integer;
begin
  for i := ArrayPos to FNoOfCardsOnTable do
  begin
    if not FCardsOnTable[i].FDisplayed then
    begin
      PutDown(FCardsOnTable[i], FCardsOnTable[i].Fx, FCardsOnTable[i].Fy,
        FCardsOnTable[i].FOwner);
      FCardsOnTable[i].FDisplayed := True;
    end;
  end;
end;

function TCardTable.StillUnder: Boolean;
//Protected method. Is Status still set to 2 ie card is still under a covering
//card. Also resets status to 1. Must be called after a final call to
//ReplaceCoveringCards.
var
  i: Integer;
begin
  Result := False;
  for i := FNoOfCardsOnTable downto 1 do
    if FCardsOnTable[i].FStatus > 1 then
    begin
      if FCardsOnTable[i].FStatus = 2 then Result := True;
      FCardsOnTable[i].FStatus := 1;
    end;
end;

procedure TCardTable.MoveTo(const ACard: TCard; const X, Y: Integer);
//See also PickUpCoveringCards comments for more details of the movement system.
//Note when you move cards from below covering cards that themselves have
//covering cards AND you attempt to move cards over these covering cards on the
//following or later moves then you may find that the buffer has been corrupted.
//This can be fixed by calling Paint before moving over the covering cards. This
//though can be slow. A nasty complex messy function.
var
  Width, Height, x1, x2, y1, y2, ArrayPos, i, CoveringNo, Speed: Integer;
  Status, CardHasShadow: Boolean;
  Deck: TCardDeck;
begin
  Deck := ACard.FOwner;
  x1 := ACard.Fx;
  y1 := ACard.Fy;
  Status := True;
  CardHasShadow := False;
  CoveringNo := 0;
  ArrayPos := 0;
  if ACard.FDisplayed then
  begin
    Speed := FCardSpeed;//Later set higher if moving out from under cards.
    repeat
    //First pick up covering cards.
    //CoveringNo is set to -2 when 0 more cards cover the moving card.
      if CoveringNo <> -2 then
        ArrayPos := PickUpCoveringCards(ACard, Deck, CoveringNo, Status);
      If Status = True then Status := False;
    //Now pick up card.
      PickUp(ACard, Deck);
      if CardHasShadow then PickUpShadow(Deck);
      ACard.FDisplayed := False;
      if ACard.Fx > X then dec(ACard.Fx, Speed)
      else if ACard.Fx < X then inc(ACard.Fx, Speed);
      if ACard.Fy > Y then dec(ACard.Fy, Speed)
      else if ACard.Fy < Y then inc(ACard.Fy, Speed);
      if (ACard.Fx<>X) and (abs(ACard.Fx - X) < Speed) then ACard.Fx := X;
      if (ACard.Fy<>Y) and (abs(ACard.Fy - Y) < Speed) then ACard.Fy := Y;
    //Put it down again.
      if (CoveringNo = 0) or (CoveringNo = -2) then
      begin
        PutDownShadow(ACard.Fx, ACard.Fy, Deck);//Note not if -1.
        CardHasShadow := True;
      end;
      PutDown(ACard, ACard.Fx, ACard.Fy, Deck);
    //Replace the covering cards. CoveringNo = -1 means 0 cards cover but STLL
    //cards need to be replaced which is why we cant yet draw a shadow.
      if (CoveringNo = -1) or (CoveringNo > 0) then
        ReplaceCoveringCards(ArrayPos+1);
    //CoveringNo = 0 & -1 mean 0 more covering cards. This is set to -2 to flag
    //to this function that we do not need to lift up any more covering cards.
      if (CoveringNo = 0) or (CoveringNo = -1) then
      begin
        CoveringNo := -2;
        Speed := FCardSpeed;
      end else
        if CoveringNo <> -2 then Speed := FCardSpeed + CoveringNo;
    //Find invalidated region.
      if x1 < ACard.Fx then
        x2 := ACard.Fx
      else
      begin
        x2 := x1;
        x1 := ACard.Fx;
      end;
      if y1 < ACard.Fy then
        y2 := ACard.Fy
      else
      begin
        y2 := y1;
        y1 := ACard.Fy;
      end;
      Width := x2-x1+Deck.FCardWidth;
      Height := y2-y1+Deck.FCardHeight;
      if CardHasShadow then
      begin
        Width := Width + Deck.FShadow01.Width;
        Height := Height + Deck.FShadow01.Width;
      end;
    //Finally copying the buffer to the screen.
      CopyBuffer(x1, y1, Width, Height);
    until (Acard.Fx = X) and (ACard.Fy = Y);
    Acard.FDisplayed := true;
    if not StillUnder then
    begin
    //Card has moved and is not under any others so move its position in array.
    //This also means that if the card IS still under that its position in the
    //CardsOnTable array is unchanged. If it has moved over others which are
    //higher in the array then expect messups. In practice shoudn`t cause
    //problems.
      if ArrayPos <> FNoOfCardsOnTable then
      begin
        for i := ArrayPos to FNoOfCardsOnTable-1 do
          FCardsOnTable[i] := FCardsOnTable[i+1];
        FCardsOnTable[FNoOfCardsOnTable] := ACard;
      end;
      if CardHasShadow then
      begin
        PickUp(ACard, Deck);
        PickUpShadow(Deck);
        ACard.FDisplayed := False;
        PutDown(ACard, ACard.Fx, ACard.Fy, Deck);
        ACard.FDisplayed := True;
        CopyBuffer(x1, y1, Width, Height);
      end;
    end;
  end;
end;

procedure TCardTable.TurnOverCard(const ACard: TCard);
//Turn over a card on the table. A much worse and slower method would be to set
//ACard.FaceUp to false & then call CardTable.DeckOrBackChanged.
var
  CoveringNo, ArrayPos: Integer;
  Deck: TCardDeck;
begin
  Deck := ACard.FOwner;
  ACard.FFaceUp := not ACard.FFaceUp;
  if ACard.FDisplayed then
  begin
    CoveringNo := 0;
    ArrayPos := PickUpCoveringCards(ACard, Deck, CoveringNo);
    PutDown(ACard, ACard.Fx, ACard.Fy, Deck, False);
    if CoveringNo <> 0 then
    begin
      ReplaceCoveringCards(ArrayPos);
      StillUnder;
    end;
    CopyBuffer(Acard.Fx, ACard.Fy, Deck.FCardWidth, Deck.FCardHeight);
  end;
end;

procedure TCardTable.Discard(const ACard: TCard);
//Send card to a visible discard pile. This movement takes no account of how the
//cards are displayed & may result in display mess ups. In this case you would
//either manually move the card nearer to the discard pile before calling
//discard or pick up the card & then discard it straight on to the discard pile.
var
  Offset, i, j: Integer;
begin
  ACard.FStatus := -1;
  inc(FNoOfDiscards);
  FDiscardArray[FNoOfDiscards] := ACard;
  Offset := FNoOfDisCards div FPlaceDeckOffset;
  if ACard.FDisplayed then
  begin
    MoveTo(ACard, FDiscardsAt.X - Offset, FDiscardsAt.Y - Offset);
    //Remove entry from CardsOnTable array.
    if FCardsOnTable[FNoOfCardsOnTable] <> ACard then
    begin
      j := 0;
      repeat
        inc(j);
      until FCardsOnTable[j] = ACard;
      for i := j to FNoOfCardsOnTable-1 do
        FCardsOnTable[i] := FCardsOnTable[i+1];
    end;
    dec(FNoOfCardsOnTable);
  end else
  begin
    PutDown(ACard, FDiscardsAt.X - Offset, FDiscardsAt.Y - Offset,
      ACard.FOwner);
    ACard.Displayed := True;
  end;
  Paint;
  RefreshLabels;
end;

procedure TCardTable.PickUpDeck(Deck: TCardDeck = nil);
//Picks up deck from the table.
var
  i: integer;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  for i := Deck.FTop+1 to NoOfCards do
    if Deck.FCardArray[i].FDisplayed then
    begin
      PickUp(Deck.FCardArray[i], Deck);
      Deck.FCardArray[i].FDisplayed := False;
    end;
  Paint;
  RefreshLabels;
end;

procedure TCardTable.PickUpDiscards;
//Picks up the discard pile not changing cards FStatus.
var
  i: Integer;
begin
  if FNoOfDiscards > 0 then
  begin
    for i := FNoOfDiscards downto 1 do
      if FDiscardArray[i].FDisplayed then
      begin
        PickUp(FDiscardArray[i], FDiscardArray[i].FOwner);
        FDiscardArray[i].FDisplayed := False;
      end;
    FNoOfDiscards := 0;
    Paint;
    RefreshLabels;
  end;
end;

procedure TCardTable.PickUpCardsOnTable;
//Note this method ONLY picks up the cards in the CardsOnTable array & not any
//displayed decks. It also does not change the cards status. To pick up all the
//cards on the table and reset the deck call PickUpAllCardsAndReset. You also
//must not forget to shuffle the deck before using again.
var
  i: Integer;
begin
  for i := FNoOfCardsOnTable downto 1 do
  begin
    PickUp(FCardsOnTable[i], FCardsOnTable[i].FOwner);
    FCardsOnTable[i].FDisplayed := False;
  end;
  FNoOfCardsOnTable := 0;
  Paint;
  RefreshLabels;
end;

procedure TCardTable.PickUpAllCardsAndReset(Deck: TCardDeck = nil);
//Picks up all cards on the table resetting all but card order. Fixed by a
//shuffle.
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  FCopyToScreen := False;
  PickUpCardsOnTable;
  PickUpDiscards;
  FCopyToScreen := True;
  PickUpDeck(Deck);
  Deck.ResetDeck;
  RefreshLabels;
end;

procedure TCardTable.SetSpeed(const Speed: Integer);
//Property method that sets FCardSpeed.
begin
  if Speed >0 then FCardSpeed := Speed;
end;

function TCardTable.MouseButtonPressed(const X, Y: Integer): TCard;
//Protected method. Is there a card under these co-ordinates? Returns either a
//TCard or nil. If the card is in the deck ie ACard.Status = 0 then the deck has
//been chosen otherwise the card is on the table. If the source is the deck then
//a card will need to be drawn from the deck unless DragAndDrop and DragFromDeck
//are enabled when the card is automatically drawn. ExternalDeck property can be
//set to another deck allowing this deck also to be selected.
var
  i, j, X1, Y1: Integer;
  ACard: TCard;
  Deck: TCardDeck;
begin
  Result := nil;
  //First check the cards on the table.
  for i := FNoOfCardsOnTable downto 1 do
  begin
    ACard := FCardsOnTable[i];
    Deck := ACard.FOwner;
    X1 := ACard.Fx;
    Y1 := ACard.Fy;
    if ((X >= X1) and (X <= X1 + Deck.FCardWidth))
      and ((Y >= Y1) and (Y <= Y1 + Deck.FCardHeight)) then
      begin
        Result := ACard;
        break;
      end;
  end;
  //Now cards in the deck.
  if Result = nil then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := NoOfCards downto 1 do
      begin
        ACard := Deck.FCardArray[i];
        if ACard.FDisplayed and (ACard.FStatus = 0) then
        begin
          X1 := ACard.Fx;
          Y1 := ACard.Fy;
          if ((X >= X1) and (X <= X1 + Deck.FCardWidth))
            and ((Y >= Y1) and (Y <= Y1 + Deck.FCardHeight)) then
            begin
              Result := ACard;
              break;
            end;
        end;
      end;
    end;
end;
{$WARNINGS OFF}
procedure TCardTable.Paint;
//Called automatically by windows and by TCardTable to draw the table. Note
//lightweight controls on the table eg TLabel are often overwritten by this and
//will need to be repainted. Set RepaintLabels property of TCardTable to True
//and the cardtable will do this task as required.
var
  i: Integer;
  ACard: TCard;
  Deck: TCardDeck;
begin
//If FBufferFix then cards on the table are picked up & put down again. This
//acts as a fix for errors possibly introduced into the buffer by the movement
//system. Caused because cards that cover cards that cover the object card but
//that dont cover the object card itself are not picked up when movement is
//performed and this results in errors within the buffer that need correcting.
  if FBufferFix then//By default True.
  begin
    if FCardDragging then//As dragged card is not in CardsOnTable array.
    begin
      ACard := FDraggingCard;
      Deck := ACard.FOwner;
      BitBlt(FBitMap.Canvas.Handle, ACard.Fx, ACard.Fy, Deck.FCardWidth,
        Deck.FCardHeight, ACard.FUnder.Canvas.Handle, 0, 0, SRCCOPY);
      ACard.FDisplayed := False;
      PickUpShadow(Deck);
    end;
    for i := FNoOfCardsOnTable downto 1 do//Pick up the cards.
      if FCardsOnTable[i].FDisplayed then//RedrawBuffer sets this to false.
      begin
        Deck := FCardsOnTable[i].FOwner;
        BitBlt(FBitMap.Canvas.Handle, FCardsOnTable[i].Fx, FCardsOnTable[i].Fy,
          Deck.FCardWidth, Deck.FCardHeight,
          FCardsOnTable[i].FUnder.Canvas.Handle, 0, 0, SRCCOPY);
        FCardsOnTable[i].FDisplayed := False;
      end;
    for I := 1 to FNoOfCardsOnTable do//And put them down again.
    begin
      Deck := FCardsOnTable[i].FOwner;
      PutDown(FCardsOnTable[i], FCardsOnTable[i].Fx, FCardsOnTable[i].Fy, Deck);
      FCardsOnTable[i].FDisplayed := True;
    end;
    if FCardDragging then
    begin
      Deck := ACard.FOwner;
      PutDownShadow(Acard.Fx, ACard.Fy, Deck);
      PutDown(ACard, ACard.Fx, ACard.Fy, Deck);
      ACard.FDisplayed := True;
    end;
  end;
//Finally copying the buffer to the form. Some methods set FCopyToScreen to
//false to stop intermediate redraws.
  if FCopyToScreen then
    CopyBuffer(0,0,FBitMap.Width,FBitMap.Height);
end;
{$WARNINGS ON}
procedure TCardTable.RedrawBuffer;
//Protected method that calls Paint forcing the buffer to be redrawn with the
//deck & cards in the correct position after the table has been resized, the
//card deck/back changed or after the tabletop picture has been changed.
var
  BufferFix, CopyToScreen: Boolean;
  i, j: Integer;
  Deck: TCardDeck;
begin
//Even if FCopyToScreen is true we dont need to copy individual changes made to
//the buffer to the screen.
  CopyToScreen := FCopyToScreen;
  FCopyToScreen := False;
//FBufferFix must be True to redraw the cards on the table.
  BufferFix := FBufferFix;
  FBufferFix := True;
//First draw any CardMarkers if used.
  PlaceCardMarkers;
//Now we need to ensure that the cards on the table are not picked up on the
//later call to paint as they will have been overdrawn by the background.
  for i := FNoOfCardsOnTable downto 1 do
    FCardsOnTable[i].FDisplayed := False;
//Redraw the internal deck & external deck in FExternalDeck if assigned.
  for i := 1 to 2 do
  begin
    if i = 1 then Deck := FCardDeck
    else
      if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
//Is the deck displayed.
    if (Deck.FCardArray[NoOfCards].FStatus = 0) and
      (Deck.FCardArray[NoOfCards].FDisplayed) then
      for j := NoOfCards downto Deck.FTop + 1 do
      begin
        Deck.FCardArray[j].FDisplayed := False;
        PutDown(Deck.FCardArray[j], Deck.FCardArray[j].Fx,
          Deck.FCardArray[j].Fy, Deck, True);
        Deck.FCardArray[j].FDisplayed := True;
      end;
  end;
//Redraw the discard pile if used.
  if FNoOfDiscards > 0 then
    for i := 1 to FNoOfDiscards do
      if FDiscardArray[i].FDisplayed then
      begin
        FDiscardArray[i].FDisplayed := False;//Else PutDown wont work.
        PutDown(FDiscardArray[i], FDiscardArray[i].Fx, FDiscardArray[i].Fy,
          FDiscardArray[i].FOwner, True);
        FDiscardArray[i].FDisplayed := true;
      end;
  FCopyToScreen := CopyToScreen;
//Redraw the cards on the table while removing flicker when loading pictures.
  if FResizedByPicture then
    Update
  else
  begin
    Paint;
    RefreshLabels;
  end;
  FBufferFix := BufferFix;
end;

procedure TCardTable.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
//Protected method that if FDragAndDrop is True and the left buton pressed the
//card is Lifted off the table. If card is within the deck and DragFromDeck is
//True then a card is drawn. If FOnCardClickEvent has been assigned then a
//TCardClickEvent is generated.
var
  ACard: TCard;
  i, CoveringNo: Integer;
  Deck: TCardDeck;
begin
  inherited;
  ACard := MouseButtonPressed(X,Y);
  if ACard <> nil then
  begin
  //Clicked on a card inform user by generating TCardClick event.
    if Assigned(FOnCardClickEvent) then
      FOnCardClickEvent(ACard, Button);
    if ((ACard.FStatus = 1) and (ACard.FFaceUp or FDragFaceDown)) or
      ((ACard.FStatus = 0) and FDragFromDeck) then
      if FDragAndDrop and (Button = mbLeft) then
      begin
        Deck := ACard.FOwner;
        if ACard.FStatus = 0 then
        begin//Dragging a card from the deck.
          ACard := Deck.Draw;
          SetDrawnCard(ACard);
          FCOTPosition := 0;
          PickUp(ACard, Deck);
          ACard.FDisplayed := False;
          dec(FNoOfCardsOnTable);
          FCardDraggedFromDeck := True;
        end else
        begin//Dragging a card on the table. Nb no longer in CardsOnTable array.
          CoveringNo := 0;
          FCOTPosition := PickUpCoveringCards(ACard, Deck, CoveringNo);
          PickUp(ACard, Deck);
          ACard.FDisplayed := False;
          if FCOTPosition <> FNoOfCardsOnTable then
            for i := FCOTPosition to FNoOfCardsOnTable-1 do
              FCardsOnTable[i] := FCardsOnTable[i+1];
          dec(FNoOfCardsOnTable);
          if CoveringNo <> 0 then
          begin
            ReplaceCoveringCards(FCOTPosition);
            StillUnder;
          end;
          FCardDraggedFromDeck := False;
        end;
        FCardDragging := True;//Setting various fields used for drag & drop.
        FDraggingCard := ACard;
        FDragOrigin.X := ACard.Fx;
        FDragOrigin.Y := ACard.Fy;
        FMouseRelX := X - ACard.Fx;
        FMouseRelY := Y - ACard.Fy;
        FPreDragCursor := Screen.Cursor;
        PutDownShadow(ACard.Fx, ACard.Fy, Deck);
        PutDown(ACard, ACard.Fx, ACard.Fy, Deck);
        ACard.FDisplayed := True;
        Paint;
        RefreshLabels;
        Screen.Cursor := crDrag;
      end;
  end;
end;

procedure TCardTable.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
//Protected method that when the left mouse button is released and CardDragging
//is true then a TCardDropEvent is raised and the user MUST then either call
//DropCard, DropCardOnShade or RestoreCard.
begin
  if FCardDragging and (Button = mbLeft) then
  begin
    FCardDragging := False;
    if Assigned(FOnCardDropEvent) then
      FOnCardDropEvent(FDraggingCard, X, Y, FMouseRelX, FMouseRelY);
    Screen.Cursor := FPreDragCursor;
  end;
end;

procedure TCardTable.MouseMove(Shift: TShiftState; X, Y: Integer);
//Protected method that if a card is being dragged by the mouse handles the
//movement of the card.
var
  ACard: TCard;
  X1, Y1, X2, Y2, Width, Height: Integer;
  Deck: TCardDeck;
begin
  inherited;
  if FCardDragging then
  begin
    ACard := FDraggingCard;
    Deck := ACard.FOwner;
    X1 := ACard.Fx;
    Y1 := ACard.Fy;
    PickUp(ACard, Deck);
    PickUpShadow(Deck);
    ACard.FDisplayed := False;
    ACard.Fx := X - FMouseRelX;
    ACard.Fy := Y - FMouseRelY;
    if ACard.Fx < 0 then ACard.Fx := 0;
    if ACard.Fy < 0 then ACard.Fy := 0;
    if ACard.Fx > Self.Width - Deck.FCardWidth then
      ACard.Fx := Self.Width - Deck.FCardWidth;
    if ACard.Fy > Self.Height - Deck.FCardHeight then
      ACard.Fy := Self.Height - Deck.FCardHeight;
    PutDownShadow(ACard.Fx, ACard.Fy, Deck);
    PutDown(ACard, ACard.Fx, ACard.Fy, Deck);
    ACard.FDisplayed := True;
//Find invalidated region.
    if X1 < ACard.Fx then
      X2 := ACard.Fx
    else begin
      X2 := X1;
      X1 := ACard.Fx;
    end;
    if Y1 < ACard.Fy then
      Y2 := ACard.Fy
    else begin
      Y2 := Y1;
      Y1 := ACard.Fy;
    end;
    Width := X2 - X1 + Deck.FCardWidth + Deck.FShadow01.Width;
    Height := Y2 - Y1 + Deck.FCardHeight + Deck.FShadow01.Width;
    CopyBuffer(x1, y1, Width, Height);
  end;
end;

procedure TCardTable.DropCard(X, Y: Integer);
//Method that user calls in response to a CardDropEvent raised by cardtable. It
//drops card at the x,y co-ordinates.
var
  Deck: TCardDeck;
begin
  Deck := FDraggingCard.FOwner;
  PickUp(FDraggingCard, Deck);
  PickUpShadow(Deck);
  FDraggingCard.FDisplayed := False;
  PutDown(FDraggingCard, X, Y, Deck, True);
  FDraggingCard.FDisplayed := True;
  inc(FNoOfCardsOnTable);
  FCardsOnTable[FNoOfCardsOnTable] := FDraggingCard;
  FDraggingCard := nil;
  //The following 2 lines allow the forced dropping of a card in code.
  FCardDragging := False;
  Screen.Cursor := FPreDragCursor;
  Paint;
  RefreshLabels;
end;

procedure TCardTable.RestoreCard;
//Called by user in response to a CardDropEvent that restores a dragged card to
//its original position on the table.
var
  i, CoveringNo: Integer;
  Deck: TCardDeck;
begin
  Deck := FDraggingCard.FOwner;
  PickUp(FDraggingCard, Deck);
  PickUpShadow(Deck);
  FDraggingCard.FDisplayed := False;
  if FCOTPosition > 0 then
  begin
    inc(FNoOfCardsOnTable);
    for i := FNoOfCardsOnTable downto FCOTPosition + 1 do
      FCardsOnTable[i] := FCardsOnTable[i-1];
    FCardsOnTable[FCOTPosition] := FDraggingCard;
    FDraggingCard.Fx := FDragOrigin.X;
    FDraggingCard.Fy := FDragOrigin.Y;
    CoveringNo := 0;
    PickUpCoveringCards(FDraggingCard, Deck, CoveringNo)
  end;
  PutDown(FDraggingCard, FDragOrigin.X, FDragOrigin.Y, Deck,
    True);
  FDraggingCard.FDisplayed := True;
  if FCOTPosition > 0 then
  begin
    if CoveringNo <> 0 then
    begin
      ReplaceCoveringCards(FCOTPosition+1);
      StillUnder;
    end
  end else
  begin
    FDraggingCard.FStatus := 0;
    dec(Deck.FTop);
  end;
  FDraggingCard := nil;
  FCardDragging := False;
  Screen.Cursor := FPreDragCursor;
  Paint;
  RefreshLabels;
end;

procedure TCardTable.LoadTableTop(const FileName: TFileName);
//Property method. A BitMap or Jpeg can be displayed on the table top. If the
//table is later resized when FStretchBackground is false then the bitmap is
//replaced by FColor. NB This MUST MUST have a fully qualified file name. ie
//c:\pic\a.jpg not just a.jpg even if in the program directory. So if in program
//dir use CardTable.Directory + '\a.jpg' as the filename.
var
  BitMap: TBitMap;
  Bounds: TRect;
begin
  if not FBitMap.Empty then//Maybe loses color depth otherwise.
  begin
    FBitMap.Free;
    FBitMap := TBitMap.Create;
  end;
  if FileExists(FileName) then
  begin
    if AnsiCompareText(ExtractFileExt(FileName), '.jpg') = 0  then
      if FFileName <> FileName then
      begin
        if not FJpeg.Empty then
        begin
          FJpeg.Free;
          FJpeg := TJpegImage.Create;
          FJpeg.Performance := FJpegPerformance;
        end;
        FJpeg.LoadFromFile(FileName);
      end;
//StretchBackground determines if the background picture determines the size of
//the tabletop or if the picture is stretched to the tabletop.
    if FStretchBackground then
    begin
      BitMap := TBitMap.Create;
      FBitMap.Height := Height;
      FBitMap.Width := Width;
      try
        if AnsiCompareText(ExtractFileExt(FileName), '.jpg') = 0  then
          BitMap.Assign(FJpeg)
        else
          BitMap.LoadFromFile(FileName);
        Bounds.Left:=0;
        Bounds.Top:=0;
        Bounds.Right:=Width;
        Bounds.Bottom:=Height;
        FBitMap.Canvas.StretchDraw(Bounds, Bitmap);
      finally
        BitMap.Free;
      end;
    end
    else begin
      if AnsiCompareText(ExtractFileExt(FileName), '.jpg') = 0  then
        FBitMap.Assign(FJpeg)
      else
        FBitMap.LoadFromFile(FileName);
      if (Height <> FBitMap.Height) or (Width <> FBitMap.Width) then
      begin
        FResizedByPicture := True;//Flagging to Resized not to change FBitmap.
        Height := FBitMap.Height;
        Width := FBitMap.Width;
      end;
    end;
    FFileName := FileName;
    FBitMap.Dormant;
    FBitMap.FreeImage;
  end else
  begin
    SetColor(FColor);
    FResizedByPicture := True;
  end;
  RedrawBuffer;//Redrawing any cards that were on the table.
  FResizedByPicture := False;//Resetting the flag.
end;

procedure TCardTable.SetColor(const Color: TColor);
//Property method that sets the color of the table if no background pic is used.
begin
  if not FBitMap.Empty then
  begin
    FBitMap.Free;
    FBitMap := TBitMap.Create;
  end;
  FBitMap.Height := Height;
  FBitMap.Width := Width;
  FColor := Color;
  FBitMap.Canvas.Brush.Style := bsSolid;
  FBitMap.Canvas.Brush.Color := FColor;
  FBitMap.Canvas.FillRect(Rect(0, 0, FBitMap.Width, FBitMap.Height));
  if FFileName <> 'None' then FFileName := 'None';
  RedrawBuffer;//Redraw any cards already on the table.
end;

procedure TCardTable.Show;
//Makes a hidden cardtable visible.
begin
  If not Parent.Visible then Parent.Show;
  inherited;
  Paint;
  RefreshLabels;
  Update;
end;

procedure TCardTable.Resize;
var
  FlagStatus: Boolean;
begin
  FlagStatus := FCopyToScreen;
  FCopyToScreen := False;
  if ((FBitMap.Width <> Width) or (FBitMap.Height <> Height))
    and (not FResizedByPicture) then
  Begin
    if FStretchBackground then
    begin
      FResizedByPicture := True;
      LoadTableTop(BackgroundPicture);
    end
    else
      SetColor(FColor);
  end;
  FCopyToScreen := FlagStatus;
end;

function TCardTable.GetCanvas: TCanVas;
//Property method surfacing the internal buffer.
begin
  Result := FBitmap.Canvas;
end;

procedure TCardTable.SetBufferFix(const Value: Boolean);
//Property method that sets FBufferFix. Used by paint.
begin
  if FBufferFix <> Value then FBufferFix := Value;
end;
{$WARNINGS OFF}
procedure TCardTable.SortHand(var Hand: Array of TCard; Const AcesLow: Boolean;
  Const SuitSort: Boolean; Mode: TMode = mAuto);
//Sorts a hand of cards of any 1..NoOfCards size. If visible then the card
//movement is displayed on screen - the cards will switch positions with each
//other. A limitation being that the cards must be arranged horizontally. If
//AcesLow is True aces are low, if False high. If SuitSort is True cards are
//sorted within suits. Mode determines if cards are moved during the switching
//process higher or lower in the Y plane than the cards in the hand. Value mAuto
//lets the method choose the direction depending upon the Y position of the
//hand. Value mHorizontal moves cards only in the X plane & should only be used
//when the cards displayed have no covering cards. Will work with unassigned
//array elements if these are higher in the array than the assigned cards. ie If
//[0..7] contain cards and [8..12] are nil than this will still work ok. Dont
//pass it an array only containing nil.
var
  StoredXY: Array[0..NoOfCards-1] of TPoint;
  i, j, k, iValue, jValue, CoveringNo, ArrayPos, Top: Integer;
  MoveCard, TempCard: TCard;
  SwapPoint: TPoint;
  Deck: TCardDeck;
begin
//First store the origin X,Y of each card as we need it later to swap the cards.
  for i := 0 to High(Hand) do
    if Assigned(Hand[i]) then
    begin
      StoredXY[i].X := Hand[i].Fx;
      StoredXY[i].Y := Hand[i].Fy;
      Top := i;
    end
    else break;
//Top points to last non nil element or High(Hand). Now sort the hand.
  for i := 0 to Top-1 do
    for j := Top downto i+1 do
    begin
      iValue := Ord(Hand[i].FValue);
      jValue := Ord(Hand[j].FValue);
      if not AcesLow then
      begin
        if iValue = 0 then iValue := 13;
        if jValue = 0 then jValue := 13;
      end;
      if SuitSort then
      begin
        iValue := iValue + Ord(Hand[i].FSuit)*13;
        jValue := jValue + Ord(Hand[j].FSuit)*13;
      end;
      if jValue < iValue then
      begin
        TempCard := Hand[i];
        Hand[i] := Hand[j];
        Hand[j] := TempCard;
      end;
  end;
  if Hand[0].FDisplayed then
//Move the cards to their sorted positions.
    if Mode = mAuto then
      if Hand[0].Fy < Height div 2 then
        Mode := mDown
      else
        Mode := mUp;
    for i := 0 to Top do
    begin
      MoveCard := Hand[i];
      Deck := MoveCard.FOwner;
      SwapPoint.X := StoredXY[i].X;//The origin we stored earlier.
      SwapPoint.Y := StoredXY[i].Y;
//MoveCard needs to be moved to the position of SwapPoint.
      if Mode <> mHorizontal then
      begin
        if Mode = mUp then j := -Deck.FCardHeight
          - Deck.FShadow01.Width//Taking account of shadow when moving up.
        else
          j := Deck.FCardHeight;
        if (MoveCard.Fx = SwapPoint.X) and (MoveCard.Fy = SwapPoint.Y) then
        begin
          CoveringNo := 0;
          ArrayPos := PickUpCoveringCards(MoveCard, Deck, CoveringNo);
          PickUp(MoveCard, Deck);
          MoveCard.FDisplayed := False;
          if ArrayPos <> FNoOfCardsOnTable then
            for k := ArrayPos to FNoOfCardsOnTable-1 do
              FCardsOnTable[k] := FCardsOnTable[k+1];
          dec(FNoOfCardsOnTable);
          if CoveringNo <> 0 then
          begin
            ReplaceCoveringCards(ArrayPos);
            StillUnder;
          end;
          PlaceCard(MoveCard, SwapPoint.X, SwapPoint.Y);
        end
        else begin
//Three stage movement bipasses limitations of the movement system.
          MoveTo(MoveCard, MoveCard.Fx, MoveCard.Fy+j);
          MoveTo(MoveCard, SwapPoint.X, MoveCard.Y);
          MoveTo(MoveCard, SwapPoint.X, SwapPoint.Y);
        end
      end
      else
        MoveTo(MoveCard, SwapPoint.X, MoveCard.Y);
    end;
//Finally fixing any buffer errors that may have been introduced.
    if Mode <> mHorizontal then
    begin
      Paint;
      RefreshLabels;
    end;
end;
{$WARNINGS ON}
procedure TCardTable.SetDeckOffset(const Offset: Integer);
//Property method that governs how a deck is displayed on the table. An offset
//of around 7 gives a reasonable 3d look to the deck.
begin
  If Offset <> 0 then FPlaceDeckOffset := Offset;
end;

procedure TCardTable.SetDragAndDrop(const Mode: Boolean);
//Property method enabling drag and drop with the mouse.
begin
  if Mode <> self.FDragAndDrop then
    FDragAndDrop := Mode;
end;

procedure TCardTable.SetDragFromDeck(const Mode: Boolean);
//Property method enabling cards to be dragged and drawn from a displayed deck.
begin
  if Mode <> FDragFromDeck then
    FDragFromDeck := Mode;
end;

procedure TCardTable.SetDragFaceDown(const Mode: Boolean);
//Property method  that if true allows face down cards to be dragged.
begin
  if Mode <> FDragFaceDown then
    FDragFaceDown := Mode;
end;

procedure TCardTable.PutDownShadow(const X, Y: Integer; const Deck: TCardDeck);
//Protected method that draws a cards shadow.
var
  X1, Y1, X2, Y2: Integer;
begin
  X1 := X + Deck.FShadow01.Width;
  Y1 := Y + Deck.FCardHeight - Deck.FShadow00.Height + Deck.FShadow01.Width;
  X2 := X + Deck.FCardWidth;
  Y2 := Y + Deck.FShadow01.Width;
  Deck.FShadow00Point.X := X1;
  Deck.FShadow00Point.Y := Y1;
  Deck.FShadow01Point.X := X2;
  Deck.FShadow01Point.Y := Y2;
  BitBlt(Deck.FShadow00Under.Canvas.Handle, 0, 0, Deck.FShadow00Under.Width,
    Deck.FShadow00Under.Height, FBitMap.Canvas.Handle, X1, Y1, SRCCOPY);
  BitBlt(FBitMap.Canvas.Handle, X1, Y1, Deck.FShadow00.Width,
    Deck.FShadow00.Height, Deck.FShadow00.Canvas.Handle, 0, 0, SRCAND);
  BitBlt(Deck.FShadow01Under.Canvas.Handle, 0, 0, Deck.FShadow01Under.Width,
    Deck.FShadow01Under.Height, FBitMap.Canvas.Handle, X2, Y2, SRCCOPY);
  BitBlt(FBitMap.Canvas.Handle, X2, Y2, Deck.FShadow01.Width,
    Deck.FShadow01.Height, Deck.FShadow01.Canvas.Handle, 0, 0, SRCAND);
end;

procedure TCardTable.PickUpShadow(const Deck: TCardDeck);
//Protected method that picks up a shadow from the buffer.
begin
  BitBlt(FBitMap.Canvas.Handle, Deck.FShadow01Point.X, Deck.FShadow01Point.Y,
    Deck.FShadow01Under.Width, Deck.FShadow01Under.Height,
    Deck.FShadow01Under.Canvas.Handle, 0, 0, SRCCOPY);
  BitBlt(FBitMap.Canvas.Handle, Deck.FShadow00Point.X, Deck.FShadow00Point.Y,
    Deck.FShadow00Under.Width, Deck.FShadow00Under.Height,
    Deck.FShadow00Under.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TCardTable.SetCardDeck(const Deck: TCardDeck);
//Property method that allows a 2nd external deck to be used with cardtable.
begin
  if Assigned(Deck) then
    FCardDeck.Assign(Deck);
end;
{$WARNINGS OFF}
procedure TCardTable.PlaceCardMarker(const CardMarker: TCardMarker; const X, Y:
  Integer; Deck: TCardDeck = nil);
//Seven CardMarkers can be placed on the table to signify where cards will be
//moved. These are named in the TCardMarker enumerated type. It is not possible
//to place more than 1 of each type on the table although it shouldn`t be too
//difficult to change this. Alternatively a second set of CardMarks can be
//created & used in an external deck which is assigned to the ExternalDeck
//property. A CardMarker or shade can be placed whilst a card is being dragged.
//The OnMouseMove event has to be caught & PlaceCardMarker called. Before
//dropping the card PickUpMarker must be called. Alternatively just call
//PlaceCardOnShade to pick up the shade & put down the card at its position.
var
  ACard: TCard;
  P: PCardMark;
  Deck2: TCardDeck;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  if not Deck.FCardMarks[CardMarker].Displayed then
  begin
    if FCardDragging then
    begin
      ACard := FDraggingCard;
      Deck2 := ACard.FOwner;
      PickUp(ACard, Deck2);
      PickUpShadow(Deck2);
      ACard.FDisplayed := False;
    end;
    P := @Deck.FCardMarks[CardMarker];
    BitBlt(P^.Under.Canvas.Handle, 0, 0, Deck.FCardWidth, Deck.FCardHeight
      ,FBitMap.Canvas.Handle, X, Y, SRCCOPY);
    BitBlt(FBitMap.Canvas.Handle, X, Y, Deck.FCardWidth, Deck.FCardHeight,
      P^.Mark.Canvas.Handle, 0, 0, SRCAND);
    P^.Position.X := X;
    P^.Position.Y := Y;
    P^.Displayed := True;
    if FCardDragging then
    begin
      PutDownShadow(ACard.Fx, ACard.Fy, Deck2);
      PutDown(ACard, ACard.Fx, ACard.Fy, Deck2);
      ACard.FDisplayed := True;
      FShadePlaced := True;
    end;
    CopyBuffer(X, Y, Deck.FCardWidth, Deck.FCardHeight);
  end;
end;

procedure TCardTable.PickUpMarker(const CardMarker: TCardMarker;
  Deck: TCardDeck = nil);
//Picks up a cardmarker from the table.
var
  ACard: TCard;
  P: PCardMark;
  Deck2: TCardDeck;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  if Deck.FCardMarks[CardMarker].Displayed then
  begin
  //As left mouse button may have been released & FCardDragging set to False.
    if Assigned(FDraggingCard) then
    begin
      ACard := FDraggingCard;
      Deck2 := ACard.FOwner;
      PickUp(ACard, Deck2);
      PickUpShadow(Deck2);
      ACard.FDisplayed := False;
    end;
    P := @Deck.FCardMarks[CardMarker];
    BitBlt(FBitMap.Canvas.Handle, P^.Position.X, P^.Position.Y, Deck.FCardWidth,
      Deck.FCardHeight, P^.Under.Canvas.Handle, 0, 0, SRCCOPY);
    P^.Displayed := False;
    if Assigned(FDraggingCard) then
    begin
      PutDownShadow(ACard.Fx, ACard.Fy, Deck2);
      PutDown(ACard, ACard.Fx, ACard.Fy, Deck2);
      ACard.FDisplayed := True;
      FShadePlaced := False;
    end;
    CopyBuffer(P^.Position.X, P^.Position.Y, Deck.FCardWidth, Deck.FCardHeight);
  end;
end;
{$WARNINGS ON}
procedure TCardTable.PickUpAllMarkers(Deck: TCardDeck = nil);
//Pick up all cardmarkers from the table. Must not be covered by any cards.
var
  i: TCardMarker;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  for i := Low(TCardMarker) to High(TCardMarker) do
    if Deck.FCardMarks[i].Displayed then PickUpMarker(i, Deck);
end;

procedure TCardTable.PlaceCardMarkers;
//Protected method that redraws any CardMarkers that have been placed.
var
  i: Integer;
  j: TCardMarker;
  P: PCardMark;
  Deck: TCardDeck;
begin
  for i := 1 to 2 do
  begin
    if i = 1 then Deck := FCardDeck
    else
      if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
    for j := Low(TCardMarker) to High(TCardMarker) do
    begin
      P := @Deck.FCardMarks[j];
      if P^.Displayed then
      begin
        P^.Displayed := False;
        PlaceCardMarker(j, P^.Position.X, P^.Position.Y, Deck);
      end;
    end;
  end;
end;

procedure TCardTable.DropCardOnShade(const CardMarker: TCardMarker;
  Deck: TCardDeck = nil);
//Picks up the shade & drops dragged card at its position.
var
  CardMark: TCardMark;
begin
  if not Assigned(Deck) then Deck := FCardDeck;
  PickUpMarker(CardMarker, Deck);
  CardMark := Deck.FCardMarks[CardMarker];
  DropCard(CardMark.Position.X, CardMark.Position.Y);
end;

procedure TCardTable.ClearTable;
//Picks up & resets internal deck & external deck referenced by ExternalDeck
//property then redraws the table.
begin
  PickUpAllCardsAndReset;
  if Assigned(FExternalDeck) then PickUpAllCardsAndReset(FExternalDeck);
  if BackgroundPicture = 'None' then
    SetColor(FColor)
  else
    LoadTableTop(BackgroundPicture);
end;

procedure TCardTable.SelectBackgroundPicture;
//Selects the background picture of the table via a form.
var
  Picture: TFileName;
begin
  with TOpenPictureDialog.Create(Application) do
    try
      Filter := 'Picture Files (*.bmp *.jpg)|*.BMP;*.JPG';
      if Execute then Picture := FileName;
    finally
      Free;
    end;
    Application.ProcessMessages;
    if Picture <> '' then BackgroundPicture := Picture;
end;

function TCardTable.DragCardWithinBounds(const Bounds: TRect): Boolean;
//Returns True if any part of a dragged card is within the Bounds given.
var
  ACard: TCard;
begin
  if FCardDragging then
  begin
    ACard := FDraggingCard;
    if (ACard.Fx + ACard.FFace.Width >= Bounds.Left) and
      (ACard.Fx <= Bounds.Right) and (ACard.Fy + ACard.FFace.Height >=
      Bounds.Top) and (ACard.Fy <= Bounds.Bottom) then
      Result := True
    else
      Result := False;
  end
  else Result := False;
end;

procedure TCardTable.ResizeTable(const Left, Top, Width, Height: Integer);
//Use this method to resize the table NOT the Width & Height properties.
var
  R: TRect;
begin
  if FCopyToScreen then
  begin
    if (Self.Width = Width) and (Self.Height = Height) then
//If not resizing its best just to use Top + Left.
    begin
      Self.Top := Top;
      Self.Left := Left;
    end
    else begin
      R.Left := Left;
      R.Top := Top;
      R.Right := Left + Width;
      R.Bottom := Top + Height;
      UpdateBoundsRect(R);
      if FStretchBackground and (FFileName <> 'None') then
      begin
        FResizedByPicture := True;
        LoadTableTop(BackgroundPicture);
      end
      else
        SetColor(FColor);
      Parent.Repaint;;
    end;
  end;
end;

procedure TCardTable.SetJpegPerformance(const Performance: TJPEGPerformance);
//Property method setting jpeg perfomance of backgroung jpeg if used.
begin
  FJpeg.Performance := Performance;
  FJpegPerformance := Performance;
end;

function TCardTable.GetCardOnTable(I: Integer): TCard;
//Surfaces the FCardsOnTable array.
begin
  if I <= FNoOfCardsOnTable then
    result := FCardsOnTable[I]
  else
    result := nil;
end;

procedure TCardTable.DeckOrBackChanged;
//As the deck is an object independent of the table changes made to the deck are
//not automatically reflected in cards already displayed on the table. So when
//changes to deck objects are made and cards owned by the deck are already on
//the table you need to call this function to reflect the changes on the table.
begin
  if BackgroundPicture = 'None' then
    SetColor(FColor)
  else
    LoadTableTop(BackgroundPicture);
end;

procedure TCardTable.RefreshLabels;
//If RepaintLabels is set to true then CardTable will repaint any light controls
//as needed. Note however also repaints controls not on the table as well.
var
  i: Integer;
  Temp: TComponent;
begin
  if FCopyToScreen and FRefreshLabels then
    with Owner do
      for I := ComponentCount - 1 downto 0 do
      begin
        Temp := Components[I];
        if (Temp is TGraphicControl) and not (Temp is TCardTable) and
          (Temp as TGraphicControl).Visible then
          (Temp as TGraphicControl).Repaint;
      end;
end;

procedure TCardTable.SaveStatus;
//Saves runtime changes made to the table such as background & card speed. Note
//only the table not the owner form.
begin
  WriteComponentResFile(FDirectory + '\Status.Dat', self);
end;

procedure TCardTable.LoadStatus;
//Reload the changes made to the table.
begin
  FCopyToScreen := False;
  if FileExists(FDirectory + '\Status.Dat') then
    ReadComponentResFile(FDirectory + '\Status.dat', self);
  FCopyToScreen := True;
  DeckOrBackChanged;
end;

//Following methods are used in the saving & loading of the table + cards.

procedure TCardTable.SaveTable(const FileName: TFileName);
//Saves Table and cards on the table (inc ExternalDeck)to the program directory.
//Note no other data eg control values or owner form data.
begin
  FSaveTable := True;
  WriteComponentResFile(FDirectory + '\' + FileName, self);
  FSaveTable := False;
end;

procedure TCardTable.LoadTable(const FileName: TFileName);
//Loads Table state & deck (internal and ExternalDeck property) status from the
//program directory. Thus you can load the table & cards and (with a bit of
//effort) 'should be' able to recreate  a game that was saved whilst cards were
//on the table. One way would be (external to CardTable) to save the cards
//values/suits as an integer (see WriteCOT and ReadCOT methods) & then on
//loading the table you could relate the card`s value & suit to its saved object
//in the CardsOnTable array. Another way would be to try and recreate the game
//thru the position of the cards in the CardsOnTable array and their relative
//positions on the table. NOTE that if a table is saved when ExternalDeck
//property is assigned that when loading again ExternalDeck MUST first point to
//an instance of TCardDeck. Likewise when loading a table saved without an
//external deck that ExternalDeck MUST first be nil.
var
  i, j, k, l, CardBack: Integer;
  Deck: TCardDeck;
begin
  FCopyToScreen := False;//Stop intermediate screen redraws.
  FSaveTable := True;//Needs to be set.
  if FileExists(FDirectory + '\' + FileName) then
    ReadComponentResFile(FDirectory + '\' + FileName, self);
  FCopyToScreen := True;
  FSaveTable := False;
//Now set the discards array if used. As discards cannot be used the order of
//cards within the array is not important so we dont need to save and load them.
//However if deck offset is used to give a 3d look we do need to place them in
//the array in order of their X values so that the pile can be displayed ok.
  FNoOfDiscards := 0;
  for l := 1 to 2 do
  begin
    if l = 1 then Deck := FCardDeck
    else
      if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
    for i := 1 to NoOfCards do
      if Deck.FCardArray[i].FStatus = -1 then
      begin
        inc(FNoOfDiscards);
        FDiscardArray[FNoOfDiscards] := nil;//Erasing any old values.
        for j := 1 to FNoOfDiscards do
          if not Assigned(FDiscardArray[j]) then
            FDiscardArray[j] := Deck.FCardArray[i]
          else
            if Deck.FCardArray[i].Fx > FDiscardArray[j].Fx then
            begin
              for k := FNoOfDiscards downto j+1 do
                FDiscardArray[k] := FDiscardArray[k-1];
              FDiscardArray[j] := Deck.FCardArray[i];
              break;
            end;
    end;
  end;
  if FNoOfDiscards <> 0 then
  begin
    FDiscardsAt.X := FDiscardArray[1].Fx;
    FDiscardsAt.Y := FDiscardArray[1].Fy;
  end;
//Now we need to change the internal deck's card face bitmaps. But it may mess
//up the card card back so.
  CardBack := FCardDeck.CardBack;
  FCardDeck.LoadDeck;
//LoadDeck sets the cardback to 1.
  if CardBack <> 1 then FCardDeck.CardBack := CardBack;
  DeckOrBackChanged;
end;

procedure TCardTable.DefineProperties(Filer: TFiler);
//Protected method that overrides method used in the streaming system. This
//allows unpublished cardtable, carddeck and card properties to be saved and
//loaded from files.
begin
  inherited;
  if not (csDesigning in ComponentState) then
  begin
    Filer.DefineProperty('cv', ReadCardValue, WriteCardValue, True);
    Filer.DefineProperty('cs', ReadCardSuit, WriteCardSuit, True);
    Filer.DefineProperty('fu', ReadFaceUp, WriteFaceUp, True);
    Filer.DefineProperty('st', ReadStatus, WriteStatus, True);
    Filer.DefineProperty('di', ReadDisplayed, WriteDisplayed, True);
    Filer.DefineProperty('cx', ReadX, WriteX, True);
    Filer.DefineProperty('cy', ReadY, WriteY, True);
    Filer.DefineProperty('ft', ReadFT, WriteFT, True);
    Filer.DefineProperty('nocot', ReadNoOfCOT, WriteNoOfCOT, True);
    Filer.DefineProperty('cot', ReadCOT, WriteCOT, True);
    Filer.DefineProperty('cms', ReadCardMarks, WriteCardMarks, True);
    if Assigned(FExternalDeck) then
    begin
      Filer.DefineProperty('d2name', ReadD2Name, WriteD2Name, True);
      Filer.DefineProperty('d2back', ReadD2Back, WriteD2Back, True);
    end;
  end;
end;

//Following methods write/read values from stream.

procedure TCardTable.WriteCardValue(Writer: TWriter);
var
  i, j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteString(GetEnumName(TypeInfo(TCardValue),
          Ord(Deck.FCardArray[i].Value)));
    end;
end;

procedure TCardTable.ReadCardValue(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].Value := TCardValue(GetEnumValue(
          TypeInfo(TCardValue), Reader.ReadString));
    end;
end;

procedure TCardTable.WriteCardSuit(Writer: TWriter);
var
  i, j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteString(GetEnumName(TypeInfo(TCardSuit),
          Ord(Deck.FCardArray[i].Suit)));
    end;
end;

procedure TCardTable.ReadCardSuit(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].Suit := TCardSuit(GetEnumValue(TypeInfo(TCardSuit),
          Reader.ReadString));
    end;
end;

procedure TCardTable.WriteFaceUp(Writer: TWriter);
var
  i, j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteBoolean(Deck.FCardArray[i].FFaceUp);
    end;
end;

procedure TCardTable.ReadFaceUp(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].FFaceUp := Reader.ReadBoolean;
    end;
end;

procedure TCardTable.WriteStatus(Writer: TWriter);
var
  i, j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteInteger(Deck.FCardArray[i].FStatus);
    end;
end;

procedure TCardTable.ReadStatus(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else                  
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].FStatus := Reader.ReadInteger;
    end;
end;

procedure TCardTable.WriteDisplayed(Writer: TWriter);
var
  i, j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
    if j = 1 then Deck := FCardDeck
    else
      if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteBoolean(Deck.FCardArray[i].FDisplayed);
    end;
end;

procedure TCardTable.ReadDisplayed(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].FDisplayed := Reader.ReadBoolean;
    end;
end;

procedure TCardTable.WriteX(Writer: TWriter);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteInteger(Deck.FCardArray[i].Fx);
    end;
end;

procedure TCardTable.ReadX(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].Fx := Reader.ReadInteger;
    end;
end;

procedure TCardTable.WriteY(Writer: TWriter);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Writer.WriteInteger(Deck.FCardArray[i].Fy);
    end;
end;

procedure TCardTable.ReadY(Reader: TReader);
var
  i, j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := 1 to NoOfCards do
        Deck.FCardArray[i].Fy := Reader.ReadInteger;
    end;
end;

procedure TCardTable.WriteFT(Writer: TWriter);
var
  j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      Writer.WriteInteger(Deck.FTop);
    end;
end;

procedure TCardTable.ReadFT(Reader: TReader);
var
  j: integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      Deck.FTop := Reader.ReadInteger;
    end;
end;

procedure TCardTable.WriteNoOfCOT(Writer: TWriter);
begin
  if FSaveTable then
    Writer.WriteInteger(FNoOfCardsOnTable);
end;

procedure TCardTable.ReadNoOfCOT(Reader: TReader);
begin
  if FSaveTable then
    FNoOfCardsOnTable := Reader.ReadInteger;
end;

procedure TCardTable.WriteCOT(Writer: TWriter);
//Writes an integer identifying the card & deck.
var
  i, a: Integer;
begin
  if FSaveTable then
    for i := 1 to NoOfCardsOnTable do
      with FCardsOnTable[i] do
      begin
        if Owner = FCardDeck then a := 0
        else
          a := NoOfCards;
        Writer.WriteInteger(a + ord(FValue) + ord(FSuit) * 13);
      end;
end;

procedure TCardTable.ReadCOT(Reader: TReader);
var
  i,j: integer;
  V: TCardValue;
  S: TCardSuit;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for i := 1 to NoOfCardsOnTable do
    begin
      j := Reader.ReadInteger;
      if j >= NoOfCards then
      begin
        j := j - NoOfCards;
        Deck := FExternalDeck;
      end
      else
        Deck := FCardDeck;
      V := TCardValue(j - (j div 13) * 13);
      S := TCardSuit(j div 13);
//Find the card with these values.
      for j := 1 to Deck.FTop do
        with Deck.FCardArray[j] do
          if (Value = V) and (Suit = S) then
          begin
            FCardsOnTable[i] := Deck.FCardArray[j];
            break;
          end;
    end;
end;

procedure TCardTable.WriteCardMarks(Writer: TWriter);
var
  i: TCardMarker;
  j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := Low(TCardMarker) to High(TCardMarker) do
      begin
        Writer.WriteBoolean(Deck.FCardMarks[i].Displayed);
        Writer.WriteInteger(Deck.FCardMarks[i].Position.X);
        Writer.WriteInteger(Deck.FCardMarks[i].Position.Y);
      end;
    end;
end;

procedure TCardTable.ReadCardMarks(Reader: TReader);
var
  i: TCardMarker;
  j: Integer;
  Deck: TCardDeck;
begin
  if FSaveTable then
    for j := 1 to 2 do
    begin
      if j = 1 then Deck := FCardDeck
      else
        if not Assigned(FExternalDeck) then break else Deck := FExternalDeck;
      for i := Low(TCardMarker) to High(TCardMarker) do
      begin
        Deck.FCardMarks[i].Displayed := Reader.ReadBoolean;
        Deck.FCardMarks[i].Position.X := Reader.ReadInteger;
        Deck.FCardMarks[i].Position.Y := Reader.ReadInteger;
      end;
    end;
end;

procedure TCardTable.WriteD2Name(Writer: TWriter);
begin
  Writer.WriteString(FExternalDeck.DeckName);
end;

procedure TCardTable.ReadD2Name(Reader: TReader);
begin
  FExternalDeck.DeckName := Reader.ReadString;
end;

procedure TCardTable.WriteD2Back(Writer: TWriter);
begin
  Writer.WriteInteger(FExternalDeck.CardBack);
end;

procedure TCardTable.ReadD2Back(Reader: TReader);
begin
  FExternalDeck.CardBack := Reader.ReadInteger;
end;





initialization

Randomize;


end.
