{********************************************************************************************************}
{                                                                                                        }
{ Carbonsoft cxDrive Disk Interrogation Toolkit Release 1                                                }
{                                                                                                        }
{ Explorer Demo                                                                                          }
{                                                                                                        }
{ Copyright  2003 Kev French, Carbonsoft. All rights reserved.                                          }
{ http://www.carbonsoft.com/cxdrive/                                                                     }
{                                                                                                        }
{ IMPORTANT INFORMATION                                                                                  }
{ You (individual or organisation) are free to use this utility and any example source code              }
{ provided by Carbonsoft for any purpose.                                                                }
{                                                                                                        }
{ Use of this utility and any example source code is entirely at your own risk.                          }
{                                                                                                        }
{ For more information please contact licensing@carbonsoft.com                                           }
{                                                                                                        }
{********************************************************************************************************}
unit ExplorerWindow;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, cxDrive10, StdCtrls, ComCtrls, ExtCtrls, ShellApi;

type
  IListviewUpdate = interface
    ['{8A13CA39-D726-46DA-9265-B39C97119515}']
  end;
  TListviewUpdate = class(TInterfacedObject, IListviewUpdate)
  private
    fCursor: TCursor;
    fListView: TListView;
  public
    constructor Create(const AListView: TListView; Clear: Boolean);
    destructor Destroy; override;
  end;

  TExplorerDialog = class(TForm)
    ExplorerPages: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    Cmd_Close: TButton;
    lTitle: TLabel;
    lRelease: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Image2: TImage;
    lvDrives: TListView;
    Cmd_Refresh: TButton;
    Cmd_Properties: TButton;
    Cmd_Features: TButton;
    procedure FormCreate(Sender: TObject);
    procedure Cmd_PropertiesClick(Sender: TObject);
    procedure Cmd_CloseClick(Sender: TObject);
    procedure Cmd_FeaturesClick(Sender: TObject);
  private
    { Private declarations }
    procedure SetupUi;
    procedure RefreshDriveInfo;
  public
    { Public declarations }
  end;

var
  ExplorerDialog: TExplorerDialog;

implementation

uses FeaturesWindow;

{$IFDEF VER150}
  {$R WindowsXp.res}
{$ENDIF VER150}

{$R *.dfm}

procedure TExplorerDialog.FormCreate(Sender: TObject);
begin
  SetupUi;
  RefreshDriveInfo;
end;

{ TListviewUpdate }

constructor TListviewUpdate.Create(const AListView: TListView;
  Clear: Boolean);
begin
  inherited Create;

  fCursor := Screen.Cursor;
  fListview := AListView;

  Screen.Cursor := crHourglass;
  fListView.Items.BeginUpdate;

  if Clear then
    fListView.Items.Clear;
end;

destructor TListviewUpdate.Destroy;
begin
  fListView.Items.EndUpdate;
  Screen.Cursor := fCursor;
  inherited;
end;

procedure TExplorerDialog.RefreshDriveInfo;
var
  i: Integer;
  lvUpdate: IListviewUpdate;
begin
  lvUpdate := TListviewUpdate.Create(lvDrives, True);
  for i := 2 to 25 do
  begin
    if cxDrive[i].Available.AsBoolean then
      with lvDrives.Items.Add do
      begin
        StateIndex := i;
        Caption := cxDrive[i].Shell.DisplayName;
        SubItems.Add(cxDrive[i].FileSystem);
        SubItems.Add(cxDrive[i].Space.BytesTotal.Formatted);
        SubItems.Add(cxDrive[i].Space.BytesUsed.Formatted);
        SubItems.Add(cxDrive[i].Space.BytesAvailable.Formatted);
      end;
  end;
end;

procedure TExplorerDialog.SetupUi;
begin
  if Screen.Fonts.IndexOf('Tahoma') >= 0 then
    Font.Name := 'Tahoma'
  else
    Font.Name := 'Arial';

  lTitle.Font.Name := Font.Name;
  lRelease.Caption := 'Release ' + cxDrive.Version;

  ExplorerPages.ActivePageIndex := 0;
end;

procedure TExplorerDialog.Cmd_PropertiesClick(Sender: TObject);
begin
  if lvDrives.Selected <> nil then
    cxDrive[lvDrives.Selected.StateIndex].ShowShellDialog;
end;

procedure TExplorerDialog.Cmd_CloseClick(Sender: TObject);
begin
  Close;
end;

procedure TExplorerDialog.Cmd_FeaturesClick(Sender: TObject);
begin
  if lvDrives.Selected <> nil then
    ShowDriveFeatures(Self,
                      cxDrive[lvDrives.Selected.StateIndex].Shell.DisplayName,
                      cxDrive[lvDrives.Selected.StateIndex].Features);
end;

end.
