unit EZCoarsL;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,EZConstL, EZMapValuesL, math;

type
  TEZCoarsL = class(TComponent)
  private
    FDistance     : TDisEv;
    FEndpoint     : TEndpoint;
    FErrorEvent   : TErrorEvent;
    FFlight       : TFlight;
    FEllipsoid    : integer;
  protected
    function  AcosF(X:Double):Double;
    function  AdjustLon(Lon:Double):Double;
    procedure CDistance(Distance,Coars : Double);
    procedure CEndpoint(Lon2,Lat2 : Double);
    Procedure ErrorLog(F : integer);
    procedure ErrorEvent(Sender: TObject; Return:integer);
  public
    Flag          : TFlag;
    ResultValues  : TCoars;
    procedure CalcDistance(Lon1,Lat1,Lon2,Lat2:Double);
    procedure CalcEndpoint(Lon1,Lat1,Coars,Distance:Double);
    procedure CalcFlightCoars(GroundCoars,FlightSpeed,WindSpeed,WindDirection:Double);
    procedure GetEllisoid(Str:TStrings);
  published
    property  Ellips: integer            read FEllipsoid       write FEllipsoid;
    property  OnError: TErrorEvent       read FErrorEvent      write FErrorEvent;
    property  OnCalcDistance: TDisEv     read FDistance        write FDistance;
    property  OnCalcEndpoint: TEndpoint  read FEndpoint        write FEndpoint;
    property  OnCalcFlight : TFlight     read FFlight          write FFlight;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('EZJ', [TEZCoarsL]);
end;
////////////////////////////////////////////////////////////////////////////////
procedure TEZCoarsL.ErrorLog(F : integer);
begin
  Flag := F;
  if Flag < 10000 then ErrorEvent(Self,Flag);                                  // 10000 might be used as a temporary value. This meens the function/procedure using it is not finished -> not error event to prevent faults
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.ErrorEvent(Sender: TObject; Return:integer);
begin
  if Assigned(FErrorEvent) then FErrorEvent(Self,Flag);
end;
//------------------------------------------------------------------------------
function TEZCoarsL.ACosF(X:Double):Double;
begin
if (abs(x) >1) then x := x/abs(x);
result := Arccos(x);
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.CDistance(Distance,Coars : Double);
begin
  if Assigned(FDistance) then FDistance(Distance,Coars);
end;
//------------------------------------------------------------------------------
function TEZCoarsL.AdjustLon(Lon:Double):Double;
var
SX,CX:Double;
begin
  SX := Sin(lon);
  CX := Cos(lon);
  if (CX < 0) and (SX > 0) then lon := Pi - ArcSin(SX);
  if (CX < 0) and (SX < 0) then lon := -Pi - ArcSin(SX);
  if (CX > 0) and (SX > 0) then lon := ArcSin(SX);
  if (CX > 0) and (SX < 0) then lon := ArcSin(SX);
  result := Lon;
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.CalcDistance(Lon1,Lat1,Lon2,Lat2:Double);
var
r,tu1,tu2,cu1,su1,cu2,s1,b1,f1,x,sx,cx,sy,cy,y,sa,c2a,cz,e,c,d,
faz,baz,s,a,f,argacos,crs12:Double;
iter :integer;
const
MAXITER=100;
begin
iter := 1;
lat1 := DegToRad(lat1);
lat2 := DegToRad(lat2);
lon1 := DegToRad(lon1);
lon2 := DegToRad(lon2);
sy := 0;
cy := 0;
cz := 0;
c2a:= 0;
e  := 0;
y := 0;

 if (FEllipsoid = MaxEllipsoid) and not ((lat1+lat2 = 0) and (abs(lon1-lon2) = Pi) and (abs(lat1) <> (Pi/180)*90)) then begin
    d := Arccos(sin(lat1)*sin(lat2)+cos(lat1)*cos(lat2)*cos(lon1-lon2));
    if ((d=0) or (lat1=-(PI/180)*90)) then crs12 := 2*Pi
    else if (lat1 = (PI/180)*90) then crs12 := Pi
    else begin
       argacos := (sin(lat2)-sin(lat1)*cos(d))/(sin(d)*cos(lat1));
       if (sin(lon2-lon1) < 0) then crs12 := AcosF(argacos)
       else crs12 := 2*Pi-AcosF(argacos);
    end;
    ErrorLog(0);
    ResultValues.Distance := d*Ellipsoid[Ellips,0];
    ResultValues.Coars    := RadToDeg(crs12);
    CDistance(d*Ellipsoid[Ellips,0],RadToDeg(crs12));
 end else if not ((lon1 = lon2) or (abs(abs(lon1-lon2)-2*PI) < 0.000000001)) then begin
    a := Ellipsoid[Ellips,0];
    f := 1/Ellipsoid[Ellips,1];
    r   := 1 - f;
    tu1 := r * tan (lat1);
    tu2 := r * tan (lat2);
    cu1 := 1 / sqrt (1 + tu1 * tu1);
    su1 := cu1 * tu1;
    cu2 := 1 / sqrt (1 + tu2 * tu2);
    s1  := cu1 * cu2;
    b1  := s1 * tu2;
    f1  := b1 * tu1;
    x   := lon2 - lon1;
    d   := x + 1; // force one pass

    while ((abs(d - x) > 0.000000001) and (iter < MAXITER)) do begin
        iter  := iter+1;
        sx    := sin(x);
        cx    := Cos(x);
        tu1   := cu2 * sx;
        tu2   := b1 - su1 * cu2 * cx;
        sy    := sqrt(tu1 * tu1 + tu2 * tu2);
        cy    := s1 * cx + f1;
        y     := Arctan2(sy, cy);
        sa    := s1 * sx / sy;
        c2a   := 1 - sa * sa;
        cz    := f1 + f1;
        if (c2a > 0) then cz := cy - cz / c2a;
        e     := cz * cz * 2 - 1;
        c     := ((-3 * c2a + 4) * f + 4) * c2a * f / 16;
        d     := x;
        x     := ((e * cy * c + cz) * sy * c + y) * sa;
        x     := (1 - c) * x * f + lon2 - lon1;
     end;
     if Iter = MAXITER then ErrorLog(11) else begin
        faz := AdjustLon(Arctan2(tu1, tu2));
        x   := sqrt ((1 / (r * r) - 1) * c2a + 1);
        x   := X + 1;
        x   := (x - 2) / x;
        c   := 1 - x;
        c   := (x * x / 4 + 1) / c;
        d   := (0.375 * x * x - 1) * x;
        x   := e * cy;
        s   := ((((sy*sy*4-3)*(1-e-e)*cz*d/6-x)*d/4+cz)*sy*d+y)*c*a*r;
        ResultValues.Distance := s;
        ResultValues.Coars    := RadToDeg(faz);
        CDistance(s,RadToDeg(faz));
        ErrorLog(0);
      end;
 end else ErrorLog(10);
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.CEndpoint(Lon2,Lat2 : Double);
begin
  if Assigned(FEndpoint) then Fendpoint(Lon2,Lat2);
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.CalcEndpoint(Lon1,Lat1,Coars,Distance:Double);
var
dlon,lat,lon,d12,crs12:Double;
r,tu,sf,cf,b,cu,su,sa,c2a,x,c,d,y,sy,cy,cz,e,lat2,lon2,faz,f,s,a:Double;
begin
Lon1 := DegToRad(Lon1);
Lat1 := DegToRad(Lat1);
crs12:= DegToRad(-Coars);
d12  := Distance/(Ellipsoid[Ellips,0]);
s    := Distance;
faz  := DegToRad(Coars);
sy := 0;
Cz := 0;
e  := 0;
cy := 0;

  if (FEllipsoid = MaxEllipsoid) and not ((abs(cos(lat1))< 0.000000001) and (abs(sin(crs12))<0.000000001)) then begin
    lat := Arcsin(sin(lat1)*cos(d12)+cos(lat1)*sin(d12)*cos(crs12));
    if (abs(cos(lat))< 0.000000001) then lon := 0 //endpoint a pole
    else begin
      dlon := Arctan2(sin(crs12)*sin(d12)*cos(lat1),cos(d12)-sin(lat1)*sin(lat));
      lon := AdjustLon((lon1-dlon+Pi)-Pi);
    end;
    ErrorLog(0);
    ResultValues.Lon := RadToDeg(lon);
    ResultValues.Lat := RadToDeg(lat);
    CEndpoint(RadToDeg(lon),RadToDeg(lat));
   end else if not ((abs(cos(lat1))< 0.000000001) and (abs(sin(crs12))<0.000000001)) then begin
    a   := Ellipsoid[Ellips,0];
    f   := 1/Ellipsoid[Ellips,1];
    r   := 1 - f;
    tu  := r * tan (lat1);
    sf  := sin(faz);
    cf  := cos(faz);
    if (cf=0) then b := 0
    else b := 2*Arctan2(tu,cf);

    cu  := 1 / sqrt(1 + tu * tu);
    su  := tu * cu;
    sa  := cu * sf;
    c2a := 1 - sa * sa;
    x   := 1 + sqrt(1 + c2a * (1 / (r * r) - 1));
    x   := (x - 2) / x;
    c   := 1 - x;
    c   := (x * x / 4 + 1) / c;
    d   := (0.375 * x * x - 1) * x;
    tu  := s / (r * a * c);
    y   := tu;
    c   := y + 1;
  while (abs (y - c) > 1E-8) do begin
      sy := sin(y);
      cy := cos(y);
      cz := cos(b + y);
      e  := 2 * cz * cz - 1;
      c  := y;
      x  := e * cy;
      y  := e + e - 1;
      y  := (((sy * sy * 4 - 3) * y * cz * d / 6 + x) *
              d / 4 - cz) * sy * d + tu;
   end;

    b    := cu * cy * cf - su * sy;
    c    := r * sqrt(sa * sa + b * b);
    d    := su * cy + cu * sy * cf;
    lat2 := ArcSin(Sin(Arctan2 (d, c)));
    c    := cu * cy - su * sy * cf;
    x    := Arctan2(sy * sf, c);
    c    := ((-3 * c2a + 4) * f + 4) * c2a * f / 16;
    d    := ((e * cy * c + cz) * sy * c + y) * sa;
    lon2 := AdjustLon(lon1 + x - (1 - c) * d * f);	// fix date line problems
    ResultValues.Lon := RadToDeg(lon2);
    ResultValues.Lat := RadToDeg(lat2);
    CEndpoint(RadToDeg(lon2),RadToDeg(lat2));
  end else ErrorLog(12);
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.GetEllisoid(Str:TStrings);
var
i : integer;
begin
Str.Clear;
for i := 0 to MaxEllipsoid - 1 do begin
    Str.Add(EllipsoidName[i]);
end;
end;
//------------------------------------------------------------------------------
procedure TEZCoarsL.CalcFlightCoars(GroundCoars,FlightSpeed,WindSpeed,WindDirection:Double);
var
S,W,WS,GC,FD,GS:Double;
begin
    S   := FlightSpeed;
    W   := (Round(WindDirection) + 180) mod 360;
    WS  := WindSpeed;
    GC  := GroundCoars;
  if (WS > S) then ErrorLog(13)
  else if (S > 0.01) then begin
    FD  := GC+RadToDeg(ArcSin(WS/S)*Sin(DegToRad(GC - W)));
    GS  := WS*Cos(DegToRad(GC - W)) + S*cos(DegToRad(GC - FD));
    ResultValues.FlightDir := FD;
    ResultValues.GroudSpd  := GS;
    if Assigned(FFlight) then FFlight(FD,GS);
    ErrorLog(0);
  end else ErrorLog(14);
end;
//------------------------------------------------------------------------------
end.

