unit EZConvertL;

interface

uses
  SysUtils, Classes, Graphics, Controls, Math, EZConstL, EZMapValuesL;

type
  TEZConvertL = class(TComponent)
  private
    FErrorEvent : TErrorEvent;
    FGetValues  : TGetValues;
    FFormat     : TFormat;
    FAlternative: boolean;
    Zone        : TZone;
    procedure SetZone(Value : TZone);
    function  GetZone(Zone : string):integer;
    procedure CalcGeoToUTM(Theta,Phi,FE,FN,LongOr,Scale,es,eccSquared:Double);
    procedure CalcUTMtoGeo(East,North,FE,FN,LongOr,Scale,es,eccSquared:Double);
    Procedure ErrorLog(F : integer);
    procedure ErrorEvent(Sender: TObject; Return:integer);
    procedure Translate(FromWGS84:boolean;Longitude,Latitude,dx,dy,dz,Ellips,Flat:Double;datumInd:integer); overload;
    // Alternative conversions
    procedure NLToRD(l,f : Double);
    procedure RDToNL(X, Y : Double);
    procedure NLToLam(Theta,Phi : Double);
    procedure LamToNL(X,Y : Double);
  protected
  public
    FloatPoint  : TFloatPoint;
    WGSPoint    : TFloatPoint;
    StringPoint : TStringPoint;
    Flag        : TFlag;
    UserDat     : TUser;
    procedure AddUserEllips(Radius,Flat,Scale:Double;UserZone:TZone;Name:string);
    Procedure AddUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;UserZone:TZone;Name:string);
    Procedure AddUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,Scale,CentralLong:Double;Name:string);
    function  CalcZone(Theta,Phi:Double):TZone;
    function  CheckTxt(Txt:string):string;
    function  ConvertDistance(Distance:Double;From,ToDistance:TDistance):Double;
    function  ConvertSpeed(Speed:Double;From,ToSpeed:TSpeed):Double;
    Procedure DatumUTMtoGeo(East,North:Double;DatumID:integer);
    procedure DeleteUserEllips(Index:integer);
    procedure DeleteUserDatum(Index:integer);
    procedure DeleteUserNatGrid(Index:integer);
    procedure FromTo(InputX,InputY:double;FromType,ToType,FromIndex,ToIndex:integer);
    procedure GetDatumUTM(Str:TStrings);
    procedure GetNatGrid(Str:TStrings);
    procedure GetUTM(Str:TStrings);
    procedure GetValues(TypeGrid,Index:integer);
    Procedure GeoToDatumUTM(Theta,Phi:Double;DatumID:integer);
    Procedure GeoToNationalGrid(Theta,Phi:Double;GridID:integer);
    Procedure GeoToTxt(Theta,Phi : Double);
    Procedure GeoToUTM(Theta,Phi:Double;EllipsoidID:integer);
    procedure LoadUserDatFromFile(FileName:TFileName);
    procedure LoadUserDatFromStream(Stream: TStream);
    Procedure NationalGridToGeo(East,North:Double;GridID:integer);
    procedure SaveUserDatToFile(FileName:TFileName);
    procedure SaveUserDatToStream(Stream: TStream);
    procedure SetUserEllipsoid(Radius,Flat,Scale:Double;UserZone:TZone;Name:string;Index:integer);
    procedure SetUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;UserZone:TZone;Name:string;Index:integer);
    procedure SetUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,Scale,CentralLong:Double;Name:string;Index:integer);
    procedure Translate(FromWGS84:boolean;Longitude,Latitude:Double;DatumID:integer); overload;
    Procedure TxtToGeo(S1,S2 : string);
    Procedure UTMtoGeo(East,North:Double;EllipsoidID:integer);
  published
    property  TxtFormat:TFormat          read FFormat          write FFormat;
    property  UTMZone:TZone              read Zone             write SetZone;
    property  Alternative:boolean        read FAlternative     write FAlternative     default false;
    property  OnGetValues:TGetValues     read FGetValues       write FGetValues;
    property  OnError: TErrorEvent       read FErrorEvent      write FErrorEvent;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('EZJ', [TEZConvertL]);
end;
//------------------------------------------------------------------------------
function TEZConvertL.ConvertDistance(Distance:Double;From,ToDistance:TDistance):Double;
var
Mid:Double;

 function GetValue(F:TDistance):Double;
 begin
  result := 1; // returs in meters
  if F = dsFathom      then result := 1.8288;{*}
  if F = dsFoot        then result := 0.3048;{*}
  if F = dsFootUS      then result := 0.3048006096;
  if F = dsInch        then result := 0.0254;{*}
  if F = dsLink        then result := 0.201168;
  if F = dsMile        then result := 1609.344;{*}
  if F = dsMileNautUK  then result := 1853.184;
  if F = dsMileNautInt then result := 1852;{*}
  if F = dsMileNautSea then result := 1828.8;{*}
  if F = dsPerch       then result := 5.0292;{*}
  if F = dsYard        then result := 0.9144;{*}
  if F = dsMeter       then result := 1;{*}
  if F = dsKilometer   then result := 0.001;{*}
 end;{* = exact}

begin
Mid    := GetValue(From)*Distance;                                              // convert first to meter
result := Mid/GetValue(ToDistance);                                             // Convert from meter
end;
//------------------------------------------------------------------------------
function TEZConvertL.ConvertSpeed(Speed:Double;From,ToSpeed:TSpeed):Double;
var
mid : Double;

 function Getvalue(F:TSpeed):Double;
 begin
  result := 1;
  if F = spKmh      then result := 1/3.6;{*}
  if F = spMs       then result := 1;{*}
  if F = spMileh    then result := 1.609344/3.6;{*}
  if F = spKnotsUK  then result := 0.5147733;
  if F = spKnots    then result := 0.514444;
 end; {* = exact}

begin

Mid    := GetValue(From)*Speed;                                                 // convert first to meters/s
Result := Mid/GetValue(ToSpeed);                                                // Convert from m/s

end;
//------------------------------------------------------------------------------
procedure TEZConvertL.ErrorLog(F : integer);
begin
  Flag := F;
  if Flag < 10000 then ErrorEvent(Self,Flag);                                   // 10000 might be used as a temporary value. This meens the function/procedure using it is not finished -> not error event to prevent faults
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.ErrorEvent(Sender: TObject; Return:integer);
begin
  if Assigned(FErrorEvent) then FErrorEvent(Self,Flag);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SetZone(Value : TZone);
const
  CharSet = ['A'..'Z'];
  DegSet  = ['0'..'6'];
  NumSet  = ['0'..'9'];
begin
  if Length(Value) <> 3 then Zone := 'N31'
  else begin
     if not (Value[1] in CharSet) then Value[1] := 'N';
     if not (Value[2] in DegSet)  then Value[2] := '3';
     if not (Value[3] in NumSet)  then Value[3] := '1';
     Zone := Value;
  end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.GetValues(TypeGrid,Index:integer);
var
I,J : integer;
begin
J := 0;
if TypeGrid = 0 then J := MaxEllipsoid-1 + UserDat.NE;
if TypeGrid = 1 then J := MaxDatum-1 + UserDat.ND;
if TypeGrid = 2 then J := MaxNationalGrid-1 + UserDat.NN;
 if Assigned(FGetValues) then begin
    if ((TypeGrid = 0) and (Index > MaxEllipsoid-1)) and (Index < J) then begin
       I := Index-MaxEllipsoid;
       FGetValues(UserDat.Ellips[I].Radius,
       UserDat.Ellips[I].Flattening,
       0{dX},0{dY},0{dZ},
       UserDat.Ellips[I].Scale,
       0{FE},0{FN},0{L},Zone);
    end else if ((TypeGrid = 1) and (Index > MaxDatum-1)) and (Index < J) then begin
       I := Index-MaxDatum;
       FGetValues(UserDat.Datum[I].Radius,
       UserDat.Datum[I].Flattening,
       UserDat.Datum[I].dX,UserDat.Datum[I].dY,UserDat.Datum[I].dZ,
       UserDat.Datum[I].Scale,
       0{FE},0{FN},0{L},Zone);
    end else if ((TypeGrid = 2) and (Index > MaxNationalGrid-1)) and (Index < J) then begin
       I := Index-MaxNationalGrid;
       FGetValues(UserDat.NatGrid[I].Radius,
       UserDat.NatGrid[I].Flattening,
       0,0,0,
       UserDat.NatGrid[I].Scale,
       UserDat.NatGrid[I].FalseEast,
       UserDat.NatGrid[I].FalseNorth,
       UserDat.NatGrid[I].CenterLon,
       '000');
    end else if (TypeGrid = 0) then FGetValues(Ellipsoid[Index,0],Ellipsoid[Index,1],0,0,0,0.9996,0,0,0,Zone)
    else if (TypeGrid = 1) then FGetValues(Ellipsoid[DatumEllips[Index],0],Ellipsoid[DatumEllips[Index],1],
                                Datum[Index,0],Datum[Index,1],Datum[Index,2],0.9996,0,0,0,zone)
    else if (TypeGrid = 2) then FGetValues(Ellipsoid[NatGridEllips[Index],0],Ellipsoid[NatGridEllips[Index],1],
                                0,0,0,NatGrid[Index,2],NatGrid[Index,0],NatGrid[Index,1],NatGrid[Index,3],'-');
 end;
end;
//------------------------------------------------------------------------------
function TEZConvertL.CheckTxt(Txt:string):string;
var
i,t : integer;
St, s : string;
c : char;
function CharIsNum(c : char) : boolean;
begin
if (c = '0') or (c = '1') or (c = '2') or (c = '3') or (c = '4') or (c = '5') or
   (c = '6') or (c = '7') or (c = '8') or (c = '9') or (c = '.') or (c = ',')
then result := true else Result := false;
end;

function CharIsSpacer(c,s: char):Char;
begin
 if (s = '.') or (s = ',') then result := c
 else result := s;
end;

begin
t := 0;
s := FloatToStr(1/2);
c := s[2];
st := '';
if Length(txt) > 0 then begin
if txt[1] = '-' then st := '-';
for i := 1 to Length(txt) do begin
     if ((txt[i] = '.') or (txt[i] = ',')) and (i = 1) then begin
        t := t + 1;
        st := '0';
     end else if ((txt[i] = '.') or (txt[i] = ',')) then t := t + 1;
     if t < 2 then begin
        txt[i] := CharIsSpacer(c,txt[i]);
        if CharIsNum(txt[i]) then st := st + txt[i];
     end;
end;
end else st := '0';
if (Length(st) = 1) and (st[1] = '-') then st := '0';
result := st;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.GeoToTxt(Theta,Phi : Double);
var
minf, minl : integer;
gradf, gradl,mf, ml, secf, secl : Double;
S : TStringPoint;
hn, he : char;
begin

gradf := trunc(abs(Theta));
mf    := (abs(Theta)-gradf)*60;
minf  := trunc(mf);
secf  := (trunc((mf-minf)*60*100))/100;

gradl := trunc(abs(Phi));
ml    := (abs(Phi)-gradl)*60;
minl  := trunc(ml);
secl  := (trunc((ml-minl)*60*100))/100;

if Theta >= 0 then S.S1 := ' '
else S.S1 := '-';

if Phi >= 0 then S.S2 := ' '
else S.S2 := '-';

if Theta >= 0 then he := 'E'
else he := 'W';

if Phi >= 0 then hn := 'N'
else hn := 'S';

//when FFormat = h* then use 'heading' else keep 'sign'
if (FFormat = hddd_dddddd) or (FFormat = hdddmm_mmmm) or (FFormat = hdddmmss_ss) then begin
  S.S1[1] := he;
  S.S2[1] := hn;
end;

if FFormat = d_dddddddd then begin                                          // Format like 5.04863244
  S.S1 := FormatFloat('0.00000000',Theta);
  S.S2 := FormatFloat('0.00000000',Phi);
end else if (FFormat = sddd_dddddd) or (FFormat = hddd_dddddd) then begin   // Format like W005,048632 or -005,048632
  S.S1 := S.S1 + FormatFloat('000.000000',abs(Theta)) + '';
  S.S2 := S.S2 + FormatFloat('000.000000',abs(Phi)) + '';
end else if (FFormat = sdddmm_mmmm) or (FFormat = hdddmm_mmmm) then begin   // Format like E00502,9179' or 00502,9179'
  S.S1 := S.S1 + FormatFloat('000',abs(gradf));
  S.S2 := S.S2 + FormatFloat('000',abs(gradl));
  S.S1 := S.S1 + '' + FormatFloat('00.0000',mf) + #039;
  S.S2 := S.S2 + '' + FormatFloat('00.0000',ml) + #039;
end else if (FFormat = sdddmmss_ss) or (FFormat = hdddmmss_ss) then begin   // Format like N00502'55,07" or 00502'55,07"
  S.S1 := S.S1 + FormatFloat('000',abs(gradf));
  S.S2 := S.S2 + FormatFloat('000',abs(gradl));
  S.S1 := S.S1 + '' + FormatFloat('00',minf) + #039;
  S.S2 := S.S2 + '' + FormatFloat('00',minl) + #039;
  S.S1 := S.S1 + FormatFloat('00.00',abs(secf)) + '"';
  S.S2 := S.S2 + FormatFloat('00.00',abs(secl)) + '"';
end;

StringPoint := S;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.TxtToGeo(S1,S2 : string);
var
i, j, t, n,sign1,sign2 : integer;
GradF, Gradl : Double;
St, s : string;
c : char;
function CharIsNum(c : char) : boolean;
begin
if (c = '0') or (c = '1') or (c = '2') or (c = '3') or (c = '4') or (c = '5') or
   (c = '6') or (c = '7') or (c = '8') or (c = '9') or (c = '.') or (c = ',')
then result := true else Result := false;
end;

function CharIsSpacer(c,s: char):Char;
begin
 if (s = '.') or (s = ',') then result := c
 else result := s;
end;

begin
gradf := 0;
gradl := 0;
sign1 := 1;
sign2 := 1;
t := 0;
n := 0;
s := FloatToStr(1/2);
c := s[2];
st := '';
if Length(S1) > 0 then begin
if (S1[1] = '-') or (S1[1] = 'W') then sign1 := -1 else sign1 := 1;
for i := 1 to Length(S1) do begin
    if S1[i] = '' then t := 1;
    if S1[i] = #39 then t := 2;
    if S1[i] = #34 then t := 3;
     if (n = t) then begin
        S1[i] := CharIsSpacer(c,S1[i]);
        if CharIsNum(S1[i]) then st := st + S1[i];
     end else begin
        if n = 0 then gradf := gradf + StrToFloat(st);
        if n = 1 then gradf := gradf + StrToFloat(st)/60;
        if n = 2 then gradf := gradf + StrToFloat(st)/3600;
        st := '';
     end;
    n := t;
end;
if (st <> '') and (n = 0) then gradf := StrToFloat(st);
end else Gradf := 0;


t := 0;
n := 0;
st := '';
if Length(S2) > 0 then begin
if (S2[1] = '-') or (S2[1] = 'S') then sign2 := -1 else sign2 := 1;
for j := 1 to Length(S2) do begin
    if S2[j] = '' then t := 1;
    if S2[j] = #39 then t := 2;
    if S2[j] = #34 then t := 3;
     if (n = t) then begin
        S2[j] := CharIsSpacer(c,S2[j]);
        if CharIsNum(S2[j]) then st := st + S2[j];
     end else begin
        if n = 0 then gradl := gradl + StrToFloat(st);
        if n = 1 then gradl := gradl + StrToFloat(st)/60;
        if n = 2 then gradl := gradl + StrToFloat(st)/3600;
        st := '';
     end;
    n := t;
end;
if (st <> '') and (n = 0) then gradl := StrToFloat(st);
end else gradl := 0;
FloatPoint.X := sign1*gradf;
FloatPoint.Y := sign2*gradl;
end;
//////////////////////////// UTM calculations  /////////////////////////////////
procedure TEZConvertL.Translate(FromWGS84:boolean;Longitude,Latitude,dx,dy,dz,Ellips,Flat:Double;datumInd:integer);
const
WGSa    = 6378137.0;
WGSinvf	= 298.257223563;
PI      = 3.14159265358979323846;
var
Phi,Theta,a0,b0,es0,f0,{a1,b1,}es1,f1,Psi,x,y,z,psi1 : Double;
begin
	Phi   := latitude* PI / 180.0;
	Theta := longitude * PI / 180.0;

    if (datumInd = DefDatum) then begin                           // do nothing if current datum is WGS84
        FloatPoint.x := Phi;
        Floatpoint.Y := Theta;
    end else begin

	if fromWGS84 then begin		       			    // convert from WGS84 to new datum */
		a0 := WGSa;					    // WGS84 semimajor axis */
		f0 := 1/WGSinvf;				    // WGS84 flattening */
	       //	a1 := Ellips;
                f1 := 1/Flat;
	end else begin                 	                            // convert from datum to WG84 */
		a0 := Ellips;	                                    // semimajor axis */
		f0 := 1/Flat;	                                    // flattening */
	      //	a1 := WGSa;		 		    // WGS84 semimajor axis */
		f1 := 1/WGSinvf;				    // WGS84 flattening */
		dx := -dx;
		dy := -dy;
		dz := -dz;
        end;

        	b0 := a0 * (1 - f0); 		        	    // semiminor axis for input datum */
        	es0 := 2 * f0 - f0*f0;			            // eccentricity^2 */
               //	b1 := a1 * (1 - f1);		        	    // semiminor axis for output datum */
        	es1 := 2 * f1 - f1*f1;			            // eccentricity^2
                                                                    // Convert geodedic latitude to geocentric latitude, psi */
	if (latitude = 0.0) or (latitude = 90.0) or (latitude = -90.0) then psi := phi
	else psi := ArcTan((1 - es0) * Tan(phi));

	if (longitude = 90.0) or (longitude = -90.0) then begin  // Calculate x and y axis coordinates with respect to the original ellipsoid */
		x := 0.0;
		y := abs(a0 * b0 / sqrt(b0*b0 + a0*a0*power(tan(psi), 2.0)));
	end else begin
		x := abs((a0 * b0) / sqrt((1 + power(tan(Theta), 2.0)) * (b0*b0 + a0*a0 * power(tan(psi), 2.0))));
		y := abs(x * tan(Theta));
        end;

	if (longitude < -90.0) or (longitude > 90.0) then x := -x;

	if (longitude < 0.0) then y := -y;

	if (latitude = 90.0) then z := b0                        //  Calculate z axis coordinate with respect to the original ellipsoid */
	else if (latitude = -90.0) then z := -b0
	else z := tan(psi) * sqrt((a0*a0 * b0*b0) / (b0*b0 + a0*a0 * power(tan(psi), 2.0)));

	psi1 := ArcTan((z - dz) / sqrt((x - dx)*(x - dx) + (y - dy)*(y - dy)));   // Calculate the geocentric latitude with respect to the new ellipsoid */
	latitude := ArcTan(tan(psi1) / (1 - es1)) * 180.0 / PI; // Convert to geocentric latitude and save return value */
	longitude := ArcTan((y - dy) / (x - dx)) * 180.0 / PI;  // Calculate the longitude with respect to the new ellipsoid */

	if (x-dx < 0.0) then begin                           // Correct the resultant for negative x values */
		if (y-dy > 0.0) then longitude := 180.0 + longitude
		else longitude := -180.0 + longitude;
        end;
    end;
FloatPoint.X := longitude;
Floatpoint.Y := latitude;
end;
//------------------------------------------------------------------------------
function TEZConvertL.CalcZone(Theta,Phi:Double):TZone;
var
Lat, Lon : Double;
ZoneNumber, LongTemp : integer;
LetterDesignator : char;
begin
Lon := Theta;
Lat := Phi;
LongTemp := Trunc(180 + Lon);
ZoneNumber    := (LongTemp div 6) + 1;
      if (84 < Lat)                    Then LetterDesignator := 'Z'
 else if((84 >= Lat) and (Lat >= 72))  Then LetterDesignator := 'X'
 else if((72 > Lat)  and (Lat >= 64))  Then LetterDesignator := 'W'
 else if((64 > Lat)  and (Lat >= 56))  Then LetterDesignator := 'V'
 else if((56 > Lat)  and (Lat >= 48))  Then LetterDesignator := 'U'
 else if((48 > Lat)  and (Lat >= 40))  Then LetterDesignator := 'T'
 else if((40 > Lat)  and (Lat >= 32))  Then LetterDesignator := 'S'
 else if((32 > Lat)  and (Lat >= 24))  Then LetterDesignator := 'R'
 else if((24 > Lat)  and (Lat >= 16))  Then LetterDesignator := 'Q'
 else if((16 > Lat)  and (Lat >= 8))   Then LetterDesignator := 'P'
 else if(( 8 > Lat)  and (Lat >= 0))   Then LetterDesignator := 'N'
 else if(( 0 > Lat)  and (Lat >= -8))  Then LetterDesignator := 'M'
 else if((-8 > Lat)  and (Lat >= -16)) Then LetterDesignator := 'L'
 else if((-16 > Lat) and (Lat >= -24)) Then LetterDesignator := 'K'
 else if((-24 > Lat) and (Lat >= -32)) Then LetterDesignator := 'J'
 else if((-32 > Lat) and (Lat >= -40)) Then LetterDesignator := 'H'
 else if((-40 > Lat) and (Lat >= -48)) Then LetterDesignator := 'G'
 else if((-48 > Lat) and (Lat >= -56)) Then LetterDesignator := 'F'
 else if((-56 > Lat) and (Lat >= -64)) Then LetterDesignator := 'E'
 else if((-64 > Lat) and (Lat >= -72)) Then LetterDesignator := 'D'
 else if((-72 > Lat) and (Lat >= -80)) Then LetterDesignator := 'C'
 else LetterDesignator := 'A';

 if ZoneNumber < 10 then Result := LetterDesignator + '0' + IntTostr(ZoneNumber)
 else Result := LetterDesignator + IntTostr(ZoneNumber);
//Zone   := Result;
end;
//------------------------------------------------------------------------------
function TEZConvertL.GetZone(Zone : string):integer;
var
i : integer;
s : string;
begin
if (zone = '') or (Length(zone) <> 3) then begin
   ErrorLog(20); result := 31;
end else begin
 s := Zone[2] + Zone[3];
 i := 0;
 case zone[1] of 'A'..'M' : i := 0;
                 'N'..'Z' : i := 1;
 end;
 i := i*100 + StrToInt(s);
 result := i;
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.CalcGeoToUTM(Theta,Phi,FE,FN,LongOr,Scale,es,eccSquared:Double);
var
Lon, Lat, LongOrigin, LongOriginRad, N, T, C, A, M,
eccPrimeSquared, UTMNorthing, UTMEasting, K0, LatRad, LongRad,LongTemp : Double;
begin
Lon := Theta;
Lat := Phi;
  K0         := Scale;
  eccSquared := 2/eccSquared  -1/sqr(eccSquared); //eccentricitySquared
  LatRad     := DegToRad(Lat);
  LongTemp := (Lon+180)-Trunc((Lon+180)/360)*360-180;
  LongRad    := DegToRad(LongTemp);
        LongOrigin      := LongOr;
	LongOriginRad   := DegToRad(LongOrigin);       
	eccPrimeSquared := (eccSquared)/(1-eccSquared);

	N := es/sqrt(1-eccSquared*sin(LatRad)*sin(LatRad));
	T := tan(LatRad)*tan(LatRad);
	C := eccPrimeSquared*cos(LatRad)*cos(LatRad);
	A := cos(LatRad)*(LongRad-LongOriginRad);

	M := es*((1 - eccSquared/4 - 3*eccSquared*eccSquared/64
                  - 5*eccSquared*eccSquared*eccSquared/256)*LatRad
                  - (3*eccSquared/8 + 3*eccSquared*eccSquared/32
                  + 45*eccSquared*eccSquared*eccSquared/1024)*sin(2*LatRad)
                  + (15*eccSquared*eccSquared/256
                  + 45*eccSquared*eccSquared*eccSquared/1024)*sin(4*LatRad)
		  - (35*eccSquared*eccSquared*eccSquared/3072)*sin(6*LatRad));

       	UTMEasting := (k0*N*(A+(1-T+C)*Power(A,3)/6
                      + (5-18*T+T*T+72*C-58*eccPrimeSquared)*Power(A,5)/120)
                      + FE);

	UTMNorthing:= (k0*(M+N*tan(LatRad)*(A*A/2+(5-T+9*C+4*C*C)*Power(A,4)/24
		      + (61-58*T+T*T+600*C-330*eccPrimeSquared)*Power(A,5)/720))) + FN;

FloatPoint.X := UTMEasting;
FloatPoint.Y := UTMNorthing;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.CalcUTMtoGeo(East,North,FE,FN,LongOr,Scale,es,eccSquared:Double);
var
N1, T1, C1, R1, D, M, X, Y,K0,eccPrimeSquared,e1, mu, phi1Rad, Lat, Long : Double;
begin
x := East - FE; //remove 500,000 meter offset for longitude
y := North - FN;

  eccSquared := 2/eccSquared  -1/sqr(eccSquared);
  e1 := (1-sqrt(1-eccSquared))/(1+sqrt(1-eccSquared));
  eccPrimeSquared := (eccSquared)/(1-eccSquared);
  K0 := Scale;

      M  := y / k0;
      mu := M/(es*(1-eccSquared/4-3*eccSquared*eccSquared/64-5*eccSquared*eccSquared*eccSquared/256));

      phi1Rad := mu + (3*e1/2-27*e1*e1*e1/32)*sin(2*mu) + (21*e1*e1/16-55*e1*e1*e1*e1/32)*sin(4*mu)
                    +(151*e1*e1*e1/96)*sin(6*mu);
      N1   := es/sqrt(1-eccSquared*sin(phi1Rad)*sin(phi1Rad));
      T1   := tan(phi1Rad)*tan(phi1Rad);
      C1   := eccPrimeSquared*cos(phi1Rad)*cos(phi1Rad);
      R1   := es*(1-eccSquared)/power(1-eccSquared*sin(phi1Rad)*sin(phi1Rad), 1.5);
      D    := x/(N1*k0);
      Lat  := phi1Rad - (N1*tan(phi1Rad)/R1)*(D*D/2-(5+3*T1+10*C1-4*C1*C1-
              9*eccPrimeSquared)*Power(D,4)/24+(61+90*T1+298*C1+45*T1*T1-
              252*eccPrimeSquared-3*C1*C1)*Power(D,5)/720);
      Lat  := RadToDeg(Lat);
      Long := (D-(1+2*T1+C1)*Power(D,3)/6+(5-2*C1+28*T1-3*C1*C1+8*eccPrimeSquared+24*T1*T1)*
               Power(D,5)/120)/cos(phi1Rad);
      Long := LongOr + RadToDeg(Long);

FloatPoint.X := Long;
FloatPoint.Y := Lat;
end;
//------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////
Procedure TEZConvertL.GeoToUTM(Theta,Phi:Double;EllipsoidID:integer);
Var
Lat,Lon,FN,LongTemp : Double;
ZoneNumber,I,J : Integer;
begin
Lon := Theta;
Lat := Phi;
I := EllipsoidID-MaxEllipsoid;
J := MaxEllipsoid -1 + UserDat.NE -1;
if (EllipsoidID < 0) or (EllipsoidID > J) then begin EllipsoidID := 0; ErrorLog(21); end;
//  Zone := CalcZone(Theta,Phi);
  ZoneNumber    := StrToInt(Zone[2] + Zone[3]);
  LongTemp := (Lon+180)-Trunc((Lon+180)/360)*360-180;
  if (Lat >= 56.0) and (Lat < 64.0) and (LongTemp >= 3.0) and (LongTemp < 12.0) then ZoneNumber := 32;
  // Special zones for Svalbard
  if (Lat >= 72.0) and (Lat < 84.0) then begin
      if (LongTemp >= 0.0) and (LongTemp <  9.0) then ZoneNumber := 31
      else if (LongTemp >= 9.0 ) and (LongTemp < 21.0) then ZoneNumber := 33
      else if (LongTemp >= 21.0) and (LongTemp < 33.0) then ZoneNumber := 35
      else if (LongTemp >= 33.0) and (LongTemp < 42.0) then ZoneNumber := 37;
  end;
  if Lat < 0 then FN := 10000000.0 else FN := 0;
   if (EllipsoidID > MaxEllipsoid-1) then CalcGeoToUTM(Theta,Phi,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,
      UserDat.Ellips[I].Scale,UserDat.Ellips[I].Radius,UserDat.Ellips[I].Flattening)
  else CalcGeoToUTM(Theta,Phi,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,0.9996,Ellipsoid[EllipsoidID,0],Ellipsoid[EllipsoidID,1]);
end;
//------------------------------------------------------------------------------
Procedure TEZConvertL.GeoToDatumUTM(Theta,Phi:Double;DatumID:integer);
Var
Lat,Lon,FN,LongTemp : Double;
ZoneNumber,I,J : Integer;
begin
I := DatumID-MaxDatum;
J := MaxDatum -1 + UserDat.ND -1;
if (DatumID < 0) or (DatumID > J) then ErrorLog(22)
else begin
Lon := Theta;
Lat := Phi;
FloatPoint.X := Theta;
FloatPoint.Y := Phi;
if (DatumID > MaxDatum-1) then Translate(False,Lon,Lat,UserDat.Datum[I].dX,UserDat.Datum[I].dY,
     UserDat.Datum[I].dZ,UserDat.Datum[I].Radius,UserDat.Datum[I].Flattening,DatumID)
else Translate(false,Lon,Lat,Datum[DatumID,0],Datum[DatumID,1],Datum[DatumID,2],
     Ellipsoid[DatumEllips[DatumID],0],Ellipsoid[DatumEllips[DatumID],1],DatumID);

Theta   := FloatPoint.X;
Phi := FloatPoint.Y;
Lon := Theta;
Lat := Phi;
//  Zone := CalcZone(Theta,Phi);
  ZoneNumber    := StrToInt(Zone[2] + Zone[3]);
  LongTemp := (Lon+180)-Trunc((Lon+180)/360)*360-180;
  if (Lat >= 56.0) and (Lat < 64.0) and (LongTemp >= 3.0) and (LongTemp < 12.0) then ZoneNumber := 32;
  // Special zones for Svalbard
  if (Lat >= 72.0) and (Lat < 84.0) then begin
      if (LongTemp >= 0.0) and (LongTemp <  9.0) then ZoneNumber := 31
      else if (LongTemp >= 9.0 ) and (LongTemp < 21.0) then ZoneNumber := 33
      else if (LongTemp >= 21.0) and (LongTemp < 33.0) then ZoneNumber := 35
      else if (LongTemp >= 33.0) and (LongTemp < 42.0) then ZoneNumber := 37;
  end;
  if Lat < 0 then FN := 10000000.0 else FN := 0;
  if (DatumID > MaxDatum-1) then CalcGeoToUTM(Theta,Phi,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,
      UserDat.Datum[I].Scale,Ellipsoid[20,0],Ellipsoid[20,1])
  else CalcGeoToUTM(Theta,Phi,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,0.9996,Ellipsoid[20,0],Ellipsoid[20,1]);
end;
end;
//------------------------------------------------------------------------------
Procedure TEZConvertL.GeoToNationalGrid(Theta,Phi:Double;GridID:integer);
var
I,J : integer;
begin
I := GridID-MaxNationalGrid;
J := MaxNationalGrid -1 + UserDat.NN -1;
if (GridID < 0) or (GridID > J) then ErrorLog(23)
else if (GridID <= MaxNationalGrid-1) then
  if (FAlternative) and (GridID = 0) then NLToRD(Theta,Phi)
  else if (FAlternative) and (GridID = 1) then NLToLam(Theta,Phi)
  else CalcGeoToUTM(Theta,Phi,NatGrid[GridID,0],NatGrid[GridID,1],
             NatGrid[GridID,3],NatGrid[GridID,2],
             Ellipsoid[NatGridEllips[GridID],0],
             Ellipsoid[NatGridEllips[GridID],1])
else
  CalcGeoToUTM(Theta,Phi,UserDat.NatGrid[I].FalseEast,UserDat.NatGrid[I].FalseNorth,
             UserDat.NatGrid[I].CenterLon,UserDat.NatGrid[I].Scale,
             UserDat.NatGrid[I].Radius,UserDat.NatGrid[I].Flattening);
end;
//------------------------------------------------------------------------------
Procedure TEZConvertL.UTMtoGeo(East,North:Double;EllipsoidID:integer);
Var
FN : Double;
ZoneNumber,NorthernHemisphere,I,J : Integer;
begin
I := EllipsoidID - MaxEllipsoid;
J := MaxEllipsoid -1 + UserDat.NE -1;
if (Zone <> '') or (Zone <> '0') then begin
if (EllipsoidID < 0) or (EllipsoidID > J) then ErrorLog(24)
else if (EllipsoidID > MaxEllipsoid-1) then begin
  NorthernHemisphere := GetZone(Zone);
  if Trunc(NorthernHemisphere/100) = 0 then FN := 10000000.0 else FN := 0;
  ZoneNumber := NorthernHemisphere - Trunc(NorthernHemisphere/100)*100;
  CalcUTMtoGeo(East,North,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,UserDat.Ellips[I].Scale,
               UserDat.Ellips[I].Radius,UserDat.Ellips[I].Flattening);
end else if (EllipsoidID < MaxEllipsoid-1) and (EllipsoidID >= 0) then begin
  NorthernHemisphere := GetZone(Zone);
  if Trunc(NorthernHemisphere/100) = 0 then FN := 10000000.0 else FN := 0;
  ZoneNumber := NorthernHemisphere - Trunc(NorthernHemisphere/100)*100;
  CalcUTMtoGeo(East,North,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,0.9996,
               Ellipsoid[EllipsoidID,0],Ellipsoid[EllipsoidID,1]);
end else ErrorLog(25);
end;
end;
//------------------------------------------------------------------------------
Procedure TEZConvertL.DatumUTMtoGeo(East,North:Double;DatumID:integer);
Var
FN : Double;
ZoneNumber,NorthernHemisphere,I,J : Integer;
begin
I := DatumID - MaxDatum;
J := MaxDatum -1 + UserDat.ND -1;
if (Zone <> '') or (Zone <> '0') then begin
if (DatumID < 0) or (DatumID > J) then ErrorLog(26)
else if (DatumID > MaxDatum-1) then begin
  NorthernHemisphere := GetZone(Zone);
  FN := 0;
  if Trunc(NorthernHemisphere/100) = 0 then FN := 10000000.0;
  ZoneNumber := NorthernHemisphere - Trunc(NorthernHemisphere/100)*100;
  CalcUTMtoGeo(East,North,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,UserDat.Datum[I].Scale,Ellipsoid[20,0],Ellipsoid[20,1]);
  Translate(True,FloatPoint.X,FloatPoint.Y,UserDat.Datum[I].dX,UserDat.Datum[I].dY,UserDat.Datum[I].dZ,
            UserDat.Datum[I].Radius,UserDat.Datum[I].Flattening,DatumID);

end else if (DatumID < MaxDatum-1) and (DatumID >= 0) then begin
  NorthernHemisphere := GetZone(Zone);
  FN := 0;
  if Trunc(NorthernHemisphere/100) = 0 then FN := 10000000.0;
  ZoneNumber := NorthernHemisphere - Trunc(NorthernHemisphere/100)*100;
  CalcUTMtoGeo(East,North,500000.0,FN,(ZoneNumber - 1)*6 - 180 + 3,0.9996,Ellipsoid[20,0],Ellipsoid[20,1]);
  Translate(True,FloatPoint.X,FloatPoint.Y,Datum[DatumID,0],Datum[DatumID,1],Datum[DatumID,2],
          Ellipsoid[DatumEllips[DatumID],0],Ellipsoid[DatumEllips[DatumID],1],DatumID);
end;
end else ErrorLog(27);
end;
//------------------------------------------------------------------------------
Procedure TEZConvertL.NationalGridToGeo(East,North:Double;GridID:integer);
var
I,J:integer;
begin
I := GridID - MaxNationalGrid;
J := MaxNationalGrid -1 + UserDat.NN -1;
if (GridID < 0) or (GridID > J) then ErrorLog(28)
else if (GridID > MaxNationalGrid-1) then begin
       CalcUTMToGeo(East,North,UserDat.NatGrid[I].FalseEast,UserDat.NatGrid[I].FalseNorth,UserDat.NatGrid[I].CenterLon,UserDat.NatGrid[I].Scale,
             UserDat.NatGrid[I].Radius,UserDat.NatGrid[I].Flattening);
end else if (GridID < MaxNationalGrid-1) and (GridID >= 0) then begin
  if (FAlternative) and (GridID = 0) then RDToNL(East,North)
  else if (FAlternative) and (GridID = 1) then LamToNL(East,North)
  else CalcUTMToGeo(East,North,NatGrid[GridID,0],NatGrid[GridID,1],
             NatGrid[GridID,3],NatGrid[GridID,2],
             Ellipsoid[NatGridEllips[GridID],0],
             Ellipsoid[NatGridEllips[GridID],1]);
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.Translate(FromWGS84:boolean;Longitude,Latitude:Double;DatumID:integer);
var
I,J:integer;
begin
I := DatumID - MaxDatum;
J := MaxDatum -1 + UserDat.ND -1;
if (DatumID > MaxDatum-1) and (DatumID <= J)then
    Translate(FromWGS84,Longitude,Latitude,UserDat.Datum[I].dX,UserDat.Datum[I].dY,UserDat.Datum[I].dZ,
          UserDat.Datum[I].Radius,UserDat.Datum[I].Flattening,DatumID)
else if (DatumID < MaxDatum-1) and (DatumID >= 0) then
    Translate(FromWGS84,Longitude,Latitude,Datum[DatumID,0],Datum[DatumID,1],Datum[DatumID,2],
          Ellipsoid[DatumEllips[DatumID],0],Ellipsoid[DatumEllips[DatumID],1],DatumID)
else begin
   FloatPoint.X := Longitude;
   FloatPoint.Y := Latitude;
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.GetUTM(Str:TStrings);
var
i,j : integer;
begin
Str.Clear;
j := UserDat.NE - 1;
for i := 0 to MaxEllipsoid - 1 do begin
    Str.Add(EllipsoidName[i]);
end;
for i := 0 to j do begin
    Str.Add(UserDat.Ellips[i].Name);
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.GetDatumUTM(Str:TStrings);
var
i,j : integer;
begin
Str.Clear;
j := UserDat.ND - 1;
for i := 0 to MaxDatum - 1 do begin
    Str.Add(DatumName[i]);
end;
for i := 0 to j do begin
    Str.Add(UserDat.Datum[i].Name);
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.GetNatGrid(Str:TStrings);
var
i,j : integer;
begin
Str.Clear;
j := UserDat.NN - 1;
for i := 0 to MaxNationalGrid - 1 do begin
    Str.Add(NatGridName[i]);
end;
for i := 0 to j do begin
    Str.Add(UserDat.NatGrid[i].Name);
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.FromTo(InputX,InputY:double;FromType,ToType,FromIndex,ToIndex:integer);
begin
{if ((FromType = ToType) and (FromIndex = ToIndex)) then begin                   // no need for conversion if both input and output are the same
   FloatPoint.X := InputX;
   FloatPoint.Y := InputY;
end else begin                  }                                                //Convert to WGS84
   if FromType = 0 then Translate(false,InputX,InputY,FromIndex);
   if FromType = 1 then UTMtoGeo(InputX,InputY,FromIndex);
   if FromType = 2 then DatumUTMtoGeo(InputX,InputY,FromIndex);
   if FromType = 3 then NationalGridToGeo(InputX,InputY,FromIndex);

   WGSPoint := FloatPoint;
                                                                                //Convert from WGS84
   if ToType   = 0 then Translate(True,FloatPoint.X,FloatPoint.Y,ToIndex);
   if ToType   = 1 then GeoToUTM(FloatPoint.X,FloatPoint.Y,ToIndex);
   if ToType   = 2 then GeoToDatumUTM(FloatPoint.X,FloatPoint.Y,ToIndex);
   if ToType   = 3 then GeoToNationalGrid(FloatPoint.X,FloatPoint.Y,ToIndex);
//end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SetUserEllipsoid(Radius,Flat,Scale:Double;UserZone:TZone;Name:string;Index:integer);
begin
if (Index < 0) or (Index > UserDat.NE -1) then ErrorLog(30)
else begin
UserDat.Ellips[Index].Name := Name;
UserDat.Ellips[Index].Radius := Radius;
UserDat.Ellips[Index].Flattening := Flat;
UserDat.Ellips[Index].Scale := Scale;
UserDat.Ellips[Index].Zone := UserZone;
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SetUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;UserZone:TZone;Name:string;Index:integer);
begin
if (Index < 0) or (Index > UserDat.ND -1) then ErrorLog(31)
else begin
UserDat.Datum[Index].Name := Name;
UserDat.Datum[Index].Radius := Radius;
UserDat.Datum[Index].Flattening := Flat;
UserDat.Datum[Index].dX := dX;
UserDat.Datum[Index].dY := dY;
UserDat.Datum[Index].dZ := dZ;
UserDat.Datum[Index].Scale := Scale;
UserDat.Datum[Index].Zone := UserZone;
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SetUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,Scale,CentralLong:Double;Name:string;Index:integer);
begin
if (Index < 0) or (Index > UserDat.NN -1) then ErrorLog(32)
else begin
UserDat.NatGrid[Index].Name := Name;
UserDat.NatGrid[Index].Radius := Radius;
UserDat.NatGrid[Index].Flattening := Flat;
UserDat.NatGrid[Index].FalseEast := FalseEasting;
UserDat.NatGrid[Index].FalseNorth := FalseNorthing;
UserDat.NatGrid[Index].Scale := Scale;
UserDat.NatGrid[Index].CenterLon := CentralLong;
end;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.AddUserEllips(Radius,Flat,Scale:Double;UserZone:TZone;Name:string);
begin
if UserDat.NE < 10 then begin
  UserDat.NE := UserDat.NE + 1;
  UserDat.Ellips[UserDat.NE-1].Name := Name;
  UserDat.Ellips[UserDat.NE-1].Radius := Radius;
  UserDat.Ellips[UserDat.NE-1].Flattening := Flat;
  UserDat.Ellips[UserDat.NE-1].Scale := Scale;
  UserDat.Ellips[UserDat.NE-1].Zone := UserZone;
  ErrorLog(0);
end else ErrorLog(33);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.AddUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;UserZone:TZone;Name:string);
begin
if UserDat.ND < 10 then begin
  UserDat.ND := UserDat.ND + 1;
  UserDat.Datum[UserDat.ND-1].Name := Name;
  UserDat.Datum[UserDat.ND-1].Radius := Radius;
  UserDat.Datum[UserDat.ND-1].Flattening := Flat;
  UserDat.Datum[UserDat.ND-1].dX := dX;
  UserDat.Datum[UserDat.ND-1].dY := dY;
  UserDat.Datum[UserDat.ND-1].dZ := dZ;
  UserDat.Datum[UserDat.ND-1].Scale := Scale;
  UserDat.Datum[UserDat.ND-1].Zone := UserZone;
  ErrorLog(0);
end else ErrorLog(34);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.AddUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,Scale,CentralLong:Double;Name:string);
begin
if UserDat.NN < 10 then begin
  UserDat.NN := UserDat.NN + 1;
  UserDat.NatGrid[UserDat.NN-1].Name := Name;
  UserDat.NatGrid[UserDat.NN-1].Radius := Radius;
  UserDat.NatGrid[UserDat.NN-1].Flattening := Flat;
  UserDat.NatGrid[UserDat.NN-1].FalseEast := FalseEasting;
  UserDat.NatGrid[UserDat.NN-1].FalseNorth := FalseNorthing;
  UserDat.NatGrid[UserDat.NN-1].Scale := Scale;
  UserDat.NatGrid[UserDat.NN-1].CenterLon := CentralLong;
  ErrorLog(0);
end else ErrorLog(35);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.DeleteUserEllips(Index:integer);
var
i : integer;
begin
if (Index < UserDat.NE) and (Index >= 0) then begin
if index < UserDat.NE -1 then begin
  for i := index to UserDat.NE  - 1 do begin
    UserDat.Ellips[I].Name       := UserDat.Ellips[I+1].Name;
    UserDat.Ellips[I].Radius     := UserDat.Ellips[I+1].Radius;
    UserDat.Ellips[I].Flattening := UserDat.Ellips[I+1].Flattening;
    UserDat.Ellips[I].Scale      := UserDat.Ellips[I+1].Scale;
    UserDat.Ellips[I].Zone       := UserDat.Ellips[I+1].Zone;
  end;
    UserDat.Ellips[UserDat.NE -1].Name       := '';
    UserDat.Ellips[UserDat.NE -1].Radius     := 0;
    UserDat.Ellips[UserDat.NE -1].Flattening := 0;
    UserDat.Ellips[UserDat.NE -1].Scale      := 0;
    UserDat.Ellips[UserDat.NE -1].Zone       := '';
  ErrorLog(0);
  UserDat.NE := UserDat.NE - 1;
end else if index = UserDat.NE -1 then begin
    UserDat.Ellips[Index].Name       := '';
    UserDat.Ellips[Index].Radius     := 0;
    UserDat.Ellips[Index].Flattening := 0;
    UserDat.Ellips[Index].Scale      := 0;
    UserDat.Ellips[Index].Zone       := '';
   UserDat.NE := UserDat.NE - 1;
   ErrorLog(0);
end;
end else ErrorLog(36);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.DeleteUserDatum(Index:integer);
var
i : integer;
begin
if (Index < UserDat.ND) and (Index >= 0) then begin
if index < UserDat.ND -1 then begin
  for i := index to UserDat.ND  - 1 do begin
    UserDat.Datum[I].Name       := UserDat.Datum[I+1].Name;
    UserDat.Datum[I].Radius     := UserDat.Datum[I+1].Radius;
    UserDat.Datum[I].Flattening := UserDat.Datum[I+1].Flattening;
    UserDat.Datum[I].dX         := UserDat.Datum[I+1].dX;
    UserDat.Datum[I].dY         := UserDat.Datum[I+1].dY;
    UserDat.Datum[I].dZ         := UserDat.Datum[I+1].dZ;
    UserDat.Datum[I].Scale      := UserDat.Datum[I+1].Scale;
    UserDat.Datum[I].Zone       := UserDat.Datum[I+1].Zone;
  end;
    UserDat.Datum[UserDat.ND -1].Name       := '';
    UserDat.Datum[UserDat.ND -1].Radius     := 0;
    UserDat.Datum[UserDat.ND -1].Flattening := 0;
    UserDat.Datum[UserDat.ND -1].dX         := 0;
    UserDat.Datum[UserDat.ND -1].dY         := 0;
    UserDat.Datum[UserDat.ND -1].dZ         := 0;
    UserDat.Datum[UserDat.ND -1].Scale      := 0;
    UserDat.Datum[UserDat.ND -1].Zone       := '';
  ErrorLog(0);
  UserDat.ND := UserDat.ND - 1;
end else if index = UserDat.ND -1 then begin
    UserDat.Datum[Index].Name       := '';
    UserDat.Datum[Index].Radius     := 0;
    UserDat.Datum[Index].Flattening := 0;
    UserDat.Datum[Index].dX         := 0;
    UserDat.Datum[Index].dY         := 0;
    UserDat.Datum[Index].dZ         := 0;
    UserDat.Datum[Index].Scale      := 0;
    UserDat.Datum[Index].Zone       := '';
   UserDat.ND := UserDat.ND - 1;
   ErrorLog(0);
end;
end else ErrorLog(37);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.DeleteUserNatGrid(Index:integer);
var
i : integer;
begin
if (Index < UserDat.ND) and (Index >= 0) then begin
if index < UserDat.ND -1 then begin
  for i := index to UserDat.ND  - 1 do begin
    UserDat.NatGrid[I].Name       := UserDat.NatGrid[I+1].Name;
    UserDat.NatGrid[I].Radius     := UserDat.NatGrid[I+1].Radius;
    UserDat.NatGrid[I].Flattening := UserDat.NatGrid[I+1].Flattening;
    UserDat.NatGrid[I].FalseEast  := UserDat.NatGrid[I+1].FalseEast ;
    UserDat.NatGrid[I].FalseNorth := UserDat.NatGrid[I+1].FalseNorth;
    UserDat.NatGrid[I].Scale      := UserDat.NatGrid[I+1].Scale;
    UserDat.NatGrid[I].CenterLon  := UserDat.NatGrid[I+1].CenterLon;
  end;
    UserDat.NatGrid[UserDat.NN-1].Name       := '';
    UserDat.NatGrid[UserDat.NN-1].Radius     := 0;
    UserDat.NatGrid[UserDat.NN-1].Flattening := 0;
    UserDat.NatGrid[UserDat.NN-1].FalseEast  := 0;
    UserDat.NatGrid[UserDat.NN-1].FalseNorth := 0;
    UserDat.NatGrid[UserDat.NN-1].Scale      := 0;
    UserDat.NatGrid[UserDat.NN-1].CenterLon  := 0;
  ErrorLog(0);
  UserDat.ND := UserDat.ND - 1;
end else if index = UserDat.ND -1 then begin
    UserDat.NatGrid[index].Name       := '';
    UserDat.NatGrid[index].Radius     := 0;
    UserDat.NatGrid[index].Flattening := 0;
    UserDat.NatGrid[index].FalseEast  := 0;
    UserDat.NatGrid[index].FalseNorth := 0;
    UserDat.NatGrid[index].Scale      := 0;
    UserDat.NatGrid[index].CenterLon  := 0;
   UserDat.ND := UserDat.ND - 1;
   ErrorLog(0);
end;
end else ErrorLog(38);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SaveUserDatToStream(Stream: TStream);
begin
Stream.Write(UserDat,SizeOf(TUser));
ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.LoadUserDatFromStream(Stream: TStream);
begin
Stream.Read(UserDat,SizeOf(TUser));
ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.SaveUserDatToFile(FileName:TFileName);
var
Stream : TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.Write(UserDat,SizeOf(TUser));
Stream.SaveToFile(FileName);
Stream.Free;
ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.LoadUserDatFromFile(FileName:TFileName);
var
Stream : TMemoryStream;
begin
Stream := TMemoryStream.Create;
Stream.LoadFromFile(FileName);
Stream.Read(UserDat,SizeOf(TUser));
Stream.Free;
ErrorLog(0);
end;
//////////////////////////   Alternative    ////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZConvertL.NLToRD(l,f : Double);
const
x0 = 155000.00;
y0 = 463000.00;

f0 = 52.15616056;
l0 = 5.38763889;

c01 = 190066.98903;  d10 = 309020.31810;
c11 = -11830.85831;  d02 =   3638.36193;
c21 =   -114.19754;  d12 =   -157.95222;
c03 =    -32.38360;  d20 =     72.97141;
c31 =     -2.34078;  d30 =     59.79734;
c13 =     -0.60639;  d22 =     -6.43481;
c23 =      0.15774;  d04 =      0.09351;
c41 =     -0.04158;  d32 =     -0.07379;
c05 =     -0.00661;  d14 =     -0.05419;
                     d40 =     -0.03444;
var
df , dl, dx, dy, x, y : Double;
begin
df := (f - F0)*0.36;
dl := (l - l0)*0.36;

dx := c01*dl + c11*df*dl + c21*power(df,2)*dl + c03*power(dl,3) +
      c31*power(df,3)*dl + c13*df*power(dl,3) + c23*power(df,2)*power(dl,3) +
      c41*power(df,4)*dl + c05*power(dl,5);

x  := x0 + dx;
x  := round(100*x)/100;

dy := d10*df + d20*power(df,2) + d02*power(dl,2) + d12*df*power(dl,2) +
      d30*power(df,3) + d22*power(df,2)*power(dl,2) + d40*power(df,4) +
      d04*power(dl,4) + d32*power(df,3)*power(dl,2) + d14*df*power(dl,4);
y  := y0 + dy;
y  := round(100*y)/100;
FloatPoint.X := x;
FloatPoint.Y := y;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.RDToNL(X, Y : Double);
const
x0  = 155000.000;
y0  = 463000.000;

f0  = 52.156160556;
l0  =  5.387638889;

a01 = 3236.0331637; b10 = 5261.3028966;
a20 =  -32.5915821; b11 =  105.9780241;
a02 =   -0.2472814; b12 =    2.4576469;
a21 =   -0.8501341; b30 =   -0.8192156;
a03 =   -0.0655238; b31 =   -0.0560092;
a22 =   -0.0171137; b13 =    0.0560089;
a40 =    0.0052771; b32 =   -0.0025614;
a23 =   -0.0003859; b14 =    0.0012770;
a41 =    0.0003314; b50 =    0.0002574;
a04 =    0.0000371; b33 =   -0.0000973;
a42 =    0.0000143; b51 =    0.0000293;
a24 =   -0.0000090; b15 =    0.0000291;

var
dx, dy : Double;
df, dl, f, l : Double;
begin

dx :=(x-x0)*power(10,-5);
dy :=(y-y0)*power(10,-5);

df := a01*dy + a20*power(dx,2) + a02*power(dy,2) + a21*power(dx,2)*dy +
      a03*power(dy,3) + a40*power(dx,4) + a22*power(dx,2)*power(dy,2) +
      a04*power(dy,4) + a41*power(dx,4)*dy + a23*power(dx,2)*power(dy,3) +
      a42*power(dx,4)*power(dy,2) + a24*power(dx,2)*power(dy,4);
f  := f0 + df/3600;

dl := b10*dx + b11*dx*dy + b30*power(dx,3) + b12*dx*power(dy,2) + b31*power(dx,3)*dy +
      b13*dx*power(dy,3) + b50*power(dx,5) + b32*power(dx,3)*power(dy,2) +
      b14*dx*power(dy,4) + b51*power(dx,5)*dy + b33*power(dx,3)*power(dy,3) + b15*dx*power(dy,5);
l  := l0 + dl/3600;

FloatPoint.X := l;                                                              // Convert radians to degrees
FloatPoint.Y := f;
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.NLToLam(Theta,Phi : Double);
var
  f,lambda0,p,k,alfa,delta_x,delta_y,xx,yy,
  e,grad_rad,tanhp,r:Double;
begin
  f        := 297;
  lambda0  := 4.35693972;
  p        := 0.7716421928;
  k        := 11565915.812935;
  alfa     := 0.00813847;
  delta_x  := 0.01256;
  delta_y  := 88.4378;

  f := 1 / f;
  e := 2* f - f * f;

  e        := sqrt(e);
  grad_rad := PI/180;
  alfa     := alfa    * grad_rad;
  lambda0  := lambda0 * grad_rad;
  phi      := phi     * grad_rad;
  Theta    := Theta   * grad_rad;

  xx := (1 + e * sin(phi)) / (1 - e * sin(phi));
  yy := e/2;
  xx := power(xx,yy);

  yy := PI/4 - phi/2;
  yy := tan(yy);
  tanhp := yy * xx;

  r := k* power(tanhp,p);

  xx := p * (Theta - lambda0) - alfa;
  FloatPoint.X :=  150000 + delta_x + r * sin(xx);
  FloatPoint.Y := 5400000 + delta_y - r * cos(xx);
end;
//------------------------------------------------------------------------------
procedure TEZConvertL.LamToNL(X,Y : Double);
var
  f,lambda0,p,k,alfa,delta_x,delta_y,lambda,rho1,xx,yy,
  e,grad_rad,rad_grad,x1,y1,phi,r,psi,rho:Double;
begin
  f        := 297;
  lambda0  := 4.35693972;
  p        := 0.7716421928;
  k        := 11565915.812935;
  alfa     := 0.00813847;
  delta_x  := 0.01256;
  delta_y  := 88.4378;

  rho1   := 0;

  f := 1 / f;
  e := 2* f - f * f;

  e        := sqrt(e);
  grad_rad := PI/180;
  rad_grad := 180/PI;
  alfa     := alfa * grad_rad;
  lambda0  := lambda0 * grad_rad;

  x1 := x - delta_x  - 150000;
  y1 := - y + delta_y +5400000;
  xx :=  x1 / y1;

  xx     := ArcTan(xx);
  phi    := xx + alfa;
  lambda := lambda0 + phi / p;
  lambda := lambda * rad_grad;

  r := x1 * x1 + y1 * y1;
  r := sqrt(r);

  xx  := r/k;
  xx  := log10(xx) / p;
  xx  := exp(xx);
  psi := 2 * ArcTan(xx);
  rho := PI / 2 - psi;

  while (rho <> rho1) do begin
    rho  := rho1;
    yy   := (1 - e * sin(rho)) / (1+e*sin(rho));
    yy   := e / 2 * log10(yy);
    yy   := xx * exp(yy);
    yy   := ArcTan(yy);
    rho1 := (PI/2) - (2*yy);
  end;

  rho := rho* rad_grad;

  FloatPoint.X := lambda;
  FloatPoint.Y := rho;
end;
//------------------------------------------------------------------------------
end.
