unit EZMapImageL;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, math,
  ExtCtrls, Menus, EZMapValuesL, EZProjectionL, EZCoarsL, EZConstL;

type
  TMapSettings = Class(TPersistent)
  private
    FGridValues   : boolean;
    Smooth        : boolean;
    FColorE       : TColor;
    FColorG       : TColor;
    FColorCD      : TColor;
    FColorMap     : TBitmap;
    FDColor       : boolean;
    FDCoast       : boolean;
    FDGrid        : boolean;
    FEZProjectionL: TEZProjectionL;
    FEZCoarsL     : TEZCoarsL;
    FIterate      : integer;
    FLineW        : double;
    FUseEmpty     : boolean;
    procedure SetEZProjection(const Value: TEZProjectionL);
    procedure SetEZCoars(const Value: TEZCoarsL);
    Procedure SetDrawColorMap(value:boolean);
    procedure SetColorMap(value:TBitmap);
  public
    constructor Create;
  published
    property AntiAlias : boolean          Read Smooth         write Smooth;
    property EZProjection: TEZProjectionL read FEZProjectionL write SetEZProjection;
    property EZCoars: TEZCoarsL           read FEZCoarsL      write SetEZCoars;
    property CoarsIterations : integer    read FIterate       write FIterate;
    property ColorEmpty : TColor          Read FColorE        write FColorE;
    property ColorGrid : TColor           Read FColorG        write FColorG;
    property GridValues : boolean         read FGridValues    write FGridValues;
    property ColorCoast: TColor           Read FColorCD       write FColorCD;
    property DrawCoast : boolean          Read FDCoast        write FDCoast;
    property DrawGrid : boolean           Read FDGrid         write FDGrid;
    property DrawColorMap : boolean       Read FDColor        write SetDrawColorMap;
    property LineWidth : double           read FLineW         write FLineW;
    property ColorMap :TBitmap            read FColorMap      write SetColorMap;
    property UseColorEmpty: boolean       read FUseEmpty      write FUseEmpty;
  end;


  TEZCustomMapImage = class(TCustomControl)
  private
    FOnPaint    : TEZPaintEvent;
    FMainMap    : TBitmap;
    procedure SetMainMap(Value:TBitmap);
  protected
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
  public
    OffsetX   : integer;
    OffsetY   : integer;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Paint; override;
    procedure SetOffset(X,Y:integer);
    property Color;
  published
    property OnPaint:TEZPaintEvent read FOnPaint write FOnPaint;
    property MainMap:TBitmap       read FMainMap write SetMainMap;
    property Anchors;
    property Constraints;
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property PopupMenu;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

  TEZMapImageL = class(TEZCustomMapImage)
  private
    FProgress     : TProgress;
    FMouseMove    : TMouseMove;
    FMouseDown    : TMouseEvent;
    FMouseUp      : TMouseEvent;
    CoastLine     : TCoastLine;
    CoastExists   : boolean;
    FErrorEvent   : TErrorEvent;
    Flag          : TFlag;
    FMapSettings  : TMapSettings;
    MarkerDrawn   : boolean;
    PluginOcc     : boolean;
    DrawStatus    : TDrawStatus;
    CurPos        : TPoint;
    IsDrawing     : boolean;
    function  ColorToRGBTriple(Color:  TColor):  TRGBTriple;
    function  RGBTripleToColor(Trip:TRGBTriple):  TColor;
    Procedure ErrorLog(F : integer);
    procedure ErrorEvent(Sender: TObject; Return:integer);
    function  IsCPAssigned: boolean;
    function  IsCoPAssigned: boolean;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure Progress(Max,Position:integer);
  public
    BackupMap   : TBitmap;
    procedure BackupImage;
    procedure CenterTo(Lon,Lat:double);
    procedure ClearBackupImage;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DrawAll;
    procedure DrawCC;
    procedure DrawCoars(Theta1,Phi1,Theta2,Phi2:Double;Color:TColor);
    procedure DrawGrid;
    procedure DrawRefMarkers;
    procedure DrawRotateGrid;
    function  Drawing:boolean;
    procedure ReloadCoast;
    procedure ResetImage;
    procedure SetCoarseEllips;
    procedure SetDemo;
    procedure SetGridProperties(East,West,North,South,Interval,Resolution:Double);
    procedure Wipe;
  published
    property  Anchors;
    property  Constraints;
    property  MapSettings : TMapSettings         read FMapSettings   write FMapSettings;
    property  OnError: TErrorEvent               read FErrorEvent    write FErrorEvent;
    property  OnMouseMove: TMouseMove            read FMouseMove     write FMouseMove;
    property  OnMouseDown: TMouseEvent           read FMouseDown     write FMouseDown;
    property  OnMouseUp  : TMouseEvent           read FMouseUp       write FMouseUp;
    property  OnProgress : TProgress             read FProgress      write FProgress;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('EZJ', [TEZMapImageL]);
end;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Base Routines  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////
constructor TEZCustomMapImage.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  TabStop := False;
  Width := 100;
  Height := 100;
  OffsetX := 0;
  OffsetY := 0;
  ParentColor := false;
  Color   := clBlack;
  FMainMap := TBitmap.Create;
end;
//------------------------------------------------------------------------------
destructor TEZCustomMapImage.Destroy;
begin
  FMainMap.Free;
  inherited Destroy;
end;
//------------------------------------------------------------------------------
procedure TEZCustomMapImage.SetMainMap(Value:TBitmap);
begin
  FMainMap.Assign(Value);
  Canvas.Brush.Color := Color;
  Canvas.Brush.Style := bsSolid;
  if Assigned(FMainMap)then if not FMainMap.Empty then begin
     OffsetX := 0;
     OffsetY := 0;
     Paint;
  end else Canvas.FillRect(Rect(0,0,width,height));
end;
//------------------------------------------------------------------------------
procedure TEZCustomMapImage.Paint;
begin
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := Color;
  if Assigned(FMainMap) then
  if not FMainMap.Empty then begin
    if (FMainMap.PixelFormat <> pf24Bit) then FMainMap.PixelFormat := pf24Bit;
    Canvas.Draw(OffsetX,OffsetY,FMainMap);
    if (OffsetX > 0) then Canvas.FillRect(Rect(0,0,OffsetX,height));
    if (MainMap.Width+OffsetX<Width) then Canvas.FillRect(Rect(MainMap.Width+OffsetX,0,Width,height));
    if (OffsetY > 0) then Canvas.FillRect(Rect(0,0,Width,OffsetY));
    if (MainMap.Height+OffsetY<Height) then Canvas.FillRect(Rect(0,MainMap.Height+OffsetY,Width,height));
  end else Canvas.FillRect(Rect(0,0,width,height));
end;
//------------------------------------------------------------------------------
procedure TEZCustomMapImage.SetOffset(X,Y:integer);
begin
  OffsetX := X;
  OffsetY := Y;
  Paint;
end;
//------------------------------------------------------------------------------
procedure TEZCustomMapImage.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  Message.Result := LRESULT(False);
end;
////////////////////////////////////////////////////////////////////////////////
constructor TMapSettings.Create;
begin
  inherited Create;
  FColorMap := TBitMap.Create;
  FDCoast   := False;
  FDColor   := false;
  FDGrid    := True;
  FColorE   := clBlack;
  FColorG   := clGray;
  FColorCD  := clLime;
  FLineW    := 1;
  FIterate  := 100;
  FUseEmpty := true;
end;
//------------------------------------------------------------------------------
constructor TEZMapImageL.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  BackupMap    := TBitMap.Create;
  FMapSettings := TMapSettings.Create;

  MarkerDrawn  := false;
  PluginOcc    := false;
  DrawStatus   := dsNone;
  IsDrawing    := false;

  ReloadCoast;
end;
//------------------------------------------------------------------------------
destructor TEZMapImageL.Destroy;
begin
  BackupMap.Free;
  FMapSettings.Free;
  ClearUndoTmp;
  inherited Destroy;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.ReloadCoast;
var
  Stream:TMemoryStream;
begin
  if FileExists(ExtractFilePath(ParamStr(0)) + 'Coast.dat') then begin
    Stream := TMemoryStream.Create;
    Stream.LoadFromFile(ExtractFilePath(ParamStr(0)) + 'Coast.dat');
    Stream.Read(CoastLine,SizeOf(TCoastLine));
    Stream.Free;
    CoastExists:= true;
  end else CoastExists := false;
end;
//------------------------------------------------------------------------------
function  TEZMapImageL.IsCPAssigned: boolean;
begin
  Result := false;
  if not Assigned(FMapSettings.FEZCoarsL) then ErrorLog(2)
  else if not Assigned(FMapSettings.FEZProjectionL) then ErrorLog(1)
  else Result := true;
end;
//------------------------------------------------------------------------------
function TEZMapImageL.IsCoPAssigned: boolean;
begin
  Result := false;
  if Assigned(MapSettings.FEZProjectionL) then begin
     if Assigned(FMapSettings.FEZProjectionL.EZConvert) then Result := true;
  end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.ErrorLog(F : integer);
begin
  Flag := F;
  if Flag < 10000 then ErrorEvent(Self,Flag);                                   // 10000 might be used as a temporary value. This meens the function/procedure using it is not finished -> not error event to prevent faults
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.ErrorEvent(Sender: TObject;Return:integer);
begin
  if Assigned(FErrorEvent) then FErrorEvent(Self,Flag);
end;
//------------------------------------------------------------------------------
procedure TMapSettings.SetDrawColorMap(value:boolean);
begin
  FDColor := value;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.BackupImage;
begin
  if Assigned(MainMap) then with BackupMap do begin
     Width := MainMap.Width;
     Height := MainMap.Height;
     Canvas.CopyMode := cmSrcCopy;
     Canvas.Draw(0,0,MainMap);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.ResetImage;
begin
  if Assigned(MainMap) and not MainMap.Empty then
  MainMap.Canvas.Draw(0,0,BackupMap);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.ClearBackupImage;
begin
  if Assigned(MainMap) then begin
     BackupMap.Free;
     BackupMap := TBitmap.Create;
  end;
end;
//------------------------------------------------------------------------------
procedure TMapSettings.SetColorMap(value:TBitmap);
begin
   if Assigned(FColorMap) then begin
      FColorMap.Assign(Value);
   end else SetDrawColorMap(false);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.Progress(Max,Position:integer);
begin
  if Assigned(FProgress) then FProgress(Self,Max,Position);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.Wipe;
begin
  if Assigned(MainMap) or (MainMap <> nil) then begin
    if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24Bit;
    MainMap.Width := Width;
    MainMap.Height := Height;
    MainMap.Canvas.Brush.Color := MapSettings.ColorEmpty;
    MainMap.Canvas.FillRect(Rect(0,0,Width,Height));
  end else begin
    MainMap := TBitmap.Create;
    if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24Bit;
    MainMap.Width := Width;
    MainMap.Height := Height;
    MainMap.Canvas.Brush.Color := MapSettings.ColorEmpty;
    MainMap.Canvas.FillRect(Rect(0,0,Width,Height));
  end;
  OffsetX := 0;
  OffsetY := 0;
  Paint;
end;
//------------------------------------------------------------------------------
procedure TMapSettings.SetEZProjection(const value : TEZProjectionL);
begin
  FEZProjectionL := value;
end;
//------------------------------------------------------------------------------
procedure TMapSettings.SetEZCoars(const value : TEZCoarsL);
begin
  FEZCoarsL := value;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.SetCoarseEllips;
begin
 if Assigned(MapSettings.FEZCoarsL) then begin
    case EllipsType[MapSettings.EZProjection.GetMapIndex] of
         1 : MapSettings.FEZCoarsL.Ellips := DefEllipsoid;                       // Lokal datum is a correction for an ellipsoid and the ezcoars nly uses ellipsoids and not datums. the translate function can be used but this has to be implemented later. I haven't checked it jet but the map projection/calculations probably already correct for any difference
         2 : MapSettings.FEZCoarsL.Ellips := DefEllipsoid;
         3 : MapSettings.FEZCoarsL.Ellips := NatGridEllips[MapSettings.EZProjection.MapInput]; // Ellipsoid corresponding with the national grid
         0 : MapSettings.FEZCoarsL.Ellips := 31;                                 // Spherical
    end;
 end;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.SetGridProperties(East,West,North,South,Interval,Resolution:Double);
begin
  if Assigned(FMapSettings.FEZProjectionL) then with FMapSettings.FEZProjectionL.MapRef do begin
    GridProp.North      := North;
    GridProp.South      := South;
    GridProp.West       := West;
    GridProp.East       := East;
    GridProp.Interval   := Interval;
    GridProp.Resolution := Resolution;
  end;
end;
//////////////////////////////////  end base  //////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  Mouse Routines  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
procedure TEZMapImageL.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Flag : Integer;
begin
  inherited;
  X := X-OffsetX;
  Y := Y-OffsetY;

  if Assigned(FMapSettings.FEZProjectionL) then begin
  if MainMap.Empty = false then begin
    FMapSettings.FEZProjectionL.CalculateInv(X,Y);
    Flag := FMapSettings.EZProjection.Flag;
    if Assigned(FMouseMove) and (Flag = 0) then FMouseMove(Self,Shift,X,Y,FMapSettings.FEZProjectionL.FloatPoint.X,FMapSettings.FEZProjectionL.FloatPoint.Y,0)
    else if Assigned(FMouseMove) then FMouseMove(Self,Shift,X,Y,0,0,FMapSettings.FEZProjectionL.Flag);
  end;
 end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  X := X-OffsetX;
  Y := Y-OffsetY;

  CurPos.X := X;
  CurPos.Y := Y;

 if Assigned(FMapSettings.FEZProjectionL) then begin
    FMapSettings.EZProjection.CalculateInv(X,Y);
    if Assigned(FMouseDown)and (FMapSettings.EZProjection.Flag = 0) then FMouseDown(Self,Button,Shift,X,Y, FMapSettings.EZProjection.FloatPoint.X, FMapSettings.EZProjection.FloatPoint.Y,0)
    else if Assigned(FMouseDown) then FMouseDown(Self,Button,Shift,X,Y,0,0,FMapSettings.EZProjection.Flag);
 end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  X := X-OffsetX;
  Y := Y-OffsetY;
  if Assigned(FMapSettings.FEZProjectionL) then begin
  if MainMap.Empty = false then begin
  FMapSettings.EZProjection.CalculateInv(X,Y);

  if (FMapSettings.EZProjection.Flag = 0) then begin
     if Assigned(FMouseUp) then FMouseUp(Self,Button,Shift,X,Y,FMapSettings.EZProjection.FloatPoint.X,FMapSettings.EZProjection.FloatPoint.Y,0)
  end else if Assigned(FMouseUp) then FMouseUp(Self,Button,Shift,X,Y,0,0,FMapSettings.EZProjection.Flag);
  end;
 end else ErrorLog(1);
 DrawStatus := dsNone;
end;
////////////////////////////////// end mouse ///////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  Draw Rouines  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
procedure TEZMapImageL.SetDemo;
var
Ind:integer;
Ratio,small : Double;
begin
if Assigned(FMapSettings.FEZProjectionL) then begin
Ind := FMapSettings.EZProjection.GetMapIndex;
if Ind > 2 then begin
if Height < Width then begin
  Ratio := Height/440;
  small := Height/2;
end else begin
  Ratio := Width/440;
  small := Width/2;
end;
  with FMapSettings.EZProjection.MapRef do begin
    Values.Rad := DemoValue[Ind,0]*Ratio;
    Values.Th0 := DegToRad(DemoValue[Ind,1]);
    Values.Ph0 := DegToRad(DemoValue[Ind,2]);
    Values.Ph1 := DegToRad(DemoValue[Ind,3]);
    Values.Ph2 := DegToRad(DemoValue[Ind,4]);
    Values.X0  := Small;
    Values.Y0  := Small;
    Values.H0  := DemoValue[Ind,7];
    Values.Oa  := 0;
    Calculated := true;
  end;
end else FMapSettings.EZProjection.MapRef.Calculated := false;
end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawRefMarkers;
var
i,X,Y : integer;
const
C = 21;
D = 2;
begin
if Assigned(FMapSettings.FEZProjectionL) then begin
if FMapSettings.FEZProjectionL.MapRef.CountRef > 0 then
for i := 0 to FMapSettings.FEZProjectionL.MapRef.CountRef -1 do begin
  X :=  Round(FMapSettings.FEZProjectionL.MapRef.Points[i].X);
  Y :=  Round(FMapSettings.FEZProjectionL.MapRef.Points[i].Y);
  MainMap.Canvas.Pen.Width := 3;
  MainMap.Canvas.Pen.Color := clRed;
  MainMap.Canvas.Brush.Style := bsClear;
  MainMap.Canvas.Ellipse(X-C,Y-C,X+C,Y+C);
  MainMap.Canvas.Pen.Width := 1;
  MainMap.Canvas.Ellipse(X-D,Y-D,X+D+1,Y+D+1);
  Paint;
end;
end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawRotateGrid;
var
  X1,X2,Y1,Y2,A,X0,Y0:Double;
const
  L=500;
  C= 4;
  Col = clRed;
begin
if Assigned(FMapSettings.FEZProjectionL) then begin
  with MainMap do begin
   Canvas.Font.Name  := 'Arial';
   Canvas.Font.Color := clBlue;
   Canvas.Font.Size  := 10;
   Canvas.Font.Style := [];
   X0 := FMapSettings.FEZProjectionL.MapRef.Values.Rx;
   Y0 := FMapSettings.FEZProjectionL.MapRef.Values.Ry;
   A  := FMapSettings.FEZProjectionL.MapRef.Values.Ra;

   Canvas.Brush.Color := clBlack;
   Canvas.Brush.Style := bsClear;
   Canvas.Ellipse(rect(round(X0-C+1),round(Y0-C+1),round(X0+C),round(Y0+C)));

   X1 :=  L;
   Y1 :=  0;
   X2 :=   Cos(A)*(X1) + Sin(A)*(Y1) + X0;
   Y2 :=  -Sin(A)*(X1) + Cos(A)*(Y1) + Y0;
   LineSmooth(Round(X0),Round(Y0),Round(X2),Round(Y2),MainMap,Col,1,MapSettings.Smooth);

   X1 :=  0;
   Y1 :=  L;
   X2 :=   Cos(A)*(X1) + Sin(A)*(Y1) + X0;
   Y2 :=  -Sin(A)*(X1) + Cos(A)*(Y1) + Y0;
   LineSmooth(Round(X0),Round(Y0),Round(X2),Round(Y2),MainMap,Col,1,MapSettings.Smooth);
   Paint;
  end;
end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawCC;
var
 x,y,i,j,max,Count,h : integer;
 fac,fat:Double;
 X1,X2,Y1,Y2,A1,A2,B1,B2,D,Da:double;
 Row,Row2:  pRGBTripleArray;
 ColCoast: TColor;
 NightDay,InitDone:boolean;
begin

 Count := 0;
 InitDone := false;

 if MainMap.Empty then begin
   MainMap := TBitmap.Create;
   if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;
   MainMap.Height := Height;
   MainMap.Width := Width;
   MainMap.Canvas.Brush.Color := FMapSettings.ColorEmpty;
   MainMap.Canvas.FillRect(rect(0,0,MainMap.Width,MainMap.Height));
   InitDone := true;
 end;

 if MapSettings.FUseEmpty and not InitDone then begin
   MainMap.Canvas.Brush.Color := FMapSettings.ColorEmpty;
   MainMap.Canvas.FillRect(rect(0,0,MainMap.Width,MainMap.Height));
   if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;
 end;

 if Assigned(FMapSettings.FEZProjectionL) then begin

 // set progress
 if (Assigned(FMapSettings.FColorMap)) and (FMapSettings.FDColor) and (FMapSettings.FColorMap.Empty = false) then Max := MainMap.Width*MainMap.Height
 else Max := 0;
 if FMapSettings.DrawCoast    and CoastExists then Max := Max + 59666;
    Progress(Max,0);

 if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;

 // Draw colormap part
 if (Assigned(FMapSettings.FColorMap)) and (FMapSettings.FDColor) and (FMapSettings.FColorMap.Empty = false) then begin
     if (FMapSettings.FColorMap.PixelFormat <> pf24Bit) then FMapSettings.FColorMap.PixelFormat := pf24bit;
     if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;
     fac := FMapSettings.FColorMap.Width/360;
     fat := FMapSettings.FColorMap.Height/180;



  for j := 0 to MainMap.Height-1 do begin
      Row := MainMap.Scanline[j];
      for i := 0 to MainMap.Width-1 do begin

      Count := Count + 1;
      if Count mod 2000 = 0 then Progress(Max,Count);

        FMapSettings.FEZProjectionL.CalculateInv(i,j-1);
        if FMapSettings.EZProjection.Flag = 0 then begin
           x := (FMapSettings.FColorMap.Width div 2 +
                 Round(FMapSettings.EZProjection.FloatPoint.X*fac)) Mod FMapSettings.FColorMap.Width;
           y := (FMapSettings.FColorMap.Height div 2 -
                 Round(FMapSettings.EZProjection.FloatPoint.Y*fat)) mod FMapSettings.FColorMap.Height;

           if (X >= 0) and (X < FMapSettings.FColorMap.Width) and (Y >= 0) and (Y < FMapSettings.FColorMap.Height) then begin
              Row2 := FMapSettings.FColorMap.ScanLine[y];
              Row[i] := ColorToRGBTriple(RGBTripleToColor(Row2[x]));
           end;
         end;
       end;
  end;

 end; // end draw colormap part
 if not Assigned(FMapSettings.FColorMap) then ErrorLog(4);


 // DrawCoast part
 if FMapSettings.FEZProjectionL.GetMapIndex > 2 then begin
 if FMapSettings.DrawCoast and CoastExists and (FMapSettings.EZProjection.MapRef.Values.Rad > 0.0001) then begin
    X1 := 0;
    Y1 := 0;
    A1 := 0;
    B1 := 0;
    for i := 0 to 59766 {MaxCoast - 1} do begin

       FMapSettings.EZProjection.CalculateFor(CoastLine[i].Lon,CoastLine[i].Lat);
       X2  := CoastLine[i].Lon;
       Y2  := CoastLine[i].Lat;
       A2  := FMapSettings.EZProjection.FloatPoint.X;
       B2  := FMapSettings.EZProjection.FloatPoint.Y;
       D   := sqrt(sqr(X1-X2)+sqr(Y1-Y2));
       Da  := sqrt(sqr(A1-A2)+sqr(B1-B2));

       Count := Count + 1;
       if Count mod 1000 = 0 then Progress(Max,Count);

       ColCoast := FMapSettings.FColorCD;

       if (d < 1) and (Da/FMapSettings.EZProjection.MapRef.Values.Rad < 0.05) then begin
          if (FMapSettings.EZProjection.FloatPoint.X >= 0) and (FMapSettings.EZProjection.FloatPoint.Y >= 0) and (FMapSettings.EZProjection.FloatPoint.X < MainMap.Width-1) and
             (FMapSettings.EZProjection.FloatPoint.Y < MainMap.Height-1) then
          LineSmooth(Round(A1),Round(B1),Round(A2),Round(B2),MainMap,ColCoast,FMapSettings.FLineW,false);
       end;

       X1 := X2;
       Y1 := Y2;
       A1 := A2;
       B1 := B2;

    end;
 end;
 end else ErrorLog(3); //end draw coast part
 Progress(Max,Max);
 if MapSettings.FUseEmpty then BackupImage;
end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawGrid;
var
  i,j,m,n,{h,Wo,}Index,Ind{,Max,Count}:integer;
  East,West,North,South,Interval,Resolution:Double;
  X1,X2,Y1,Y2:double;
  B,one,first : boolean;
  D : double;
  UP:TInPut;
begin

if Assigned(FMapSettings.FEZProjectionL) and (MapSettings.FDGrid) then begin
  With FMapSettings.FEZProjectionL.MapRef do begin
    North      := GridProp.North;
    South      := GridProp.South;
    East       := GridProp.East;
    West       := GridProp.West;
    Interval   := GridProp.Interval;
    Resolution := GridProp.Resolution;
  end;

  Index := FMapSettings.EZProjection.GetMapIndex;

  B := false;
  if (Index > 2) and (FMapSettings.FEZProjectionL.MapRef.Calculated) and
     (FMapSettings.EZProjection.MapRef.Values.Rad > 0.01) then B := True
  else if (Index <= 2) then begin
    B := True;
    if ((abs(FMapSettings.EZProjection.MapRef.Values.Rad) < 1E-9) and
       (abs(FMapSettings.EZProjection.MapRef.Values.Th0) < 1E-9)) or
       ((abs(FMapSettings.EZProjection.MapRef.Values.Ph1) < 1E-9) and
       (abs(FMapSettings.EZProjection.MapRef.Values.Ph2) < 1E-9)) then B := False;
    end;

  if B then begin

    if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;
    UP    := FMapSettings.EZProjection.MapRef.Input;
    Ind   := FMapSettings.EZProjection.MapRef.InputIndex;
{    Max :=  trunc((North-South)/resolution) +  trunc((East-West)/Interval) +
            trunc((North-South)/Interval) +  trunc((East-West)/resolution);
    Count := 0;}


         if Index = 0 then FMapSettings.EZProjection.MapRef.Input := ifUTM
    else if Index = 1 then FMapSettings.EZProjection.MapRef.Input := ifDatumUTM
    else if Index = 2 then FMapSettings.EZProjection.MapRef.Input := ifNationalGrid
    else  FMapSettings.EZProjection.MapRef.Input := ifGeographic;
    FMapSettings.EZProjection.MapRef.InputIndex := FMapSettings.EZProjection.MapRef.MapIndex;

    MainMap.Canvas.Pen.Color   := FMapSettings.ColorGrid;
    MainMap.Canvas.Font        := Font;
    MainMap.Canvas.Font.Color  := FMapSettings.ColorGrid;
    MainMap.Canvas.Brush.Style := bsClear;
    D  := 0;
//    Wo := 0;
//    H := Round(MainMap.Canvas.TextHeight('I'));

    // Longtitude lines
    X1 := 0;
    Y1 := 0;
    m := trunc((North-South)/resolution);
    n := trunc((East-West)/Interval);
    for j := 0 to n  do begin
    first := true;
    for i := 0 to m  do begin
        FMapSettings.EZProjection.CalculateFor(West+j*interval,North-(i)*resolution);
        X2 := round(FMapSettings.EZProjection.FloatPoint.X);
        Y2 := round(FMapSettings.EZProjection.FloatPoint.Y);
        if (FMapSettings.EZProjection.Flag = 0) then D := sqrt((X1-X2)*(X1-X2)+(Y1-Y2)*(Y1-Y2));

        B := false;
        if (Index <= 2) and (FMapSettings.EZProjection.Flag = 0) then begin
          if (D < Interval) then B := True;
        end else if (FMapSettings.EZProjection.Flag = 0) then begin
          if (D/FMapSettings.EZProjection.MapRef.Values.Rad < 0.1) then B := true;
        end;

        if B and not first then LineSmooth(Round(X2),Round(Y2),Round(X1),Round(Y1),MainMap,FMapSettings.FColorG,FMapSettings.FLineW,FMapSettings.Smooth);
        First := false;
        X1 := X2;
        Y1 := Y2;
    end;
    end;

    // Lattitude lines
    X1 := 0;
    Y1 := 0;
    m := trunc((North-South)/Interval);
    n := trunc((East-West)/resolution);
    for j := 0 to m  do begin
    first := true;
    for i := 0 to n  do begin
        FMapSettings.EZProjection.CalculateFor(West+(i)*resolution,North-j*interval);
        X2 := round(FMapSettings.EZProjection.FloatPoint.X);
        Y2 := round(FMapSettings.EZProjection.FloatPoint.Y);
        if (FMapSettings.EZProjection.Flag = 0) then D   := sqrt(sqr(X1-X2)+sqr(Y1-Y2));

        B := false;
        if (Index <= 2) and (FMapSettings.EZProjection.Flag = 0) then begin
          if (D < Interval) then B := True;
        end else if (FMapSettings.EZProjection.Flag = 0) then begin
          if (D/FMapSettings.EZProjection.MapRef.Values.Rad < 0.1) then B := true;
        end;

        if B and not first then LineSmooth(Round(X2),Round(Y2),Round(X1),Round(Y1),MainMap,FMapSettings.FColorG,FMapSettings.FLineW,FMapSettings.Smooth);
        first := false;
        X1 := X2;
        Y1 := Y2;
    end;
    end;

   // Text to go with the grid
   if FMapSettings.FGridValues then begin                                                             // only start if the values must be shown
    MainMap.Canvas.Pen.Color   := FMapSettings.ColorGrid;
    MainMap.Canvas.Font.Color  := FMapSettings.ColorGrid;
    MainMap.Canvas.Brush.Style := bsClear;

    m := trunc((North-South)/Interval);
    n := trunc((East-West)/Interval);
    one := true;
    X2 := 0; Y2 := 0;
    for j := 0 to m  do for i := 0 to n  do begin                               // determine the most centred Lon/Lat lines
          FMapSettings.EZProjection.CalculateFor(West+i*interval,North-j*interval);
          X1 := round(FMapSettings.EZProjection.FloatPoint.X);
          Y1 := round(FMapSettings.EZProjection.FloatPoint.Y);
        if one and (X1 > MainMap.Width div 2) and (Y1 > MainMap.Height div 2)
        and (X1 < MainMap.Width) and (Y1 < MainMap.Height) then begin
          X2 := West+i*interval;
          Y2 := North-j*interval;
          One := false;
        end;
    end;

    for j := 0 to m  do begin                                                     // draw all lattitude values
       FMapSettings.EZProjection.CalculateFor(X2,North-j*interval);
       X1 := round(FMapSettings.EZProjection.FloatPoint.X);
       Y1 := round(FMapSettings.EZProjection.FloatPoint.Y);
       if (X1 > 0) and (X1 < MainMap.Width) and (Y1 > 0) and (Y1 < MainMap.Height)
       and (North-j*interval <> Y2) and (FMapSettings.EZProjection.Flag = 0) then // only draw if the values can be seen and if they do not overlap with the longtitude values
       MainMap.Canvas.TextOut(Round(X1)+2,Round(Y1)+2,FloatToStr(North-j*interval));
    end;

    for i := 0 to n  do begin                                                     // Draw all longtitude lines
       FMapSettings.EZProjection.CalculateFor(West+i*interval,Y2);
       X1 := round(FMapSettings.EZProjection.FloatPoint.X);
       Y1 := round(FMapSettings.EZProjection.FloatPoint.Y);
       if (X1 > 0) and (X1 < MainMap.Width) and (Y1 > 0) and (Y1 < MainMap.Height)
       and (West+i*interval <> X2) and (FMapSettings.EZProjection.Flag = 0) then// Draw only is visible or if they do not overlap with the lattitude lines
       MainMap.Canvas.TextOut(Round(X1)+2,Round(Y1)+2,FloatToStr(West+i*interval));
    end;

    FMapSettings.EZProjection.CalculateFor(X2,Y2);                             // Calculate overlap point
    X1 := round(FMapSettings.EZProjection.FloatPoint.X);
    Y1 := round(FMapSettings.EZProjection.FloatPoint.Y);
    if (FMapSettings.EZProjection.Flag = 0) then
    MainMap.Canvas.TextOut(Round(X1)+2,Round(Y1)+2,FloatToStr(X2)+ '/' +            // draw overlap point
    FloatToStr(Y2));
 end;

 FMapSettings.FEZProjectionL.MapRef.Input      := UP;
 FMapSettings.FEZProjectionL.MapRef.InputIndex := Ind;

 end else ErrorLog(5);
 end else ErrorLog(1);
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawCoars(Theta1,Phi1,Theta2,Phi2:Double;Color:TColor);
var
  C,D,Inc:Double;
  I : integer;
  X1,X2,Y1,Y2,Da:double;
  Ref:boolean;
begin

  if IsCPAssigned then if
     (FMapSettings.FEZProjectionL.MapRef.Calculated) and
     (FMapSettings.EZProjection.MapRef.Values.Rad > 0.01) then begin

  SetCoarseEllips;                                                                // Set current ellpsoid of EZProjection to EZCoars

  if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24bit;
  MainMap.Canvas.Pen.Color := Color;
  FMapSettings.FEZCoarsL.CalcDistance(Theta1,Phi1,Theta2,Phi2);
  D := FMapSettings.FEZCoarsL.ResultValues.Distance;
  C := FMapSettings.FEZCoarsL.ResultValues.Coars;
  Inc := D/MapSettings.FIterate;
  Ref := false;
   X1 := 0;
   Y1 := 0;
  For I := 0 to MapSettings.FIterate do begin

     FMapSettings.FEZCoarsL.CalcEndpoint(Theta1,Phi1,C,I*Inc);
     FMapSettings.FEZProjectionL.CalculateFor(FMapSettings.FEZCoarsL.ResultValues.Lon,
       FMapSettings.FEZCoarsL.ResultValues.Lat);
     X2 := FMapSettings.FEZProjectionL.FloatPoint.X;
     Y2 := FMapSettings.FEZProjectionL.FloatPoint.Y;
     Da := sqrt(sqr(X1-X2)+sqr(Y1-Y2));
     if (Da/FMapSettings.FEZProjectionL.MapRef.Values.Rad < 0.1) and (Ref) and
        (FMapSettings.FEZProjectionL.Flag = 0) then begin
          LineSmooth(Round(X2),Round(Y2),Round(X1),Round(Y1),MainMap,Color,FMapSettings.FLineW,FMapSettings.Smooth);
     end;

     X1 := X2;
     Y1 := Y2;
     if FMapSettings.FEZProjectionL.Flag = 0 then Ref := true  else Ref := false;
  end;
  end else ErrorLog(5);
  Paint;
end;
//------------------------------------------------------------------------------
function  TEZMapImageL.ColorToRGBTriple(Color:  TColor):  TRGBTriple;
begin
  with result do begin
    r   := GetRValue(Color);
    g   := GetGValue(Color);
    b   := GetBValue(Color)
  end
end;
//------------------------------------------------------------------------------
function  TEZMapImageL.RGBTripleToColor(Trip:TRGBTriple):  TColor;
begin
  with Trip do begin
    Result := RGB(r,g,b);
  end
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.DrawAll;
begin
 if not IsDrawing then begin
  IsDrawing := true;

  ResetImage;
  if (MainMap.PixelFormat <> pf24Bit) then MainMap.PixelFormat := pf24Bit;

  DrawCC;
  DrawGrid;
  Paint;

  IsDrawing := false;
 end;
end;
//------------------------------------------------------------------------------
function TEZMapImageL.Drawing:boolean;
begin
 Result := IsDrawing;
end;
//------------------------------------------------------------------------------
procedure TEZMapImageL.CenterTo(Lon,Lat:double);
var
 P : TPoint;
begin
  if IsCoPAssigned then if FMapSettings.FEZProjectionL.MapRef.Calculated then begin
    MapSettings.FEZProjectionL.CalculateFor(Lon,lat);
    P := Point(Trunc(MapSettings.FEZProjectionL.FloatPoint.X),Trunc(MapSettings.FEZProjectionL.FloatPoint.Y));

    OffsetX := (Width div 2)-P.X;
    OffSetY := (Height div 2)-P.Y;
  end;
end;
////////////////////////////////////////////////////////////////////////////////
end.

