unit Example;
//Demonstrating using a cardtable.
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, CardTable, ExtCtrls;

type
  TForm1 = class(TForm)
    Button3: TButton;
    Button5: TButton;
    Label1: TLabel;
    Button1: TButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Button3Click(Sender: TObject);
    procedure CardClicked(ACard: TCard; Button: TMouseButton);
    procedure FormCreate(Sender: TObject);
    procedure CardDropped(ACard: TCard; X, Y, RX, RY:Integer);
    procedure OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure Button5Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure FormCanResize(Sender: TObject; var NewWidth,
      NewHeight: Integer; var Resize: Boolean);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

var
  gCardTable: TCardTable;
  gHand: Array[1..8] of TCard;
  gSetByMarker: TCardMarker;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if gCardTable <> nil then
  begin
    gCardTable.Free;//Not really needed as form calls it.
    gCardTable := nil;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
//Create the table.
  gCardTable := TCardTable.Create(self);
//If exception is raised then it is probably because CardTable cannot find the
//cardsets. Ensure it is in program directory or file named config.txt in
//program directory holds the fully qualified path to it.
//Now make the table the size of the form. To ensure it is always the size of
//the form you also need to do this when the form resizes. It is better to do
//this in the forms OnCanResize (pre resize) and not OnResize (post resize)
//event so as to stop flicker as shown below. Always use ResizeTable NOT width
//or height properties.
  gCardTable.ResizeTable(0, 0, ClientWidth, ClientHeight);
//I want to Stretch any background picture to the size of the table else the
//table will be the size of the background picture.
  gCardTable.StretchBackground := True;
//Enable drag & drop with the mouse.
  gCardTable.DragAndDrop := True;
//If we want to in someway control dragged cards or draw card markers then we
//need to associate a procedure with the CardTable`s OnMouseMove event.
  gCardTable.OnMouseMove := OnMouseMove;
//Likewise to do something when a card is clicked or dropped we need to catch
//these events too. If you enable DragAndDrop you must catch the OnCardDropEvent
//to process the drop.
  gCardTable.OnCardClickEvent := CardClicked;
  gCardTable.OnCardDropEvent := CardDropped;
//Lets give a pseudo 3d look to the deck when its displayed.
  gCardTable.PlaceDeckOffset := 7;
//As there is a lightweight control, TLabel, on the table (TCardTable is set
//to be below all others) we set RepaintLabels to True to automatically repaint
//all lightweight controls as needed otherwise they will disappear. (Nb. The
//display of lightweight controls on the table is far from perfect with nasty
//flicker so if at all possible place them off the table.)
  gCardTable.RepaintLabels := True;
//Shuffle the internal deck before use.
  gCardTable.CardDeck.Shuffle;
//Make the table green.
  gCardTable.Color := clGreen;
//Place some card markers on the table.
  gCardTable.PlaceCardMarker(cmClub, 130, 20);
  gCardTable.PlaceCardMarker(cmDiamond, 230, 20);
  gCardTable.PlaceCardMarker(cmHeart, 330, 20);
  gCardTable.PlaceCardMarker(cmSpade, 430, 20);
//FormCreate isn`t the best place to place cards and decks as the form isn`t yet
//displayed. If you wish to display cards & decks on startup then FormActivate
//seems to work better, see below.
end;

procedure TForm1.FormActivate(Sender: TObject);
var
  i: Integer;
begin
//We need to ensure the table is displayed before the deck is placed otherwise
//it can look a bit ugly.
  Update;
//Now we can place the deck.
  gCardTable.PlaceDeck(35,20);
//Now lets deal a hand of cards. First changing the CardSpeed.
  gCardTable.CardSpeed := 5;
  for i := 1 to 8 do
  begin
//Drawing a card from a displayed deck. Nb although CardTable is concerned with
//displaying cards and CardDeck with controlling the deck sometimes CardTable
//has methods that make the interface between the deck and the table slightly
//easier. DrawCardFromDeck is one such method. Alternatively you would have to
//call CardDeck.Draw and CardTable.SetDrawnCard.
    gHand[i] := gCardTable.DrawCardFromDeck;
//Move the card to new x,y co-ordinates.
    gCardTable.MoveTo(gHand[i], (i-1)*55, 250);
//Turn over the card.
    gCardTable.TurnOverCard(gHand[i]);
  end;
//Change color of the table to SkyBlue as I`m bored with green.
  gCardTable.Color := clSkyBlue;
//Sort the cards in the hand. The other params are concerned with the value of
//aces and wheather cards are sorted within suits.
  gCardTable.SortHand(gHand, True, False);
//I`m bored with SkyBlue. Using a picture stretched to the table. You have to
//give a fully qualified file name.
  gCardTable.BackgroundPicture := gCardTable.Directory + '\TXT007.jpg';
end;

procedure TForm1.OnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
//Responding to this event allows me to draw a card marker on the table when a
//dragged card reaches a specific position on the screen.
var
  Bounds: TRect;
  i: TCardMarker;
  XPos: Integer;
begin
//Are we dragging a card.
  if gCardTable.CardDragging then
  begin
    XPos := 130;
//Check bounds for each of the dragging areas.
    for i := cmClub to cmSpade do
    begin
//Bounds specify the target area.
      Bounds.Left := XPos;
      Bounds.Top := 20;
      Bounds.Right := XPos + gCardTable.CardDeck.CardWidth;
      Bounds.Bottom:= 20 + gCardTable.CardDeck.CardHeight;
      Inc(XPos, 100);
//Is the card within these bounds.
      if gCardTable.DragCardWithinBounds(Bounds) then
      begin
//If there is NOT a mark already on the table and the dragged card is the right
//suit place one.
        if not gCardTable.ShadePlaced then
        begin
          if ((i = cmClub) and (gCardTable.DraggingCard.Suit = csClub)) or
             ((i = cmDiamond) and (gCardTable.DraggingCard.Suit = csDiamond)) or
             ((i = cmHeart) and (gCardTable.DraggingCard.Suit = csHeart)) or
             ((i = cmSpade) and (gCardTable.DraggingCard.Suit = csSpade)) then
          begin
            gCardTable.PlaceCardMarker(cmDragShade, Bounds.Left, Bounds.Top);
            gSetByMarker := i;
          end;
        end;
      end
//The card is not within the bounds so if a mark has been placed by this marker
//pick it up.
    else if gCardTable.ShadePlaced and (gSetByMarker = i) then
      gCardTable.PickUpMarker(cmDragShade);
    end;
  end;
end;

procedure TForm1.CardDropped(ACard: TCard; X, Y, RX, RY:Integer);
//Responding to a CardDropped event. You must deal with this if you enable drag
//and drop!
begin
//If a Shade/Marker has been placed it must be first picked up.
  if gCardTable.ShadePlaced then
//DropCardOnShade does this and also drops the card on the shades position.
    gCardTable.DropCardOnShade(cmDragShade)
  else
//We will just return it to its previous position.
    gCardTable.RestoreCard;
end;

procedure TForm1.CardClicked(ACard: TCard; Button: TMouseButton);
//A Card has been clicked on.
begin
//If right mouse button and card is in play on the table (ACard.Status = 1).
  if (Button = mbright) and (ACard.Status = 1) then
//We will just turn it over.
    gCardTable.TurnOverCard(ACard);
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
//Selecting a new card deck via a form.
  gCardTable.CardDeck.SelectDeck;
//We need to inform CardTable of the change.
  gCardTable.DeckOrBackChanged;
end;

procedure TForm1.Button5Click(Sender: TObject);
begin
//Selecting a new card back via a form.
  gCardTable.CardDeck.SelectBack;
//We need to inform CardTable of the change.
  gCardTable.DeckOrBackChanged;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  i, j: Integer;
  ACard: TCard;
begin
//Now pick up all the cards and reset the deck. PickUpAllCardsAndReset is the
//same as calling PickUpDeck, PickUpCardsOnTable, PickUpDiscards and
//CardDeck.ResetDeck.
  gCardTable.PickUpAllCardsAndReset;
//Shuffle the deck.
  gCardTable.CardDeck.Shuffle;
//Place on the table.
  gCardTable.PlaceDeck(35,30);
//Deal a few cards to the table.
  for i := 1 to 4 do
    for j := 3 to 5 do
    begin
      ACard := gCardTable.DrawCardFromDeck;
//Note that using co-ordinates that are not related to the current deck`s card
//width and height means that there will be no problems with moving or placing
//cards if a different deck is chosen mid game.
      gCardTable.MoveTo(ACard, 60 + i * 90, j * 55);
      gCardTable.TurnOverCard(ACard);
    end;
end;

procedure TForm1.FormCanResize(Sender: TObject; var NewWidth,
  NewHeight: Integer; var Resize: Boolean);
begin
//The form is about to resize. Lets make the table fit the form. Use this event
//and not the forms resize event so as to avoid flicker. Use ResizeTable method
//to resize the table and NOT width and height properties. Reason: This method
//takes account of the FCopyToScreen flag which determines if changes should be
//reflected on screen. Width & Height do not and result in nasty flicker.
  if (NewWidth >650) and (NewHeight > 400) then
//Make the table the size of the forms client area not entire area of the form.
  gCardTable.ResizeTable(0, 0, NewWidth - (Width-ClientWidth),
    NewHeight - (Height-ClientHeight))
  else
//Stop the resize.
    Resize := False;
end;

end.
