(*
   Firebird Library
   File:FBLDatabase.pas
   Copyright (c) 2002 Alessandro Batisti
   abatisti@tiscali.it
   http://web.tiscali.it/fblib

  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
*)

{$I fbl.inc}
unit FBLDatabase;

interface

uses
  SysUtils, Classes,
  {$IFDEF D6P}
  Types,
  {$ENDIF}
  ibase_h;


type

  IFBLDbEvent = interface
    ['{08324E34-B790-D611-9B47-AA8BBB4BE1E4}']
    procedure DoOnDataBaseDisconnect;
  end;

  TProtocolType = (ptLocal, ptTcpIp, ptNetBeui);
  TProviderInfo = (piUnknow, piInterbase, piFirebird);
  TServerInfo = (siUnKnow, siSuperServer, siClassicServer);

  TFBLDatabase = class(TComponent)
  private
    FLibraryLoaded:  Boolean;
    FDBHandle:       TISC_DB_HANDLE;
    FUser:           string;
    FPassword:       string;
    FRole:           string;
    FHost:           string;
    FTcpPort:        Word;                  // default 3050
    FProtocol:       TProtocolType;
    FDBFile:         string;
    FConnectString:  string;
    FTrans:          TList;                 // list attach transaction Objects
    FSQLDialect:     Integer;               // Sql client Dialect
    FBackoutCount:   TStringList;           // isc_info_backout_count
    FDeleteCount:    TStringList;           // isc_info_delete_count
    FExpungeCount:   TStringList;           // isc_info_expunge_count
    FInsertCount:    TStringList;           // isc_info_insert_count
    FPurgeCount:     TStringList;           // isc_info_purge_count
    FReadIdxCount:   TStringList;           // isc_info_read_idx_count
    FReadSeqCount:   TStringList;           // isc_info_read_seq_count
    FUpdateCount:    TStringList;           // isc_info_update_count
    FAutoSetClientDialect: Boolean;
    FOnConnect:      TNotifyEvent;
    FOnDisconnect:   TNotifyEvent;
    function GetDbHandle:  PISC_DB_HANDLE;
    function GetConnected: Boolean;
    procedure SetSQLDialect(Value: Integer);
    // database info
    function GetDBSQLDialect: Integer;
    function GetVersion:      string;
    function GetDBSiteName:   string;
    function GetDBFileName:   string;
    function GetLocalConnection: Boolean;
    function GetPageSize: Integer;
    function GetDBInfoInt2(isc_info: Integer; var ok: Boolean): Integer;
    function GetDBInfoInt(isc_info: Integer): Integer;
    function GetProviderInfo: TProviderInfo;
    function GetServerInfo: TServerInfo;
    function GetReads: Integer;
    function GetWrites: Integer;
    function GetFetches: Integer;
    function GetODSVersion: Integer;
    function GetODSMinorVersion: Integer;
    function GetBaseLevel: Integer;
    function GetImplementationNumber: Integer;
    function GetImplementationClass: Integer;
    function GetCurrentMemory: Integer;
    function GetMaxMemory: Integer;
    function GetAllocation: Integer;
    function GetNumBuffers: Integer;
    function GetSweepInterval: Integer;
    function GetBackoutCount: TStringList;       // isc_info_backout_count
    function GetDeleteCount: TStringList;        // isc_info_delete_count
    function GetExpungeCount: TStringList;       // isc_info_expunge_count
    function GetInsertCount: TStringList;        // isc_info_insert_count
    function GetPurgeCount: TStringList;         // isc_info_purge_count
    function GetReadIdxCount: TStringList;       // isc_info_read_idx_count
    function GetReadSeqCount: TStringList;       // isc_info_read_seq_count
    function GetUpdateCount: TStringList;        // isc_info_update_count
    function GetOperationCounts(DBInfoCommand: Integer;
      FOperation: TStringList): TStringList;
    procedure SetConnectString;
    function GetClientVersion: string;
    function GetClientMajorVersion: Integer;
    function GetClientMinorVersion: Integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddTransaction(ATransaction: IFBLDbEvent);
    procedure RemoveTransaction(ATransaction: IFBLDbEvent);
    procedure Connect;
    procedure Disconnect;
    procedure CreateDatabase(const AFilename, AUser, APassword: string;
      ADialect: Integer = 3; APagesize: Integer = 4096; ACharSet: string = '');
    property Connected: Boolean read GetConnected;
    property DBHandle: PISC_DB_HANDLE read GetDBHandle;
    //dbinfo
    property DBSqlDialect: Integer read GetDBSqlDialect;
    property Version: string read GetVersion;
    property DBSiteName: string read GetDBSiteName;
    property DBFileName: string read GetDBFileName;
    property PageSize: Integer read GetPageSize;
    property ProvaiderInfo: TProviderInfo read GetProviderInfo;
    property ServerInfo: TServerInfo read GetServerInfo;
    property IsLocalConnection: Boolean read GetLocalConnection;
    property Reads: Integer read GetReads;
    property Writes: Integer read GetWrites;
    property Fetches: Integer read GetFetches;
    property ODSVersion: Integer read GetODSVersion;
    property ODSMinorVersion: Integer read GetODSMinorVersion;
    property BaseLevel: Integer read GetBaseLevel;
    property ImplementationNumber: Integer read GetImplementationNumber;
    property ImplementationClass: Integer read GetImplementationClass;
    property CurrentMemory: Integer read GetCurrentMemory;
    property MaxMemory: Integer read GetMaxMemory;
    property Allocation: Integer read GetAllocation;
    property NumBuffers: Integer read GetNumBuffers;
    property SweepInterval: Integer read GetSweepInterval;
    property BackoutCount: TStringList read GetBackoutCount; // isc_info_backout_count
    property DeleteCount: TStringList read GetDeleteCount; // isc_info_delete_count
    property ExpungeCount: TStringList read GetExpungeCount; // isc_info_expunge_count
    property InsertCount: TStringList read GetInsertCount; // isc_info_insert_count
    property PurgeCount: TStringList read GetPurgeCount; // isc_info_purge_count
    property ReadIdxCount: TStringList read GetReadIdxCount; // isc_info_read_idx_count
    property ReadSeqCount: TStringList read GetReadSeqCount; // isc_info_read_seq_count
    property UpdateCount: TStringList read GetUpdateCount; // isc_info_update_count
    property ClientVersion: string read GetClientVersion;  //only in FB 1.5.x  result:='' in FB 1.x.x
    property ClientMajorVersion: Integer read GetClientMajorVersion;   //only in FB 1.5.x  result := 0 in FB 1.0.x
    property ClientMinorVersion: Integer read GetClientMinorVersion;    //only in FB 1.5.x  result := 0 in FB 1.0.x
  published
    property DBFile: string read FDBFile write FDBFile;
    property Host: string read FHost write FHost;
    property Protocol: TProtocolType read FProtocol write FProtocol;
    property User: string read FUser write FUser;
    property Password: string read FPassword write FPassword;
    property Role: string read FRole write FRole;
    property TcpPort: word read FTcpPort write FTcpPort default 3050;
    property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
    property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    property SQLDialect: Integer read FSQLDialect write SetSqlDialect;
    property AutoSetClientDialect: Boolean read FAutosetClientDialect  
      write FAutoSetClientDialect default True;
  end;



implementation

uses FBLExcept, FBLConst;

(* TFBLDatabase *)

constructor TFBLDatabase.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLibraryLoaded := False;
  CheckIbSqlLoaded;
  FLibraryLoaded := True;
  CheckIbSqlLoaded;
  FTrans := TList.Create;
  FDBHandle := nil;
  FSQLDialect := 1;
  FConnectString := '';
  FTcpPort := 3050;
  FAutoSetClientDialect := True;
  FBackoutCount := nil;
  FDeleteCount := nil;
  FExpungeCount := nil;
  FInsertCount := nil;
  FPurgeCount := nil;
  FReadIdxCount := nil;
  FReadSeqCount := nil;
  FUpdateCount := nil;
end;

//------------------------------------------------------------------------------

destructor TFBLDatabase.Destroy;
begin
  if FLibraryLoaded then
  begin
    if FDBHandle <> nil then Disconnect;
    FTrans.Free;
    FBackoutCount.Free;
    FDeleteCount.Free;
    FExpungeCount.Free;
    FInsertCount.Free;
    FPurgeCount.Free;
    FReadIdxCount.Free;
    FReadSeqCount.Free;
    FUpdateCount.Free;
  end;
  inherited Destroy;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDbHandle: PISC_DB_HANDLE;
begin
  Result := @FDBhandle;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetConnected;
begin
  Result := (FDBHandle <> nil);
end;


//------------------------------------------------------------------------------

procedure TFBLDatabase.SetConnectString;
begin
  case FProtocol of
    ptLocal:
      FConnectString := FDBFile;
    ptTcpIp:
      if FHost = '' then
        FBLError(E_DB_NO_HOST)
      else
        if FtcpPort <> 3050 then
          FConnectString := FHost + '/' + IntToStr(FtcpPort) + ':' + FDBFile
        else
          FConnectString := FHost + ':' + FDBFile;
        ptNetBeui:
      if FHost = '' then
        FBLError(E_DB_NO_HOST)
      else
        FConnectString := '\\' + FHost + '\' + FDBFile;
  end;
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.Connect;
var
  Status_vector: ISC_STATUS_VECTOR;
  DPB_Buffer: array [0..255] of Char;
  DPB_Index: Integer;
  DBName_cstr: array [0..255] of Char;
begin
  if FDBHandle <> nil then
    FBLError(E_DB_ALREADY_CON);
  SetConnectString;
  FillChar(DPB_Buffer, SizeOf(DPB_Buffer), #0);
  //FillChar(DBName_cstr, SizeOf(DBName_cstr), #0);
  DPB_Index := 0;
  DPB_Buffer[DPB_index] := Char(isc_dpb_version1);
  Inc(DPB_index);
  // set user name
  DPB_Buffer[DPB_Index] := Char(isc_dpb_user_name);
  Inc(DPB_index);
  DPB_Buffer[DPB_index] := Char(Length(FUser));
  Inc(DPB_index);
  StrPCopy(@DPB_Buffer[DPB_index],FUser);
  Inc(DPB_index, Length(FUser));
  // set password
  DPB_Buffer[DPB_index] := Char(isc_dpb_password);
  Inc(DPB_index);
  DPB_Buffer[DPB_index] := Char(length(FPassword));
  Inc(DPB_index);
  StrPCopy(@DPB_Buffer[DPB_index], FPassword);
  Inc(DPB_index, length(FPassword));

  if FRole <> '' then
  begin
    DPB_Buffer[DPB_index] := Char(isc_dpb_sql_role_name);
    Inc(DPB_index);
    DPB_Buffer[DPB_index] := Char(length(FRole));
    Inc(DPB_index);
    StrPCopy(@DPB_Buffer[DPB_index], FRole);
    Inc(DPB_index, Length(FRole));
  end;

  StrPCopy(DBName_cstr, FConnectString);

  isc_attach_database(@Status_vector, Length(FConnectString), @DBName_cstr, @FDBHandle,
    DPB_Index, DPB_Buffer);
  

  if (Status_vector[0] = 1) and (Status_vector[1] <> 0) then
    FBLShowError(@Status_vector);
  if FAutoSetClientDialect then SQLDialect := DBSqlDialect;
  if Assigned(FOnConnect) then
    FOnConnect(self);
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.Disconnect;
var
  Status_vector: ISC_STATUS_VECTOR;
  i: integer;
begin
  if FDBHandle = nil then
    FBLError(E_DB_NOACTIVE_CON);

  for i := 0 to FTrans.Count - 1 do
    IFBLDbEvent(FTrans[i]).DoOnDatabaseDisconnect;

  isc_detach_database(@Status_vector, @FDBhandle);

  if FDBHandle <> nil then
    FBLShowError(@Status_vector);
  if Assigned(FOnDisconnect) then
    FOnDisconnect(self);
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.CreateDatabase(const AFilename, AUser, APassword: string;
  ADialect: integer = 3; APagesize: integer = 4096;
  ACharset: string = '');
var
  Status_vector: ISC_STATUS_VECTOR;
  DBh: TISC_DB_HANDLE;
  TRh: TISC_TR_HANDLE;
  Params: string;
begin
  DBh := nil;
  TRh := nil;
  Params := 'USER ''' + AUser + ''' ' + 'PASSWORD ''' +
    APassword + ''' ' + 'PAGE_SIZE ' + IntToStr(APageSize);
  if ACharSet <> '' then
    Params := Params + ' DEFAULT CHARACTER SET ' + ACharSet;
  isc_dsql_execute_immediate(@Status_Vector, @DBh, @TRh, 0,
    PChar('CREATE DATABASE ''' + AFileName + ''' ' +
    Params), ADialect, nil);
  if (Status_vector[0] = 1) and (Status_vector[1] <> 0) then
    FBLShowError(@Status_vector);
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.SetSQLDialect(Value: Integer);
begin
  if (Value < 1) then FBLError(E_DB_SQLDIALECT_INVALID);
  if (FDBHandle = nil) then
    FSQLDialect := Value
  else if Value <= DBSQLDialect then
    FSQLDialect := Value
  else
    FBLError(E_DB_SQLDIALECT_INVALID)
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.AddTransaction(ATransaction: IFBLDbEvent);
begin
  FTrans.Add(Pointer(ATransaction));
end;

//------------------------------------------------------------------------------

procedure TFBLDatabase.RemoveTransaction(ATransaction: IFBLDbEvent);
var
  i: Integer;
begin
  i := FTrans.IndexOf(Pointer(ATransaction));
  if (i > -1) then
    FTrans.Delete(i);
end;

(* ---------------------------------------------------------------------------*)
(*                   Db info                                                  *)
(*----------------------------------------------------------------------------*)

function TFBLDatabase.GetDBSQLDialect: Integer;
var
  bOk: Boolean;
begin
  Result := GetDBInfoInt2(isc_info_db_sql_Dialect, bOk);
  if not bOk then  Result := 1;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetVersion: string;
var
  Status_vector: ISC_STATUS_VECTOR;
  Buffer: array[0..1023] of Char;
  DBInfo: Char;
begin
  DBInfo := Char(isc_info_version);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    SizeOf(buffer), Buffer) <> 0 then
    FBLShowError(@Status_vector);

  SetLength(Result, integer(buffer[4]));
  Move(Buffer[5], Result[1], Integer(Buffer[4]));
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDBSiteName: string;
var
  Status_vector: ISC_STATUS_VECTOR;
  buffer: array[0..1023] of char;
  DBInfo: Char;
  Pos, Len: Integer;
begin
  DBInfo := Char(isc_info_db_id);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    SizeOf(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);

  //in buffer[4] length of DBfilename
  //in buffer[5 + (integer(buffer[4])] length DBSiteName
  len := integer(buffer[5 + integer(buffer[4])]);
  pos := 6 + integer(buffer[4]);
  SetLength(Result, len);
  Move(Buffer[pos], Result[1], len);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDBFileName: string;
var
  Status_vector: ISC_STATUS_VECTOR;
  buffer: array[0..1023] of char;
  DBInfo: char;
begin
  DBInfo := char(isc_info_db_id);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  //in buffer[4] length of DBfilename
  SetLength(Result, integer(buffer[4]));
  Move(Buffer[5], Result[1], integer(buffer[4]));
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetLocalConnection: Boolean;
var
  Status_vector: ISC_STATUS_VECTOR;
  buffer: array[0..1023] of char;
  DBInfo: char;
begin
  DBinfo := char(isc_info_db_id);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);

  if Buffer[3] = #2 then   // '#2 for local connection
    Result := True       //  #4 for remote connection
  else
    Result := False;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDBInfoInt2(isc_info: integer; var ok: Boolean): integer;
var
  Status_vector: ISC_STATUS_VECTOR;
  buffer: array[0..511] of char;
  length: integer;
  DBInfo: char;
begin
  DBInfo := char(isc_info);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  ok := (buffer[0] = char(isc_info));
  length := isc_vax_integer(@buffer[1], 2);
  Result := isc_vax_integer(@buffer[3], length);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDBInfoInt(isc_info: integer): integer;
var
  bOk: Boolean;
begin
  Result := GetDbInfoInt2(isc_info, bOk);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetPageSize: integer;
begin
  Result := GetDBInfoInt(isc_info_page_size);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetProviderInfo: TProviderInfo;
var
  bOk: Boolean;
  Res: integer;
begin
  Result := piUnknow;
  Res := GetDBInfoInt2(isc_info_db_provider, bOk);
  if bOk then
  begin
    case Res of
      isc_info_db_code_interbase: Result := piInterBase;
      isc_info_db_code_firebird: Result := piFireBird;
    end;
  end;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetServerInfo: TServerInfo;
var
  bOk: Boolean;
  Res: Integer;
begin
  Result := siUnknow;
  Res := GetDBInfoInt2(isc_info_db_class, bOk);
  if bOk then
  begin
    case Res of
      isc_info_db_class_classic_access: Result := siClassicServer;
      isc_info_db_class_server_access: Result := siSuperServer;
    end;
  end;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetReads: Integer;
begin
  Result := GetDBInfoInt(isc_info_reads);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetWrites: Integer;
begin
  Result := GetDBInfoInt(isc_info_writes);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetFetches: Integer;
begin
  Result := GetDBInfoInt(isc_info_fetches);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetODSVersion: Integer;
begin
  Result := GetDBInfoInt(isc_info_ods_version);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetODSMinorVersion: Integer;
begin
  Result := GetDBInfoInt(isc_info_ods_minor_version);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetBaseLevel: Integer;
var
  buffer: array[0..511] of char;
  DBInfo: char;
  Status_vector: ISC_STATUS_VECTOR;
begin
  DBInfo := char(isc_info_base_level);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  Result := isc_vax_integer(@buffer[4], 1);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetImplementationNumber: Integer;
var
  buffer: array[0..511] of char;
  DBInfo: char;
  Status_vector: ISC_STATUS_VECTOR;
begin
  DBInfo := char(isc_info_implementation);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    Sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  Result := isc_vax_integer(@buffer[3], 1);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetImplementationClass: Integer;
var
  buffer: array[0..511] of char;
  DBInfo: char;
  Status_vector: ISC_STATUS_VECTOR;
begin
  DBInfo := char(isc_info_implementation);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  Result := isc_vax_integer(@buffer[4], 1);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetCurrentMemory: Integer;
begin
  Result := GetDBInfoInt(isc_info_current_memory);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetMaxMemory: Integer;
begin
  Result := GetDBInfoInt(isc_info_max_memory);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetAllocation: Integer;
begin
  Result := GetDBInfoInt(isc_info_allocation);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetNumBuffers: Integer;
begin
  Result := GetDBInfoInt(isc_info_num_buffers);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetSweepInterval: Integer;
begin
  Result := GetDBInfoInt(isc_info_sweep_interval);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetOperationCounts(DBInfocommand: Integer;
  FOperation: TStringList): TStringList;
var
  buffer: array[0..10239] of char;
  DBInfo: char;
  i, qtd_tables, id_table, qtd_operations: Integer;
  Status_vector: ISC_STATUS_VECTOR;
begin
  if FOperation = nil then FOperation := TStringList.Create;
  Result := FOperation;
  DBInfo := char(DBInfoCommand);
  if isc_database_info(@Status_Vector, @FDBHandle, 1, @DBInfo,
    sizeof(buffer), buffer) <> 0 then
    FBLShowError(@Status_vector);
  FOperation.Clear;
  // 1. 1 byte specifying the item type requested (e.g., isc_info_insert_count).
  // 2. 2 bytes telling how many bytes compose the subsequent value pairs.
  // 3. A pair of values for each table in the database on wich the requested
  //    type of operation has occurred since the database was last attached.
  // Each pair consists of:
  // 1. 2 bytes specifying the table ID.
  // 2. 4 bytes listing the number of operations (e.g., inserts) done on that table.
  qtd_tables := trunc(isc_vax_integer(@buffer[1], 2) / 6);
  for i := 0 to qtd_tables - 1 do
  begin
    id_table := isc_vax_integer(@buffer[3 + (i * 6)], 2);
    qtd_operations := isc_vax_integer(@buffer[5 + (i * 6)], 4);
    FOperation.Add(IntToStr(id_table) + '=' + IntToStr(qtd_operations));
  end;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetBackoutCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_backout_count, FBackoutCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetDeleteCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_delete_count, FDeleteCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetExpungeCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_expunge_count, FExpungeCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetInsertCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_insert_count, FInsertCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetPurgeCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_purge_count, FPurgeCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetReadIdxCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_read_idx_count, FReadIdxCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetReadSeqCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_read_seq_count, FReadSeqCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetUpdateCount: TStringList;
begin
  Result := GetOperationCounts(isc_info_update_count, FUpdateCount);
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetClientVersion: string;
var
  Buffer: array[0..255] of Char;
begin
  Result := '';
  if GetIbSqlClientVersion = 7 then
  begin
    isc_get_client_version(Buffer);
    Result := string(Buffer);
  end;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetClientMajorVersion: Integer;
begin
  Result := 0;
  if GetIbSqlClientVersion = 7 then
    Result := isc_get_client_major_version;
end;

//------------------------------------------------------------------------------

function TFBLDatabase.GetClientMinorVersion: Integer;
begin
  Result := 0;
  if GetIbSqlClientVersion = 7 then
    Result := isc_get_client_minor_version;
end;

//------------------------------------------------------------------------------

end.
