 (*
   Firebird Library
   File:FBLExcept.pas
   Copyright (c) 2002 Alessandro Batisti
   abatisti@tiscali.it
   http://web.tiscali.it/fblib

  * This file is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * This file is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
*)

{$I fbl.inc}
unit FBLExcept;


interface

uses
  SysUtils, ibase_h, FBLConst;


type
  EFBLError = class(Exception)
  private
    FSQLCode: Long;
    FError_code: Long;
  public
    constructor Create(ASQLCode: Long; AErrorCode: Long; AMsg: string); overload;
    constructor CreateFmt(AMsg: string; const AArgs: array of const); overload;
    constructor Create(AMsg: string); overload;
    property SQLCode: Long read FSQLCode write FSQLCode;
    property ISC_ErrorCode: Long read FError_Code;
  end;


  ISC_STATUS_VECTOR = array[0..19] of ISC_STATUS;
  PSTATUS_VECTOR = ^ISC_STATUS_VECTOR;

procedure FBLError(AErr: string; const AArgs: array of const); overload;
procedure FBLError(AErr: string); overload;
procedure FBLShowError(AStatusVectorPtr: PSTATUS_VECTOR);


implementation

procedure FBLShowError(AStatusVectorPtr: PSTATUS_VECTOR);
var
  MsgBuffer: array[0..511] of char;
  ErrMsg: string;
  LastMsg: string;
  ErrCode: ISC_STATUS;
  SQLCode: ISC_LONG;
  IbErr: integer;
begin
  ErrMsg := '';
  LastMsg := '';
  IbErr := AStatusVectorPtr[1];
  SQLCode := isc_sqlcode(@AStatusVectorPtr);
  repeat
    ErrCode := isc_interprete(@msgBuffer, @AStatusVectorPtr);
    if LastMsg <> string(MsgBuffer) then
    begin
      LastMsg := string(MsgBuffer);
      if Length(ErrMsg) <> 0 then
        ErrMsg := ErrMsg + NEW_LINE;
      ErrMsg := ErrMsg + LastMsg;
    end;
  until ErrCode = 0;
  raise EFBLError.Create(SQLCode, IbErr, ErrMsg);
end;


(* EFBLEerror *)


constructor EFBLError.Create(ASQLCode: Long; AErrorCode: Long; AMsg: string);
begin
  inherited Create(AMsg);
  FSQLCode := ASQLCode;
  FError_Code := AErrorCode;
end;

//------------------------------------------------------------------------------

constructor EFBLError.CreateFmt(AMsg: string; const AArgs: array of const);
begin
  inherited CreateFmt(AMsg, AArgs);
  FSQLCode := 0;
  FError_Code := 0;;
end;

//------------------------------------------------------------------------------

constructor EFBLError.Create(AMsg: string);
begin
  inherited Create(AMsg);
  FSQLCode := 0;
  FError_Code := 0;;
end;

//------------------------------------------------------------------------------

procedure FBLError(AErr: string; const AArgs: array of const);
begin
  raise EFBLError.CreateFmt(AErr, AArgs);
end;

//------------------------------------------------------------------------------

procedure FBLError(AErr: string);
begin
  raise EFBLError.Create(AErr);
end;

//------------------------------------------------------------------------------

end.
