unit TGReal48;

interface

type Real48=packed record
       Exponent:Byte;
       MantissaLow:LongWord;
       SignMantissaHigh:Byte;

       procedure Normalize;

       function Sign:Int64;
       function Mantissa:Int64;
       class operator Implicit(a: Real48): Double;
       class operator Implicit(a: Double): Real48;

       class operator Divide(a,b:Real48):Real48;
       class operator Multiply(a,b:Real48):Real48;
       class operator Subtract(a,b:Real48):Real48;
       class operator Add(a,b:Real48):Real48;

       class operator Negative(a:Real48):Real48;
       class operator Inc(a:Real48):Real48;
       class operator Dec(a:Real48):Real48;
       class operator Trunc(a:Real48):Int64;
       class operator Round(a:Real48):Int64;

       class operator Equal(const Left, Right: Real48): Boolean;
       class operator NotEqual(const Left, Right: Real48): Boolean;
       class operator LessThan(const Left, Right: Real48): Boolean;
       class operator LessThanOrEqual(const Left, Right: Real48): Boolean;
       class operator GreaterThan(const Left, Right: Real48): Boolean;
       class operator GreaterThanOrEqual(const Left, Right: Real48): Boolean;
       end;

implementation

var MantissaFactor:Double=0;

procedure Real48.Normalize;
begin
  if Exponent=0 then begin
     MantissaLow:=0;
     SignMantissaHigh:=0;
     end;
  end;

function Real48.Sign:Int64;
begin
  if SignMantissaHigh and $80 <> 0 then
     Result:=-1
  else
     Result:=1;
  end;

function Real48.Mantissa:Int64;
begin
  Result:=MantissaLow+Int64($100000000)*Int64(SignMantissaHigh and $7F);
  end;

class operator Real48.Implicit(a: Real48): Double;
begin
  if a.Exponent=0 then
     Result:=0
  else
     Result:=a.Sign * Math.Pow(2,a.Exponent-129) * (1+a.Mantissa/MantissaFactor);
  end;

class operator Real48.Implicit(a: Double): Real48;
var bsign,mantissa,lowmant:Int64;
begin
  if a=0 then begin
     Result.Exponent:=0;
     Result.MantissaLow:=0;
     Result.SignMantissaHigh:=0;
     Exit;
     end;
  if a<0 then begin
     bsign:=1;  // negative
     a:=-a; // make positive for further calculations
     end
  else
     bsign:=0; // positive
  Result.Exponent:=Trunc(Math.Log(a,2)+129);
  a:=a / Math.Pow(2,Result.Exponent-129);
  mantissa:=Trunc((a-1)*MantissaFactor);

  Result.MantissaLow:=mantissa mod $100000000;
  lowmant:=mantissa shr 32;
  if lowmant>127 then
     raise Exception.Create('Mantissa Too Long');
  Result.SignMantissaHigh:=lowmant + bsign*$80;
  end;

class operator Real48.Divide(a,b:Real48):Real48;
begin
  Result:=Double(a)/Double(b);
  end;

class operator Real48.Multiply(a,b:Real48):Real48;
begin
  Result:=Double(a)*Double(b);
  end;

class operator Real48.Subtract(a,b:Real48):Real48;
begin
  Result:=Double(a)-Double(b);
  end;

class operator Real48.Add(a,b:Real48):Real48;
begin
  Result:=Double(a)+Double(b);
  end;

class operator Real48.Negative(a:Real48):Real48;
begin
  Result:=a;
  if a.Sign=1 then
     Result.SignMantissaHigh:=Result.SignMantissaHigh or $80
  else
     Result.SignMantissaHigh:=Result.SignMantissaHigh and $7F;
  Result.Normalize;
  end;

class operator Real48.Inc(a:Real48):Real48;
begin
  Result:=Double(a)+1.0;
  end;

class operator Real48.Dec(a:Real48):Real48;
begin
  Result:=Double(a)-1.0;
  end;

class operator Real48.Trunc(a:Real48):Int64;
begin
  Result:=Trunc(Double(a));
  end;

class operator Real48.Round(a:Real48):Int64;
begin
  Result:=Round(Double(a));
  end;

class operator Real48.Equal(const Left, Right: Real48): Boolean;
begin
  Result:=(Left.Exponent=Right.Exponent) and
          (Left.MantissaLow=Right.MantissaLow) and
          (Left.SignMantissaHigh=Right.SignMantissaHigh);
  end;

class operator Real48.NotEqual(const Left, Right: Real48): Boolean;
begin
  Result:=not (Left=Right);
  end;

class operator Real48.LessThan(const Left, Right: Real48): Boolean;
begin
  Result:=Double(Left)<Double(Right);
  end;

class operator Real48.LessThanOrEqual(const Left, Right: Real48): Boolean;
begin
  Result:=Double(Left)<=Double(Right);
  end;

class operator Real48.GreaterThan(const Left, Right: Real48): Boolean;
begin
  Result:=Double(Left)>Double(Right);
  end;

class operator Real48.GreaterThanOrEqual(const Left, Right: Real48): Boolean;
begin
  Result:=Double(Left)>=Double(Right);
  end;

initialization
  MantissaFactor:=Math.Pow(2,39);
  end.
