unit MainUnit;
{

 Raw Disk Access Demonstration Application
 =========================================
 Created by Tom Honaker. Uses the Raw Disk Access unit file,
 which is based on the Delphi Pascal (.PAS) file VWIN32,
 created by Geir Wikran. Since Geir was so kind as to release
 his code to the open-source community, Tom will do likewise.
 :-)

 unit       : RawDiskAccess Demo Application
 version    : 1.0
 last update: 2004/03/25
 written for: Delphi 7+ (but should work on 6)
 written by : Tom Honaker
 e-mail     : general@e-f-w.com

 Please note that while RDA is made freely available, Geir's
 license terms apply to RDA - and to this component - as well:
 "This source code is freeware. You may use, change, and
 distribute without charge the source code as you like. This
 unit can be used freely in any commercial applications.
 However, it may not be sold as a standalone product and the
 source code may not be included in a commercial product.
 This unit is provided as is with no warrent [sic] or
 support. Make sure to read relevant information and
 documentation from Microsoft before using this unit."
 In other words, you may use RDA for any commercial or
 private project, but you must NOT include the source itself
 in any commercial project, nor can you sell the source.


 About RawDiskAccess

 RawDiskAccess is a combined, "universal" raw disk access
 provider for Delphi 7 applications. RDA expands on
 VWIN32.PAS by providing the same services for NT-kernel
 Windows versions as VWIN32.PAS does for 9x-kernel versions.
 This system automatically determines the kernel family and
 specific OS version information (down to the service pack
 number), and from there chooses and uses the proper methods
 for direct disk access. Where possible, RDA uses Win32 API
 calls to get information, so as to not invoke the Vwin32
 VxD, or open the disk directly, unnecessarily.

 This application uses RDA directly, and also uses the RDA
 frontend VCL component.
 
}
interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics,
  Controls, Forms, Dialogs, ComCtrls, FileCtrl, StdCtrls,
  Grids, MPHexEditor, MPHexEditorEx, StrUtils, Spin,
  ExtCtrls, RawDiskAccessComponent;

type
  TMainForm = class(TForm)
    MainPageControl: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet3: TTabSheet;
    ClusterPageControl: TPageControl;
    TabSheet4: TTabSheet;
    TabSheet5: TTabSheet;
    DCDriveInfo: TMemo;
    DCDrive: TDriveComboBox;
    DCClusterContents: TMPHexEditor;
    Label1: TLabel;
    SaveClusterButton: TButton;
    Label2: TLabel;
    DCCluster: TSpinEdit;
    GroupBox1: TGroupBox;
    OSInfo: TMemo;
    GroupBox2: TGroupBox;
    Memo1: TMemo;
    DCScroll: TScrollBar;
    Image1: TImage;
    Label3: TLabel;
    ReadFailedPanel: TPanel;
    RawDiskAccessComponent: TRawDiskAccessComponent;
    procedure DCDriveChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure DCClusterChange(Sender: TObject);
    procedure SaveClusterButtonClick(Sender: TObject);
    procedure DCScrollChange(Sender: TObject);
    procedure DCClusterContentsKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    StartingUp: Boolean;
  public
    { We don't need no steenkin' declarations! }
  end;

var
  MainForm: TMainForm;

implementation

uses RawDiskAccess;

{$R *.dfm}

procedure TMainForm.DCDriveChange(Sender: TObject);
var
  // Drive IDing
  tDrvLetter: string;
  tDrv: Char;

  // Various types we'll be using
  bNTFS: TNTFSBootRecord;
  bFAT: TFAT32BootRecord;
// NOTE: The following commented-out code is needed if not
// using the RDA frontend component.
//
//  MID: TMediaIdentifier;
//  DSI: TDiskSizingInformation;

  // Drive cluster size, as derived from conventional API
  // calls
  ClusSize: Integer;

begin
  // Bail if the app's starting up...
  if (StartingUp) then Exit;

  // Grab the drive letter...
  tDrvLetter := AnsiUppercase(DCDrive.Drive);
  tDrv := tDrvLetter[1];

  // Set the RDA frontend's drive letter - this will trigger
  // it to obtain the drive's basic information...
  RawDiskAccessComponent.LogicalDrive := DCDrive.Drive;

// NOTE: The following commented-out code is needed if not
// using the RDA frontend component.
//
//  // Acquire the drive's info...
//  DSI := RDA.RDAFindDiskSizeInformation(DCDrive.Drive + ':\');
//  MID := RDA.RDAGetMediaIdentifier(DCDrive.Drive);
//  if (MID.VolumeLabel = '') then MID.VolumeLabel := '<none>';

  // Clear the list..
  DCDriveInfo.Clear;

  // First, add basic info on the drive...
  DCDriveInfo.Lines.Append('Drive Info'); 
  DCDriveInfo.Lines.Append('  Volume Label: ' + RawDiskAccessComponent.VolumeName {MID.VolumeLabel});
  DCDriveInfo.Lines.Append('  Serial Number: ' + Format('%8.8x', [RawDiskAccessComponent.SerialNumber {MID.SerialNumber}]));
  DCDriveInfo.Lines.Append('  Filesystem: ' + RawDiskAccessComponent.FileSystem {MID.FileSysType});

  // Next, add info taken from the boot sector of the drive...
  DCDriveInfo.Lines.Append(' ');
  DCDriveInfo.Lines.Append('Boot Sector Info');

  // UDF filesystem - CD/DVD data track...
  if (AnsiContainsText(RawDiskAccessComponent.FileSystem {MID.FileSysType}, 'UDF')) then
  begin
    DCDriveInfo.Lines.Append('  CD-/DVD-ROM filesystem.');
    DCDriveInfo.Lines.Append('  Boot sector not available, but raw clusters can still be read.');
  end;

  // Unspecified filesystem - nothing in the drive...
  if (RawDiskAccessComponent.FileSystem {MID.FileSysType} = '') then
    DCDriveInfo.Lines.Append('  No disk present in this drive.');

  // FATxx filesystem...
  if (AnsiContainsText(RawDiskAccessComponent.FileSystem {MID.FileSysType}, 'FAT')) then
  begin
    // Note that we're performing two checks - one to get the
    // kernel family, and the other to see if we can read the
    // first cluster of the drive to get its particulars.
    if ((RDA.RDAKernelFamily = 'NT') and (RDA.RDAReadNTSectors((Ord(DCDrive.Drive) - 96), 0, 1, @bFAT))) or
       ((RDA.RDAKernelFamily = '9x') and (RDA.RDARead9xSectors((Ord(DCDrive.Drive) - 96), 0, 1, @bFAT)))
    then
    begin
      DCDriveInfo.Lines.Append('  OEM Name:  ' + bFAT.OemName);
      DCDriveInfo.Lines.Append('  Bytes/Sector:  ' + IntToStr(bFAT.BytesPerSec));
      DCDriveInfo.Lines.Append('  Sectors/Cluster:  ' + IntToStr(bFAT.SecPerClust));
      DCDriveInfo.Lines.Append('  Reserved Sectors:  ' + IntToStr(bFAT.ReservedSectors));
      DCDriveInfo.Lines.Append('  # of FAT Copies:  ' + IntToStr(bFAT.NumOfFatCopies));
      DCDriveInfo.Lines.Append('  Root Entry Count Limit:  ' + IntToStr(bFAT.MaxRootEntries));
      DCDriveInfo.Lines.Append('  Old Sectors:  ' + IntToStr(bFAT.SectorsOld));
      DCDriveInfo.Lines.Append('  Media Descriptor:  ' + IntToStr(bFAT.MediaDescriptor));
      DCDriveInfo.Lines.Append('  Old Sectors/FAT:  ' + IntToStr(bFAT.SectorsPerFatOld));
      DCDriveInfo.Lines.Append('  Sectors/Track:  ' + IntToStr(bFAT.SectorsPerTrack));
      DCDriveInfo.Lines.Append('  Heads:  ' + IntToStr(bFAT.NumOfHeads));
      DCDriveInfo.Lines.Append('  Hidden Sectors:  ' + IntToStr(bFAT.HiddenSectors));
      DCDriveInfo.Lines.Append('  Sector Count:  ' + IntToStr(bFAT.SectorsSize));
      DCDriveInfo.Lines.Append('  Sectors/FAT:  ' + IntToStr(bFAT.SectorsPerFat));
      DCDriveInfo.Lines.Append('  Flags:  ' + IntToStr(bFAT.Flags));
      DCDriveInfo.Lines.Append('  Version:  ' + IntToStr(bFAT.Version));
      DCDriveInfo.Lines.Append('  Root Directory Start Cluster:  ' + IntToStr(bFAT.RootDirStartClust));
      DCDriveInfo.Lines.Append('  Filesystem Info Sector:  ' + IntToStr(bFAT.FileSystemInfoSec));
      DCDriveInfo.Lines.Append('  FAT Backup Sector:  ' + IntToStr(bFAT.BackupSec));
      DCDriveInfo.Lines.Append('  Drive Number:  ' + IntToStr(bFAT.DriveNumber));
      DCDriveInfo.Lines.Append('  Extra Signature:  ' + IntToStr(bFAT.ExtraSignature));
      DCDriveInfo.Lines.Append('  Drive Ext. Serial Number:  ' + Format('%8.8x', [bFAT.SerialNumber]));
      DCDriveInfo.Lines.Append('  Volume Name:  ' + bFAT.VolumeName);
      DCDriveInfo.Lines.Append('  FAT Name:  ' + bFAT.FATName);

      DCCluster.MaxValue := Round(bFAT.SectorsSize / bFAT.SecPerClust);
      DCScroll.Max := DCcluster.MaxValue;
    end;
  end;

  // NTFS filesystem...
  if (AnsiContainsText(RawDiskAccessComponent.FileSystem {MID.FileSysType}, 'NTFS')) then
  begin
    if (RDA.RDAReadNTSectors((Ord(DCDrive.Drive) - 96), 0, 1, @bNTFS)) then
    begin
      DCDriveInfo.Lines.Append('  OEM Name:  ' + bNTFS.OemName);
      DCDriveInfo.Lines.Append('  Bytes/Sector:  ' + IntToStr(bNTFS.BytesPerSec));
      DCDriveInfo.Lines.Append('  Sectors/Cluster:  ' + IntToStr(bNTFS.SecPerClust));
      DCDriveInfo.Lines.Append('  Reserved Sectors:  ' + IntToStr(bNTFS.ReservedSectors));
      DCDriveInfo.Lines.Append('  Media Descriptor:  ' + IntToStr(bNTFS.MediaDescriptor));
      DCDriveInfo.Lines.Append('  Sectors/Track:  ' + IntToStr(bNTFS.SectorsPerTrack));
      DCDriveInfo.Lines.Append('  Heads:  ' + IntToStr(bNTFS.NumOfHeads));
      DCDriveInfo.Lines.Append('  Hidden Sectors:  ' + IntToStr(bNTFS.HiddenSectors));
      DCDriveInfo.Lines.Append('  Total Sector Count:  ' + IntToStr(bNTFS.SectorSize));
      DCDriveInfo.Lines.Append('  Drive Ext. Serial Number:  ' + Format('%8.8x', [bNTFS.SerialNumber]));
      DCDriveInfo.Lines.Append('  Boot Block Checksum:  ' + Format('%8.8x', [bNTFS.Checksum]));
      DCDriveInfo.Lines.Append('  ');
      DCDriveInfo.Lines.Append('  Master File Table Info');
      DCDriveInfo.Lines.Append('    Starting Cluster:  ' + IntToStr(bNTFS.MTFClust));
      DCDriveInfo.Lines.Append('    Mirror Starting Cluster:  ' + IntToStr(bNTFS.MTFMirrotClust));
      DCDriveInfo.Lines.Append('    File Record Size (Clusters):  ' + IntToStr(bNTFS.ClustersPerFileRec));
      DCDriveInfo.Lines.Append('    Clusters/Index Block:  ' + IntToStr(bNTFS.ClustersPerIndexBlock));

      DCCluster.MaxValue := Round(bNTFS.SectorSize / bNTFS.SecPerClust);
      DCScroll.Max := DCcluster.MaxValue;
    end else begin
      DCDriveInfo.Lines.Append('  ERROR - Unable to read boot sector!');
    end;
  end;

  // Add the cluster count to the clusters list, and enable it...
  DCCluster.MinValue := 0;
  DCCluster.Value := 0;
  DCCluster.Enabled := True;
  DCClusterChange(Self);
end;

procedure TMainForm.FormCreate(Sender: TObject);
var
  OSV: TWinVersion;

begin
  StartingUp := True;

  try
    OSV := RDA.RDAAcquireOSVersion;
  except
    ;
    // Yes, we're eating any errors. The reason being that
    // there are hundreds of things that can throw an exception
    // in the OS version detect, and almost none of them prevent
    // an accurate result.
  end;

  // Start populating our OS info list...
  OSInfo.Clear;
  OSInfo.Lines.Append(' Detected OS:  ' + OSV.WinName);
  OSInfo.Lines.Append(' Kernel Family:  ' + OSV.KernelType);
  OSInfo.Lines.Append(' Service Pack:  ' + OSV.ServicePack);
  OSInfo.Lines.Append(' Version/Build:  ' + IntToStr(OSV.MajorVer) + '.' + IntToStr(OSV.MinorVer) + '.' + IntToStr(OSV.BuildNum));

  // Force the tabs to specific pages...
  MainPageControl.ActivePageIndex := 0;
  ClusterPageControl.ActivePageIndex := 0;
end;

procedure TMainForm.FormActivate(Sender: TObject);
begin
  StartingUp := False;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  StartingUp := True;
end;

procedure TMainForm.DCClusterChange(Sender: TObject);
// NOTE: The following commented-out code is needed if not
// using the RDA frontend component.
//
//var
//  DSI: TDiskSizingInformation;
//  ClusStream: TMemoryStream;
//  bDrv: Byte;
//  TempStr: string;

begin
  // For this portion of the demonstration, we will use the
  // RDA frontend component. However, we'll also include code
  // for doing the same thing via RDA itself without using the
  // frontend component, although that code will be commented
  // out.


  RawDiskAccessComponent.LogicalDrive := DCDrive.Drive;

// NOTE: The following commented-out code is needed if not
// using the RDA frontend component.
//
//  // Get the basic sizing info off the drive...
//  DSI := RDA.RDAFindDiskSizeInformation(DCDrive.Drive + ':\');

  // Wrap the spinedit...
  if DCCluster.Value < 0 then
    DCCluster.Value := RawDiskAccessComponent.TotalClusters {DSI.TotalClusters} - 1;
  if DCCluster.Value > RawDiskAccessComponent.TotalClusters {DSI.TotalClusters} then
    DCCluster.Value := 0;

  // Update the scrollbar...
  DCScroll.Position := DCCluster.Value;

  // Disable the save-cluster button...
  SaveClusterButton.Enabled := False;

  // Set the RDA frontend to read one cluster, starting at
  // the proper cluster nubber...
  RawDiskAccessComponent.ClusterNumber := DCCluster.Value;
  RawDiskAccessComponent.ClusterCount := 1;

  // Read the cluster...
  RawDiskAccessComponent.ReadDrive;

  // Check to see if the read succeeded...
  if (RawDiskAccessComponent.StatusCode = 9999) then
  begin
    // Yep, read was successful, so dump the disk buffer into
    // the hex editor...
    DCClusterContents.Enabled := False;
    DCClusterContents.LoadFromStream(RawDiskAccessComponent.DiskBufferStream);
    DCClusterContents.Enabled := True;
  end else begin
    // Nope, the read failed, so clear the hex editor...
    DCClusterContents.Enabled := False;
    DCClusterContents.AsText := '';
    DCClusterContents.Enabled := True;

    // Show the read-failed panel...
    ReadFailedPanel.Visible := True;
    ReadFailedPanel.BringToFront;
  end;

// NOTE: The following commented-out code is needed if not
// using the RDA frontend component.
//
  {

  // How to do the same thing without the RDA frontend:


  // Get the drive letter as a byte. Note that the letter is
  // passed out of the DriveComboBox as a lowercase
  // character...
  bDrv := Ord(DCDrive.Drive) - 96;

  // Create the memory stream that will be used as our buffer...
  ClusStream := TMemoryStream.Create;
  ClusStream.SetSize(DSI.SectorSize * DSI.SectorsPerCluster);

  // Get the cluster's contents...
  if (RDA.RDAReadClusters(bDrv, DCCluster.Value, 1, ClusStream.Memory)) then
  begin
    // If the read succeeded, throw the contents into the hex
    // editor...
    DCClusterContents.Enabled := False;
    DCClusterContents.LoadFromStream(ClusStream);
    DCClusterContents.Enabled := True;

    // Make sure our read-failed panel isn't visible...
    ReadFailedPanel.Visible := False;
  end else begin
    // Clear the hex editor...
    DCClusterContents.Enabled := False;
    DCClusterContents.AsText := '';
    DCClusterContents.Enabled := True;

    // Show the read-failed panel...
    ReadFailedPanel.Visible := True;
    ReadFailedPanel.BringToFront;
  end;

  // Free up the memory stream...
  ClusStream.Free;
  }
end;

procedure TMainForm.SaveClusterButtonClick(Sender: TObject);
var
  MBret: Integer;
  ClusStream: TMemoryStream;
  bDrv: Byte;

begin
  // Note that here we'll show how to accomplish a direct disk
  // write with RDA alone. It'd take less code to use the RDA
  // frontend control and let it manage all the minutae, but
  // that'd not demonstrate the process for pure RDA, would it?
  // <grin>
  

  // Ask if this is a good idea...
  MBret := Application.MessageBox('You are about to DIRECTLY WRITE TO A DRIVE. Are you sure you want to do this?', 'Are You Sure?', MB_ICONQUESTION or MB_YESNO);

  // If the answer is "oh s**!" we bail without touching
  // anything...
  if MBret = ID_NO then Exit;

  // Create the memory stream...
  ClusStream := TMemoryStream.Create;

  // Get the revised cluster contents from the hex editor...
  DCClusterContents.SaveToStream(ClusStream);

  // Get the drive letter as a byte. Note that the letter is
  // passed out of the DriveComboBox as a lowercase
  // character...
  bDrv := Ord(DCDrive.Drive) - 96;

  // Write the stream to disk, and report the operation's
  // result...
  if (RDA.RDAWriteClusters(bDrv, DCCluster.Value, 1, ClusStream.Memory)) then
  begin
    Application.MessageBox('Direct write succeeded.', 'RDA Direct Write', MB_ICONINFORMATION);
    DCClusterChange(Self);
  end else begin
    Application.MessageBox('Direct write failed.', 'RDA Direct Write', MB_ICONEXCLAMATION);
  end;

  // Free up the memory stream...
  ClusStream.Free;
end;

procedure TMainForm.DCScrollChange(Sender: TObject);
begin
  DCCluster.Value := DCScroll.Position;
end;

procedure TMainForm.DCClusterContentsKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (not StartingUp) then SaveClusterButton.Enabled := True;
end;

end.
