/********************************************************
              The part of Labit library
       Copyright (c) 2002-2004, Labit Corporation.
                  All rights reserved.
********************************************************/
// frausCommon.h
#pragma once

// windows includes
#include <windows.h>
#include <unknwn.h>

extern const GUID IID_IatException;
extern const GUID IID_IatUserDumpStreamWriter;
extern const GUID IID_IatUserDumpStreamReader;

#define	SV_FAILONEXCEPT		1		/**/
#define	SV_MAKELOG				2		/**/
#define  SV_UNWINDSTACK       4     /*Forces FailureAgent to Unwind stack information on exception*/
#define  SV_SENDLOG           8     /*Forces FailureAgent to Send Exception Log to developer*/

///////////////////////////////////////////
// interface IatUserDumpStreamWriter : public IUnknown
interface IatUserDumpStreamWriter : public IUnknown
{
	virtual HRESULT __stdcall addRawData(UINT_PTR data, long size) = 0;
};

//////////////////////////////////////////////////////
// interface IatUserDumpStreamReader : public IUnknown
interface IatUserDumpStreamReader : public IUnknown
{
   // returns the size of stream in bytes
   virtual LONG      __stdcall getStreamSize() = 0;
   // returns the pointer to memory location of raw data of users data stream
   virtual UINT_PTR  __stdcall getRawData() = 0;
   //
   virtual IUnknown* __stdcall getStream(const GUID & guid) = 0;
};

///////////////////////////////////////////
// interface IatException : public IUnknown
interface IatException : public IUnknown
{
	// return raw data about exception
	virtual UINT_PTR	__stdcall GetRawData() = 0;
	// returns the number of itsme on the stack
	virtual int			__stdcall GetStackSize() = 0;
	// returns the code of exception
	virtual int			__stdcall GetExceptionCode() = 0;
	// returns the type of exception ( general, user, etc )
	virtual int			__stdcall GetExceptionType() = 0;
	// returns the address of exception
	virtual UINT_PTR	__stdcall GetExceptionAddress() = 0;
	// returns the exception's ebp
	virtual UINT_PTR	__stdcall GetEbp() = 0;
	// returns the exception's esp
	virtual UINT_PTR	__stdcall GetEsp() = 0;
	// returns the exception's eip
	virtual UINT_PTR	__stdcall GetEip() = 0;
	// returns the exception's flag
	virtual int			__stdcall GetExceptionFlag() = 0;
	// returns the address fro the stack by its index
	virtual UINT_PTR	__stdcall GetStackItem(int index) = 0;
	// returns the thread's Id there the exception occured
	virtual long		__stdcall GetThreadId() = 0;
	// returns the stack base for the exception's thread
	virtual UINT_PTR	__stdcall GetStackBase() = 0;
	// returns if the exception is a nested exception
	virtual long		__stdcall IsNestedException() = 0;
};

extern const GUID IID_IatExceptionListener;
////////////////////////////////////////////////
// interface IatExceptionListener : public IUnknown
interface IatExceptionListener : public IUnknown
{
	// this method will be executed when FailureAgent.dll was unloaded
	virtual void    __stdcall OnFailureAgentUnloaded() = 0;
	// some exception has been raised
	virtual HRESULT __stdcall OnException(IatException * pException /*[in]*/) = 0;
	// unhandled exception occured
	virtual HRESULT __stdcall OnUnhandledException(IatException *pException) = 0;
	// the filter will generate memory dump, and want to get extended information to save in
	virtual HRESULT __stdcall OnDump( IatUserDumpStreamWriter* pUserDumpStream) = 0;
};

extern const GUID IID_IatExceptionfilter;
////////////////////////////////////////////////
// interface IatExceptionfilter : public IUnknown
interface IatExceptionfilter : public IUnknown
{
	// returns the number of locks
	virtual int			__stdcall GetLockCount() = 0;
	// Lock the exception filter and return a value that shows
	// current lock count
	virtual int			__stdcall Lock() = 0;
	// Unlick the exception filter and return value that shows
	// current lock countt
	virtual int			__stdcall Unlock() = 0;
	// adds a safe module ( the module that must be skipped from exception handling
	virtual HRESULT	__stdcall AddSafeModule(BSTR moduleName) = 0;
	// add s safe exception this exception will be skipped without any handling
	virtual HRESULT   __stdcall AddSafeException(int nExceptionCode) = 0;
	// delete a safe module
	virtual HRESULT	__stdcall DeleteSafeModule(BSTR moduleName) = 0;
	// delete a safe module
	virtual HRESULT	__stdcall DeleteSafeExceptionCode(int nExceptionCode) = 0;
	// clear safe module list
	virtual HRESULT   __stdcall ClearSafeModules() = 0;
	// clear safe exception list
	virtual HRESULT   __stdcall ClearSafeExceptionCodes() = 0;

	// register exception listener
	virtual HRESULT   __stdcall RegisterExceptionListener(IatExceptionListener *pListener) = 0;
	virtual HRESULT   __stdcall UnregisterExceptionListener(IatExceptionListener *pListener) = 0;
	// this method overrides setting that is stored in "FalureAgent.ini" file
	virtual HRESULT   __stdcall OverrideSettings(LONG settings) = 0;
	// this method returs currect settings
	virtual LONG      __stdcall GetCurrentSettings() = 0;
	// set/get proxy name
	virtual BSTR      __stdcall GetMailServer() = 0;
	virtual HRESULT   __stdcall SetMailServer(BSTR name) = 0;
	// set/get mail address
	virtual BSTR      __stdcall GetMailAddress() = 0;
	virtual HRESULT   __stdcall SetMailAddress(BSTR name) = 0;
	// set/get company name
	virtual BSTR      __stdcall GetCompanyName() = 0;
	virtual HRESULT   __stdcall SetCompanyName(BSTR name) = 0;
   // set/get message subject
   virtual BSTR      __stdcall GetMessageSubject() = 0;
   virtual HRESULT   __stdcall SetMessageSubject(BSTR subject) = 0;
};

typedef HRESULT (__stdcall *TGetExceptionFilter)(IatExceptionfilter** ppIntfs);
typedef BOOL (__stdcall *TSetExceptionFilter)(IatExceptionfilter* ppIntfs);