unit frausCommon;

interface

uses
  Windows, SysUtils;

const	SV_FAILONEXCEPT		     : Integer = 1;    //
const	SV_MAKELOG				     : Integer = 2;    //
const SV_UNWINDSTACK            : Integer = 4;    // Forces FailureAgent to Unwind stack information on exception
const SV_SENDLOG                : Integer = 8;    // Forces FailureAgent to Send Exception Log to developer

const EH_DELPHIEXCEPTIONS       : Integer = $EEDFADE;

const FAILUREAGENTDLLNAME    : PCHAR = 'failureagent.dll';
const GETEXCEPTIONFILTERNAME : PCHAR = 'GetExceptionFilter';
const SETEXCEPTIONFILTERNAME : PCHAR = 'SetExceptionFilter';

type
   ////////////////////////////////////////////
   // interface IatUserDumpStreamWriter : public IUnknown
   IatUserDumpStreamWriter = interface(IUnknown)
      ['{75B6ED85-2FD1-466a-8723-1B28FBF2E94F}']
		function addRawData(data : Cardinal; size : Cardinal) : HRESULT; stdcall;
   end;

	///////////////////////////////////////////
	// interface IatException : public IUnknown
	IatException = interface(IUnknown)
   	['{A47031DA-F978-4e6d-AFB5-8EC5F2F83719}']
   	  function getRawData() : DWORD; stdcall;
    // returns the number of itsme on the stack
      function getStackSize() : Integer; stdcall;
		// returns the code of exception
      function getExceptionCode() : Integer; stdcall;
		// returns the type of exception ( general, user, etc )
      function getExceptionType() : Integer; stdcall;
		// returns the address of exception
      function getExceptionAddress() : DWORD; stdcall;
		// returns the exception's ebp
      function getEbp() : DWORD; stdcall;
		// returns the exception's esp
      function getEsp() : DWORD; stdcall;
		// returns the exception's eip
      function getEip() : DWORD; stdcall;
		// returns the exception's flag
      function getExceptionFlag() : Integer; stdcall;
		// returns the address fro the stack by its index
      function getStackItem(index : Integer) : DWORD; stdcall;
		// returns the thread's Id there the exception occured
      function getThreadId() : DWORD; stdcall;
		// returns the stack base for the exception's thread
      function getStackBase() : DWORD; stdcall;
		// returns if the exception is a nested exception
      function isNestedException() : DWORD; stdcall;
   end;

	////////////////////////////////////////////////
	// interface IatExceptionListener : public IUnknown
   IatExceptionListener = interface(IUnknown)
   	['{168AC9D2-F35C-4dc1-9C7D-F9F05AFA4C86}']
    // this method will be executed when FailureAgent.dll was unloaded
      procedure OnFailureAgentUnloaded(); stdcall;
   	// some exception has been raised
   	  function onException ( const pException : IatException ) : HRESULT; stdcall;
		// unhandled exception occured
      function onUnhandledException( const pException: IatException ) : HRESULT; stdcall;
		// the filter will generate memory dump, and want to get extended information to save in
      function onDump( const pUserException : IatUserDumpStreamWriter ) : HRESULT; stdcall;
   end;

	////////////////////////////////////////////////
	// interface IatExceptionfilter : public IUnknown
   IatExceptionfilter = interface(IUnknown)
   	['{A61370D8-6B26-462d-B101-F0D12B181BD3}']
   	// returns the number of locks
      function getLockCount() : Integer; stdcall;
		// Lock the exception filter and return a value that shows
		// current lock count
      function lock() : Integer; stdcall;
		// Unlock the exception filter and return value that shows
		// current lock count
      function unlock() : Integer; stdcall;
      // adds a safe module ( the module that must be skipped from exception handling
      function addSafeModule( const moduleName : WideString ) : HRESULT; stdcall;
      // add safe exception this exception will be skipped without any handling
      function addSafeException(nExceptionCode : Integer) : HRESULT; stdcall;
      // delete a safe module
      function deleteSafeModule( const moduleName : WideString) : HRESULT; stdcall;
      // delete safe exception code
	   function deleteSafeExceptionCode(nExceptionCode : Integer) : HRESULT; stdcall;
	   // clear safe module list
	   function clearSafeModules() : HRESULT; stdcall;
   	// clear safe exception list
	   function clearSafeExceptionCodes() : HRESULT; stdcall;
      // register exception listener
      function registerExceptionListener(const pListener : IatExceptionListener) : HRESULT; stdcall;
      // unregister exception listener
      function UnregisterExceptionListener(const pListener : IatExceptionListener) : HRESULT; stdcall;
      // this method overrides setting that is stored in "FalureAgent.ini" file
      function overrideSettings(settings : DWORD) : HRESULT; stdcall;
	   // this method returs currect settings
      function getCurrentSettings() : DWORD; stdcall;
      // returns mail server where the dump files will be sent
      function getMailServer() : WideString; stdcall;
      // set mail server where the dump files will be sent
      function setMailServer( const name : WideString ) : HRESULT; stdcall;
      // returns the mail address where the dump files will be sent
      function getMailAddress() : WideString; stdcall;
      // set mail address where the dump files will be sent
      function setMailAddress( const name : WideString ) : HRESULT; stdcall;
      // returns the product publisher company name
      function getCompanyName() : WideString; stdcall;
      // set the product publisher company name
      function setCompanyName( const name : WideString ) : HRESULT; stdcall;
      // set/get message subject
      function getMessageSubject() WideString; stdcall;
      function setMessageSubject( const subject : WideString ) : HRESULT; stdcall;
   end;

   TGetExceptionFilter = function(var EhFilter : IatExceptionfilter) : HRESULT; stdcall;
   TSetExceptionFilter = function(const EhFilter : IatExceptionfilter) : BOOL; stdcall;

implementation

end.
