unit frmMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, frausCommon, ComCtrls, Mask;

type
  TForm1 = class(TForm)
    btnAV: TButton;
    btnMath: TButton;
    btnUser: TButton;
    btnHandledAV: TButton;
    btnHandledMath: TButton;
    btnHandledUser: TButton;
    btnSkipAV: TButton;
    btnSkipMath: TButton;
    btnSkipUser: TButton;
    btnLoadFalureAgent: TButton;
    btnUnloadFalureAgent: TButton;
    GroupBox1: TGroupBox;
    cbMakeLog: TCheckBox;
    cbUseUserDll: TCheckBox;
    cbFailOnExcept: TCheckBox;
    btnAplySettings: TButton;
    btnRegHandler: TButton;
    btnUnRegHandler: TButton;
    ListView1: TListView;
    Label1: TLabel;
    Label2: TLabel;
    btnAddSafeModule: TButton;
    btnRemoveSafeModule: TButton;
    Label3: TLabel;
    ceModule: TEdit;
    Label4: TLabel;
    ceEhType: TEdit;
    ListView2: TListView;
    btnClearTypeFilter: TButton;
    btnClearModuleFilter: TButton;
    btnRestoreAV: TButton;
    btnRestoreMath: TButton;
    btnRestoreUser: TButton;
    NestedHandledException: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnLoadFalureAgentClick(Sender: TObject);
    procedure btnUnloadFalureAgentClick(Sender: TObject);
    procedure btnAVClick(Sender: TObject);
    procedure btnMathClick(Sender: TObject);
    procedure btnUserClick(Sender: TObject);
    procedure btnHandledAVClick(Sender: TObject);
    procedure btnHandledMathClick(Sender: TObject);
    procedure btnHandledUserClick(Sender: TObject);
    procedure btnAplySettingsClick(Sender: TObject);
    procedure btnRegHandlerClick(Sender: TObject);
    procedure btnUnRegHandlerClick(Sender: TObject);
    procedure btnAddSafeModuleClick(Sender: TObject);
    procedure btnRemoveSafeModuleClick(Sender: TObject);
    procedure btnSkipAVClick(Sender: TObject);
    procedure btnSkipMathClick(Sender: TObject);
    procedure btnRestoreAVClick(Sender: TObject);
    procedure btnRestoreMathClick(Sender: TObject);
    procedure btnClearModuleFilterClick(Sender: TObject);
    procedure btnClearTypeFilterClick(Sender: TObject);
    procedure NestedHandledExceptionClick(Sender: TObject);
    procedure btnSkipUserClick(Sender: TObject);
    procedure btnRestoreUserClick(Sender: TObject);
  private
	 FFalureAgentModule  : THandle;
    FLoaded             : BOOL;
    FGetExceptionFilter : TGetExceptionFilter;
    FIatExceptionfilter : IatExceptionfilter;
  	 FExceptionListener  : IatExceptionListener;

    procedure unloadFalureAgent();
    procedure loadFalureAgent();
    procedure updateSettingsInfo();
    procedure changeControlsState(newState : Boolean);
    procedure cleanListView(lv : TListView);
    procedure addItemToListView(lv : TListView; const item : String);
    procedure cleanSafeModules();
    procedure cleanSafeExceptions();
    //
    procedure makeAVException();
    procedure makeIntegerDivisionException();
    procedure makeUserException();
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses ExceptionImpl;

const CFalureAgent = '..\..\..\..\bin\labitstudio\failureAgent.dll';
//const CFalureAgent = 'failureAgent.dll';
const CMsgExceptionHandled = 'Exception handled';

// some constsnt
const CAVExceptionCode   = $C0000005;
const CMathExceptionCode = $C0000094;

//error messages
const CErrFalureAgentLoading = 'Could not load FailureAgent.dll';
const CErrDoubleLoading      = 'The FailureAgent.dll is loaded, but loading requested.';
const CErrDoubleUnloading    = 'The FailureAgent.dll was not loaded, but unloading requested.';
const CErrWrongDataTyped     = 'Wrong data was spcified for safe exception type. Please type integer value.';

{$R *.DFM}

procedure TForm1.changeControlsState(newState : Boolean);
begin
    btnUnloadFalureAgent.Enabled := newState;
    GroupBox1.Enabled            := newState;
    cbMakeLog.Enabled            := newState;
    cbUseUserDll.Enabled         := newState;
    cbFailOnExcept.Enabled       := newState;
    btnAplySettings.Enabled      := newState;
    btnRegHandler.Enabled        := newState;
    ListView1.Enabled            := newState;
    ListView2.Enabled            := newState;
    Label1.Enabled               := newState;
    Label2.Enabled               := newState;
    btnAddSafeModule.Enabled     := newState;
    btnRemoveSafeModule.Enabled  := newState;
    Label3.Enabled               := newState;
    ceModule.Enabled             := newState;
    Label4.Enabled               := newState;
    ceEhType.Enabled             := newState;
    btnClearTypeFilter.Enabled   := newState;
    btnClearModuleFilter.Enabled := newState;
    btnSkipAV.Enabled            := newState;
    btnSkipMath.Enabled          := newState;
    btnSkipUser.Enabled          := newState;
    btnRestoreAV.Enabled         := newState;
    btnRestoreMath.Enabled       := newState;
    btnRestoreUser.Enabled       := newState;
    btnClearTypeFilter.Enabled   := newState;
    btnClearModuleFilter.Enabled := newState;

    cleanSafeModules();
    cleanSafeExceptions();
end;

procedure TForm1.cleanListView(lv : TListView);
begin
	lv.Items.Clear;
end;

procedure TForm1.cleanSafeModules();
begin
	cleanListView(ListView1);
end;

procedure TForm1.cleanSafeExceptions();
begin
	cleanListView(ListView2);
end;

procedure TForm1.addItemToListView(lv : TListView; const item : String);
begin
	lv.Items.Add().Caption := item;
end;

procedure TForm1.unloadFalureAgent();
begin
   Assert(FLoaded = true, CErrDoubleUnloading);
   if ( FLoaded = false ) then exit;
   btnUnRegHandler.Enabled := false;
   FIatExceptionfilter     := nil;
   FGetExceptionFilter     := nil;
   FreeLibrary(FFalureAgentModule);
   FFalureAgentModule      := 0;
	FLoaded                 := false;
   changeControlsState(FLoaded);
end;

procedure TForm1.loadFalureAgent();
begin
   Assert(FLoaded = false, CerrDoubleLoading);
   if ( FLoaded = true ) then exit;
   FFalureAgentModule := LoadLibrary(CFalureAgent);
   Assert(FFalureAgentModule <> 0, CErrFalureAgentLoading);

   FGetExceptionFilter := GetProcAddress(FFalureAgentModule, GETEXCEPTIONFILTERNAME);
   Assert( @FGetExceptionFilter <> nil );

   FGetExceptionFilter(FIatExceptionfilter);
   Assert(FIatExceptionfilter <> nil);

   updateSettingsInfo();
   FLoaded := true;
   changeControlsState(FLoaded);
end;

procedure TForm1.makeAVException();
begin
   asm
      mov   eax, 0
      mov   eax, dword ptr [eax];
   end;
end;

procedure TForm1.makeIntegerDivisionException();
begin
   asm
      mov   eax, 0
      div   eax
   end;
end;

procedure TForm1.makeUserException();
begin
   raise EXception.Create('Sample:User Exception');
end;

procedure TForm1.updateSettingsInfo();
var settings : DWORD;
begin
   Assert(FIatExceptionfilter <> nil);
   settings := FIatExceptionfilter.GetCurrentSettings();

   cbFailOnExcept.Checked    := ( ( settings and SV_FAILONEXCEPT ) <> 0 );
   cbMakeLog.Checked         := ( ( settings and SV_MAKELOG ) <> 0 );
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
	FExceptionListener := TPTExceptionListener.Create as IatExceptionListener;
   FLoaded := false;
   changeControlsState(FLoaded);
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
	if ( FLoaded = TRUE ) then
      unloadFalureAgent();
	FExceptionListener := nil;
end;

procedure TForm1.btnLoadFalureAgentClick(Sender: TObject);
begin
   loadFalureAgent();
end;

procedure TForm1.btnUnloadFalureAgentClick(Sender: TObject);
begin
   unloadFalureAgent();
end;

procedure TForm1.btnAVClick(Sender: TObject);
begin
   makeAVException();
end;

procedure TForm1.btnMathClick(Sender: TObject);
begin
   makeIntegerDivisionException();
end;

procedure TForm1.btnUserClick(Sender: TObject);
begin
   makeUserException();
end;

procedure TForm1.btnHandledAVClick(Sender: TObject);
begin
   try
     makeAVException();
   except
      ShowMessage(CMsgExceptionHandled);
   end;
end;

procedure TForm1.btnHandledMathClick(Sender: TObject);
begin
   try
      makeIntegerDivisionException();
   except
      ShowMessage(CMsgExceptionHandled);
   end;
end;

procedure TForm1.btnHandledUserClick(Sender: TObject);
begin
   try
      makeUserException();
   except
      ShowMessage(CMsgExceptionHandled);
   end;
end;

procedure TForm1.btnAplySettingsClick(Sender: TObject);
var settings : Integer;
begin
   settings := 0;
   if ( TRUE = cbFailOnExcept.Checked ) then
   	settings := settings or SV_FAILONEXCEPT;
   if ( TRUE = cbMakeLog.Checked ) then
   	settings := settings or SV_MAKELOG;
   Assert(FIatExceptionfilter <> nil);
   FIatExceptionfilter.OverrideSettings( settings );
   updateSettingsInfo();
end;

procedure TForm1.btnRegHandlerClick(Sender: TObject);
begin
   Assert(FIatExceptionfilter <> nil);
   FIatExceptionfilter.RegisterExceptionListener(FExceptionListener);
   btnUnRegHandler.Enabled := btnRegHandler.Enabled;
   btnRegHandler.Enabled   := false;
end;

procedure TForm1.btnUnRegHandlerClick(Sender: TObject);
begin
   Assert(FIatExceptionfilter <> nil);
   // Unimplemented
   FIatExceptionfilter.UnregisterExceptionListener(FExceptionListener);
   btnRegHandler.Enabled := btnUnRegHandler.Enabled;
   btnUnRegHandler.Enabled := false;
end;

procedure TForm1.btnAddSafeModuleClick(Sender: TObject);
begin
   if ( Length(ceModule.Text) <> 0 ) then
   begin
		addItemToListView(ListView1, ceModule.Text);
      FIatExceptionfilter.AddSafeModule( ceModule.Text );
   end;
end;

procedure TForm1.btnRemoveSafeModuleClick(Sender: TObject);
var ehCode : Integer;
begin
	FIatExceptionfilter.Lock();
   try
      if ( Length(ceEhType.Text) <> 0 ) then
      begin
         ehCode := StrToInt(ceEhType.Text);
         addItemToListView(ListView2, Format('0x%x', [ehCode]));
         FIatExceptionfilter.AddSafeException(ehCode);
      end;
   except
     on EConvertError do ShowMessage(CErrWrongDataTyped);
   end;
   FIatExceptionfilter.Unlock();
end;

procedure TForm1.btnSkipAVClick(Sender: TObject);
begin
   FIatExceptionfilter.AddSafeException(Integer(CAVExceptionCode));
end;

procedure TForm1.btnSkipMathClick(Sender: TObject);
begin
   FIatExceptionfilter.AddSafeException(Integer(CMathExceptionCode));
end;

procedure TForm1.btnRestoreAVClick(Sender: TObject);
begin
   FIatExceptionfilter.DeleteSafeExceptionCode(Integer(CAVExceptionCode));
end;

procedure TForm1.btnRestoreMathClick(Sender: TObject);
begin
   FIatExceptionfilter.DeleteSafeExceptionCode(Integer(CMathExceptionCode));
end;

procedure TForm1.btnClearModuleFilterClick(Sender: TObject);
begin
   FIatExceptionfilter.ClearSafeModules();
   cleanSafeModules();
end;

procedure TForm1.btnClearTypeFilterClick(Sender: TObject);
begin
   FIatExceptionfilter.ClearSafeExceptionCodes();
   cleanSafeExceptions();
end;

procedure TForm1.NestedHandledExceptionClick(Sender: TObject);
begin
   try
      makeAVException();
   except
      try
         makeIntegerDivisionException
      except
      end;
   end;
end;

procedure TForm1.btnSkipUserClick(Sender: TObject);
begin
   FIatExceptionfilter.AddSafeException(EH_DELPHIEXCEPTIONS);
end;

procedure TForm1.btnRestoreUserClick(Sender: TObject);
begin
   FIatExceptionfilter.DeleteSafeExceptionCode(EH_DELPHIEXCEPTIONS);
end;

end.
