{
----------------------------------------------------------
MAS-CompMaker was used to generate this code
MAS-CompMaker, 2000-2002 MASoft, Mats Asplund
----------------------------------------------------------

Component Name: TmRegApp
       Author: MASoft, Mats Asplund
     Creation: 2002-08-03
      Version: 4.0
  Description: An application registration component.
      Credits: Thanks to Oleg Petkov, http://olegpetkov.dir.bg/
               for providing the code/decode-routines.
               Thanks to Ferruccio Accalai, for providing
               the single computer-workstation Registration option.
       E-mail: mdp@telia.com
         Site: http://hem.passagen.se/matasp
 Legal issues: All rights reserved 2002-2003 by MAsoft, Mats Asplund

Usage:
 This software is provided 'as-is', without any express or
 implied warranty.  In no event will the author be held liable
 for any  damages arising from the use of this software.

 Permission is granted to anyone to use this software for any
 purpose, including commercial applications, and to alter it
 and redistribute it freely, subject to the following
 restrictions:

 1. The origin of this software must not be misrepresented,
    you must not claim that you wrote the original software.
    If you use this software in a product, an acknowledgment
    in the product documentation would be appreciated but is
    not required.

 2. Altered source versions must be plainly marked as such, and
    must not be misrepresented as being the original software.

 3. This notice may not be removed or altered from any source
    distribution.

 4. If you decide to use this software in any of your applications.
    Send me an EMail and tell me about it.

Quick Reference:
 TmRegApp inherits from TComponent.

 Key-Properties:
   PLock: Locks the password. Set PLock to false at designtime, before entering
     the password.

   MachineDependent: Set this to true, before entering the password, if the
     installation should be a one-machine-only installation.

   P: The password. Automatically coded. Otherwise it could be read in the
     resource part of the exe-file. Note! Set password at designtime. Otherwise
     it will not work.

   InstallKey: Used when the installation should be Machinedependent. The
     user should send this Installkey to the vendor, and get an unique
     ResultPwd in return. Use the RegKey Creator application to generate the
     ResultPwd.

   CodeKey: Key used for coding/decoding. Change this to anything you like.

   DaysToUse: The number of days before registering is needed.

   FakeGUID: A false GUID used as key in Registry. You should change this,
     otherwise others with knowledge of this component could find it.

 Key-Methods:
   Init: Should be called each time the application starts executing. Returns
     the days left before registrering is needed. If the app. is registered
     -1 is returned.

   ClearReg: Deletes all Registry-entries made. (Deletes the FakeGUID-key from
     Registry.)

   CheckPassWord: Returns true if password is correct. If so the values:
     'DaysToUse', 'Expire', 'LastUse' will be deleted from Registry. The value:
     'Registered' will be set true.

   GetMoveBack: Read this to see if user has manipulated the clock. If true,
     the user has moved the clock back more then one day compared with
     the lastuse date. In this case 0 will be returned (and will
     continue to do so until registered), when calling the Init function.

   GetExpireDate: Returns Expiredate as a string.

   GetInstallationDate: Returns Installationdate as a string.

   GetLastUseDate: Returns the date of last use as a string.

   GetInstallKey: Returns the InstallKey as a string.


--------------------------------------------------------------------------------
}
unit mRegApp;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls;

type
  TDType = (dtInstall, dtLastUse, dtExpire);

  TRegAppCfgData = string;

  TmRegApp = class(TComponent)
  private
    fAbout: string;
    fP: string;
    fCodeKey: string;
    fDaysToUse: integer;
    fFakeGUID: string;
    fMachineDependent: Boolean;
    fPLocked: Boolean;
    fInstallKey: string;
    procedure SetCop(Value: string);
    procedure SetP(const Value: string);
    function code(Text: string): shortstring;
    function Decode(CodedText: string): shortstring;
    function GetDates(DateType: TDType): string;
    function GetVolumeSerialNumber: string;
    procedure RegisterApp;
    function GetConfigData: TRegAppCfgData;
    procedure SetConfigData(const Value: TRegAppCfgData); //added by Ferruccio Accalai
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Init: integer;
    function GetExpireDate: string;
    function GetInstallKey: string;
    //added by Ferruccio Accalai: the coded VolumeSerialNumber
    function GetInstallationDate: string;
    function GetLastUseDate: string;
    function GetMoveBack: Boolean;
    procedure ClearReg;
    function CheckPassWord(PassWord: string): Boolean;
    property P: string read fP write SetP; // The password
    property PLocked: Boolean read fPLocked write fPLocked; // The password
    property CodeKey: string read fCodeKey write fCodeKey;
    property InstallKey: string read GetInstallKey write fInstallKey;
    property DaysToUse: integer read fDaysToUse write fDaysToUse;
    property FakeGUID: string read fFakeGUID write fFakeGUID;
    property MachineDependent: Boolean read fMachineDependent write
      fMachineDependent;
  published
    property ConfigData: TRegAppCfgData read GetConfigData write SetConfigData;
    property About: string read fAbout write SetCop;
  end;

procedure Register;

const
  RegKey = '\Software\CLASSES\CLSID\';

implementation
uses DateUtils, Registry;

procedure Register;
begin
  RegisterComponents('MASoft', [TmRegApp]);
end;

constructor TmRegApp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fCodeKey:= 'MA-89D7-02';
  fDaysToUse:= 31;
  fFakeGUID:= '{A1159D67-71E0-11D4-BE52-39A481D50110}';
  fPLocked:= true;
  fAbout:= 'Version 4.0, 2002-2003 MASoft';
end;

destructor TmRegApp.Destroy;
begin
  inherited Destroy;
end;

function TmRegApp.Init: integer; //DaysLeft. -1 if registered
var
  Reg: TRegistry;
  LastUse, ExpireDate: TDateTime;
begin
  Reg:= TRegistry.Create;
  try
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
    begin
      Reg.OpenKey(RegKey + fFakeGUID, false);
      if Reg.ReadBool(code('Registered')) then
        result:= -1
      else
      begin
        LastUse:= StrToDateTime(Decode(Reg.ReadString(code('LastUse'))));
        if (Reg.ReadBool(code('MoveBack'))) or (CompareDate(Now, LastUse) < 0)
          then
        begin
          Reg.WriteBool(code('MoveBack'), true);
          result:= 0; // Time moved back
          Exit;
        end;
        Reg.WriteString(code('LastUse'), code(DateTimeToStr(Now)));
        ExpireDate:= StrToDateTime(Decode(Reg.ReadString(code('Expire'))));
        if CompareDate(ExpireDate, Now) > 0 then
          result:= DaysBetween(ExpireDate, Now)
        else
          result:= 0; // Expired
      end
    end
    else
    begin // First time started
      Reg.OpenKey(RegKey + fFakeGUID, true);
      if fMachineDependent then
        Reg.WriteString(code('InstallKey'), code(GetVolumeSerialNumber));
      Reg.WriteString(code('DaysToUse'), code(IntToStr(fDaysToUse)));
      Reg.WriteString(code('Install'), code(DateTimeToStr(Now)));
      Reg.WriteString(code('Expire'), code(DateTimeToStr(IncDay(Now,
        fDaysToUse))));
      Reg.WriteString(code('LastUse'), code(DateTimeToStr(Now)));
      Reg.WriteBool(code('MoveBack'), false);
      Reg.WriteBool(code('Registered'), false);
      Reg.CloseKey;
      result:= fDaysToUse;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TmRegApp.ClearReg;
var
  Reg: TRegistry;
begin
  Reg:= TRegistry.Create;
  try
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
      Reg.DeleteKey(RegKey + fFakeGUID);
  finally
    Reg.Free;
  end;
end;

function TmRegApp.GetLastUseDate: string;
begin
  result:= GetDates(dtLastUse);
end;

function TmRegApp.GetExpireDate: string;
begin
  result:= GetDates(dtExpire);
end;

function TmRegApp.GetInstallationDate: string;
begin
  result:= GetDates(dtInstall);
end;

function TmRegApp.GetMoveBack: Boolean;
var
  Reg: TRegistry;
begin
  result:= false;
  Reg:= TRegistry.Create;
  try
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
    begin
      Reg.OpenKey(RegKey + fFakeGUID, false);
      result:= Reg.ReadBool(code('MoveBack'));
    end;
  finally
    Reg.Free;
  end;
end;

function TmRegApp.CheckPassWord(PassWord: string): Boolean;
var
  Reg: TRegistry;
begin
  Reg:= TRegistry.Create;
  try
    result:= false;
    {modified by Ferruccio Accalai
     Now Password is combined with VolumeSerialNumber for
     single computer-workstation Registrations}
    if fMachineDependent then
    begin
      if PassWord = code(Decode(fP) + GetVolumeSerialNumber) then
      begin
        RegisterApp;
        result:= true;
      end
      else
        result:= false;
    end
    else
    begin
      if PassWord = Decode(fP) then
      begin
        RegisterApp;
        result:= true;
      end
      else
        result:= false;
    end;
  finally
    Reg.Free;
  end;
end;

procedure TmRegApp.SetCop(Value: string);
begin
  Exit;
end;

procedure TmRegApp.SetP(const Value: string);
begin
  fP:= Value;
  if (csDesigning in ComponentState) and not (fPLocked) then // Could only be set at designtime
  begin
    fP:= code(fP);
    ClearReg;
    Init;
    fInstallKey:= GetInstallKey;
    fPLocked:= true;
  end;
end;

function TmRegApp.GetDates(DateType: TDType): string;
var
  Reg: TRegistry;
  DT: string;
begin
  Reg:= TRegistry.Create;
  try
    result:= '';
    case DateType of
      dtInstall: DT:= 'Install';
      dtLastUse: DT:= 'LastUse';
      dtExpire: DT:= 'Expire';
    end;
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
    begin
      Reg.OpenKey(RegKey + fFakeGUID, false);
      result:= Decode(Reg.ReadString(code(DT)));
    end;
  finally
    Reg.Free;
  end;
end;

function TmRegApp.Decode(CodedText: string): shortstring;
var
  DataSize, q, Q1, k, Len: short;
  P: PChar;

  function HexaToInt(ch: Char): short;
  begin
    case ch of
      '0'..'9': result:= StrToInt(ch);
      'A': result:= 10;
      'B': result:= 11;
      'C': result:= 12;
      'D': result:= 13;
      'E': result:= 14;
      'F': result:= 15;
    else
      result:= 7;
    end;
  end;
begin
  result:= '';
  k:= 1;
  P:= PChar(CodedText);
  DataSize:= Length(CodedText);
  Len:= Length(fCodeKey);
  if Len > DataSize then
    Len:= DataSize;
  while DataSize > 0 do
  begin
    case P^ of
      #10, #13, #32: Inc(P);
    else
      Q1:= HexaToInt(P^);
      Inc(P);
      q:= HexaToInt(P^);
      Inc(P);
      q:= (Q1 * 16) + q;
      q:= q xor (Ord(fCodeKey[k]));
      Inc(k);
      if k > Len then
        k:= 1;
      result:= result + Chr(q);
    end;
    Dec(DataSize, 2);
  end;
end;

function TmRegApp.code(Text: string): shortstring;
var
  DataSize, i, q, k: short;
begin
  result:= '';
  DataSize:= Length(Text);
  k:= 1;
  for i:= 1 to DataSize do
  begin
    q:= Ord(Text[i]) xor Ord(fCodeKey[k]);
    Inc(k);
    if k > Length(fCodeKey) then
      k:= 1;
    result:= result + IntToHex(q, 2);
  end;
end;

procedure TmRegApp.RegisterApp;
var
  Reg: TRegistry;
begin
  Reg:= TRegistry.Create;
  try
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
    begin
      Reg.OpenKey(RegKey + fFakeGUID, false);
      Reg.WriteBool(code('Registered'), true);
      Reg.DeleteValue(code('DaysToUse'));
      Reg.DeleteValue(code('Expire'));
      Reg.DeleteValue(code('LastUse'));
      if fMachineDependent then
        Reg.DeleteValue(code('InstallKey'));
    end;
  finally
    Reg.Free;
  end;
end;

{functions added by Ferruccio Accalai}

function TmRegApp.GetVolumeSerialNumber: string;
var
  NotUsed, VolFlags: Cardinal;
  VolSN: DWORD;
  VolumeSerialNumber: string;
begin
  GetVolumeInformation(PChar('C:\'), nil, 0, @VolSN, NotUsed, VolFlags, nil, 0);
  VolumeSerialNumber:= AnsiUpperCase(IntToHex(VolSN, 8));
  result:= Copy(VolumeSerialNumber, 1, 4) + ':' +
    Copy(VolumeSerialNumber, 5, 4);
end;

function TmRegApp.GetInstallKey: string;
var
  Reg: TRegistry;
begin
  Reg:= TRegistry.Create;
  try
    result:= '';
    Reg.RootKey:= HKEY_LOCAL_MACHINE;
    if Reg.KeyExists(RegKey + fFakeGUID) then
    begin
      Reg.OpenKey(RegKey + fFakeGUID, false);
      if Reg.ReadString(code('InstallKey')) = '' then
        result:= 'Not used'
      else
        result:= Reg.ReadString(code('InstallKey'));
    end
    else
      result:= 'No data';
  finally
    Reg.Free;
  end;
end;

function TmRegApp.GetConfigData: TRegAppCfgData;
var
  dummy: TStringList;
begin
//showmessage('GetConfigData');
  dummy:= TStringList.Create;
  try
    dummy.Add(fCodeKey);
    dummy.Add(IntToStr(fDaysToUse));
    dummy.Add(fFakeGUID);
    dummy.Add(fInstallKey);
    if fPLocked then
      dummy.Add('1')
    else
      dummy.Add('0');
    dummy.Add(fP);
    if fMachineDependent then
      dummy.Add('1')
    else
      dummy.Add('0');
    Result:= dummy.Text;
  finally
    dummy.Free;
  end;
end;

procedure TmRegApp.SetConfigData(const Value: TRegAppCfgData);
var
  dummy: TStringList;
begin
  if Value <> '' then
  begin
    dummy:= TStringList.Create;
    try
      dummy.Text:= Value;
      fCodeKey:= dummy[0];
      fDaysToUse:= StrToInt(dummy[1]);
      fFakeGUID:= dummy[2];
      fInstallKey:= dummy[3];
      fPLocked:= dummy[4] = '1';
      SetP(dummy[5]);
      fMachineDependent:= dummy[6] = '1';
    finally
      dummy.Free;
    end;
  end;
end;

end.

