{-------------------------------------------------------------------------------
 TEZProjectionL is a component to convert between lattitude and longtitude and the
 X/Y  position on a map.
 EZProjectionL uses TEZConvert to convert between UTM grid and Lat/Lon.

 Projections have been adapted from the General Cartographic Transformation Package.
--------------------------------------------------------------------------------}
unit EZProjectionL;

interface

uses
  SysUtils, Classes, Graphics, Controls, Math, EZConstL, EZConvertL, EZMapValuesL, Forms;

type
//  for TRef see EZMapalues
  TEZProjectionL = class(TComponent)
  private
    FErrorEvent   : TErrorEvent;
    FForEvent     : TForEvent;
    FInvEvent     : TInvEvent;
    FGetInfoEvent : TGetInfoEvent;
    FGetConstants : TGetConstants;
    FIOFormat     : TIOFormat;
    FEZConvertL   : TEZConvertL;
    FTruncUTM     : boolean;
    TempVal1      : Double;
    TempVal2      : Double;
    CutOff        : boolean;                                                    // Because the inverse iterations are not capable of getting all possible x,y for a map, these area are not shown in forward calculations. No cutoff if 'ForCutOff = false;'.
    FloatTemp     : TFloatPoint;

    procedure ConvertFor(Theta,Phi : Double);                                   // Convert UTM, Datum UTM or National Grid to WGS84
    procedure ConvertInv(X,Y : Double);
                                                                                // Projections with no reference calculations
                                                                                // All following projection computations use var FloatPoint as input and output
    procedure EckertTFor;
    procedure EckertTInv;
    procedure StereographicFor;
    procedure StereographicInv;
    procedure AlbertsCEAFor;
    procedure AlbertsCEAInv;
    procedure GvnspFor;
    procedure GvnspInv;
    procedure ConicalEDFor;
    procedure ConicalEDInv;
    procedure IMollweideFor;
    procedure IMollweideInv;
    procedure WagnerVIIFor;
    procedure WagnerVIIInv;
    procedure WagnerIVFor;
    procedure WagnerIVInv;
    procedure MollweideFor;
    procedure MollweideInv;
    procedure MercatorFor;
    procedure MercatorInv;
    procedure CassiniSFor;
    procedure CassiniSInv;
    procedure TMercatorFor;
    procedure TMercatorInv;
    procedure GnomonicFor;
    procedure GnomonicInv;
    procedure LaCConicFor;
    procedure LaCConicInv;
    procedure RobinsonFor;
    procedure RobinsonInv;
    procedure PolyconicFor;
    procedure PolyconicInv;
    procedure GrintenFor;
    procedure GrintenInv;
    procedure LaAzimuthalEAFor;
    procedure LaAzimuthalEAInv;
    procedure AzimuthalEAFor;
    procedure AzimuthalEAInv;
    procedure HammerFor;
    procedure HammerInv;
    procedure BStereoFor;
    procedure BStereoInv;
    procedure EquirectFor;
    procedure EquirectInv;
    procedure SinusoidalFor;
    procedure SinusoidalInv;
    procedure WinkelIFor;
    procedure WinkelIInv;
    procedure WinkelIIFor;
    procedure WinkelIIInv;
    procedure BonneFor;
    procedure BonneInv;
    procedure LamCylindricalEAFor;
    procedure LamCylindricalEAInv;
    procedure BerCylindricalEAFor;
    procedure BerCylindricalEAInv;
    procedure TristanEdwardsFor;
    procedure TristanEdwardsInv;
    procedure PetersFor;
    procedure PetersInv;
    procedure GallOrthFor;
    procedure GallOrthInv;
    procedure BalthasartFor;
    procedure BalthasartInv;
    procedure CollignonFor;
    procedure CollignonInv;
    procedure EckertIFor;
    procedure EckertIInv;
    procedure EckertIIIFor;
    procedure EckertIIIInv;
    procedure PutninsFor;
    procedure PutninsInv;
    procedure KavraiskyVIIFor;
    procedure KavraiskyVIIInv;
    procedure WagnerVIFor;
    procedure WagnerVIInv;
    procedure CrasterFor;
    procedure CrasterInv;
    procedure EckertIVFor;
    procedure EckertIVInv;
    procedure FaheyFor;
    procedure FaheyInv;
    procedure FoucautSFor;
    procedure FoucautSInv;
    procedure GallFor;
    procedure GallInv;
    procedure EckertVIFor;
    procedure EckertVIInv;
    procedure McBrydeFor;
    procedure McBrydeInv;
    procedure LoximuthalFor;
    procedure LoximuthalInv;
    procedure MillerFor;
    procedure MillerInv;
    procedure EckertVFor;
    procedure EckertVInv;
    procedure NellFor;
    procedure NellInv;
    procedure FoucautFor;
    procedure FoucautInv;
    procedure KavraiskyVFor;
    procedure KavraiskyVInv;
    procedure McBrydeSinFor;
    procedure McBrydeSinInv;
    procedure QuarticAuFor;
    procedure QuarticAuInv;
    procedure WagnerIIFor;
    procedure WagnerIIInv;
    procedure WagnerIIIFor;
    procedure WagnerIIIInv;
    procedure UrmaevFor;
    procedure UrmaevInv;
    procedure WagnerIFor;
    procedure WagnerIInv;
    procedure UrmaevVFor;
    procedure UrmaevVInv;
    procedure McBrydeSin2For;
    procedure McBrydeSin2Inv;
    procedure McBrydePFor;
    procedure McBrydePInv;
    procedure McBrydeQFor;
    procedure McBrydeQInv;
//    procedure Putnins2For;
//    procedure Putnins2Inv;
                                                                                    // With reference calculations
    Procedure UTMRef;
    procedure UTMFor;
    procedure UTMInv;
    procedure OrthRef;
    procedure OrthFor;
    procedure OrthInv;

   {procedure HammerRef;                                                        // Still needs some testing....
    procedure BStereoRef;
    procedure EquirectRef;
    procedure SinusoidalRef;
    procedure WinkelIRef;
    procedure WinkelIIRef;}

    function  RMinor:Double;                                                    // Calculate the Minor radius from the current map ellipsoid (index)
    function  OrthRange(Theta,Phi: Double):boolean;                             // Determines if a coordinate is visible in an Orthographic projection
    procedure Rotate(Theta,Phi:Double;Forw:boolean);                            // Rotate (Oblate) a coordinate (uses MapRef.Oa,Ot,Op)

    procedure phi1z(eccent,qs:Double);                                          // Computes Phi for Albers Conical Equal-Area
    procedure Phi2z(eccent,ts:Double);                                          // Computes the latitude angle for Lambert Conformal Conic and Polar Stereographic
    procedure phi3z(ml,e0,e1,e2,e3:Double);                                     // Computes the latitude for Equidistant Conic
    procedure Phi4z(eccent,e0,e1,e2,e3,a,b:Double);                             // Computes the latitude for Polyconic
    function  ASinZ(con : Double): Double;                                      // Safe ArcSin function
    function  E0fn(x:Double):Double;                                            // Computes the constants e0 for calculating the distance along a meridian
    function  E1fn(x:Double):Double;                                            // Computes the constants e1 for calculating the distance along a meridian
    function  E2fn(x:Double):Double;                                            // Computes the constants e2 for calculating the distance along a meridian
    function  E3fn(x:Double):Double;                                            // Computes the constants e4 for calculating the distance along a meridian
    function  qsfnz(eccent,sinphi,cosphi:Double):Double;                        // Computes e4 used for Polar Stereographic
    function  mlfn(e0,e1,e2,e3,phi:Double):Double;                              // Computes M, the distance along a meridian
//    function  MlfnInv(arg,es:double):double;
    function  Msfnz(eccent,sinphi,cosphi:Double):Double;                        // Computes the constant small m for Oblique Equal Area and for conical projections
    function  Tsfnz(eccent,phi,sinphi:Double):Double;                           // Computes the small t for Lambert Conformal Conic and Polar Stereographic

    Procedure ErrorLog(F : integer);                                            // Envokes the ErrorEvent with the apropriate description
    procedure ErrorEvent(Sender: TObject;Return:integer);
    procedure ForEvent(Sender: TObject; X,Y:Double;Return:integer);             // Event  triggered by a forward calculation
    procedure InvEvent(Sender: TObject; Theta,Phi:Double;Return:integer);       // Same for inverse
    procedure GetInfoEvent(Sender: TObject; MapName: string; TypeGeo,
                       Reference:boolean;EllipsType,MapNumber:integer);         // Even triggerd by procedure GetInfo. Gives some basic information about the current projection
    procedure SetEZConvert(Value:TEZConvertL);                                   // Assign a TEZConnvert. TEZConvert  is needed to convert between Lat,Lon and UTM like grids
    procedure DoTruncUTM;                                                       // Because the conversions are not absolutly accurate the decimal part of all UTM outputs are trunced if property TruncUTM is set to true
    procedure Protect(Index:integer);                                           // prevent errors. Sets mapindex to default for projection if value is not apropriate
    function  GetTemp:TFloatPoint;
    procedure CylEA(S:double);                                                  // Cylindrical Equal-Area computation depending on standard latitude
    function  CylEAInv(S:double):byte;
    function  Asqrt(D:double):Double;
    procedure EcT(cx,cy,a,b:double);
    function  EcTI(cx,cy,a,b:double):byte;
    procedure EcTV(n,m:double);
    function  EcTVI(n,m:double):byte;
    procedure Fouc(p,q:double;m:byte);
    function  FoucI(p,q:double;m:byte):byte;

    procedure PrepFor1;
    procedure PrepFor2;
    procedure PrepInv1;
    procedure PrepInv2(E,F:integer);
  protected
  public
    FloatPoint  : TFloatPoint;                                                  // The backbone for most calculations
    Flag        : TFlag;                                                        // Error flag
    MapRef      : TRef;                                                         // All variables needed for the projection
    procedure AddRefPoint(X,Y,Theta,Phi:Double);                                // Add a point (lon,lat,x,y) to the reference list
    Procedure CalculateFor(Theta,Phi : Double);                                 // calculate X,Y from Lon,Lat using the values in var MapRef
    Procedure CalculateInv(X,Y : Double);                                       // calculate Lon,Lat from X,Y using the values in var MapRef
    Procedure CalculateRef;                                                     // calculate map constants for var MapRef. This is a way of getting the makeup of the map to do all projection calculations
    procedure ClearRef;                                                         // Delete all reference points from MapRef.Points
    constructor Create(AOwner: TComponent); override;                           // Creates component and sets default values if none has been assigned
    procedure DeleteRefPoint(index : integer);                                  // Delete a point from the reference list
    procedure GetConstants;                                                     // Envokes the OnGetConstants event. This to get the basic map values (constants)
    procedure GetEllipsoid(Str:TStrings);                                        // Get the list (like for a listbox) of all ellipsoids from EZEllips
    procedure GetDatum(Str:TStrings);                                           // Get the list (like for a listbox) of all Datums from EZDatum
    procedure GetInfo;                                                          // Triggers an onGetInfo event. The event will provide some basic information about the currrent projection
    procedure GetIOFormat;                                                      // Triggers an onGetIOForrmat event. The event will give the format type and subtype (ellipsoids e.a.) for both in and output
    function  GetMapIndex:integer;                                              // Returns the index value representing the  current projection type (like mpOrthographic = 3)
    procedure GetNatGrid(Str:TStrings);                                         // Get the list (like for a listbox) of all National Grids from EZDatum
    function  GetNorth(Theta,Phi:Double):Double;                                // Returns the angle in degrees in wich direction North could be found on the currently used map image (if you would use any) from point Theta,Phi.
    procedure GetProjectionList(Str:TStrings;GridInclude:boolean);              // Get the list (like for a listbox) of all national grids from EZDatum
    function  GetMapName:string;                                                // Returnes Mapref.MapName
    procedure LoadMapRefFromFile(FileName:TFileName);                           // Loads all values for MapRef form a file with same structure as Mapref
    procedure LoadMapRefFromStream(Stream: TStream);                            // Loads all values for MapRef from a Stream
    procedure SaveMapRefToFile(FileName:TFileName);                             // Saves map constants, reference points, grid information,.. to the file FileName
    procedure SaveMapRefToStream(Stream: TStream);                              // Stores map constants, reference points, grid information,.. to a stream
    procedure SetEllipsoid(Index:integer);                                      // Sets the map index (an ellipsoid) to MapRf.MapIndex
    procedure SetGridConstants(A1,A2,A3,B1,B2,B3:Double);                       // Sets grid properties to MMapRef.GridProp
    procedure SetMapName(Text:string);                                          // Stores the name of the map to MapRef.MapName
    procedure SetMapProjection(MapInd:integer);                                 // Sets current projection type from an index numberr
    procedure SetPrOblique(Oblique:Double);                                     // Sets Oblique angle
    procedure SetPrObliqueCenter(Theta,Phi:Double);                             // Sets rotation point for oblique rotation. The Theta here will be corrected with the Theta0 of the projection constants
    procedure SetPrOther(Other:Double);                                         // Sets additional properties. For 'General Vertical Near-side Perspective' it will be the height and for 'Equidistant Conical' the mode (<1 = mode A, >= 1 = mode B)
    procedure SetPrPhi0(Phi0:Double);                                           // Sets main lattitude for a projection if used
    procedure SetPrPhi1(Phi1:Double);                                           // Sets latitude of first standard parallel for a conical projection
    procedure SetPrPhi2(Phi2:Double);                                           // Sets latitude of second standard parallel for a conical projection
    procedure SetPrRadius(Radius:Double);                                       // Sets the radius of projection. The radius unit are pixels
    procedure SetPrTheta(Theta:Double);                                         // Sets main longtitude for all projection types
    procedure SetPrX0(X0:Double);                                               // Sets x offset in pixls for a map image
    procedure SetPrY0(Y0:Double);                                               // Sets y offset in pixls for a map image
    procedure SetScanRotation(X,Y,Angle:Double);                                // Sets map properties for a scanned map (if not scanned Angle = 0) to correct for a possible 'image tilt' due to scanning
    procedure SetUserEllipsoid(Radius,Flat,Scale:Double;UserZone,Name:string;Index:integer);                          // Changes the values of an existing user defined ellipsoid. If the ellipsoid does not exist create one by calling EZConvert.AddUserEllips(...);
    procedure SetUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;UserZone,Name:string;Index:integer);                          // Changes the values of an existing user defined Datum. If the ellipsoid does not exist create one by calling EZConvert.AddUserDatum(...);
    procedure SetUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,Scale,CentralLong:Double;Name:string;Index:integer);       // Changes the values of an existing user defined national grid. If the ellipsoid does not exist create one by calling EZConvert.AddUserNatGid(...);
  published
    property  EZConvert: TEZConvertL          read FEZConvertL        write SetEZConvert;          // Assigned EZConvert to EZProjectionL
    property  ForwardCutOff:boolean           read CutOff             write CutOff;                // Because the inverse iterations are not capable of getting all possible x,y for a map, these area are not shown in forward calculations. No cutoff if 'ForCutOff = false;'.
    property  InputFormat:TInPut              read MapRef.Input       write MapRef.Input;          // Input format: Geographic, UTM,....
    property  InputIndex:integer              read MapRef.InputIndex  write MapRef.InputIndex;     // What subtype is used for input. Like a lokal UTM it could used 'North American 1927:Alberta,British Columbia' with value of 119.
    property  MapInput:integer                read MapRef.MapIndex    write MapRef.MapIndex;       // What subtype is used for the  projection. Like Mercator uses an ellipsoid. So '23' would be 'Krassovsky 1940'.
    property  Projection:TProjection          read MapRef.Projection  write MapRef.Projection;     // Projection type used (npOrthographic, mpMercator, mpUTM,....)
    property  TruncUTM:boolean                read FTruncUTM          write FTruncUTM;             // Because the conversions are not absolutly accurate the decimal part of all UTM like outputs are trunced if set to true
    property  OutputFormat:TOutPut            read MapRef.Output      write MapRef.Output;         // Output format: Geographic, UTM,....
    property  OutputIndex:integer             read MapRef.OutputIndex write MapRef.OutputIndex;    // Same as InputIndex
    property  OnError: TErrorEvent            read FErrorEvent        write FErrorEvent;           // If any calculation has been made. if Return = 0 (computation = OK) else a number with corresponding description will be given. If no event is needed, use Flag.F/S to get the information on any errors (Flag.F  = 0 => Flag.S = 'OK')
    property  OnForward: TForEvent            read FForEvent          write FForEvent;             // If a forward computation has been made. Returns the map X,Y and error information. Is used by EZMapImage for onMousEvents
    property  OnGetInfo: TGetInfoEvent        read FGetInfoEvent      write FGetInfoEvent;         // Trriggerd by procedure GetInfo.
    property  OnGetConstants: TGetConstants   read FGetConstants      write FGetConstants;         // Triggered by procedure GetConstants (note: does not return the oblique rotaion coordinates);
    property  OnGetIOFormat : TIOFormat       read FIOFormat          write FIOFormat;             // Triggered by procedure GetIOFormat. Returns information on both In/Output formats
    property  OnInverse: TInvEvent            read FInvEvent          write FInvEvent;             // Same as onForEvent but for inverse computations
  end;

procedure Register;

implementation


procedure Register;
begin
  RegisterComponents('EZJ', [TEZProjectionL]);
end;
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
constructor TEZProjectionL.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ClearRef;                                                                     // Set all map values to default;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.ClearRef;
begin
 With MapRef do begin
   CountRef        := 0;
   Input           := ifGeographic;                                             // Input as default geographic (WGS84)
   Output          := ofGeographic;                                             // Output as default geographic (WGS84)
   Projection      := mpOrthographic;                                           // Default projection is set to Orthographic
   MapIndex        := DefDatum;                                                 // Geographic default (datum) -> WGS84
   InputIndex      := DefDatum;                                                 // Inpute default (datum) -> WGS84
   OutputIndex     := DefDatum;                                                 // Output default (datum) -> WGS84
   Calculated      := false;                                                    // Not calculated! This means no computation can be done. This is just a precoution! Example -> use procedure SetDemo to set all values and calculated.
   MapName         := '';                                                       // Can store path to a map image. This value is used by TEZMapImage and TEZMap
   Values.Zone     := 'N31';                                                    // Just a zone near equator
   Values.Rad      := 0;                                                        // All following values are needed for drawing (see declaration above);
   Values.Th0      := 0;
   Values.Ph0      := 0;
   Values.Ph1      := 0;
   Values.Ph2      := 0;
   Values.X0       := 0;
   Values.Y0       := 0;
   Values.Oa       := 0;
   Values.Ot       := 0;
   Values.Op       := 0;
   Values.Rx       := 0;
   Values.Ry       := 0;
   Values.Ra       := 0;
   GridProp.East   := 180;
   GridProp.West   := -180;
   GridProp.North  := 90;
   GridProp.South  := -90;
   GridProp.Interval:= 15;
   GridProp.Resolution:=1;
  end;
  Mapref.Version   := 1;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetMapProjection(MapInd:integer);                       // Set map projection type by integer
begin
Case MapInd of
	3	 : MapRef.Projection := 	mpAlbertsCEA	;
	4	 : MapRef.Projection := 	mpAzimuthalEA	;
	5	 : MapRef.Projection := 	mpBalthasart	;
	6	 : MapRef.Projection := 	mpBerCylindricalEA;
	7	 : MapRef.Projection := 	mpBonne	        ;
	8	 : MapRef.Projection := 	mpBStereo	;
	9	 : MapRef.Projection := 	mpCassiniS	;
	10	 : MapRef.Projection := 	mpCollignon	;
	11	 : MapRef.Projection := 	mpCraster	;
	12	 : MapRef.Projection := 	mpEckertI	;
	13	 : MapRef.Projection := 	mpEckertII	;
	14	 : MapRef.Projection := 	mpEckertIII	;
	15	 : MapRef.Projection := 	mpEckertIV	;
	16	 : MapRef.Projection := 	mpEckertV	;
	17	 : MapRef.Projection := 	mpEckertVI	;
	18	 : MapRef.Projection := 	mpEDConical	;
	19	 : MapRef.Projection := 	mpEquirectangular;
	20	 : MapRef.Projection := 	mpFahey	        ;
	21	 : MapRef.Projection := 	mpFoucaut	;
	22	 : MapRef.Projection := 	mpFoucautS	;
	23	 : MapRef.Projection := 	mpGallOrth	;
	24	 : MapRef.Projection := 	mpGall	        ;
	25	 : MapRef.Projection := 	mpGvnsp	        ;
	26	 : MapRef.Projection := 	mpGnomonic	;
	27	 : MapRef.Projection := 	mpHammer	;
	28	 : MapRef.Projection := 	mpIMollweide	;
	29	 : MapRef.Projection := 	mpKavraiskyV	;
	30	 : MapRef.Projection := 	mpKavraiskyVII	;
	31	 : MapRef.Projection := 	mpLaAzimuthalEA	;
	32	 : MapRef.Projection := 	mpLaConformalConic;
	33	 : MapRef.Projection := 	mpLamCylindricalEA;
	34	 : MapRef.Projection := 	mpLoximuthal	;
	35	 : MapRef.Projection := 	mpMcBrydeP	;
	36	 : MapRef.Projection := 	mpMcBrydeQ	;
	37	 : MapRef.Projection := 	mpMcBrydeSin	;
	38	 : MapRef.Projection := 	mpMcBryde	;
	39	 : MapRef.Projection := 	mpMcBrydeSin2	;
	40	 : MapRef.Projection := 	mpMercator	;
	41	 : MapRef.Projection := 	mpMiller	;
	42	 : MapRef.Projection := 	mpMollweide	;
	43	 : MapRef.Projection := 	mpNell	        ;
	44	 : MapRef.Projection := 	mpOrthographic	;
	45	 : MapRef.Projection := 	mpPeters	;
	46	 : MapRef.Projection := 	mpPolyconic	;
	47	 : MapRef.Projection := 	mpPutnins	;
	48	 : MapRef.Projection := 	mpQuarticAu	;
	49	 : MapRef.Projection := 	mpRobinson	;
	50	 : MapRef.Projection := 	mpSinusoidal	;
	51	 : MapRef.Projection := 	mpStereographic	;
	52	 : MapRef.Projection := 	mpTMercator	;
	53	 : MapRef.Projection := 	mpTristanEdwards;
	54	 : MapRef.Projection := 	mpUrmaev	;
	55	 : MapRef.Projection := 	mpUrmaevV	;
	56	 : MapRef.Projection := 	mpGrinten	;
	57	 : MapRef.Projection := 	mpWagnerI	;
	58	 : MapRef.Projection := 	mpWagnerII	;
	59	 : MapRef.Projection := 	mpWagnerIII	;
	60	 : MapRef.Projection := 	mpWagnerIV	;
	61	 : MapRef.Projection := 	mpWagnerVI	;
	62	 : MapRef.Projection := 	mpWagnerVII	;
	63	 : MapRef.Projection := 	mpWinkelI	;
	64	 : MapRef.Projection := 	mpWinkelII	;
end;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.GetMapIndex:integer;                                     // Get the integer representation of the used projection type
var
R : integer;
begin
R := 0;
Case MapRef.Projection of
  	mpAlbertsCEA	 : R := 	3	;
  	mpAzimuthalEA	 : R := 	4	;
  	mpBalthasart	 : R := 	5	;
  	mpBerCylindricalEA:R := 	6	;
  	mpBonne	         : R := 	7	;
  	mpBStereo	 : R := 	8	;
  	mpCassiniS	 : R := 	9	;
  	mpCollignon	 : R := 	10	;
  	mpCraster	 : R := 	11	;
  	mpEckertI	 : R := 	12	;
  	mpEckertII	 : R := 	13	;
  	mpEckertIII	 : R := 	14	;
  	mpEckertIV	 : R := 	15	;
  	mpEckertV	 : R := 	16	;
  	mpEckertVI	 : R := 	17	;
  	mpEDConical	 : R := 	18	;
  	mpEquirectangular: R := 	19	;
  	mpFahey	         : R := 	20	;
  	mpFoucaut	 : R := 	21	;
  	mpFoucautS	 : R := 	22	;
  	mpGallOrth	 : R := 	23	;
  	mpGall	         : R := 	24	;
  	mpGvnsp	         : R := 	25	;
  	mpGnomonic	 : R := 	26	;
  	mpHammer	 : R := 	27	;
  	mpIMollweide	 : R := 	28	;
  	mpKavraiskyV	 : R := 	29	;
  	mpKavraiskyVII	 : R := 	30	;
  	mpLaAzimuthalEA	 : R := 	31	;
  	mpLaConformalConic:R := 	32	;
  	mpLamCylindricalEA:R := 	33	;
  	mpLoximuthal	 : R := 	34	;
  	mpMcBrydeP	 : R := 	35	;
  	mpMcBrydeQ	 : R := 	36	;
  	mpMcBrydeSin	 : R := 	37	;
  	mpMcBryde	 : R := 	38	;
  	mpMcBrydeSin2	 : R := 	39	;
  	mpMercator	 : R := 	40	;
  	mpMiller	 : R := 	41	;
  	mpMollweide	 : R := 	42	;
  	mpNell	         : R := 	43	;
  	mpOrthographic	 : R := 	44	;
  	mpPeters	 : R := 	45	;
  	mpPolyconic	 : R := 	46	;
  	mpPutnins	 : R := 	47	;
  	mpQuarticAu	 : R := 	48	;
  	mpRobinson	 : R := 	49	;
  	mpSinusoidal	 : R := 	50	;
  	mpStereographic	 : R := 	51	;
  	mpTMercator	 : R := 	52	;
  	mpTristanEdwards : R := 	53	;
  	mpUrmaev	 : R := 	54	;
  	mpUrmaevV	 : R := 	55	;
  	mpGrinten	 : R := 	56	;
  	mpWagnerI	 : R := 	57	;
  	mpWagnerII	 : R := 	58	;
  	mpWagnerIII	 : R := 	59	;
  	mpWagnerIV	 : R := 	60	;
  	mpWagnerVI	 : R := 	61	;
  	mpWagnerVII	 : R := 	62	;
  	mpWinkelI	 : R := 	63	;
  	mpWinkelII	 : R := 	64	;

end;
result := R;
end;
//------------------------------------------------------------------------------
Procedure TEZProjectionL.CalculateRef;                                           // Calculate the map values from reference points
begin
Case MapRef.Projection of
  mpUTM              : UTMRef;
  mpDatumUTM         : UTMRef;
  mpNationalGrid     : UTMRef;
  mpOrthographic     : OrthRef;
  mpWinkelI          : ErrorLog(139);//WinkelIRef;                              // These still need some testing
  mpWinkelII         : ErrorLog(169);//WinkelIIRef;
  mpSinusoidal       : ErrorLog(179);//SinusoidalRef;
  mpEquirectangular  : ErrorLog(189);//EquirectRef;
  mpBStereo          : ErrorLog(199);//BStereoRef;
  mpHammer           : ErrorLog(219);//HammerRef;
  mpAzimuthalEA      : ErrorLog(229);
  mpLaAzimuthalEA    : ErrorLog(239);
  mpGrinten          : ErrorLog(249);
  mpPolyconic        : ErrorLog(259);
  mpRobinson         : ErrorLog(269);
  mpLaConformalConic : ErrorLog(279);
  mpGnomonic         : ErrorLog(289);
  mpTMercator        : ErrorLog(299);
  mpCassiniS         : ErrorLog(309);
  mpMercator         : ErrorLog(319);
  mpMollweide        : ErrorLog(329);
  mpWagnerIV         : ErrorLog(339);
  mpWagnerVII        : ErrorLog(349);
  mpIMollweide       : ErrorLog(359);
  mpEDConical        : ErrorLog(369);
  mpGvnsp            : ErrorLog(209);
  mpAlbertsCEA       : ErrorLog(379);
  mpStereographic    : ErrorLog(389);
  mpEckertII         : ErrorLog(399);
  mpBonne            : ErrorLog(400);
  mpLamCylindricalEA : ErrorLog(410);
  mpBerCylindricalEA : ErrorLog(420);
  mpTristanEdwards   : ErrorLog(430);
  mpPeters           : ErrorLog(440);
  mpGallOrth         : ErrorLog(450);
  mpBalthasart       : ErrorLog(460);
  mpCollignon        : ErrorLog(470);
  mpEckertI          : ErrorLog(480);
  mpEckertIII        : ErrorLog(490);
  mpPutnins          : ErrorLog(500);
  mpKavraiskyVII     : ErrorLog(510);
  mpWagnerVI         : ErrorLog(520);
  mpEckertV          : ErrorLog(530);
  mpCraster          : ErrorLog(540);
  mpEckertIV         : ErrorLog(550);
  mpFahey            : ErrorLog(560);
  mpFoucautS         : ErrorLog(570);
  mpGall             : ErrorLog(580);
  mpEckertVI         : ErrorLog(590);
  mpMcBryde          : ErrorLog(600);
  mpLoximuthal       : ErrorLog(610);
  mpMiller           : ErrorLog(620);
  mpNell             : ErrorLog(630);
  mpFoucaut          : ErrorLog(640);
  mpKavraiskyV       : ErrorLog(650);
  mpMcBrydeSin       : ErrorLog(660);
  mpQuarticAu        : ErrorLog(670);
  mpWagnerII         : ErrorLog(680);
  mpWagnerIII        : ErrorLog(690);
  mpUrmaev           : ErrorLog(700);
  mpWagnerI          : ErrorLog(710);
  mpUrmaevV          : ErrorLog(720);
  mpMcBrydeSin2      : ErrorLog(730);
  mpMcBrydeP         : ErrorLog(740);
  mpMcBrydeQ         : ErrorLog(750);
  mpPutnins2         : ErrorLog(760);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.CalculateFor(Theta,Phi : Double);                       // calculate Geo or UTM to image X,Y;
var                                                                             // All calculations will have there results in FloatPoint
X1,Y1:Double;
begin
Flag       := 0;
FloatPoint.X := 0;                                                              // Just in case I forgot something
FloatPoint.Y := 0;
ConvertFor(Theta,Phi);                                                          // Converts Input format to map format

Case MapRef.Projection of
  mpUTM              : UTMFor;
  mpDatumUTM         : UTMFor;
  mpNationalGrid     : UTMFor;
  mpOrthographic     : OrthFor;
  mpWinkelI          : WinkelIFor;
  mpWinkelII         : WinkelIIFor;
  mpSinusoidal       : SinusoidalFor;
  mpEquirectangular  : EquirectFor;
  mpBStereo          : BStereoFor;
  mpHammer           : HammerFor;
  mpAzimuthalEA      : AzimuthalEAFor;
  mpLaAzimuthalEA    : LaAzimuthalEAFor;
  mpGrinten          : GrintenFor;
  mpPolyconic        : PolyconicFor;
  mpRobinson         : RobinsonFor;
  mpLaConformalConic : LaCConicFor;
  mpGnomonic         : GnomonicFor;
  mpTMercator        : TMercatorFor;
  mpCassiniS         : CassiniSFor;
  mpMercator         : MercatorFor;
  mpMollweide        : MollweideFor;
  mpWagnerIV         : WagnerIVFor;
  mpWagnerVII        : WagnerVIIFor;
  mpIMollweide       : IMollweideFor;
  mpEDConical        : ConicalEDFor;
  mpGvnsp            : GvnspFor;
  mpAlbertsCEA       : AlbertsCEAFor;
  mpStereographic    : StereographicFor;
  mpEckertII         : EckertTFor;
  mpBonne            : BonneFor;
  mpLamCylindricalEA : LamCylindricalEAFor;
  mpBerCylindricalEA : BerCylindricalEAFor;
  mpTristanEdwards   : TristanEdwardsFor;
  mpPeters           : PetersFor;
  mpGallOrth         : GallOrthFor;
  mpBalthasart       : BalthasartFor;
  mpCollignon        : CollignonFor;
  mpEckertI          : EckertIFor;
  mpEckertIII        : EckertIIIFor;
  mpPutnins          : PutninsFor;
  mpKavraiskyVII     : KavraiskyVIIFor;
  mpWagnerVI         : WagnerVIFor;
  mpEckertV          : EckertVFor;
  mpCraster          : CrasterFor;
  mpEckertIV         : EckertIVFor;
  mpFahey            : FaheyFor;
  mpFoucautS         : FoucautSFor;
  mpGall             : GallFor;
  mpEckertVI         : EckertVIFor;
  mpMcBryde          : McBrydeFor;
  mpLoximuthal       : LoximuthalFor;
  mpMiller           : MillerFor;
  mpNell             : NellFor;
  mpFoucaut          : FoucautFor;
  mpKavraiskyV       : KavraiskyVFor;
  mpMcBrydeSin       : McBrydeSinFor;
  mpQuarticAu        : QuarticAuFor;
  mpWagnerII         : WagnerIIFor;
  mpWagnerIII        : WagnerIIIFor;
  mpUrmaev           : UrmaevFor;
  mpWagnerI          : WagnerIFor;
  mpUrmaevV          : UrmaevVFor;
  mpMcBrydeSin2      : McBrydeSin2For;
  mpMcBrydeP         : McBrydePFor;
  mpMcBrydeQ         : McBrydeQFor;
//  mpPutnins2         : Putnins2For;
end;                                                                            // If the image is scanned it can be rotated

                                                                                // The resulting x,y will be rotated by MapRef.Values.Ra radins
X1 :=  FloatPoint.X - MapRef.Values.Rx;                                         // on the point MapRef.Values.Rx,y. Uses x' := cos(a)x + sin(a)y + b; y' := -sin(a)x + cos(a)y + c;
Y1 :=  FloatPoint.Y - MapRef.Values.Ry;
FloatPoint.X :=   Cos(MapRef.Values.Ra)*(X1) + Sin(MapRef.Values.Ra)*(Y1) +
                      MapRef.Values.Rx;
FloatPoint.Y :=  -Sin(MapRef.Values.Ra)*(X1) + Cos(MapRef.Values.Ra)*(Y1) +
                      MapRef.Values.Ry;
                                                                                // the flag = 10000 might be a reminant of an unsuccesfull computation -> no event
if (Flag < 10000) and (Flag = 0) then
    ForEvent(Self,FloatPoint.X,FloatPoint.Y,Flag)
else if (Flag < 10000) then ForEvent(Self,0,0,Flag);                            // faild computation -> Lat,Lon = 0
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.CalculateInv(X,Y : Double);                             // calculate image X,Y to Geo or UTM
var
X1,Y1:Double;
begin                                                                           // If the image is scanned it can be rotated
X1 :=  X - MapRef.Values.Rx;                                                    // The resulting x,y will be rotated by -MapRef.Values.Ra radins (inverse computation so angle inverse)
Y1 :=  Y - MapRef.Values.Ry;                                                    // on the point MapRef.Values.Rx,y. Uses x' := cos(a)x + sin(a)y + b; y' := -sin(a)x + cos(a)y + c;
FloatPoint.X :=   Cos(-MapRef.Values.Ra)*(X1) + Sin(-MapRef.Values.Ra)*(Y1) +
                       MapRef.Values.Rx;
FloatPoint.Y :=  -Sin(-MapRef.Values.Ra)*(X1) + Cos(-MapRef.Values.Ra)*(Y1) +
                       MapRef.Values.Ry;

Flag       := 0;

Case MapRef.Projection of                                                       // All calculations will have there results in FloatPoint
  mpUTM              : UTMInv;
  mpDatumUTM         : UTMInv;
  mpNationalGrid     : UTMInv;
  mpOrthographic     : OrthInv;
  mpWinkelI          : WinkelIInv;
  mpWinkelII         : WinkelIIInv;
  mpSinusoidal       : SinusoidalInv;
  mpEquirectangular  : EquirectInv;
  mpBStereo          : BStereoInv;
  mpHammer           : HammerInv;
  mpAzimuthalEA      : AzimuthalEAInv;
  mpLaAzimuthalEA    : LaAzimuthalEAInv;
  mpGrinten          : GrintenInv;
  mpPolyconic        : PolyconicInv;
  mpRobinson         : RobinsonInv;
  mpLaConformalConic : LaCConicInv;
  mpGnomonic         : GnomonicInv;
  mpTMercator        : TMercatorInv;
  mpCassiniS         : CassiniSInv;
  mpMercator         : MercatorInv;
  mpMollweide        : MollweideInv;
  mpWagnerIV         : WagnerIVInv;
  mpWagnerVII        : WagnerVIIInv;
  mpIMollweide       : IMollweideInv;
  mpEDConical        : ConicalEDInv;
  mpGvnsp            : GvnspInv;
  mpAlbertsCEA       : AlbertsCEAInv;
  mpStereographic    : StereographicInv;
  mpEckertII         : EckertTInv;
  mpBonne            : BonneInv;
  mpLamCylindricalEA : LamCylindricalEAInv;
  mpBerCylindricalEA : BerCylindricalEAInv;
  mpTristanEdwards   : TristanEdwardsInv;
  mpPeters           : PetersInv;
  mpGallOrth         : GallOrthInv;
  mpBalthasart       : BalthasartInv;
  mpCollignon        : CollignonInv;
  mpEckertI          : EckertIInv;
  mpEckertIII        : EckertIIIInv;
  mpPutnins          : PutninsInv;
  mpKavraiskyVII     : KavraiskyVIIInv;
  mpWagnerVI         : WagnerVIInv;
  mpEckertV          : EckertVInv;
  mpCraster          : CrasterInv;
  mpEckertIV         : ECkertIVInv;
  mpFahey            : FaheyInv;
  mpFoucautS         : FoucautSInv;
  mpGall             : GallInv;
  mpEckertVI         : EckertVIInv;
  mpMcBryde          : McBrydeInv;
  mpLoximuthal       : LoximuthalInv;
  mpMiller           : MillerInv;
  mpNell             : NellInv;
  mpFoucaut          : FoucautInv;
  mpKavraiskyV       : KavraiskyVInv;
  mpMcBrydeSin       : McBrydeSinInv;
  mpQuarticAu        : QuarticAuInv;
  mpWagnerII         : WagnerIIInv;
  mpWagnerIII        : WagnerIIIInv;
  mpUrmaev           : UrmaevInv;
  mpWagnerI          : WagnerIInv;
  mpUrmaevV          : UrmaevVInv;
  mpMcBrydeSin2      : McBrydeSin2Inv;
  mpMcBrydeP         : McBrydePInv;
  mpMcBrydeQ         : McBrydeQInv;
//  mpPutnins2         : Putnins2Inv;
end;
ConvertInv(FloatPoint.X,FloatPoint.Y);                                          // Converts map format to output format
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.ErrorLog(F : integer);
begin
  Flag := F;                                                                    // So you can get the last error
  if Flag < 10000 then ErrorEvent(Self,Flag);                                   // 10000 might be used as a temporary value. This meens the function/procedure using it is not finished -> no error event to prevent faults
end;                                                                            // Do error event
//------------------------------------------------------------------------------
procedure TEZProjectionL.ErrorEvent(Sender: TObject;Return:integer);
begin
  if Assigned(FErrorEvent) then FErrorEvent(Self,Flag);                         // If and error event was assigned do errorevent
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.ForEvent(Sender: TObject; X,Y:Double;Return:integer);
begin
  if Assigned(FForEvent) then FForEvent(Self,X,Y,Flag);                         // if a forward computation has been made returns forrward computation values together with any error information.
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.InvEvent(Sender: TObject; Theta,Phi:Double;Return:integer);
begin
  if Assigned(FInvEvent) then FInvEvent(Self,Theta,Phi,Flag);                   // Same as forward but for an inverse event
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetInfoEvent(Sender: TObject; MapName: string;     // Trigggerd by procedure GetInfo
TypeGeo,Reference:boolean;EllipsType,MapNumber:integer);
begin
  if Assigned(FGetInfoEvent) then FGetInfoEvent(Self,MapName,TypeGeo,
     Reference,EllipsType,MapNumber);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetIOFormat;                                            // Get both In/Output information of MapRef
var
 i,j,k,l:integer;
begin
 i := 0;
 j := 0;
 k := MapRef.InputIndex;                                                         // In and Output sub types are already represented by an integer
 l := MapRef.OutputIndex;
                                                                                // the Input and Output types are represented by an integer
 if MapRef.Input  = ifGeographic   then i := 0;
 if MapRef.Input  = ifUTM          then i := 1;
 if MapRef.Input  = ifDatumUTM     then i := 2;
 if MapRef.Input  = ifNationalGrid then i := 3;

 if MapRef.Output = ofGeographic   then j := 0;
 if MapRef.Output = ofUTM          then j := 1;
 if MapRef.Output = ofDatumUTM     then j := 2;
 if MapRef.Output = ofNationalGrid then j := 3;

 if Assigned(FIOFormat) then FIOFormat(Self,i,j,k,l);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetEZConvert(Value:TEZConvertL);
begin
   FEZConvertL := Value;                                                         // Assign a TEZConvert component to EZConvert property
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetScanRotation(X,Y,Angle:Double);                      // Set map tilt (if scanned)
begin
 MapRef.Values.Rx := X;
 MapRef.Values.Ry := Y;
 MapRef.Values.Ra := DegToRad(Angle);                                           // Convert angle in degrees to radians (easier to use)
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.GetInfo;                                                // Triggers GetInfoEvent
var
 I,E:integer;
 T,R:boolean;
 M:string;
begin
 I := GetMapIndex;                                                              // Uses procedure GetMapIndex to get the integer representation of the pprojection type
 E := EllipsType[I];                                                            // Integer representation of ellpsoid type used by current pprojection (like mpMercator: 1 = Ellipsoid, mpNationalGrid: 3 = National Grid). 0: none, 1: ellipsoid, 2: datum, 3: national grid.
 M := MapName[I];                                                               // Full name of projection
 T := TypeGeoraphic[I];                                                         // Is the grid of the projection in Lat,Lon? -> UTM, DatumUTM and NationalGrid are false. the rest are true.
 R := RefPre [I];                                                               // Is there procedure to calculate the map-constants from reference points. If true procedure CalcRef can be called if reefeerencee points are added.
 GetInfoEvent(Self,M,T,R,E,I);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrRadius(Radius:Double);                             // Sets the radius of projection. The radius unit are pixels
begin
 MapRef.Values.Rad := Radius;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrTheta(Theta:Double);                               // Sets center longtitude for all projection types
begin
 MapRef.Values.Th0 := DegToRad(Theta);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrPhi0(Phi0:Double);                                 // Sets center lattitude for a projection
begin
 MapRef.Values.Ph0 := DegToRad(Phi0);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrPhi1(Phi1:Double);                                 // Sets latitude of first standard parallel for a conical projection
begin
 MapRef.Values.Ph1 := DegToRad(Phi1);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrPhi2(Phi2:Double);                                 // Sets latitude of second standard parallel for a conical projection
begin
 MapRef.Values.Ph2 := DegToRad(Phi2);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrX0(X0:Double);                                     // Sets x offset in pixls for the projection (could read map image)
begin
 MapRef.Values.X0 := X0;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrY0(Y0:Double);                                     // Sets y offset in pixls for the projection (could read map image)
begin
 MapRef.Values.Y0 := Y0;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrOblique(Oblique:Double);                           // Sets Oblique angle// Sets Oblique angle
begin
 MapRef.Values.Oa := DegToRad(Oblique);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrObliqueCenter(Theta,Phi:Double);                   // Sets rotation point for oblique rotation. The Theta here will be corrected with the Theta0 of the projection constants
begin
 MapRef.Values.Ot := DegToRad(Theta);
 MapRef.Values.Op := DegToRad(Phi);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetPrOther(Other:Double);                               // Sets additional properties. For 'General Vertical Near-side Perspective' it will be the height and for 'Equidistant Conical' the mode (<1 = mode A, >= 1 = mode B)
begin
 MapRef.Values.H0 := Other;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetEllipsoid(Index:integer);                            // Sets the map index (an ellipsoid) to MapRf.MapIndex
begin
  MapRef.MapIndex:=Index;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetUserEllipsoid(Radius,Flat,Scale:Double;UserZone,
  Name:string;Index:integer);
begin                                                                          // Changes the values of an existing user defined ellipsoid. If the ellipsoid does not exist create one by calling EZConvert.AddUserEllips(...);
 if Assigned(FEZConvertL) then FEZConvertL.SetUserEllipsoid(Radius,Flat,Scale,
    UserZone,Name,Index)
 else ErrorLog(401);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetUserDatum(Radius,Flat,dX,dY,dZ,Scale:Double;
  UserZone,Name:string;Index:integer);
begin                                                                           // Changes the values of an existing user defined Datum. If the ellipsoid does not exist create one by calling EZConvert.AddUserDatum(...);
if Assigned(FEZConvertL) then FEZConvertL.SetUserDatum(Radius,Flat,dX,dY,dZ,
  Scale,UserZone,Name,Index)
else ErrorLog(401);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetUserNatGrid(Radius,Flat,FalseEasting,FalseNorthing,
  Scale,CentralLong:Double;Name:string;Index:integer);
begin                                                                           // Changes the values of an existing user defined national grid. If the ellipsoid does not exist create one by calling EZConvert.AddUserNatGid(...);
if Assigned(FEZConvertL) then FEZConvertL.SetUserNatGrid(Radius,Flat,FalseEasting,
  FalseNorthing,Scale,CentralLong,Name,Index)
else ErrorLog(401);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.GetMapName:string;                                   // Returnes Mapref.FilePath. File path is used to store the path to the corresponding map for the loaded map constants.. Is used by EZMapImage
begin
  Result := MapRef.MapName;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetMapName(Text:string);                            // Stores a file path of a map to MapRef.FilePath
begin
  MapRef.MapName := Text;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetConstants;                                           // Envokes the OnGetConstants event. This to get the basic map values (constants)
var
A,B,C,D,E,F,G,H,Oblique:Double;
Ind : integer;
begin
  Ind := GetMapIndex;

  if Ind <= 2 then begin                                                        // Set values for UTM like projection
     A := MapRef.Values.Rad;
     B := MapRef.Values.Th0;
     C := MapRef.Values.Ph0;
     D := MapRef.Values.Ph1;
     E := MapRef.Values.Ph2;
     F := MapRef.Values.X0;
     G := 0;
     H := 0;
     Oblique := 0;
  end else begin                                                                // Set values for geographic projection
     A := MapRef.Values.Rad;
     B := RadToDeg(MapRef.Values.Th0);
     C := RadToDeg(MapRef.Values.Ph0);
     D := RadToDeg(MapRef.Values.Ph1);
     E := RadToDeg(MapRef.Values.Ph2);
     F := MapRef.Values.X0;
     G := MapRef.Values.Y0;
     Oblique := RadToDeg(MapRef.Values.Oa);
     H := MapRef.Values.H0;
  end;

  FGetconstants(Self,A,B,C,D,E,F,G,Oblique,H);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SetGridConstants(A1,A2,A3,B1,B2,B3:Double);             // Sets grid properties to MMapRef.GridProp
begin
 MapRef.Values.Rad := A1;
 MapRef.Values.Th0 := A2;
 MapRef.Values.Ph0 := A3;
 MapRef.Values.Ph1 := B1;
 MapRef.Values.Ph2 := B2;
 MapRef.Values.X0 := B3;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetProjectionList(Str:TStrings;GridInclude:boolean);    // Get the list (like for a listbox) of all national grids from EZDatum
var
I,J : integer;
begin
 if GridInclude then J := 0 else J := 3;                                        // Do the UTM like projections need to be included? if not start interation at the point after UTM like grids
 Str.Clear;                                                                     // Clear assigned stringlist
 for I := J to Max_Pr do begin                                                      // Add sequential all projections to assigned stringlist
   Str.Add(MapName[I]);
 end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetEllipsoid(Str:TStrings);                              // Get the list (like for a listbox) of all ellipsoids from EZEllips
var
i,J : integer;
begin
 Str.Clear;                                                                     // Clear assigned stringlist
 for i := 0 to MaxEllipsoid - 1 do Str.Add(EllipsoidName[i]);                   // Add sequentially all ellipsoids till maximum number of ellipsoids reached

 if Assigned(FEZConvertL) then begin                                             // is EZConvert assigned. If so, User defined ellipsoids might be present
    j := FEZConvertL.UserDat.NE - 1;                                             // Get number of user defined ellipsoids
    for i := 0 to j do Str.Add(FEZConvertL.UserDat.Ellips[i].Name);              // Add user defined ellipsoids to stringlist
 end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetDatum(Str:TStrings);                                 // Get the list (like for a listbox) of all Datums from EZDatum
var
i,j : integer;
begin
 Str.Clear;                                                                     // Clear assigned stringlist
 for i := 0 to MaxDatum - 1 do Str.Add(DatumName[i]);                           // Add sequentially all datums till maximum number of datums reached

 if Assigned(FEZConvertL) then begin                                             // is EZConvert assigned. If so, User defined datums might be present
    j := FEZConvertL.UserDat.ND - 1;                                             // Get number of user defined datums
    for i := 0 to j do Str.Add(FEZConvertL.UserDat.Datum[i].Name);               // Add user defined datums to stringlist
 end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GetNatGrid(Str:TStrings);                               // Get the list (like for a listbox) of all National Grids from EZDatum
var
i,j : integer;
begin
 Str.Clear;                                                                     // Clear assigned stringlist
 for i := 0 to MaxNationalGrid - 1 do Str.Add(NatGridName[i]);                  // Add sequentially all national grids till maximum number of grids reached

 if Assigned(FEZConvertL) then begin                                             // is EZConvert assigned. If so, User defined national grids might be present
    j := FEZConvertL.UserDat.NN - 1;                                             // Get number of user defined national grids
    for i := 0 to j do Str.Add(FEZConvertL.UserDat.NatGrid[i].Name);             // Add user defined national grids to stringlist
 end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SaveMapRefToStream(Stream: TStream);                    // Stores map constants, reference points, grid information,.. to a stream
begin
 Stream.Write(MapRef,SizeOf(TRef));
 ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.LoadMapRefFromStream(Stream: TStream);                  // Loads all values for MapRef from a Stream
begin
 Stream.Read(MapRef,SizeOf(TRef));
 ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.SaveMapRefToFile(FileName:TFileName);                   // Saves map constants, reference points, grid information,.. to the file FileName
var
Stream : TMemoryStream;
begin
 Stream := TMemoryStream.Create;                                                // Create stream (TMemoryStream can save easly to a file)
 Stream.Write(MapRef,SizeOf(TRef));                                             // Write constants to stream (can be done whithout iteration because TRef has no dynamic arrays)
 Stream.SaveToFile(FileName);                                                   // Save stream to file
 Stream.Free;                                                                   // Free stream
 ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.LoadMapRefFromFile(FileName:TFileName);                 // Loads all values for MapRef from a Stream
var
Stream : TMemoryStream;
begin
 Stream := TMemoryStream.Create;                                                // Create stream
 Stream.LoadFromFile(FileName);                                                 // Load File to stream
 Stream.Read(MapRef,SizeOf(TRef));                                              // Read constants from stream
 Stream.Free;                                                                   // FFrree stream
 ErrorLog(0);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.AddRefPoint(X,Y,Theta,Phi:Double);                      // Add a point (lon,lat,x,y) to the reference list
begin
if MapRef.CountRef < 30 then begin                                              // If maximum amount of reference points of 30 is not reached
  MapRef.CountRef := MapRef.CountRef + 1;                                       // Get current amount
  MapRef.Points[MapRef.CountRef-1].X := X;                                      // Set values in new lokation
  MapRef.Points[MapRef.CountRef-1].Y := Y;
  MapRef.Points[MapRef.CountRef-1].Theta := Theta;
  MapRef.Points[MapRef.CountRef-1].Phi   := Phi;
  ErrorLog(0);                                                                  // All OK
end else ErrorLog(120);                                                         // If maximum amount of reference point have been reached
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.DeleteRefPoint(index : integer);                        // Delete a point from the rreference list
var
i : integer;
begin
if (index < 0) and (Index > MapRef.CountRef -1) then ErrorLog(122)              // Does the refernce point exist? If not give error
else if index < MapRef.CountRef -1 then begin                                   // It does exist and is not last point in list (array)
  for i := index to MapRef.CountRef  - 1 do begin                               // Move all points above index point one position lower -> index pointt gets replaced
    MapRef.Points[i].X     := MapRef.Points[i+1].X;
    MapRef.Points[i].Y     := MapRef.Points[i+1].Y;
    MapRef.Points[i].Theta := MapRef.Points[i+1].Theta;
    MapRef.Points[i].Phi   := MapRef.Points[i+1].Phi;
  end;
  MapRef.Points[MapRef.CountRef-1].X  := 0;                                     // Set last point in list (is now double due to replacement) to zero
  MapRef.Points[MapRef.CountRef-1].Y  := 0;
  MapRef.Points[MapRef.CountRef-1].Theta  := 0;
  MapRef.Points[MapRef.CountRef-1].Phi  := 0;
  MapRef.CountRef := MapRef.CountRef - 1;                                       // One point less
  ErrorLog(0);
end else if index = MapRef.CountRef -1 then begin                               // Index point is last in list. AAll nneed to do is set zero to delete
   MapRef.Points[MapRef.CountRef-1].X  := 0;
   MapRef.Points[MapRef.CountRef-1].Y  := 0;
   MapRef.Points[MapRef.CountRef-1].Theta  := 0;
   MapRef.Points[MapRef.CountRef-1].Phi  := 0;
   MapRef.CountRef := MapRef.CountRef - 1;                                      // One point less
   ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.GetTemp:TFloatPoint;
begin
  if (Assigned(FEZConvertL)) and (MapRef.MapIndex > MaxEllipsoid -1)
     and (MapRef.MapIndex <> 220) then begin                                    // Get user defined Ellipsoid
    Result.X   := FEZConvertL.UserDat.Ellips[MapRef.MapIndex].Radius;
    Result.Y   := RMinor/Result.X;
  end else if (MapRef.MapIndex > MaxEllipsoid-1) then begin                     // Or get WGS84 ellipsoid if EZConvert has not been assigned or if MapIndex is 220 (WGS84)
    Result.X   := Ellipsoid[30,0];
    Result.Y   := RMinor/Result.X;
  end else begin                                                                // Or just get existing ellipsoid
    Result.X   := Ellipsoid[MapRef.MapIndex,0];
    Result.Y   := RMinor/Result.X;
  end;
end;
//////////////////////  Conversion Geo <-> Other  //////////////////////////////
function TEZProjectionL.GetNorth(Theta,Phi:Double):Double;                       // Returns the angle in degrees in wich direction North could be found on the currently used map image (if you would use any) from point Theta,Phi. The marker in TEZMapImage uses this function.
var
i,f : integer;
X1,Y1,X2,Y2,Des,An,Ac:Double;
begin                                                                           // Set all to zero to prevent problems
X1 := 0;
X2 := 0;
Y1 := 0;
Y2 := 0;
f  := 0;
Flag := 0;
Des:= 0;

for i := 0 to 1 do begin                                                        // Calculate twice: First current coordinat and second coordinate 1/10 degree down. from this the angle can be calculated
  if Assigned(FEZConvertL) then begin                                            // Is EZConvert assigned to do any grid conversions

     if i = 1 then Phi := Phi - 0.1;                                            // Second run calculate second point 1/10 degree down
     CalculateFor(Theta,Phi);                                                   // Calculate
     f := Flag + f;                                                           // Store any errors from first and second run. If any have an error total <> 0

     if i = 0 then begin                                                        // Store points from first and second run
        X1 := FloatPoint.X;
        Y1 := FloatPoint.Y;
     end else begin
        X2 := FloatPoint.X;
        Y2 := FloatPoint.Y;
     end;
  end else ErrorLog(401);
end;
  if f = 0 then Des := sqrt(sqr(X2-X1) + sqr(Y2-Y1));                           // If no errors occured calculate distance betweeen points

  if Des > 0.000000001 then begin                                               // If points do not coincide calculate angle
     An := (X2-X1)/Des;
     Ac := (Y2-Y1)/Des;
     if Ac < 0 then Result := -RadToDeg(Pi-ArcSin(An))                          // Calculate angle
     else Result := -RadToDeg(ArcSin(An));
     FloatPoint.X := X1;                                                        // Store first (position on map) point. Else the Marker in EZMapImage shows a shift.
     FloatPoint.Y := Y1;
  end else begin                                                                // Oops. They do coincide
     FloatPoint.X := 0;
     FloatPoint.Y := 0;
     Result := 0;
     ErrorLog(400);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.DoTruncUTM;                                             // If property TruncUTM is set to true it truncs all (excessive) decimals from CalculateInv.
begin
 if FTruncUTM then begin
   FloatPoint.X := Trunc(FloatPoint.X);
   FloatPoint.Y := Trunc(FloatPoint.Y);
 end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.Protect(Index:integer);
var
E,D,N:integer;
begin
 if Assigned(FEZConvertL) then begin
    E := FEZConvertL.UserDat.NE;
    D := FEZConvertL.UserDat.ND;
    N := FEZConvertL.UserDat.NN;
 end else begin
    E := 0;
    D := 0;
    N := 0;
 end;
     if (Index = 0) and (EllipsType[GetMapIndex] = 1) then
        if (MaxEllipsoid + E < MapRef.MapIndex) or (MapRef.MapIndex < 0) then
           MapRef.MapIndex := DefEllipsoid;

     if (Index = 0) and (EllipsType[GetMapIndex] = 0) then
        if (MaxDatum + D < MapRef.MapIndex) or (MapRef.MapIndex < 0) then
           MapRef.MapIndex := DefDatum;

     if (Index = 1) then
        if (MaxEllipsoid + E < MapRef.MapIndex) or (MapRef.MapIndex < 0) then
           MapRef.MapIndex := DefEllipsoid;

     if (Index = 2) then
        if (MaxDatum + D < MapRef.MapIndex) or (MapRef.MapIndex < 0) then
           MapRef.MapIndex := DefDatum;

     if (Index = 3) then
        if (MaxNationalGrid + N < MapRef.MapIndex) or (MapRef.MapIndex < 0) then
           MapRef.MapIndex := DefNationalGrid;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.ConvertFor(Theta,Phi : Double);                         // Adjust the Lon,Lat to the aropriate format of the projection
var
FromType,ToType:integer;
begin

  if Assigned(FEZConvertL) then begin                                            // If EZConvert has not been assigned conversion between UTM,Datum UTM, National gridss and geographics can not take place
     FEZConvertL.UTMZone := MapRef.Zone;                                         // Set zone in EZConvert to current zone used

     FromType := 0;                                                             // Get an integer value for both Input format and map index to convert using EZConver.FromTo
     if MapRef.Input = ifGeographic   then FromType := 0;
     if MapRef.Input = ifUTM          then FromType := 1;
     if MapRef.Input = ifDatumUTM     then FromType := 2;
     if MapRef.Input = ifNationalGrid then FromType := 3;

     ToType := 0;
     if MapRef.Projection = mpUTM          then ToType := 1;
     if MapRef.Projection = mpDatumUTM     then ToType := 2;
     if MapRef.Projection = mpNationalGrid then ToType := 3;

     Protect(ToType);
     if (EllipsType[GetMapIndex] = 1) and (ToType = 0) then                     // If the map uses an ellipsoid so the conversion is not bidirectional (only onversion for input to WGS84 needed). if this is not done the transverse mecator usig a spheroid will not give the proper projetion
       FEZConvertL.Translate(false,Theta,Phi,MapRef.InputIndex)
     else FEZConvertL.FromTo(Theta,Phi,FromType,ToType,MapRef.InputIndex,MapRef.MapIndex);


     Theta := FEZConvertL.FloatPoint.X;                                          // This is a little overdone  but it is needed for the following conversion of DegToRad
     Phi   := FEZConvertL.FloatPoint.Y;

     MapRef.Zone := FEZConvertL.UTMZone;                                         // Set zone back
  end;

if (MapRef.Projection = mpUTM) or (MapRef.Projection = mpDatumUTM) or
   (MapRef.Projection = mpNationalGrid) then begin                              // Set values from conversion to (backbone) FloatPoint
  FloatPoint.X := Theta;
  FloatPoint.Y := Phi;
end else begin                                                                  // Convert degrees to radians and set values from conversion to (backbone) FloatPoint
  FloatPoint.X := DegToRad(Theta);
  FloatPoint.Y := DegToRad(Phi);
  if abs(MapRef.Values.Oa) > 0.0001 then Rotate(FloatPoint.X,FloatPoint.Y,true);// If oblate angle > 0.0001 do oblate. aves some time if  no oblation is used
end;

end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.ConvertInv(X,Y : Double);                               // Adjust the format of the projection to the aropriate Lat,Lon
var
SX,CX : Double;
FromType,ToType:integer;
begin
if (MapRef.Projection = mpUTM) or (MapRef.Projection = mpDatumUTM) or           // No oblate if projection type is UTM like
   (MapRef.Projection = mpNationalGrid) then begin
  FloatPoint.X := X;
  FloatPoint.Y := Y;
end else begin
 if abs(MapRef.Values.Oa) > 0.0001 then Rotate(X,Y,false) else begin           // Oblate if angle is bigger than 0.0001
     FloatPoint.X := X;
     FloatPoint.Y := Y;
  end;
  SX := Sin(FloatPoint.X);                                                      // Blunt way of keeping the values for Lattitude and Longtitude between 90/-90 and 180 and -180 degrees (but then in radians)
  CX := Cos(FloatPoint.X);
  if (CX < 0) and (SX > 0) then X := Pi - ArcSin(SX);
  if (CX < 0) and (SX < 0) then X := -Pi - ArcSin(SX);
  if (CX > 0) and (SX > 0) then X := ArcSin(SX);
  if (CX > 0) and (SX < 0) then X := ArcSin(SX);
  Y := ArcSin(Sin(FloatPoint.Y));

  FloatPoint.X := RadToDeg(X);                                                  // Convert radians to degrees
  FloatPoint.Y := RadToDeg(Y);
end;
     FromType := 0;                                                             // Just to prevent warnings
     ToType := 0;

  if Assigned(FEZConvertL) then begin                                            // If EZConvert has been assigned any conversions can take place
     FEZConvertL.UTMZone := MapRef.Zone;                                         // Assign zone to EZConvert

     if MapRef.Projection = mpUTM          then FromType := 1;                  // Get integer values for both projection type and output type. This is needed for EZConvert.FromTo procedure
     if MapRef.Projection = mpDatumUTM     then FromType := 2;
     if MapRef.Projection = mpNationalGrid then FromType := 3;

     Protect(FromType);

     if MapRef.Output = ofGeographic   then ToType := 0;
     if MapRef.Output = ofUTM          then ToType := 1;
     if MapRef.Output = ofDatumUTM     then ToType := 2;
     if MapRef.Output = ofNationalGrid then ToType := 3;

{     if ToType = 0 then ToIndex := DefDatum                                     // EZConvert will see the MapIndex as Datum and not as ellipsoid. Also the projection takes care of the ellips tranlation. So no translation needed -> index needs to be the default datum, WGS84, and will be ignored.
     else ToIndex := MapRef.OutputIndex;}


//     if (EllipsType[GetMapIndex] = 1) and (ToType = 0) then                     // If the map uses an ellipsoid so the conversion is not bidirectional (only onversion for input to WGS84 needed). if this is not done the transverse mecator usig a spheroid will not give the proper projetion
//       FEZConvertL.Translate(true,FloatPoint.X,FloatPoint.Y,MapRef.InputIndex)
//     else FEZConvertL.FromTo(FloatPoint.X,FloatPoint.Y,FromType,ToType,MapRef.MapIndex,MapRef.OutPutIndex);

     FEZConvertL.FromTo(FloatPoint.X,FloatPoint.Y,FromType,ToType,               // Convert projection fformat to output format
                       MapRef.MapIndex,MapRef.OutputIndex);

     FloatPoint.X := FEZConvertL.FloatPoint.X;                                   // Assign EZConverts FloatPoint to EZProjectionLs FloatPoint
     FloatPoint.Y := FEZConvertL.FloatPoint.Y;                                   // They are not the same -> declared twice because EZConvert does not have to be assigned to make EZProjectionL work.

     MapRef.Zone := FEZConvertL.UTMZone;                                         // Set zone back
  end;

  if ToType <> 0 then DoTruncUTM;

if (Flag < 10000) and (Flag = 0) then InvEvent(Self,FloatPoint.X,FloatPoint.Y,Flag)
else if (Flag < 10000) then InvEvent(Self,0,0,Flag);
end;
////////////////////////////////  Projections   ////////////////////////////////
//------------------------------------------------------------------------------     Putnins2


/////////////////////////////////  McBrydeP  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeQFor;
var
  th,c,p : double;
  i : integer;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(751) else begin
    PrepFor1;

    c := 1.70710678118654752440*Sin(FloatPoint.Y);
    for I := 0 to 20 do begin
      th := (sin(FloatPoint.Y/2)+Sin(FloatPoint.Y)-c)/(Cos(FloatPoint.Y/2)/2 + Cos(FloatPoint.Y));
      FloatPoint.Y := FloatPoint.Y-th;
      if (abs(th) < 1E-7) then break;
    end;
    p := FloatPoint.Y;
    FloatPoint.X := 0.31245971410378249250*FloatPoint.X*(1+2*Cos(p)/Cos(p/2));
    FloatPoint.Y := 1.87475828462269495505*Sin(p/2);

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeQInv;
var
  F : byte;
  t,p : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(752) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F :=3;

    T := 1;
    p := FloatPoint.Y*0.53340209679417701685;
    if (abs(p) >= 1) then begin
       if (abs(p) > 1) then F := 4
       else if (p < 0) then begin
         t := -1;
         p := -Pi;
       end else begin
         t := 1;
         p := Pi;
       end;
    end else begin
      t := p;
      p := 2*ASinZ(t);
    end;
    FloatPoint.X := 3.20041258076506210122*FloatPoint.X/(1+2*Cos(p)/Cos(p/2));
    FloatPoint.Y := 0.58578643762690495119*(t+Sin(p));
    if (abs(FloatPoint.Y) >= 1) then begin
       if (abs(FloatPoint.Y) > 1) then F := 5
       else if (FloatPoint.Y < 0) then FloatPoint.Y := -Pi/2
       else FloatPoint.Y := Pi/2;
    end else FloatPoint.Y := ASinZ(FloatPoint.Y);

    PrepInv2(752,F);
  end;
end;
/////////////////////////////////  McBrydeP  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydePFor;
var
  p : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(741) else begin
    PrepFor1;

    P := ASinZ(0.95257934441568037152*Sin(FloatPoint.Y));
    FloatPoint.X := 0.92582009977255146156*FloatPoint.X*(2*Cos(2/3*P)-1);
    FloatPoint.Y := 3.40168025708304504493*Sin(P/3);

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydePInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(742) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F :=3;

    FloatPoint.Y := FloatPoint.Y/3.40168025708304504493;
    if (abs(FloatPoint.Y) >= 1) then begin
       if (abs(FloatPoint.Y) > 1) then F := 4
       else if (FloatPoint.Y < 0) then FloatPoint.Y := -Pi/2
       else FloatPoint.Y := Pi/2;
    end else FloatPoint.Y := ASinZ(FloatPoint.Y);
    FloatPoint.Y := FloatPoint.Y*3;
    FloatPoint.X := FloatPoint.X/(0.92582009977255146156*(2*Cos(2/3*FloatPoint.Y)-1));
    FloatPoint.Y := Sin(FloatPoint.Y)/0.95257934441568037152;
    if (abs(FloatPoint.Y) >= 1) then begin
       if (abs(FloatPoint.Y) > 1) then F := 5
       else if (FloatPoint.Y < 0) then FloatPoint.Y := -Pi/2
       else FloatPoint.Y := Pi/2;
    end else FloatPoint.Y := ASinZ(FloatPoint.Y);

    PrepInv2(742,F);
  end;
end;
/////////////////////////////////  McBrydeSin2  ////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeSin2For;
var
 k,v,t : double;
 i : integer;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(731) else begin
    PrepFor1;

    k := 1.41546*Sin(FloatPoint.Y);
    for I := 0 to 10 do begin
       t := FloatPoint.Y/1.36509;
       v := (0.45503*Sin(t) + Sin(FloatPoint.Y) -k)/(1/3*Cos(t) + Cos(FloatPoint.Y));
       FloatPoint.Y := FloatPoint.Y - v;
       if (abs(V) < 1E-7) then break;
    end;
    t := FloatPoint.Y/1.36509;
    FloatPoint.X := 0.22248*FloatPoint.X*(1+3*Cos(FloatPoint.Y)/Cos(t));
    FloatPoint.Y := 1.44492*Sin(t);

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeSin2Inv;
var
  F : byte;
  t : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(732) else begin
    PrepInv1;
    F := 0;
    t := ASinZ(FloatPoint.Y/1.44492);
    FloatPoint.Y := 1.36509*t;
    if (abs(FloatPoint.Y) > Pi/2) then F :=3;
    FloatPoint.X := FloatPoint.X /(0.22248*(1+3*Cos(FloatPoint.Y)/Cos(t)));
    FloatPoint.Y := ASinZ((0.45503*Sin(t) + Sin(FloatPoint.Y))/1.41546);

    PrepInv2(732,F);
  end;
end;
/////////////////////////////////  Urmaev V  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.UrmaevVFor;
var
  t,m : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(721) else begin
    PrepFor1;

    t := ASinZ(0.8*Sin(FloatPoint.Y));
    m := 2*Power(3,0.25)/3;
    FloatPoint.X := m*FloatPoint.X*Cos(FloatPoint.Y);
    FloatPoint.Y := t*(1+0.414524*sqr(t)/3)/(m*0.8);

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.UrmaevVInv;
var
  F : byte;
  t,m,a,y,c : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(722) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F :=3;

    m := 2*Power(3,0.25)/3;
    a := 0.414524/3;
    y := FloatPoint.Y*m*0.8;
    c := Power((108*y+12*sqrt(3)*sqrt((4+27*sqr(y)*a)/a))*sqr(a),1/3);

    t := (1/6)*(c/a) - (2/c);

    FloatPoint.Y := ASinZ(Sin(t)/0.8);
    FloatPoint.X := (FloatPoint.X/Cos(FloatPoint.Y))/m;

    PrepInv2(722,F);
  end;
end;
/////////////////////////////////  Wagner I  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(711) else begin
    PrepFor1;

    FloatPoint.Y := ASinZ(0.8660254037844386467637231707*Sin(FloatPoint.Y));
    FloatPoint.X := 0.8773826753*FloatPoint.X*Cos(FloatPoint.Y);
    FloatPoint.Y := 1.139753528477*FloatPoint.Y;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIInv;
var
  F : byte;
  Y : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(712) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F :=3;

    Y := FloatPoint.Y/1.139753528477;
    FloatPoint.Y := ASinZ(Sin(Y)/0.8660254037844386467637231707);
    FloatPoint.X := FloatPoint.X/(0.8773826753*Cos(Y));

    PrepInv2(712,F);
  end;
end;
/////////////////////////////////  Urmaev  /////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.UrmaevFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(701) else begin
    PrepFor1;

    if (MapRef.Values.H0 <= 0) then MapRef.Values.H0 := 0.0001;
    if (MapRef.Values.H0 > 1)  then MapRef.Values.H0 := 1;
    FloatPoint.Y := ASinZ(MapRef.Values.H0*Sin(FloatPoint.Y));
    FloatPoint.X := 0.8773826753*FloatPoint.X*Cos(FloatPoint.Y);
    FloatPoint.Y := 1.139753528477*FloatPoint.Y;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.UrmaevInv;
var
  F : byte;
  Y : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(702) else begin
    PrepInv1;
    F := 0;

    if (MapRef.Values.H0 <= 0) then MapRef.Values.H0 := 0.0001;
    if (MapRef.Values.H0 > 1)  then MapRef.Values.H0 := 1;
    Y := FloatPoint.Y/1.139753528477;
    if (abs(Y) > Pi/2) then F :=3;
    FloatPoint.Y := ASinZ(Sin(Y)/MapRef.Values.H0);
    FloatPoint.X := FloatPoint.X/(0.8773826753*Cos(Y));

    PrepInv2(702,F);
  end;
end;
/////////////////////////////////  Wagner III  /////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIIIFor;
var
  ts : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(691) else begin
    PrepFor1;

    ts := Cos(MapRef.Values.H0)/Cos(2*MapRef.Values.H0/3);

    FloatPoint.X := ts*FloatPoint.X*Cos(2/3*FloatPoint.Y);
    FloatPoint.Y := FloatPoint.Y;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIIIInv;
var
  F : byte;
  ts : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(692) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F := 3;

    ts := Cos(MapRef.Values.H0)/Cos(2*MapRef.Values.H0/3);

    FloatPoint.Y := FloatPoint.Y;
    FloatPoint.X := FloatPoint.X/(ts*Cos(2/3*FloatPoint.Y));

    PrepInv2(692,F);
  end;
end;
/////////////////////////////////  WagnerII  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(681) else begin
    PrepFor1;

    FloatPoint.Y := ASinZ(0.88022*Sin(0.88550*FloatPoint.Y));
    FloatPoint.X := 0.92483*FloatPoint.X*Cos(FloatPoint.Y);
    FloatPoint.Y := FloatPoint.Y*1.38725;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIIInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(682) else begin
    PrepInv1;
    F := 0;
    if (abs(FloatPoint.Y) > Pi/2) then F := 3;

    FloatPoint.Y := FloatPoint.Y/1.38725;
    FloatPoint.X := FloatPoint.X/(0.92483*Cos(FloatPoint.Y));
    FloatPoint.Y := AsinZ(Sin(FloatPoint.Y)/0.88022)/0.88550;

    PrepInv2(682,F);
  end;
end;
/////////////////////////////////  QuarticAu  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.QuarticAuFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(671) else begin
    PrepFor1;
    Fouc(2,2,0);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.QuarticAuInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(672) else begin
    PrepInv1;
    F := FoucI(2,2,0);
    PrepInv2(672,F);
  end;
end;
////////////////////////////////  McBrydeSin  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeSinFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(661) else begin
    PrepFor1;
    Fouc(1.48875,1.36509,0);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeSinInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(662) else begin
    PrepInv1;
    F := FoucI(1.48875,1.36509,0);
    PrepInv2(662,F);
  end;
end;
////////////////////////////////  KavraiskyV  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.KavraiskyVFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(651) else begin
    PrepFor1;
    Fouc(1.50488,1.35439,0);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.KavraiskyVInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(652) else begin
    PrepInv1;
    F := FoucI(1.50488,1.35439,0);
    PrepInv2(652,F);
  end;
end;
///////////////////////////////////  Foucaut  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.FoucautFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(641) else begin
    PrepFor1;
    Fouc(2,2,1);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.FoucautInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(642) else begin
    PrepInv1;
    F := FoucI(2,2,1);
    PrepInv2(642,F);
  end;
end;
////////////////////////////////////  Nell  ////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.NellFor;
var
 K,V : double;
 I : integer;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(631) else begin
    PrepFor1;

    K := 2*Sin(FloatPoint.Y);
    V := sqr(FloatPoint.Y);
    FloatPoint.Y := FloatPoint.Y*(1.00371+V*(-0.0935382+V*-0.011412));
    for I := 0 to 10 do begin
      V := (FloatPoint.Y+Sin(FloatPoint.Y)-K)/(1+Cos(FloatPoint.Y));
      FloatPoint.Y := FloatPoint.Y-V;
      if abs(V) < 1E-9 then Break;
    end;

    FloatPoint.X := FloatPoint.X*(1+Cos(FloatPoint.Y))/2;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.NellInv;
var
  F : byte;
  X : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(632) else begin
    PrepInv1;

    X := 2*FloatPoint.X/(1+Cos(FloatPoint.Y));
    FloatPoint.Y := ASinZ((FloatPoint.Y+Sin(FloatPoint.Y))/2);
    FloatPoint.X := X;

    F := 0;
    PrepInv2(632,F);
  end;
end;
///////////////////////////////////  Miller  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.MillerFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(621) else begin
    PrepFor1;

    FloatPoint.X := FloatPoint.X;
    FloatPoint.Y := Ln(Tan(Pi/4+FloatPoint.Y*0.4))*1.25;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.MillerInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(622) else begin
    PrepInv1;

    FloatPoint.X := FloatPoint.X;
    FloatPoint.Y := 2.5*(ArcTan(exp(0.8*FloatPoint.Y))-Pi/4);

    F := 0;
    PrepInv2(622,F);
  end;
end;
////////////////////////////////  Loximuthal  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.LoximuthalFor;
var
 X,Y : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(611) else begin
   PrepFor1;

    Y := FloatPoint.Y-MapRef.Values.Ph0;
    if (abs(Y) < 1E-9) then X := FloatPoint.X*Cos(MapRef.Values.Ph0)
    else begin
       X := Pi/4+FloatPoint.Y/2;
       if (abs(X) < 1E-9) or (abs(abs(X)-Pi/2) < 1E-9) then X := 0
       else X := FloatPoint.X*Y/ln(tan(X)/tan(pi/4+MapRef.Values.Ph0/2));
    end;
    FloatPoint.X := X;
    FloatPoint.Y := Y;

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.LoximuthalInv;
var
  F : byte;
  P,T,Tp,Tt : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(612) else begin
    PrepInv1;

    Tp := tan(pi/4+MapRef.Values.Ph0/2);
    P := FloatPoint.Y+MapRef.Values.Ph0;
    if (abs(FloatPoint.Y) < 1E-9) then T := FloatPoint.X/Cos(MapRef.Values.Ph0)
    else begin
       T := Pi/4+P/2;
       if (abs(T) < 1E-9) or (abs(abs(T)-Pi/2) < 1E-9) or (abs(T) > 1E30) then T := 0
       else begin
         Tt := tan(T);
         if (Tt/Tp > 0) then T := FloatPoint.X*Ln(Tt/Tp)/FloatPoint.Y
         else T := 0;
       end;
    end;

    FloatPoint.X := T;
    FloatPoint.Y := P;
    F := 0;
    PrepInv2(612,F);
  end;
end;
/////////////////////////////////  McBryde  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(601) else begin
    PrepFor1;
    EcTV(1.785398163397448309615660845,0.5);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.McBrydeInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(602) else begin
    PrepInv1;
    F := EcTVI(1.785398163397448309615660845,0.5);
    PrepInv2(602,F);
  end;
end;
/////////////////////////////////  Eckert VI  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertVIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(591) else begin
    PrepFor1;
    EcTV(2.570796326794896619231321691,1);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertVIInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(592) else begin
    PrepInv1;
    F := EcTVI(2.570796326794896619231321691,1);
    PrepInv2(592,F);
  end;
end;
///////////////////////////////////  Gall  /////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.GallFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(581) else begin
    PrepFor1;
    FloatPoint.X := FloatPoint.X*0.70710678118654752440;
    FloatPoint.Y := 1.70710678118654752440*Tan(FloatPoint.Y/2);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GallInv;
var
  F : byte;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(582) else begin
    PrepInv1;
    FloatPoint.X := 1.41421356237309504880*FloatPoint.X;
    FloatPoint.Y := 2*ArcTan(0.58578643762690495119*FloatPoint.Y);
    PrepInv2(582,F);
  end;
end;
/////////////////////////////////  Foucaut  ////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.FoucautSFor;
var
  T,N : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(571) else begin
    PrepFor1;

    N := MapRef.Values.H0;
    T := Cos(FloatPoint.Y);
    FloatPoint.X := FloatPoint.X*T/(n + (1-n)*T);
    FloatPoint.Y := n*FloatPoint.Y+(1-n)*Sin(FloatPoint.Y);

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.FoucautSInv;
var
  F : byte;
  I : integer;
  V,N : double;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(572) else begin
    PrepInv1;

    N := MapRef.Values.H0;
    if (N < 0) or (N > 2) then F := 3;
    if N = 1 then for I := 0 to 15 do begin
        V := N*FloatPoint.Y+(1-N)*Sin(FloatPoint.Y)-FloatPoint.Y;
        FloatPoint.Y := FloatPoint.Y-V;
        if abs(V) < 1E-9 then Break;
    end else FloatPoint.Y := AsinZ(FloatPoint.Y);
    V := Cos(FloatPoint.Y);
    FloatPoint.X := FloatPoint.X*(N+ (1-N)*V)/V;

    PrepInv2(572,F);
  end;
end;
///////////////////////////////////  Fahey  ////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.FaheyFor;
var
  X : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(561) else begin
    PrepFor1;

    X := Tan(0.5*FloatPoint.Y);
    FloatPoint.Y := 1.819152*X;
    FloatPoint.X := 0.819152*FloatPoint.X*asqrt(1-sqr(X));

    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.FaheyInv;
var
  F : byte;
  P,T : double;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(562) else begin
    PrepInv1;

    FloatPoint.Y := FloatPoint.Y/1.819152;
    P := 2*ArcTan(FloatPoint.Y);
    FloatPoint.Y := 1-sqr(FloatPoint.Y);
    if FloatPoint.Y < 1E-9 then begin
       F := 3;
       T := 0;
    end else T := FloatPoint.X/(0.819152*sqrt(FloatPoint.Y));


    if (T > Pi) or (T < -Pi) then F := 1;
    if (P > Pi/2) or (P < -Pi/2) then F := 2;

    if (F = 0) then begin
      ErrorLog(0);
      FloatPoint.X := T+MapRef.Values.Th0;
      FloatPoint.Y := P;
    end else ErrorLog(562+F);
  end;
end;
/////////////////////////////////  Eckert IV  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIVFor;
var
  P,V,S,C : double;
  I : integer;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(551) else begin
    PrepFor1;

    P := 3.57079632679489661922*Sin(FloatPoint.Y);
    V := sqr(FloatPoint.Y);
    FloatPoint.Y := FloatPoint.Y*(0.895168+V*(0.0218849+V*0.00826809));

    for I := 0 to 10 do begin
       C := Cos(FloatPoint.Y);
       S := Sin(FloatPoint.Y);
       V := (FloatPoint.Y+S*(C+2)-P)/(1+C*(C+2)-S*S);
       FloatPoint.Y := FloatPoint.Y-V;
       if (abs(V) < 1E-9) then Break;
    end;

    FloatPoint.X := 0.42223820031577120149*FloatPoint.X*(1+cos(FloatPoint.Y));
    FloatPoint.Y := 1.32650042817700232218*sin(FloatPoint.Y);
    PrepFor2;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIVInv;
var
  F : byte;
  C,D : double;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(552) else begin
    PrepInv1;

    FloatPoint.Y := AsinZ(FloatPoint.Y/1.32650042817700232218);
    C := cos(FloatPoint.Y);
    if (abs(C) < 1E-9) then F := 3;
    FloatPoint.X := FloatPoint.X/(0.42223820031577120149*(1+(C)));
    D := (FloatPoint.Y + sin(FloatPoint.Y)*(C+2))/3.57079632679489661922;
    FloatPoint.Y := AsinZ(D);

    PrepInv2(552,F);
  end;
end;
/////////////////////////////////  Craster  ////////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.CrasterFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(541) else begin
    if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
    if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
    FloatPoint.X := FloatPoint.X-MapRef.Values.Th0;

    FloatPoint.Y := FloatPoint.Y/3;
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(0.97720502380583984317*FloatPoint.X*(2*cos(2*FloatPoint.Y)-1));
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*(3.06998012383946546542*Sin(FloatPoint.Y));

    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.CrasterInv;
var
  F : byte;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(542) else begin

    if (abs((MapRef.Values.Y0-FloatPoint.Y)*0.32573500793527994772/MapRef.Values.Rad) > 1) then F := 3
    else begin
      FloatPoint.Y := 3*ArcSin((MapRef.Values.Y0-FloatPoint.Y)*0.32573500793527994772/MapRef.Values.Rad);
      if abs(MapRef.Values.Rad*(2*cos(2*FloatPoint.Y/3)-1)) < 1E-9 then F := 4
      else FloatPoint.X := 1.02332670794648848847*(FloatPoint.X-MapRef.Values.X0)/(MapRef.Values.Rad*(2*cos(2*FloatPoint.Y/3)-1));
    end;

    PrepInv2(542,F);
  end;
end;
/////////////////////////////////  Eckert V  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertVFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(531) else begin
    if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
    if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*FloatPoint.X*0.44101277172455148219*(1+cos(FloatPoint.Y));
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*0.88202554344910296438*FloatPoint.Y;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertVInv;
var
  F : byte;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(532) else begin
    FloatPoint.Y := 1.13375401361911319568*(MapRef.Values.Y0-FloatPoint.Y)/MapRef.Values.Rad;
    FloatPoint.X := 2.26750802723822639137*(FloatPoint.X-MapRef.Values.X0)/(MapRef.Values.Rad*(1+cos(FloatPoint.Y)));
    PrepInv2(532,F);
  end;
end;
////////////////////////////////  Wagner VI  ///////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerVIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(521) else begin
    EcT(0.94745,0.94745,0,0.30396355092701331433);
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerVIInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(522) else begin
    F := EcTI(0.94745,0.94745,0,0.30396355092701331433);
    if (F = 0) then begin
      ErrorLog(0);
      FloatPoint.X := FloatPoint.X+MapRef.Values.Th0;
    end else ErrorLog(522+F);
  end;
end;
////////////////////////////////  KavraiskyVII  ////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.KavraiskyVIIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(511) else begin
    EcT(0.8660254037844,1,0,0.30396355092701331433);
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.KavraiskyVIIInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(512) else begin
    F := EcTI(0.8660254037844,1,0,0.30396355092701331433);
    if (F = 0) then begin
      ErrorLog(0);
      FloatPoint.X := FloatPoint.X+MapRef.Values.Th0;
    end else ErrorLog(512+F);
  end;
end;
//////////////////////////////////  Putnins P1  ////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.PutninsFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(501) else begin
    EcT(1.89490,0.94745,-0.5,0.30396355092701331433);
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PutninsInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(502) else begin
    F := EcTI(1.89490,0.94745,-0.5,0.30396355092701331433);
    if (F = 0) then begin
      ErrorLog(0);
      FloatPoint.X := FloatPoint.X+MapRef.Values.Th0;
    end else ErrorLog(502+F);
  end;
end;
//////////////////////////////////  Eckert III  ////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIIIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(491) else begin
    EcT(0.42223820031577120149,0.84447640063154240298,1,0.4052847345693510857755);
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIIIInv;
var
  F : byte;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(492) else begin
    F := EcTI(0.42223820031577120149,0.84447640063154240298,1,0.4052847345693510857755);
    if (F = 0) then begin
      ErrorLog(0);
      FloatPoint.X := FloatPoint.X+MapRef.Values.Th0;
    end else ErrorLog(492+F);
  end;
end;
//////////////////////////////////  Eckert I  //////////////////////////////////
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIFor;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(481) else begin
    PrepFor1;
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*0.92131773192356127802*FloatPoint.X*(1-0.31830988618379067154*abs(FloatPoint.Y));
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*0.92131773192356127802*FloatPoint.Y;
    ErrorLog(0);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertIInv;
var
  F : byte;
  phi : double;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(482) else begin
     phi          := (MapRef.Values.Y0-FloatPoint.Y)/(MapRef.Values.Rad*0.9213177319235612780);
     FloatPoint.X := (FloatPoint.X-MapRef.Values.X0)/(MapRef.Values.Rad*0.9213177319235612780*(1-0.31830988618379067154*abs(phi)));
     FloatPoint.Y := Phi;

     PrepInv2(482,F);
  end;
end;
//------------------------------------------------------------------------------
/////////////////////////////////  Collignon  //////////////////////////////////
procedure TEZProjectionL.CollignonFor;
var
  Y : double;
begin
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(471) else begin
    if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
    if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
    Y := 1 - Sin(FloatPoint.Y);
    if (Y <= 0) then Y := 0 else Y := sqrt(Y);
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*1.12837916709551257390*(FloatPoint.X-MapRef.Values.Th0)*Y;
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*1.77245385090551602729*(1-Y);
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.CollignonInv;
var
  Lon,Lat : double;
  F : byte;
begin
  F := 0;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(472) else begin
     Lat := (MapRef.Values.Y0-FloatPoint.Y)/(MapRef.Values.Rad*1.77245385090551602729)-1;
     if (Lat > 0) then F := 1;
     Lat := 1-sqr(Lat);
     if (abs(Lat) <= 1) then Lat := ASinZ(Lat) else F := 2;

     Lon := 1-sin(Lat);
     Lon := (FloatPoint.X-MapRef.Values.X0)/(MapRef.Values.Rad*1.12837916709551257390 * sqrt(Lon));
     if (Lon > Pi) or (Lon < -Pi) then F := 3;

     if (F = 0) then begin
       FloatPoint := SetFloatPoint(Lon+MapRef.Values.Th0,Lat);
       ErrorLog(0);
     end else ErrorLog(472+F);
  end;
end;
//------------------------------------------------------------------------------
////////////////////  cylindrical equal-area projections ///////////////////////
procedure TEZProjectionL.BerCylindricalEAFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(421) else begin
  CylEA(DegToRad(30));
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.BerCylindricalEAInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(422) else begin
   B := CylEAInv(DegToRad(30));
   if (B = 0) then ErrorLog(0) else ErrorLog(422+B);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.TristanEdwardsFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(431) else begin
  CylEA(DegToRad(37.383));
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.TristanEdwardsInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(432) else begin
   B := CylEAInv(DegToRad(37.383));
   if (B = 0) then ErrorLog(0) else ErrorLog(432+B);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PetersFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(441) else begin
  CylEA(DegToRad(44.138));
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PetersInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(442) else begin
   B := CylEAInv(DegToRad(44.138));
   if (B = 0) then ErrorLog(0) else ErrorLog(442+B);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GallOrthFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(451) else begin
  CylEA(DegToRad(45));
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.GallOrthInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(452) else begin
   B := CylEAInv(DegToRad(45));
   if (B = 0) then ErrorLog(0) else ErrorLog(452+B);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.BalthasartFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(461) else begin
  CylEA(DegToRad(50));
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.BalthasartInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(462) else begin
   B := CylEAInv(DegToRad(50));
   if (B = 0) then ErrorLog(0) else ErrorLog(462+B);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.LamCylindricalEAFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(411) else begin
  CylEA(0);
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.LamCylindricalEAInv;
var
 B : byte;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(412) else begin
   B := CylEAInv(0);
   if (B = 0) then ErrorLog(0) else ErrorLog(412+B);
end;
end;
//------------------------------------------------------------------------------
////////////////////////////////////  Bonne  ///////////////////////////////////
procedure TEZProjectionL.BonneFor;
var
  rh,e:double;
begin
  if (abs(MapRef.Values.Ph0) < 1E-9) then begin
    SinusoidalFor;
    Exit;
  end;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then Flag := 411 // If radius is near zero it will give rise to a lot of errors
  else begin
    if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
    if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
    rh  := Cot(MapRef.Values.Ph0)+MapRef.Values.Ph0-FloatPoint.Y;
    e   := (FloatPoint.X-MapRef.Values.Th0)*Cos(FloatPoint.Y)/rh;
    if (abs(rh) > 1E-11) then begin
    	FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*rh*sin(e);
	FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*(Cot(MapRef.Values.Ph0)-rh*cos(e));
    end else FloatPoint := SetFloatPoint(MapRef.Values.X0,MapRef.Values.Y0);
    Flag := 0;
    if (Cot(MapRef.Values.Ph0)-Cot(MapRef.Values.Ph0)-rh*cos(e) > 0) and CutOff then begin
      Flag := 414;
      FloatPoint := SetFloatPoint(-1,-1);
    end;
  end;
  ErrorLog(Flag);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.BonneInv;
var
  rh,X,Y,C : double;
  Po : TFloatPoint;
  B : boolean;
begin
  if (abs(MapRef.Values.Ph0) < 1E-9) then begin
    SinusoidalInv;
    Exit;
  end;
  Po.X := -1;
  Po.Y := -1;
  if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(412) // If radius is near zero it will give rise to a lot of errors
  else begin
    y  := (MapRef.Values.Y0 - FloatPoint.Y)/MapRef.Values.Rad;
    x  := (FloatPoint.X - MapRef.Values.X0)/MapRef.Values.Rad;                                     // Correct for offset
    rh := sqrt(sqr(x)+sqr(Cot(MapRef.Values.Ph0)-Y));
    if (MapRef.Values.Ph0 > 1E-9) then B := True else B := False;
    if (not B and (Cot(MapRef.Values.Ph0)-Y < 0)) or (B and (Cot(MapRef.Values.Ph0)-Y > 0)) then begin
      C := Cot(MapRef.Values.Ph0)+MapRef.Values.Ph0;
      if B then Po.Y := C-rh else Po.Y := C+rh;
      C := (rh/Cos(Po.Y))*ArcTan(X/(Cot(MapRef.Values.Ph0)-Y));
      if B then Po.X := C+MapRef.Values.Th0 else Po.X := -C+MapRef.Values.Th0;
      if (Po.Y > Pi/2) or (Po.Y < -Pi/2) or (Po.X > MapRef.Values.Th0+Pi) or (Po.X < MapRef.Values.Th0-Pi) then ErrorLog(415) else
      ErrorLog(0);
    end else ErrorLog(413);
  end;
  if (Flag = 0) then FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////////////  Eckert Trapezium   ///////////////////////////////
procedure TEZProjectionL.EckertTFor;
var
 dLon,D,S : Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(390) // If radius is near zero it will give rise to a lot of errors
else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;

  dlon := FloatPoint.X - MapRef.Values.Th0;
  D    := sqrt(4-3*sin(abs(FloatPoint.Y)));
  if (abs(FloatPoint.Y) > 1E-9) then S := abs(FloatPoint.Y)/FloatPoint.Y else S := 1;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*0.46065886596178063902*dLon*D;
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*1.44720250911653531871*(2-D)*S;
  ErrorLog(0);
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;          // If an Error occured reset FloatPoint
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EckertTInv;
var
 X,Y : Double;
 F : byte;
 Po : TFloatPoint;
begin
Po.X := -1;
Po.Y := -1;
F    := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(391) // If radius is to small devision by zero might occure
else begin
  x     := (FloatPoint.X - MapRef.Values.X0)/(MapRef.Values.Rad*0.46065886596178063902);
  y     := (MapRef.Values.Y0 - FloatPoint.Y)/(MapRef.Values.Rad*1.44720250911653531871);
  Po.Y  := 2-abs(y);
  Po.X  := x/Po.Y;
  Po.Y  := (4-sqr(Po.Y))/3;
  if (abs(Po.Y) > 1) then F := 1
  else Po.Y := ArcSin(Po.Y);
  if (y < 0) then Po.Y := -Po.Y;
  if (Po.X > Pi) or (Po.X < -Pi) then F := 2;
  Po.X := Po.X+MapRef.Values.Th0;
end;
if (F = 0) then FloatPoint := Po else ErrorLog(392);                            // If error then Po = -1,-1 and FloatPoint to
end;
//------------------------------------------------------------------------------
//////////////////////////////  Stereographic  /////////////////////////////////
procedure TEZProjectionL.StereographicFor;
var
lon,lat : Double;
dlon,coslon,g,ksp:Double;
begin

if FloatPoint.X-MapRef.Values.Th0 < -Pi then                                    // this is needed to keep the map proportions correct. Else when changing Theta0 the map just drifts
   FloatPoint.X := 2*Pi + FloatPoint.X;
if FloatPoint.X-MapRef.Values.Th0 > Pi  then
   FloatPoint.X := FloatPoint.X - 2*Pi;

lon := FloatPoint.X;
lat := FloatPoint.Y;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(380) // Prevent devision by zero. If not by this computation, then by following
else begin
  dlon   := lon - MapRef.Values.Th0;
  coslon := cos(dlon);
  g      := sin(MapRef.Values.Ph0) * sin(lat) + cos(MapRef.Values.Ph0) *        // If g is constant the projection will be an orthographic projection (see through though)
            cos(lat) * coslon;
  if (abs(g + 1.0) <= 0.000000001) then ErrorLog(382) else begin                // Prevent devision by zero
    ksp := 2.0 / (1.0 + g);
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad * ksp *
                    cos(lat) * sin(dlon);
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad * ksp *
                    (cos(MapRef.Values.Ph0) * sin(lat) -
                    sin(MapRef.Values.Ph0) * cos(lat) * coslon);
    ErrorLog(0);
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.StereographicInv;
var
rh,z,lon,lat,con,X,Y:Double;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(381) // Prevent devision by zero. If not by this computation, then by following
else begin
 x      := X - MapRef.Values.X0;                                                // Adjust to offset X0,Y0
 y      := MapRef.Values.Y0 - Y;
 rh     := sqrt(x * x + y * y);                                                 // Distance to offset
 z      := 2*ArcTan(rh/(2*MapRef.Values.Rad));
 lon    := MapRef.Values.Th0;
 if (abs(rh) <= 0.000000001) then  lat := MapRef.Values.Ph0                     // X,Y coincide with point where phi = Phi0
 else begin
    lat := ArcSin(cos(z)*sin(MapRef.Values.Ph0)+(y*sin(z)*
                  cos(MapRef.Values.Ph0))/rh);
    con := abs(MapRef.Values.Ph0) - Pi/2;
    if (abs(con) <= 0.000000001) then begin                                     // Is the projection polar
      if (MapRef.Values.Ph0 >= 0.0) then begin
        lon := MapRef.Values.Th0 + Arctan2(x, -y);                              // for NorthPole
      end else begin
        lon := MapRef.Values.Th0 - Arctan2(-x, y);                              // For SouthPole
      end;
    end else begin
      con := cos(z) - sin(MapRef.Values.Ph0) * sin(lat);
      if ((abs(con) < 0.000000001) and (abs(x) < 0.000000001)) then             // Do nothing. Lon has already been calculated
      else lon := MapRef.Values.Th0 + Arctan2((x * sin(z) *                     // Calculate Lon if not polar
                  cos(MapRef.Values.Ph0)), (con * rh));
     end;
 end;
  Po.X := lon;
  Po.Y := lat;
  ErrorLog(0);
 end;
  FloatPoint := Po;                                                             // If error then Po = -1,-1 and FloatPoint to
end;
//------------------------------------------------------------------------------
//////////////////////////  Alberts Conical Equal Area  ////////////////////////
procedure TEZProjectionL.AlbertsCEAFor;
var
lat1,lat2,R,temp,es,e3,sinpo,cospo,ms1,qs1,ms2,qs2,qs0,ns0,c,rh,qs,rh1,
con,lat,lon,theta:Double;
begin
lat  := ArcSin(Sin(FloatPoint.Y));                                              // Blunt way of keeping abs(Phi) < 90
lon  := FloatPoint.X;
lat1 := MapRef.Values.Ph1;
lat2 := MapRef.Values.Ph2;
if FloatPoint.X-MapRef.Values.Th0 < -Pi then lon := 2*Pi + FloatPoint.X;        // this is needed to keep the map proportions correct. Else when changing Theta0 the map just drifts
if FloatPoint.X-MapRef.Values.Th0 > Pi  then lon := FloatPoint.X - 2*Pi;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(370) // Prevent devision by zero. If not by this computation, then by following
else if (abs(lat1 + lat2) < 0.000000001) then ErrorLog(371)                     // A Conic projection can not be Cylindrical (both lattitudes on both sides of the equator)
else begin

FloatTemp := GetTemp;
R := FloatTemp.X;
Temp := FloatTemp.Y;

es         := 1 - sqr(temp);
e3         := sqrt(es);
sinpo      := sin(lat1);
cospo      := cos(lat1);
con        := sinpo;
ms1        := msfnz(e3,sinpo,cospo);
qs1        := qsfnz(e3,sinpo,cospo);
sinpo      := sin(lat2);
cospo      := cos(lat2);
ms2        := msfnz(e3,sinpo,cospo);
qs2        := qsfnz(e3,sinpo,cospo);
qs0        := qsfnz(e3,sin(0),cos(0));

if (abs(lat1 - lat2) > 0.000000001) then ns0 := (ms1*ms1-ms2*ms2)/(qs2-qs1)
else ns0 := con;

c            := ms1 * ms1 + ns0 * qs1;
rh           := R * sqrt(c - ns0 * qs0)/ns0;
qs           := qsfnz(e3,sin(lat),cos(lat));
rh1          := R * sqrt(c - ns0 * qs)/ns0;
theta        := ns0 * (lon - MapRef.Values.Th0);
if (lat1+lat2 < 0) then begin                                                   // SouthPole view
  theta := -theta;
  FloatPoint.X := MapRef.Values.X0 - (rh1 * sin(theta))*(MapRef.Values.Rad/R);
  FloatPoint.Y := MapRef.Values.Y0/2 - (rh - rh1 * cos(theta))*(MapRef.Values.Rad/R);
end else begin                                                                  // NorthPole view
  FloatPoint.X := (rh1 * sin(theta))*(MapRef.Values.Rad/R) + MapRef.Values.X0;
  FloatPoint.Y := MapRef.Values.Y0 - (rh - rh1 * cos(theta))*(MapRef.Values.Rad/R);
end;
ErrorLog(0);
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;          // If an Error occured reset FlaotPoint to prevent false

end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.AlbertsCEAInv;
var
Po : TFloatPoint;
lat1,lat2,R,temp,es,e3,sinpo,cospo,ms1,qs1,ms2,qs2,qs0,ns0,c,rh,qs,rh1,con,lat,
lon,Theta,X,Y:Double;
begin
  X          := FloatPoint.X;
  Y          := FloatPoint.Y;
  Po.X       := 0;
  Po.Y       := 0;
  Flag     := 10000;
  lat1       := MapRef.Values.Ph1;
  lat2       := MapRef.Values.Ph2;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(372) // Prevent devision by zero. If not by this computation, then by following
else if (abs(lat1 + lat2) <= 0.000000001) then ErrorLog(373)                    // A Conic projection can not be Cylindrical (both lattitudes on both sides of the equator)
else begin

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es         := 1 - sqr(temp);
  e3         := sqrt(es);


  sinpo      := sin(lat1);
  cospo      := cos(lat1);
  con        := sinpo;
  ms1        := msfnz(e3,sinpo,cospo);
  qs1        := qsfnz(e3,sinpo,cospo);

  sinpo      := sin(lat2);
  cospo      := cos(lat2);
  ms2        := msfnz(e3,sinpo,cospo);
  qs2        := qsfnz(e3,sinpo,cospo);

  qs0        := qsfnz(e3,sin(0),cos(0));

  x            := (MapRef.Values.X0 - X)/(MapRef.Values.Rad/R);

if (abs(lat1 - lat2) > 0.000000001) then ns0 := (ms1* ms1-ms2*ms2)/(qs2-qs1)
else ns0 := con;

c            := ms1 * ms1 + ns0 * qs1;
rh           := R * sqrt(c - ns0 * qs0)/ns0;

if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then y := rh + (y -              // SothPole view
    MapRef.Values.Y0/2)/(MapRef.Values.Rad/R)
else y := rh + (y - MapRef.Values.Y0)/(MapRef.Values.Rad/R);                    // NorthPole view

if (ns0 >= 0) then begin                                                        // Values for NorthPole view
   rh1 := sqrt(x * x + y * y);
   con := 1.0;
end else begin                                                                  // SouthPole View
   rh1 := -sqrt(x * x + y * y);
   con := -1.0;
end;

theta := 0.0;
if (rh1 <> 0.0) then theta := ArcTan2(con * x, con * y);
con := rh1 * ns0 / R;
qs  := (c - con * con) / ns0;
if (e3 >= 0.0000000001) then begin
   con := 1 - 0.5 * (1.0 - es) * ln((1.0 - e3) / (1.0 + e3))/e3;
   if (abs(abs(con) - abs(qs)) > 0.0000000001 ) then begin
      phi1z(e3,qs);
      lat := TempVal1;
   end else begin
      if (qs >= 0) then lat := 0.5 * PI
      else lat := -0.5 * PI;
   end;
end else begin
   phi1z(e3,qs);
   lat := TempVal1;
end;

if (Flag = 0) and (Theta/ns0 <= Pi) and (Theta/ns0 >= -Pi) then begin         // Confine area inside projection
  lon := theta/ns0 - MapRef.Values.Th0;
  ErrorLog(0);
  Po.Y := lat;
  Po.X := -lon;
end else ErrorLog(374);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////  General Vertical Near-Side Perspective  /////////////////////
procedure TEZProjectionL.GvnspFor;
var
R,P,g,dLon,ksp,lon,lat : Double;
begin

lon := FloatPoint.X;
lat := FloatPoint.Y;

if FloatPoint.X-MapRef.Values.Th0 < -Pi then                                    // this is needed to keep the map proportions correct. Else when changing Theta0 the map just drifts
   FloatPoint.X := 2*Pi + FloatPoint.X;
if FloatPoint.X-MapRef.Values.Th0 > Pi  then
   FloatPoint.X := FloatPoint.X - 2*Pi;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) or
   (abs(MapRef.Values.Rad + MapRef.Values.H0) < 0.01) then ErrorLog(200)        // Prevent devision by zero. If not by this computation, then by following
else begin
  R      := MapRef.Values.Rad;
  p      := 1.0 + MapRef.Values.H0 / R;
  dlon   := (lon - MapRef.Values.Th0);
  g      := sin(MapRef.Values.Ph0)*sin(lat)+cos(MapRef.Values.Ph0)              // With constant g projection turns orthographic
            *cos(lat)*cos(dlon);
  if (g < (1.0/ p)) or (abs(p-g) < 0.000000001) then ErrorLog(202)              // Outside projection area
  else begin
    ksp := (p - 1.0)/(p - g);
    FloatPoint.X := MapRef.Values.X0 + R*ksp*cos(lat)*sin(dlon);
    FloatPoint.Y := MapRef.Values.Y0 - R*ksp*(cos(MapRef.Values.Ph0)*sin(lat)
                    -sin(MapRef.Values.Ph0)*cos(lat)*cos(dlon));
    ErrorLog(0);
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;          // Reset FloatPoint when Error
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.GvnspInv;
var
R,P,rh,con,com,z,sinz,X,Y:Double;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(201) // Prevent devision by zero. If not by this computation, then by following
else begin
R      := MapRef.Values.Rad;
p      := 1.0 + MapRef.Values.H0 / R;
x      := X - MapRef.Values.X0;
y      := MapRef.Values.Y0 - Y;
rh     := sqrt(x * x + y * y);
r      := rh / R;
con    := p - 1.0;
com    := p + 1.0;
if (MapRef.Values.H0 < 0) then ErrorLog(204)                                    // Height can not be negative (devision by zero by con/com)
else if (r > sqrt(con/com)) then ErrorLog(203)                                  // Ouside projection area
else begin
  Po.X := MapRef.Values.Th0;                                                    // Set Po to center lon,lat
  Po.Y := MapRef.Values.Ph0;
  if (abs(rh) >= 0.000000001) then begin                                        // else center Lon,Lat. Has already been set
    sinz := (p - sqrt(1.0 - (r * r * com) / con)) / (con / r + r/con);
    z    := asinz(sinz);
    Po.Y := AsinZ(cos(z) * sin(MapRef.Values.Ph0) + ( y * sin(z) * cos(MapRef.Values.Ph0))/rh);
    con  := abs(MapRef.Values.Ph0) - Pi/2;
    if (abs(con) <= 0.000000001) and (MapRef.Values.Ph0 >= 0.0) then begin      // NorthPole view
        Po.X := MapRef.Values.Th0 + ArcTan2(x, -y);
        Flag := 0;
    end else begin                                                              // SouthPole view
        Po.X := MapRef.Values.Th0 - ArcTan2(-x, y);
        Flag := 0;
    end;
    con := cos(z) - sin(MapRef.Values.Ph0) * sin(Po.Y);                         // When x,y coincide with center Lat,Lon or when it is a polar view
    if ((abs(con) < 0.000000001) and (abs(x) < 0.000000001)) or
       (abs(MapRef.Values.Ph0) >= Pi/2 - 0.00000001) then else begin
      Po.X  := MapRef.Values.Th0 + ArcTan2((x * sin(z) *
               cos(MapRef.Values.Ph0)),(con * rh));
      Flag := 0;
    end;
  end;
end;

end;
FloatPoint := Po;
if Flag = 0 then ErrorLog(0);                                                 // No Errors, than OK
end;
//------------------------------------------------------------------------------
////////////////////////  Equidistant Conical    ///////////////////////////////
procedure TEZProjectionL.ConicalEDFor;
var
R,Temp,es,e,e0,e1,e2,e3,ms1,ms2,ml1,ml2,ns,lat1,lat2,g,ml0,rh,rh1,lon,
lat,ml,theta:Double;
mode : integer;
begin

FloatPoint.Y := ArcSin(Sin(FloatPoint.Y));                                      // Blunt way of keeping abs(Phi) < 90

if FloatPoint.X-MapRef.Values.Th0 < -Pi then                                    // this is needed to keep the map proportions correct. Else when changing Theta0 the map just drifts
   FloatPoint.X := 2*Pi + FloatPoint.X;
if FloatPoint.X-MapRef.Values.Th0 > Pi  then
   FloatPoint.X := FloatPoint.X - 2*Pi;


  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es         := 1 - sqr(temp);
  e          := sqrt(es);
  e0         := e0fn(es);
  e1         := e1fn(es);
  e2         := e2fn(es);
  e3         := e3fn(es);
  lat        := FloatPoint.Y;
  lon        := FloatPoint.X;
  lat1       := MapRef.Values.Ph1;
  lat2       := MapRef.Values.Ph2;
  ns         := 0;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(360);
end else begin
ms1 := msfnz(e,sin(lat1),cos(lat1));
ml1 := mlfn(e0, e1, e2, e3, lat1);
if MapRef.Values.H0 >= 1 then mode := 1
else mode := 0;
if (mode <> 0) then begin
 if (abs(lat1 + lat2) < 0.000000001) then ErrorLog(361) else begin
   ms2     := msfnz(e,sin(lat2),cos(lat2));
   ml2     := mlfn(e0, e1, e2, e3, lat2);
   if (abs(lat1 - lat2) >= 0.000000001) then
      ns   := (ms1 - ms2) / (ml2 - ml1)
   else ns := sin(lat2);
   Flag  := 0;
 end;
end else begin
   ns:= sin(lat1);
   Flag := 0;
end;

if Flag = 0 then begin
  g            := ml1 + ms1/ns;
  ml0          := mlfn(e0, e1, e2, e3, 0);
  rh           := R*(g - ml0);
  ml           := mlfn(e0, e1, e2, e3, lat);
  rh1          := R*(g - ml);
  theta        := ns * (lon - MapRef.Values.Th0);
  if MapRef.Values.Ph2 < MapRef.Values.Ph1 then FloatPoint.Y := (MapRef.Values.Rad/R)*(rh-rh1*cos(theta)) + MapRef.Values.Y0 else
  FloatPoint.Y := MapRef.Values.Y0 - (MapRef.Values.Rad/R)*(rh-rh1*cos(theta));
  FloatPoint.X := (MapRef.Values.Rad/R)*rh1*sin(theta) + MapRef.Values.X0;
  ErrorLog(0);
end else ErrorLog(368);

end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.ConicalEDInv;
var
R,Temp,es,e,e0,e1,e2,e3,ms1,ms2,ml1,ml2,ns,lat1,lat2,g,ml0,rh,rh1,ml,theta,con,
X,Y:Double;
mode : integer;
Po : TFloatPoint;
begin
  X          := FloatPoint.X;
  Y          := FloatPoint.Y;
  Po.X       := 0;
  Po.Y       := 0;
  ns         := 0;
  Flag     := 10000;

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es         := 1 - sqr(temp);
  e          := sqrt(es);
  e0         := e0fn(es);
  e1         := e1fn(es);
  e2         := e2fn(es);
  e3         := e3fn(es);
  lat1       := MapRef.Values.Ph1;
  lat2       := MapRef.Values.Ph2;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(362)
//else if F > MapRef.Values.Rad*Pi/2 then ErrorLog(367)
else begin
ms1 := msfnz(e,sin(lat1),cos(lat1));
ml1 := mlfn(e0, e1, e2, e3, lat1);
if MapRef.Values.H0 >= 1 then mode := 1
else mode := 0;
if (mode <> 0) then begin
 if (abs(lat1 + lat2) < 0.000000001) then ErrorLog(363) else begin
   ms2     := msfnz(e,sin(lat2),cos(lat2));
   ml2     := mlfn(e0, e1, e2, e3, lat2);
   if (abs(lat1 - lat2) >= 0.000000001) then
      ns   := (ms1 - ms2) / (ml2 - ml1)
   else ns := sin(lat2);
   Flag  := 0;
 end;
end else begin
   ns:= sin(lat1);
   Flag := 0;
end;
if Flag = 0 then begin
  g            := ml1 + ms1/ns;
  ml0          := mlfn(e0, e1, e2, e3, 0);
  rh           := (g - ml0)*R;
  x            := (MapRef.Values.X0 - X)/(MapRef.Values.Rad/R);
  if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then y := rh + (y - MapRef.Values.Y0)/(MapRef.Values.Rad/R)
  else y := rh + (y - MapRef.Values.Y0)/(MapRef.Values.Rad/R);

  if (ns >= 0) then begin
     rh1 := sqrt(x * x + y * y);
     con := 1.0;
  end else begin
     rh1 := -sqrt(x * x + y * y);
     con := -1.0;
  end;

  if (rh1 <> 0.0) then begin
     theta := ArcTan2(con * x, con * y);
     ml    := g - rh1 /R;
     phi3z(ml,e0,e1,e2,e3);
     if (Flag = 0) and (Theta/ns <= Pi) and (Theta/ns >= -Pi) and (abs(ml) <= Pi/2) then begin
       Po.Y  := TempVal1;
       Po.X  := -(theta/ns - MapRef.Values.Th0);
      ErrorLog(0);
     end else ErrorLog(367);
  end;
end else ErrorLog(367);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////////////   Interupted Mollweide  ////////////////////////////
procedure TEZProjectionL.IMollweideFor;
var
con,dtheta,lat,lon,dlon,SX,CX,theta : Double;
i,region : integer;
Po : TFloatPoint;
const
c1 = Pi;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(350) else begin
  Po.X := FloatPoint.X - MapRef.Values.Th0 + Pi;

  SX := Sin(Po.X);
  CX := Cos(Po.X);
  if (CX < 0) and (SX > 0) then Po.X := Pi - ArcSin(SX);
  if (CX < 0) and (SX < 0) then Po.X := -Pi - ArcSin(SX);
  if (CX > 0) and (SX > 0) then Po.X := ArcSin(SX);
  if (CX > 0) and (SX < 0) then Po.X := ArcSin(SX);

  lat := FloatPoint.Y;
  lon := Po.X;

  if (lat >= 0) then begin
    if (lon >= 0.34906585) and (lon < 1.91986217719) then region := 0
    else if ((lon >= 1.919862177) and (lon <= (Pi + 0.000000001)) or
             (lon >= (-Pi-0.000000001)) and (lon < -1.745329252)) then region := 1
    else region := 2;
  end else begin
    if (lon >= 0.34906585) and (lon < 2.44346095279) then region := 3
    else if ((lon >= 2.4434609527) and (lon <= (Pi+0.000000001)) or
             (lon >= (-Pi-0.000000001)) and (lon < -1.2217304764)) then region := 4
    else region := 5;
  end;

  con := Pi*sin(lat);
  if (lon >= 1.919862177) and (lon <= Pi) then dlon := lon-Molw[region,0]-2*Pi
  else dlon  := lon-Molw[region,0];
  if (lat < 0) and (lon >= 1.919862177) and (lon <= 2.44346095279) then dlon := lon-Molw[region,0];
  Theta := lat;

  for i := 0 to 50 do begin
     dtheta := -(theta + sin(theta) - con)/ (1.0 + cos(theta));
     theta  := Theta + dtheta;
     if (abs(dTheta) < 0.00000001) then break;
  end;
  if (i = 50) then ErrorLog(352) else Flag := 0;
  Theta := Theta/2;

 if (PI/2 - abs(lat) < 0.00000001) then dlon := 0;
 if Flag = 0 then begin
   FloatPoint.X := MapRef.Values.Rad * (molw[region,1] + 0.900316316158 * dlon * cos(theta)) + MapRef.Values.X0;
   FloatPoint.Y := MapRef.Values.Y0 - 1.4142135623731 * MapRef.Values.Rad * sin(theta);
   ErrorLog(0);
 end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.IMollweideInv;
var
Po : TFloatPoint;
Theta,X,Y : Double;
region : integer;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
x    := X - MapRef.Values.X0;
y    := MapRef.Values.Y0 - Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(351)
else if (abs(y) > sqrt(2)*MapRef.Values.Rad) then ErrorLog(353)
else begin
  Flag := 10000;
  if (y >= 0.0) then begin
     if (x <= MapRef.Values.Rad * -1.41421356248)     then region := 0
     else if (x <= MapRef.Values.Rad * 0.942809042)   then region := 1
     else                                         region := 2;
  end else begin
     if (x <= MapRef.Values.Rad * -0.942809042)       then region := 3
     else if (x <= MapRef.Values.Rad * 1.41421356248) then region := 4
     else                                         region := 5;
  end;

  x := x-molw[region,1]*MapRef.Values.Rad;

  theta  := AsinZ(y / (1.4142135623731 * MapRef.Values.Rad));
  Po.X   := molw[region,0] + (x / (0.900316316158*MapRef.Values.Rad * cos(theta)));
  Po.Y   := AsinZ((2.0 * theta + sin(2.0 * theta)) / PI);

  if (region = 0) and ((Po.X < 0.34906585) or (Po.X > 1.91986217719))    then ErrorLog(353);
  if (region = 1) and (((Po.X < 1.91986217719) and (Po.X < -Pi-1.2217304764))  or
                      ((Po.X > -1.74532925199){ and (Po.X < 0.34906585)})) then ErrorLog(353);
  if (region = 2) and ((Po.X < -1.745329252) or (Po.X > 0.34906585))     then ErrorLog(353);
  if (region = 3) and ((Po.X < 0.34906585) or (Po.X > 2.44346095279))    then ErrorLog(353);
  if (region = 4) and (((Po.X < 2.44346095279) and (Po.X < -Pi-0.34906585*2))  or
                      ((Po.X > -1.2217304764){ and (Po.X < 0.34906585)}))  then ErrorLog(353);
  if (region = 5) and ((Po.X < -1.2217304764) or (Po.X > 0.34906585))    then ErrorLog(353);

  if Flag = 10000 then ErrorLog(0);
end;
Po.X := Po.X-Pi+MapRef.Values.Th0;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////////////////////// Wagner VII  ////////////////////////////////
procedure TEZProjectionL.WagnerVIIFor;
var
dLon,s,c0,c1,lat,lon : Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(340) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lat := FloatPoint.Y;
  lon := FloatPoint.X;
  dlon := lon - MapRef.Values.Th0;
  s    := 0.90631 * sin(lat);
  c0   := sqrt(1-s*s);
  c1   := sqrt(2.0 / (1.0 + c0 * cos(dlon/3)));
  FloatPoint.X := MapRef.Values.X0 + 2.66723 * MapRef.Values.Rad * c0 * c1 * sin(dlon/3);
  FloatPoint.Y := MapRef.Values.Y0 - 1.24104 * MapRef.Values.Rad * s * c1;
  ErrorLog(0);
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerVIIInv;
var
Po : TFloatPoint;
P,T1,T2,C,F,X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
F := sqrt( sqr((X-MapRef.Values.X0)*sqrt(2)/4)+sqr((MapRef.Values.Y0-Y)*sqrt(2)/2.5))/MapRef.Values.Rad;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(341)
else if F > 1 then ErrorLog(343)
else begin
  x     := X - MapRef.Values.X0;
  y     := MapRef.Values.Y0 - Y;

  t1    := sqr(x / 2.66723);
  t2    := sqr(y / 1.24104);
  p     := sqrt(t1 + t2);
  c     := 2.0 * asinz(p / (2.0 * MapRef.Values.Rad));
  if abs(y) < 0.00000001 then Po.Y := 0
  else Po.Y  := asinz(y * sin(c) / (1.24104 * 0.90631 * p));
  if abs(x) < 0.000000001 then Po.X  := MapRef.Values.Th0
  else Po.X  := MapRef.Values.Th0 + 3 * ArcTan2(x * tan(c), 2.66723 * p);

  if (abs(Po.Y) > Pi/2 - 0.000000001) then ErrorLog(343) else
  if (x > 0) and (Po.X - MapRef.Values.Th0 > Pi) then ErrorLog(343)
  else if (x < 0) and (Po.X - MapRef.Values.Th0 < -Pi) then ErrorLog(343)
  else ErrorLog(0);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////////////////////// Wagner IV  /////////////////////////////////
procedure TEZProjectionL.WagnerIVFor;
var
dLon,dTheta,lat,lon,con,theta : Double;
i : integer;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(330) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lat := FloatPoint.Y;
  lon := FloatPoint.X;
  dlon  := lon - MapRef.Values.Th0;
  theta := lat;
  con   := 2.9604205062*sin(lat);

  for i := 0 to 30 do begin
    dtheta := -(theta + sin(theta) - con) / (1.0 + cos(theta));
    theta  := Theta + dtheta;
    if (abs(dtheta) < 0.000000001) then break;
  end;
  if (i = 30) then ErrorLog(332) else Flag := 0;
  theta := theta/2;
  if Flag = 0 then begin
    FloatPoint.X := MapRef.Values.X0 + 0.86310 * MapRef.Values.Rad * dlon * cos(theta);
    FloatPoint.Y := MapRef.Values.Y0 - 1.56548 * MapRef.Values.Rad * sin(theta);
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.WagnerIVInv;
var
Po : TFloatPoint;
Theta,X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(331)
else begin
  x     := X - MapRef.Values.X0;
  y     := MapRef.Values.Y0 - Y;
  theta := ASinz(y/(1.56548 * MapRef.Values.Rad));
  Po.X  := MapRef.Values.Th0 + (x / (0.86310 * MapRef.Values.Rad * cos(theta)));
  Po.Y  := AsinZ((2.0 * theta + sin(2.0 * theta)) / 2.9604205062);
  if (abs(y) > MapRef.Values.Rad*1.3575) then ErrorLog(333)
  else if (x > 0) and (Po.X - MapRef.Values.Th0 > Pi) then ErrorLog(333)
  else if (x < 0) and (Po.X - MapRef.Values.Th0 < -Pi) then ErrorLog(333)
  else ErrorLog(0);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////////////////////   Mollweide  //////////////////////////////////
procedure TEZProjectionL.MollweideFor;
var
con,dtheta,lat,lon,dlon,theta : Double;
i : integer;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(320) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lat := FloatPoint.Y;
  lon := FloatPoint.X;
  con := Pi*sin(lat);
  dlon := lon-MapRef.Values.Th0;

  {Iterate using the Newton-Raphson method to find theta}
  Theta := lat;
  for i := 0 to 50 do begin
     dtheta := -(theta + sin(theta) - con)/ (1.0 + cos(theta));
     theta  := Theta + dtheta;
     if (abs(dTheta) < 0.000000001) then break;
  end;
  if (i = 50) then ErrorLog(322) else Flag := 0;
  Theta := Theta/2;

{ If the latitude is 90 deg, force the x coordinate to be "0 + false easting"
   this is done here because of precision problems with "cos(theta)"}

 if (PI/2 - abs(lat) < 0.00000001) then dlon := 0;
 if Flag = 0 then begin
   FloatPoint.X := 0.900316316158 * MapRef.Values.Rad * dlon * cos(theta) + MapRef.Values.X0;
   FloatPoint.Y := MapRef.Values.Y0 - 1.4142135623731 * MapRef.Values.Rad * sin(theta);
   ErrorLog(0);
 end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.MollweideInv;
var
Po : TFloatPoint;
arg,Theta,F,X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
F := sqrt( sqr((X-MapRef.Values.X0)*sqrt(2)/4)+sqr((MapRef.Values.Y0-Y)*sqrt(2)/2))/MapRef.Values.Rad;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(321)
else if F > 1 then ErrorLog(323)
else begin
  x    := X - MapRef.Values.X0;
  y    := MapRef.Values.Y0 - Y;
  arg  := y /(1.4142135623731 * MapRef.Values.Rad);
{Because of division by zero problems, 'arg' can not be 1.0.  Therefore
   a number very close to one is used instead.}
  if (abs(arg) > 0.999999999999) then arg := 0.999999999999;
  theta := ArcSin(arg);
  Po.X  := MapRef.Values.Th0 + (x / (0.900316316158 * MapRef.Values.Rad * cos(theta)));

  arg := (2*theta+sin(2*theta))/Pi;
  if (abs(arg) > 1) then arg := 1;
  Po.Y := ArcSin(arg);
  ErrorLog(0);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////////////  Transverse Mercator  /////////////////////////////
procedure TEZProjectionL.MercatorFor;
var
temp,es,e,lon,lat,ts,R:Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(310);
end else begin

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e    := sqrt(es);
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lon  := FloatPoint.X;
  lat  := FloatPoint.Y;
  if (abs(abs(lat) - Pi/2)  <= 0.00000001) then begin
   ErrorLog(315);
  end else begin
   ts := tsfnz(e,lat,sin(lat));
   FloatPoint.X := MapRef.Values.X0 + R*MapRef.Values.Rad/R *(lon - MapRef.Values.Th0);
   FloatPoint.Y := MapRef.Values.Y0 + R*MapRef.Values.Rad/R * ln(ts);
   ErrorLog(0);
  end;

end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.MercatorInv;
var
temp,es,e,ts,X,Y:Double;
Po : TFloatPoint;
begin
X    := FloatPoint.X;
Y    := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
x    := X - MapRef.Values.X0;
y    := MapRef.Values.Y0 - Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(311)
else if (abs(x) > Pi*MapRef.Values.Rad) then ErrorLog(316)
else begin

  FloatTemp := GetTemp;
  Temp := FloatTemp.Y;


  es   := 1 - sqr(temp);
  e    := sqrt(es);
  ts   := exp(-y/MapRef.Values.Rad);
  Po.X := MapRef.Values.Th0 + x/MapRef.Values.Rad;
  phi2z(e,ts);
  Po.Y := TempVal1;
  if (Flag <> 0) then ErrorLog(316)
  else ErrorLog(0);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
//////////////////////////////  Cassini-Soldner  ///////////////////////////////
procedure TEZProjectionL.CassiniSFor;
var
temp,es,e0,e1,e2,e3,ml0,esp,con,lon,lat,ml,R,al,c,tq,t,n,ind,b,con1:Double;
co : boolean;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(300);
end else begin

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e0   := e0fn(es);
  e1   := e1fn(es);
  e2   := e2fn(es);
  e3   := e3fn(es);
  ml0  := R*mlfn(e0,e1,e2,e3,0);
  esp  := sqr(es)/(1-sqr(es));

  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lon  := FloatPoint.X;
  lat  := FloatPoint.Y;
  if (es < 0.00001) then ind := 1 else ind := 0;
  if ind <> 0 then begin
    b := cos(lat) * sin(lon - MapRef.Values.Th0);
    if ((abs(abs(b) - 1)) < 0.000000001) then begin
     ErrorLog(308);
    end else begin
     FloatPoint.X := MapRef.Values.X0 +  MapRef.Values.Rad*ArcSin(b);
     con1 := cos(lat) * cos(lon - MapRef.Values.Th0)/sqrt(1 - b*b);
     if (abs(con1) > 1) then con1 := 1;
       con := ArcCos(con1);
       if (lat < 0) then con := -con;
       FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad* (con);
       ErrorLog(0);
    end;
  end else begin

    al  := cos(lat)*(Lon-MapRef.Values.Th0);
    c   := esp * sqr(cos(lat));
    tq  := tan(lat);
    t   := sqr(tq);
    con := 1 - es * sqr(sin(lat));
    n   := R/ sqrt(con);
    ml  := R*mlfn(e0, e1, e2, e3, lat);

    if (abs(lon - MapRef.Values.Th0) > 0.9) and (CutOff = true) then Co := true else co := false;

    if (abs(lon - MapRef.Values.Th0) > Pi-1) then ErrorLog(308)
    else if co = true then ErrorLog(305)
    else begin

     FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad/R*n*(al - t*Power(al,3)/6 -
                     (8 - t + 8*c)*t*Power(al,5)/120);

     FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad/R*(ml - ml0 + n*tq*(sqr(al)/2 +
                     (5 - t + 6*c)*Power(al,4)/24));
     ErrorLog(0);
    end;
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.CassiniSInv;
var
temp,es,e0,e1,e2,e3,ml0,ind,con,lon,lat,R,t,n,f,g,h,j,Phi,delta_phi,d,ds,
X,Y:Double;
i : integer;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
lon := 0;
lat := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(301)
else begin

  x    := X - MapRef.Values.X0;
  y    := MapRef.Values.Y0 - Y;

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e0   := e0fn(es);
  e1   := e1fn(es);
  e2   := e2fn(es);
  e3   := e3fn(es);
  ml0  := R*mlfn(e0,e1,e2,e3,0);

  if (es < 0.00001) then ind := 1 else ind := 0;

   if (abs(x) < Pi/2*MapRef.Values.Rad - 0.00001) and (abs(y) < MapRef.Values.Rad*Pi - 0.00001) then begin
     f    := Cos(x/MapRef.Values.Rad);
     j    := Sin(x/MapRef.Values.Rad);
     g    := (j/f);
     temp := y/MapRef.Values.Rad;
     h    := cos(temp);
     con  := sqrt((1 - h * h)/(1 + g * g));
     lat  := AsinZ(con);

     if (temp < 0) then lat := -lat;

     if ((g = 0) and (h = 0)) then begin
       lon := MapRef.Values.Th0;
       ErrorLog(0);
     end else begin
       lon := ArcTan2(g,h) + MapRef.Values.Th0;
       ErrorLog(0);
     end;
   end else ErrorLog(307);

   if (ind = 0) and (Flag = 0) and (abs(lon - MapRef.Values.Th0) < 0.9) then begin
   con := (ml0 + y /(MapRef.Values.Rad/R)) /R;
   phi := con;
   for i := 0 to 6 do begin
      delta_phi := ((con+e1*sin(2*phi)-e2*sin(4*phi)+e3*sin(6*phi))/e0)-phi;
      phi       := Phi + delta_phi;
      if (abs(delta_phi) <= 0.000000001) then break;
      if (i = 6) then ErrorLog(306);
   end;

   if (abs(phi) < Pi) and (Flag = 0) then begin
      t    := sqr(tan(phi));
      con  := 1 - es * sqr(sin(phi));
      n    := R / sqrt(con);
      r    := n * (1 - es) / con;
      d    := x / (n * MapRef.Values.Rad/R);
      ds   := sqr(d);
      lat  := phi-(n*tan(phi)/r)*(ds/2-(1+3*t)*ds*ds/24);
      lon  := MapRef.Values.Th0 + (d-t*d*ds/3+(1+3*t)*t*ds*ds*d/15)/cos(phi);
      ErrorLog(0);
   end else begin
      lat  := Pi*y/2;
      lon  := MapRef.Values.Th0;
      ErrorLog(0);
   end;
  end else if (ind = 0) then ErrorLog(306);
end;
Po.X := lon;
Po.Y := lat;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////////////  Transverse Mercator  /////////////////////////////
procedure TEZProjectionL.TMercatorFor;
var
temp,es,e0,e1,e2,e3,ml0,esp,ind,con,lon,lat,ml,R,b,al,c,tq,t,n,con1:Double;
co : boolean;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(290);
end else begin

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e0   := e0fn(es);
  e1   := e1fn(es);
  e2   := e2fn(es);
  e3   := e3fn(es);
  ml0  := R*mlfn(e0,e1,e2,e3,0);
  esp  := sqr(es)/(1-sqr(es));

  if (es < 0.00001) then ind := 1 else ind := 0;
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lon  := FloatPoint.X;
  lat  := FloatPoint.Y;

  if ind <> 0 then begin
    b := cos(lat) * sin(lon - MapRef.Values.Th0);
    if ((abs(abs(b) - 1)) < 0.000000001) then begin
     ErrorLog(298);
    end else begin
     FloatPoint.X := MapRef.Values.X0 +  R/2 * MapRef.Values.Rad/R * ln((1+b)/(1-b));
     con1 := cos(lat) * cos(lon - MapRef.Values.Th0)/sqrt(1 - b*b);
     if (abs(con1) > 1) then con1 := 1;
     con := ArcCos(con1);
     if (lat < 0) then con := -con;
     FloatPoint.Y := MapRef.Values.Y0 - R * MapRef.Values.Rad/R * (con);
     ErrorLog(0);
    end;
  end else begin
    al  := cos(lat)*(Lon-MapRef.Values.Th0);
    c   := esp * sqr(cos(lat));
    tq  := tan(lat);
    t   := sqr(tq);
    con := 1 - es * sqr(sin(lat));
    n   := R/ sqrt(con);
    ml  := R*mlfn(e0, e1, e2, e3, lat);

    if (abs(lon - MapRef.Values.Th0) > Pi/3.5) and (CutOff = true) then Co := true else co := false;

    if (abs(lon - MapRef.Values.Th0) > Pi-1) then ErrorLog(298)
    else if co = true then ErrorLog(295)
    else begin
     FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad/R*n*(al +  (1 - t + c)*Power(al,3)/6 +
                     (5 - 18*t + sqr(t) + 72*c - 58*esp)*Power(al,5)/120 );

     FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad/R*(ml - ml0 + n*tq*(sqr(al)/2 +
                     (5 - t + 9*c + 4*sqr(c))*Power(al,4)/24 +
                     (61 - 58*t + sqr(t) + 600*C - 330*esp)*Power(al,6)/720 ));
     ErrorLog(0);
    end;
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.TMercatorInv;
var
temp,es,e0,e1,e2,e3,ml0,esp,ind,con,lon,lat,R,c,t,n,f,g,h,Phi,delta_phi,cs,ts,d,ds,
X,Y:Double;
i : integer;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
lon := 0;
lat := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(291)
else begin
  x    := X - MapRef.Values.X0;
  y    := MapRef.Values.Y0 - Y;

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e0   := e0fn(es);
  e1   := e1fn(es);
  e2   := e2fn(es);
  e3   := e3fn(es);
  ml0  := R*mlfn(e0,e1,e2,e3,0);
  esp  := sqr(es)/(1-sqr(es));

  if (es < 0.00001) then ind := 1 else ind := 0;

   if (abs(y) < MapRef.Values.Rad*Pi) then begin
     f    := exp(x/MapRef.Values.Rad);
     g    := (f - 1/f)/2;
     temp := y/MapRef.Values.Rad;
     h    := cos(temp);
     con  := sqrt((1 - h * h)/(1 + g * g));
     lat  := AsinZ(con);

     if (temp < 0) then lat := -lat;

     if ((g = 0) and (h = 0)) then begin
       lon := MapRef.Values.Th0;
       ErrorLog(0);
     end else begin
       lon := ArcTan2(g,h) + MapRef.Values.Th0;
       ErrorLog(0);
     end;
   end else ErrorLog(297);

  if (ind = 0) and (Flag = 0) and (abs(lon - MapRef.Values.Th0) < Pi/3.5) then begin
   con := (ml0 + y /(MapRef.Values.Rad/R)) /R;
   phi := con;
   for i := 0 to 6 do begin
      delta_phi := ((con+e1*sin(2*phi)-e2*sin(4*phi)+e3*sin(6*phi))/e0)-phi;
      phi       := Phi + delta_phi;
      if (abs(delta_phi) <= 0.000000001) then break;
      if (i = 6) then ErrorLog(296);
   end;

   if (abs(phi) < Pi) and (Flag = 0) then begin
      c    := esp * sqr(cos(phi));
      cs   := sqr(c);
      t    := sqr(tan(phi));
      ts   := sqr(t);
      con  := 1 - es * sqr(sin(phi));
      n    := R / sqrt(con);
      r    := n * (1 - es) / con;
      d    := x / (n * MapRef.Values.Rad/R);
      ds   := sqr(d);
      lat  := phi-(n*tan(phi)*ds/r) * (0.5-ds/24*(5+3*t+10*c-4*cs-9*esp-ds/30*(61+90*t+298*c+45*ts-252*esp-3*cs)));
      lon  := MapRef.Values.Th0 + (d*(1-ds/6*(1+2*t+c-ds/20*(5-2*c+28*t-3*cs+8*esp+24.0*ts)))/cos(phi));
      ErrorLog(0);
   end else begin
      lat  := Pi*y/2;
      lon  := MapRef.Values.Th0;
      ErrorLog(0);
   end;
  end else if (ind = 0) then ErrorLog(296);
end;
Po.X := lon;
Po.Y := lat;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////////////////////// Gnomonic  //////////////////////////////////
procedure TEZProjectionL.GnomonicFor;
var
 g,ksp,Theta1 : Double;
begin

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(280) else begin
  Theta1 := FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then Theta1 := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then Theta1 := FloatPoint.X - 2*Pi;

  g := sin(MapRef.Values.Ph0)*sin(FloatPoint.Y)+cos(MapRef.Values.Ph0)*cos(FloatPoint.Y)*cos(Theta1-MapRef.Values.Th0);
  if (g <= 0.0000001) then ErrorLog(282) else Flag := 0;
  if Flag = 0 then begin
    ksp := 1/g;
    FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*ksp*cos(FloatPoint.Y)*sin(Theta1-MapRef.Values.Th0);
    FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*ksp*(cos(MapRef.Values.Ph0)*sin(FloatPoint.Y)-
                               sin(MapRef.Values.Ph0)*cos(FloatPoint.Y)*cos(Theta1-MapRef.Values.Th0));
    ErrorLog(0);
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.GnomonicInv;
var
rh,z,con,X,Y : Double;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(281) else begin
  x  := X - MapRef.Values.X0;
  y  := MapRef.Values.Y0 - Y;
  rh := sqrt(x*x+y*y);
  z  := arctan(rh/MapRef.Values.Rad);
  Po.X := MapRef.Values.Th0;
  if (abs(rh) <= 0.000000001) then begin
    Po.Y := MapRef.Values.Ph0;
  end else begin
     Po.Y := AsinZ(cos(z)*sin(MapRef.Values.Ph0)+(y*sin(z)*cos(MapRef.Values.Ph0))/rh);
     con := abs(MapRef.Values.Ph0) - Pi/2;
    if (abs(con) <= 0.000000001) then begin
      if (MapRef.Values.Ph0 >= 0) then begin
        Po.X := MapRef.Values.Th0 + ArcTan2(x,-y);
      end else begin
        Po.X := MapRef.Values.Th0 + ArcTan2(-x,y);
      end;
    end;
 end;
 con := cos(z) - sin(MapRef.Values.Ph0)*sin(Po.Y);
 Po.X := MapRef.Values.Th0 + ArcTan2((x*sin(z)*cos(MapRef.Values.Ph0)),(con*rh));
 ErrorLog(0);
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////////  Lambert Conformal Conic  /////////////////////////////
procedure TEZProjectionL.LaCConicFor;
var
temp,es,e,sinpo,cospo,centerlon,centerlat,con,ms1,ts1,ms2,ts2,ts0,
rh,ns,f0,lat,Lon,sinphi,ts,rh1,Ttheta,R : Double;
begin
Flag := 10000;

if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;


  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es         := 1 - sqr(temp);
  e          := sqrt(es);

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(270);
end else if (abs(MapRef.Values.Ph1 + MapRef.Values.Ph2) < 0.0000001) then begin
   ErrorLog(271);
end else begin
  ns         := 0;
  rh1        := 0;
  rh         := 0;
  centerlon  := MapRef.Values.Th0;
  centerlat  := Pi/2;
  sinpo      := Sin(MapRef.Values.Ph1);
  cospo      := Cos(MapRef.Values.Ph1);
  con        := sinpo;
  ms1        := msfnz(e,sinpo,cospo);
  ts1        := tsfnz(e,MapRef.Values.Ph1,sinpo);
  sinpo      := Sin(MapRef.Values.Ph2);
  cospo      := Cos(MapRef.Values.Ph2);
  ms2        := msfnz(e,sinpo,cospo);
  ts2        := tsfnz(e,MapRef.Values.Ph2,sinpo);
  sinpo      := sin(centerlat);
  ts0        := tsfnz(e,centerlat,sinpo);
  lat        := FloatPoint.Y;
  lon        := FloatPoint.X;

      if (abs(Pi/2 - MapRef.Values.Ph2) < 0.0000001)     then ns := 1
 else if (abs(MapRef.Values.Ph1 - MapRef.Values.Ph2) < 0.0000001) then ns := con
 else if (abs(MapRef.Values.Ph1 + MapRef.Values.Ph2) > 0.0000001) then ns := (Ln(ms1/ms2))/(Ln(ts1/ts2));
 if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then ns := -ns;

 if ns < 0.001 then ErrorLog(272)
 else begin
   f0 := ms1/(ns*power(ts1,ns));
   rh := R*f0*power(ts0,ns);
   con  := abs(abs(lat)-Pi/2);


   if (con > 0.000000001) then begin
     sinphi := sin(lat);
     ts     := tsfnz(e,lat,sinphi);
     if MapRef.Values.Ph2 < MapRef.Values.Ph1 then rh1 := R*f0*power(ts,-ns)
     else rh1 := R*f0*power(ts,ns);
     Flag := 0;
   end else begin
     con    := lat*ns;
     rh1    := 0;
     Flag := 0;
     if (con <= 0) then ErrorLog(272);
   end;
 end;
 if Flag = 0 then begin
    Ttheta       := ns*(lon-centerlon);
    FloatPoint.X := (MapRef.Values.Rad/R)*rh1*sin(Ttheta) + MapRef.Values.X0;
    if MapRef.Values.Ph2 < MapRef.Values.Ph1 then FloatPoint.Y := (MapRef.Values.Rad/R)*(rh-rh1*cos(Ttheta)) + MapRef.Values.Y0
    else FloatPoint.Y := (MapRef.Values.Rad/R)*(rh1*cos(Ttheta)-rh) + MapRef.Values.Y0;
    ErrorLog(0);
 end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.LaCConicInv;
var
temp,es,e,sinpo,cospo,centerlat,con,ms1,ts1,ms2,ts2,ts0,         //   centerlon  := MapRef.Values.Th0;
rh,ns,f0,ts,rh1,R,theta,X,Y : Double;
Po : TFloatPoint;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
Po.X := 0;
Po.Y := 0;
Flag := 10000;

  FloatTemp := GetTemp;
  R := FloatTemp.X;
  Temp := FloatTemp.Y;

  es         := 1 - sqr(temp);
  e          := sqrt(es);

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin  ErrorLog(275);
end else if (abs(MapRef.Values.Ph1 + MapRef.Values.Ph2) < 0.0000001) then begin ErrorLog(273);
end else begin
  ns         := 0;
  f0         := 1;
  centerlat  := pi/2;
  sinpo      := Sin(MapRef.Values.Ph1);
  cospo      := Cos(MapRef.Values.Ph1);
  con        := sinpo;
  ms1        := msfnz(e,sinpo,cospo);
  ts1        := tsfnz(e,MapRef.Values.Ph1,sinpo);
  sinpo      := Sin(MapRef.Values.Ph2);
  cospo      := Cos(MapRef.Values.Ph2);
  ms2        := msfnz(e,sinpo,cospo);
  ts2        := tsfnz(e,MapRef.Values.Ph2,sinpo);
  sinpo      := sin(centerlat);
  ts0        := tsfnz(e,centerlat,sinpo);

      if (abs(Pi/2 - MapRef.Values.Ph2) < 0.0000001)     then ns := 1
 else if (abs(MapRef.Values.Ph1 - MapRef.Values.Ph2) < 0.0000001) then ns := con
 else if (abs(MapRef.Values.Ph1 + MapRef.Values.Ph2) > 0.0000001) then ns := (Ln(ms1/ms2))/(Ln(ts1/ts2));
 if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then ns := -ns;

 if ns < 0.001 then ErrorLog(274)
 else begin
    f0 := ms1/(ns*power(ts1,ns));
    rh := R*f0*power(ts0,ns);
    x   := X - MapRef.Values.X0;
     if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then y := rh - y + MapRef.Values.Y0
     else y := rh + y - MapRef.Values.Y0;
 end;
 if (ns > 0) then begin
    rh1 := sqrt(sqr(x)+sqr(y));
    con := 1;
 end else begin
    rh1 := -sqrt(sqr(x)+sqr(y));
    con := -1;
 end;

 theta := 0;
 if (rh1 <> 0) then theta := ArcTan2((con*x),(con*y));
 if (rh1 <> 0) or (ns > 0) then begin
    con  := 1/ns;
    ts   := power((rh1/(MapRef.Values.Rad*f0)),con);
    phi2z(e,ts);
    Po.Y := TempVal1;
    if (Flag <> 0) then ErrorLog(277);
 end else begin
    Po.Y := -Pi/2;
 end;
 if (Flag = 0) and (Theta/ns <= Pi) and (Theta/ns >= -Pi) then begin Po.X := theta/ns + MapRef.Values.Th0; ErrorLog(0); end else ErrorLog(276);
 if (MapRef.Values.Ph1 + MapRef.Values.Ph2 < 0) then Po.Y := -Po.Y;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
//////////////////////////////   Robinson  /////////////////////////////////////
procedure TEZProjectionL.RobinsonFor;
var
p2,lat,dLon : Double;
ip1 : integer;
begin

if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(260);
end else begin
  Lat := FloatPoint.Y;
  p2  := abs(lat/5/0.01745329252);
  ip1 := Round(p2 - 0.0000001);
  p2  := P2-ip1;
  dLon:= FloatPoint.X-MapRef.Values.Th0;
  FloatPoint.X := MapRef.Values.Rad*(xlr[ip1+2]+p2*(xlr[ip1+3]-xlr[ip1+1])/2+p2*p2*(xlr[ip1+3]-
      2*xlr[ip1+2]+xlr[ip1+1])/2)*dLon+MapRef.Values.X0;
  if (lat >= 0) then FloatPoint.Y := -MapRef.Values.Rad*(pr[ip1+2]+p2*(pr[ip1+3]-pr[ip1+1])/2+
      p2*p2*(pr[ip1+3]-2*pr[ip1+2]+pr[ip1+1])/2)*Pi/2+MapRef.Values.Y0
  else FloatPoint.Y := MapRef.Values.Rad*(pr[ip1+2]+p2*(pr[ip1+3]-pr[ip1 +1])/2+p2*p2*(pr[ip1+3]-
      2.0*pr[ip1+2]+pr[ip1+1])/2)*Pi/2+MapRef.Values.Y0;
  if (abs(MapRef.Values.Y0-FloatPoint.Y)/1.35 >= MapRef.Values.Rad) and (CutOff = true) then ErrorLog(262)
  else ErrorLog(0);
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.RobinsonInv;
var
yy,Phid,P2,y1,u,v,c,t,xx,SX,CX,X,Y : Double;
ip1,i : integer;
Po : TFloatPoint;
begin
X        := FloatPoint.X;
Y        := FloatPoint.Y;
Po.X     := 0;
Po.Y     := 0;
Flag   := 0;
xx       := x;
X        := X - MapRef.Values.X0;
Y        := MapRef.Values.Y0 - Y;
yy       := 2*y/Pi/MapRef.Values.Rad;
phid     := yy*90;
p2       := abs(phid/5);
ip1      := Round(p2);
if (ip1 = 0) then ip1 := 1;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(261);
end else if (abs(Y)*2/Pi > MapRef.Values.Rad) then ErrorLog(267)
else if (abs(Y)/1.35 > MapRef.Values.Rad) and (abs(Y)*2/Pi < MapRef.Values.Rad) then ErrorLog(266)
else begin
  if (p2 >= 0) or (ip1 = 1) then begin
    for i := 0 to 50 do begin
        u  := pr[ip1+3]-pr[ip1+1];
        v  := pr[ip1+3]-2*pr[ip1+2]+pr[ip1+1];
        t  := 2*(abs(yy)-pr[ip1+2])/u;
        c  := v/u;
        p2 := t*(1-c*t*(1-2*c*t));
        if (y >= 0) then y1 := MapRef.Values.Rad*(pr[ip1+2]+p2*(pr[ip1+3]-pr[ip1+1])/2+
                               p2*p2*(pr[ip1+3]-2*pr[ip1+2]+pr[ip1+1])/2)*Pi/2
        else             y1 := -MapRef.Values.Rad*(pr[ip1+2]+p2*(pr[ip1+3]-pr[ip1+1])/2+
                               p2*p2*(pr[ip1+3]-2*pr[ip1+2]+pr[ip1+1])/2)*Pi/2;

        phid := phid-180*(y1-y)/Pi/MapRef.Values.Rad;

        if (i = 100) then begin
          ErrorLog(268);
        end;
        if (abs(y1 - y) < 0.00000001) then break;
    end;
  end else begin
    ip1 := ip1 - 1;
    if (ip1 < 0) then ErrorLog(267);
  end;

  if (Flag <> 267) or (Flag <> 268) then begin
    Po.Y := phid*0.01745329252;
    Po.X := MapRef.Values.Th0+x/MapRef.Values.Rad/(xlr[ip1+2]+p2*(xlr[ip1+3]-xlr[ip1+1])
            /2+p2*p2*(xlr[ip1+3]-2*xlr[ip1+2]+xlr[ip1+1])/2);
    Po.Y := Po.Y*0.91;
  end else ErrorLog(267);
end;
if (Flag = 0) then begin
  SX := Sin(Po.X);
  CX := Cos(Po.X);
  if (CX < 0) and (SX > 0) then Po.X := Pi - ArcSin(SX);
  if (CX < 0) and (SX < 0) then Po.X := -Pi - ArcSin(SX);
  if (CX > 0) and (SX > 0) then Po.X := ArcSin(SX);
  if (CX > 0) and (SX < 0) then Po.X := ArcSin(SX);
  CalculateFor(RadToDeg(Po.X),RadToDeg(Po.Y));                                  // Altough time consuming, it is Needed to find edges
  if (abs(FloatPoint.X-XX) > MapRef.Values.Rad*Pi/2) then ErrorLog(267)
  else FloatPoint := Po;
end;
end;
//------------------------------------------------------------------------------
//////////////////////////////   Polyconic  ////////////////////////////////////
procedure TEZProjectionL.PolyconicFor;
var
  temp,es,e,e0,e1,e2,e3,ml0,con,lon,lat,ml,ms,F:Double;
begin

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
   ErrorLog(250);
end else begin
  FloatTemp := GetTemp;
  Temp := FloatTemp.Y;

  es   := 1 - sqr(temp);
  e    := sqrt(es);
  e0   := e0fn(es);
  e1   := e1fn(es);
  e2   := e2fn(es);
  e3   := e3fn(es);
  ml0  := mlfn(e0,e1,e2,e3,0);

  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  lon  := FloatPoint.X;
  lat  := FloatPoint.Y;
  con  := lon - MapRef.Values.Th0;

  if (abs(lat) < 0.00000001) then begin
     FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*con;
     FloatPoint.Y := MapRef.Values.Y0 + MapRef.Values.Rad*ml0;
  end else begin
     ml  := mlfn(e0,e1,e2,e3,lat);
     ms  := msfnz(e,sin(lat),cos(lat));
     con := con*sin(lat);
     FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*ms* sin(con)/sin(lat);
     FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*(ml-ml0+ms*(1-cos(con))/sin(lat));
  end;
  F := sqrt(sqr((FloatPoint.X-MapRef.Values.X0)*Pi/2/sqrt(2))+sqr(MapRef.Values.Y0-FloatPoint.Y));
  if (F > MapRef.Values.Rad*Pi/2) and (CutOff = true) then ErrorLog(258) else ErrorLog(0);
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.PolyconicInv;
var
temp,es,e0,e1,e2,e3,ml0,al,b,F,X,Y:Double;
Po : TFloatPoint;
begin
X      := FloatPoint.X;
Y      := FloatPoint.Y;
Po.X   := 0;
Po.Y   := 0;
Flag := 10000;
F      := sqrt(sqr((X-MapRef.Values.X0)*Pi/2/sqrt(2))+sqr(MapRef.Values.Y0-Y));
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(251)
else if F > MapRef.Values.Rad*Pi/2 then ErrorLog(257)
else begin
   X    := X - MapRef.Values.X0;
   Y    := MapRef.Values.Y0 - Y;

  FloatTemp := GetTemp;
  Temp := FloatTemp.Y;

   es   := 1 - sqr(temp);
   e0   := e0fn(es);
   e1   := e1fn(es);
   e2   := e2fn(es);
   e3   := e3fn(es);
   ml0  := mlfn(e0,e1,e2,e3,0);
   al   := ml0 + y/MapRef.Values.Rad;
   b    := al*al+(x/MapRef.Values.Rad)*(x/MapRef.Values.Rad);
  if (abs(al) < 0.000000001) then begin
     Po.X := x/MapRef.Values.Rad + MapRef.Values.Th0;
     Po.Y := 0;
     Flag := 0;
  end else begin
    phi4z(es,e0,e1,e2,e3,al,b);
    if Flag = 0 then begin
      Po.X := (ASinZ(x*TempVal2/MapRef.Values.Rad) /sin(TempVal1)) + MapRef.Values.Th0;
      Po.Y := TempVal1;
      ErrorLog(0);
    end else ErrorLog(257);
  end;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
//////////////////////////////   Van der Grinten  //////////////////////////////
procedure TEZProjectionL.GrintenFor;
var
g,al,asq,msq,con,gsq,m,Theta1,Phi,theta:Double;
begin
Phi := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(240)
else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  if (Phi > Pi/2 - 0.000000001)   then FloatPoint.Y := FloatPoint.Y - 0.000000001;
  if (Phi < -Pi/2 + 0.000000001)  then FloatPoint.Y := FloatPoint.Y + 0.000000001;
  Theta := FloatPoint.X;
  Theta1 := Theta;
  Phi   := FloatPoint.Y;

  if (abs(Phi) < 0.000000001) then begin
     FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(Theta1-MapRef.Values.Th0);
     FloatPoint.Y := MapRef.Values.Y0;
     ErrorLog(0);
  end else Flag := 10000;

  if (Flag <> 0) and (abs(2.0*abs(Phi/Pi)) < 1) and (abs(2.0*abs(Phi/Pi)) > 0) Then theta := asinz(2.0*abs(Phi/Pi));
  if (abs(abs(Phi)-Pi/2) < 0) and (Flag <> 0) then begin
     FloatPoint.X := MapRef.Values.X0;
     if (Phi >= 0) then FloatPoint.Y := MapRef.Values.Y0 - Pi*MapRef.Values.Rad*tan(0.5*theta)
     else FloatPoint.Y := MapRef.Values.Y0 - Pi*MapRef.Values.Rad*-tan(0.5*theta);
     ErrorLog(0);
  end;
  if Flag <> 0 then begin
   if (abs(Theta1 - MapRef.Values.Th0) < 0.000000001) then Theta1 := Theta1 + 0.000000001;
   if (abs((sin(theta)+cos(theta)-1.0)) <  0.000000001) then Theta := Theta + 0.0000000001;
   al  := 0.5*abs((Pi/(Theta1-MapRef.Values.Th0)) - ((Theta1-MapRef.Values.Th0)/Pi));
   asq := sqr(al);
   g   := cos(theta)/(sin(theta)+cos(theta)-1.0);
   gsq := sqr(g);
   m   := g*(2.0/sin(theta)-1.0);
   msq := sqr(m);
   con := Pi*MapRef.Values.Rad*(al*(g-msq)+sqrt(asq*(g-msq)*(g-msq)-(msq+asq)*(gsq-msq)))/(msq+asq);
   if (Theta1-MapRef.Values.Th0 < 0.0)  and (Flag <> 0) then con := -con;
   if (Flag <> 0) then FloatPoint.X := MapRef.Values.X0 + con;
   con := abs(con/(Pi*MapRef.Values.Rad));

   if ((1.0-sqr(con)-2.0*al*con) > 0) and (Flag <> 0)  then begin
     if (Phi >= 0) then FloatPoint.Y := MapRef.Values.Y0-Pi*MapRef.Values.Rad*sqrt(1.0-sqr(con)-2.0*al*con)
     else FloatPoint.Y := MapRef.Values.Y0+Pi*MapRef.Values.Rad*sqrt(1.0-sqr(con)-2.0*al*con);
     ErrorLog(0);
   end;
  end;
end;
if Flag <> 0 then begin FloatPoint.X := -1; FloatPoint.Y := -1; ErrorLog(248); end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.GrintenInv;
var
con,xx,yy,xys,c1,c2,c3,d,a1,m1,H,th1,X,Y : Double;
Po : TFloatPoint;
begin
X      := FloatPoint.X;
Y      := FloatPoint.Y;
Po.X   := 0;
Po.Y   := 0;
Flag := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
      ErrorLog(241);
end else begin
   X    := X - MapRef.Values.X0;
   Y    := MapRef.Values.Y0 - Y;
   H    := sqrt(sqr(X)+sqr(Y));
  if (H > MapRef.Values.Rad*Pi - 0.00000001) then ErrorLog(247)
  else if (H < 0.000000001) then begin
   Po.X := MapRef.Values.Th0;
   Po.Y := 0;
  end else if (H > 0.000000001) then begin
   con  := Pi*MapRef.Values.Rad;
   xx   := x/con;
   yy   := y/con;
   xys  := sqr(xx)+sqr(yy);
   c1   := -abs(yy)*(1+xys);
   c2   := c1 - 2.0 * sqr(yy)+sqr(xx);
   c3   := -2*c1+1+2*sqr(yy)+sqr(xys);
   d    := (yy*yy/c3) + (2*c2*c2*c2/c3/c3/c3 - 9*c1*c2/c3/c3)/27;
   a1   := (c1-c2*c2/3/c3)/c3;
   m1   := 2*sqrt(-a1/3);
   con  := ((3*d)/a1)/m1;

   if (con >  1) then con :=  1;
   if (con < -1) then con := -1;
   th1 := ArcCos(con)/3;

   if (y >= 0) then Po.Y := (-m1*cos(th1+Pi/3)-c2/3/c3)*Pi
   else             Po.Y :=-(-m1*cos(th1+Pi/3)-c2/3/c3)*Pi;

   if (abs(xx) < 0.00000001) then Po.X := MapRef.Values.Th0
   else Po.X := MapRef.Values.Th0 + Pi*(xys-1+sqrt(1+2*(sqr(xx)-sqr(yy))+sqr(xys)))/2/xx;
   ErrorLog(0);
  end;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
////////////////////   Lambert Azimuthal Equal Area Projection  ////////////////
procedure TEZProjectionL.LaAzimuthalEAFor;
var
Phi1,Theta1,g,Ksp:Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(230) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  Theta1 := FloatPoint.X;
  Phi1   := FloatPoint.Y;
  if (Phi1 = MapRef.Values.Ph0) and (Theta1 = MapRef.Values.Th0) then begin FloatPoint.X := MapRef.Values.X0; FloatPoint.Y := MapRef.Values.Y0; end;
  g := Sin(MapRef.Values.Ph0)*Sin(Phi1)+Cos(MapRef.Values.Ph0)*Cos(Phi1)*Cos(Theta1-MapRef.Values.Th0);
  if (abs(abs(g) - 1.0) > 0.000000001) and (abs(g) > 0.000000001) then begin
       ksp := sqrt(2/(1+g));
       FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*Ksp*Cos(Phi1)*Sin(Theta1-MapRef.Values.Th0);
       FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*Ksp*(Cos(MapRef.Values.Ph0)*Sin(Phi1)-Sin(MapRef.Values.Ph0)*Cos(Phi1)*Cos(Theta1-MapRef.Values.Th0));
       ErrorLog(0);
  end else begin FloatPoint.X := MapRef.Values.X0; FloatPoint.Y := MapRef.Values.Y0; ErrorLog(238); end;
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.LaAzimuthalEAInv;
var
H,lat,lon,z,temp,X,Y : Double;
Po : TFloatPoint;
begin
X      := FloatPoint.X;
Y      := FloatPoint.Y;
Po.X   := 0;
Po.Y   := 0;
Flag := 0;
lon    := 0;
lat    := 0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
      ErrorLog(231);
end else begin
      X := X - MapRef.Values.X0;
      Y := MapRef.Values.Y0 - Y;
      H := sqrt(sqr(X) + sqr(Y));
      temp := H/(2*MapRef.Values.Rad);
  if (temp > 1) then ErrorLog(237)
  else if (H > 0.01) then begin
      z := 2*ASinZ(temp);
      lat  := asinz(Sin(MapRef.Values.Ph0)*Cos(z)+Cos(MapRef.Values.Ph0)*Sin(z)*y/H);
      temp := abs(MapRef.Values.Ph0)-Pi/2;
      if (abs(temp) > 0.000000001) then temp := (Cos(z)-Sin(MapRef.Values.Ph0)*Sin(lat));
      if (abs(temp) <> 0) then lon := MapRef.Values.Th0 + Arctan2(x*sin(z)*Cos(MapRef.Values.Ph0),temp*H)
      else if (MapRef.Values.Ph0 < 0) then lon := MapRef.Values.Th0 - Arctan2(-x,y)
      else lon := MapRef.Values.Th0 + Arctan2(x,-y);
      ErrorLog(0);
  end;
Po.X := lon;
Po.Y := lat;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
/////////////////////////   Azimuthal Equal Area Projection  ///////////////////
procedure TEZProjectionL.AzimuthalEAFor;
var
Phi1,Theta1,g,Ksp:Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(220) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  Theta1 := FloatPoint.X;
  Phi1   := FloatPoint.Y;
  if (Phi1 = MapRef.Values.Ph0) and (Theta1 = MapRef.Values.Th0) then begin FloatPoint.X := MapRef.Values.X0; FloatPoint.Y := MapRef.Values.Y0; end else begin
  g := Sin(MapRef.Values.Ph0)*Sin(Phi1)+Cos(MapRef.Values.Ph0)*Cos(Phi1)*Cos(Theta1-MapRef.Values.Th0);
  if (abs(abs(g) - 1.0) > 0.000000001) and (abs(g) > 0.000000001) then begin
   if (abs(g) < 0.000000001 ) then ksp := 1 else
     ksp := ArcCos(g)/Sin(ArcCos(g));
       FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*Ksp*Cos(Phi1)*Sin(Theta1-MapRef.Values.Th0);
       FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*Ksp*(Cos(MapRef.Values.Ph0)*Sin(Phi1)-Sin(MapRef.Values.Ph0)*Cos(Phi1)*Cos(Theta1-MapRef.Values.Th0));
       ErrorLog(0);
  end else begin FloatPoint.X := -1; FloatPoint.Y := -1; ErrorLog(228); end;
  end;
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.AzimuthalEAInv;
var
H,lat,lon,con,z,TmpP0,X,Y : Double;
Po : TFloatPoint;
begin
X     := FloatPoint.X;
Y     := FloatPoint.Y;
Po.X  := 0;
Po.Y  := 0;
TmpP0 := MapRef.Values.Ph0;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then begin
      ErrorLog(221);
end else begin
      X := X - MapRef.Values.X0;
      Y := MapRef.Values.Y0 - Y;
      H := sqrt(sqr(X) + sqr(Y));
      z := H/MapRef.Values.Rad;
      con := abs(MapRef.Values.Ph0) - Pi/2;
      lat := MapRef.Values.Ph0;
      lon := MapRef.Values.Th0;
  if H > MapRef.Values.Rad*Pi then begin
      ErrorLog(227);
  end else if H < 0.01 then begin
      Po.X := lon;
      Po.Y := lat;
      ErrorLog(0);
  end else if (abs(con) <= 1*10E-8) then begin
   if H < 1 then lat := MapRef.Values.Ph0 else lat := ASinZ(cos(z)*sin(MapRef.Values.Ph0) -
      (y*sin(z)*cos(MapRef.Values.Ph0))/H);
   if (MapRef.Values.Th0 >= 0) then begin
      lon := MapRef.Values.Th0 - ArcTan2(x,-y);
      if (Sin(TmpP0) = 1) then lon := MapRef.Values.Th0 + ArcTan2(x,-y);
      if (Sin(TmpP0) = -1) then lon := MapRef.Values.Th0 - ArcTan2(-x,y);
      Po.X := lon;
      Po.Y := lat;
      ErrorLog(0);
   end else begin
      lon := MapRef.Values.Th0 - ArcTan2(-x,y);
      ErrorLog(0);
      Po.X := lon;
      Po.Y := lat;
   end;
  end else begin
   if H <= 0.000001 then lat := MapRef.Values.Ph0
   else lat := ASinZ(cos(z)*sin(MapRef.Values.Ph0) + (y*sin(z)*cos(MapRef.Values.Ph0))/H);
   con  := cos(z) - sin(MapRef.Values.Ph0)*sin(lat);
   lon  := MapRef.Values.Th0 + ArcTan2((x*sin(z)*cos(MapRef.Values.Ph0)),(con*H));
   if (cos(TmpP0) < 0) Then Lon := lon + Pi;
   Po.X := lon;
   Po.Y := lat;
   ErrorLog(0);
  end;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------------
///////////////////////////////   Hammer Projection  /////////////////////////////////
procedure TEZProjectionL.HammerFor;
var
Phi1,Theta1:Double;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(210) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  Theta1 := FloatPoint.X;
  Phi1   := FloatPoint.Y;
  if (1+cos(Phi1)*Cos(1/2*Theta1-1/2*MapRef.Values.Th0)) > 0 then begin
     FloatPoint.X := MapRef.Values.X0+2*MapRef.Values.Rad*sqrt(2)/sqrt(1+cos(Phi1)*Cos(1/2*Theta1-1/2*MapRef.Values.Th0))*cos(Phi1)*Sin(1/2*Theta1-
                     1/2*MapRef.Values.Th0);
     FloatPoint.Y := MapRef.Values.Y0-MapRef.Values.Rad*sqrt(2)/sqrt(1+cos(Phi1)*Cos(1/2*Theta1-1/2*MapRef.Values.Th0))*Sin(Phi1);
     ErrorLog(0);
  end else ErrorLog(218);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.HammerInv;
var
Fac,T,X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 1) then ErrorLog(211) else begin
   X := X - MapRef.Values.X0;
   Y := MapRef.Values.Y0 - y;
   Flag := 10000;
   T := sqr(MapRef.Values.Rad) - sqr(x/4) - sqr(y/2);
    if (T >= 0) then begin
       fac := sqrt( sqr(MapRef.Values.Rad) - sqr(x/4) - sqr(y/2) );
       FloatPoint.X  := MapRef.Values.Th0 + 2.0 *ArcTan2((x*fac),(2.0*sqr(MapRef.Values.Rad)-sqr(x)/4-sqr(y)));
       FloatPoint.Y  := asinz(y*fac/sqr(MapRef.Values.Rad));
       ErrorLog(0);
    end else begin
       fac := 0;
       FloatPoint.X  := MapRef.Values.Th0 + 2.0 *ArcTan2((x*fac),(2.0*sqr(MapRef.Values.Rad)-sqr(x)/4-sqr(y)));
       FloatPoint.Y  := asinz(y*fac/sqr(MapRef.Values.Rad));
       ErrorLog(0);
    end;//ErrorLog(217);

    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(217);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
    if Flag = 10000 then ErrorLog(217);
end;
end;
//------------------------------------------------------------------------------
////////////////////////   Braun's Stereographic Projection  ///////////////////
procedure TEZProjectionL.BStereoFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(190) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0);
  if FloatPoint.Y <> abs(Pi) then FloatPoint.Y := MapRef.Values.Y0 - 2*MapRef.Values.Rad*Tan(FloatPoint.Y/2);
  ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.BStereoInv;
var
X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(191) else begin
      FloatPoint.X := (X-MapRef.Values.X0+MapRef.Values.Rad*MapRef.Values.Th0)/MapRef.Values.Rad;
      FloatPoint.Y := 2*ArcTan(1/2*(MapRef.Values.Y0-Y)/MapRef.Values.Rad);
    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or
       (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(197);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
end;
end;
//------------------------------------------------------------------------------------
////////////////////////   Equirectangular Projection  ///////////////////////
procedure TEZProjectionL.EquirectFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(180) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0);
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*FloatPoint.Y;
  ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.EquirectInv;
var
X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(181) else begin
      FloatPoint.X := (X-MapRef.Values.X0+MapRef.Values.Rad*MapRef.Values.Th0)/MapRef.Values.Rad;
      FloatPoint.Y := -(Y-MapRef.Values.Y0)/MapRef.Values.Rad;
    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or
       (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(187);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
end;
end;
//------------------------------------------------------------------------------------
////////////////////////   Sinusoidal Projection  /////////////////////////////////////
procedure TEZProjectionL.SinusoidalFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(170) else begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0)*Cos(FloatPoint.Y);
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*FloatPoint.Y;
  ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.SinusoidalInv;
var
X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(171) else begin
      FloatPoint.X := (X-MapRef.Values.X0+MapRef.Values.Rad*Cos((Y-MapRef.Values.Y0)/MapRef.Values.Rad)*MapRef.Values.Th0)/MapRef.Values.Rad/Cos((Y-MapRef.Values.Y0)/MapRef.Values.Rad);
      FloatPoint.Y := -(Y-MapRef.Values.Y0)/MapRef.Values.Rad;
    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(177);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
end;
end;
//------------------------------------------------------------------------------------
////////////////////////   Winkel II Projection  /////////////////////////////////////
procedure TEZProjectionL.WinkelIIFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(160) else begin

  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;

  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0)*(1+sqrt(sqr(Pi) - 4*sqr(FloatPoint.Y))/Pi)/2;
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*FloatPoint.Y;
  ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.WinkelIIInv;
var
Aa,X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(161) else begin
     Aa := (sqr(Pi)*sqr(MapRef.Values.Rad)-4*sqr(Y)+8*Y*MapRef.Values.Y0-4*sqr(MapRef.Values.Y0))/sqr(MapRef.Values.Rad);
  if Aa > 0 then begin
     FloatPoint.X := (2*X*Pi-2*MapRef.Values.X0*Pi+MapRef.Values.Rad*MapRef.Values.Th0*Pi+MapRef.Values.Rad*MapRef.Values.Th0*sqrt(Aa))/MapRef.Values.Rad/(Pi+sqrt(Aa));
      FloatPoint.Y := -(Y -MapRef.Values.Y0)/MapRef.Values.Rad;
    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(167);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
  end else ErrorLog(167);
end;
end;
//------------------------------------------------------------------------------------
////////////////////////   Winkel I Projection  /////////////////////////////////////
procedure TEZProjectionL.WinkelIFor;
begin
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(130) else begin

  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;

  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0)*(1 + Cos(FloatPoint.Y))/2;
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*FloatPoint.Y;
  ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------------
procedure TEZProjectionL.WinkelIInv;
var
X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(131) else begin
      FloatPoint.X := 2*(X-MapRef.Values.X0)/(MapRef.Values.Rad*(1+Cos((Y-MapRef.Values.Y0)/MapRef.Values.Rad))) + MapRef.Values.Th0;
      FloatPoint.Y := -(Y-MapRef.Values.Y0)/MapRef.Values.Rad;
    if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
       ErrorLog(137);
    end else begin
      if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
      if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
      ErrorLog(0);
    end;
end;
end;
//------------------------------------------------------------------------------------
////////////////////////   UTM Projection  /////////////////////////////////////
Procedure TEZProjectionL.UTMFor;
var
X, Y : Double;
begin
if MapRef.Calculated = false then ErrorLog(152)
else begin
  X := FloatPoint.X;
  Y := FloatPoint.Y;
  FloatPoint.X := MapRef.Values.Rad*X + MapRef.Values.Th0*Y + MapRef.Values.Ph0;
  FloatPoint.Y := MapRef.Values.Ph1*Y + MapRef.Values.Ph2*X + MapRef.Values.X0;
end;
end;
//------------------------------------------------------------------------------
Procedure  TEZProjectionL.UTMInv;
var
X,Y : Double;
begin
X := FloatPoint.X;
Y := FloatPoint.Y;
if MapRef.Calculated = false then begin
   ErrorLog(153);
end else begin
   FloatPoint.Y := -1/(MapRef.Values.Rad*MapRef.Values.Ph1-MapRef.Values.Ph2*MapRef.Values.Th0)*(-MapRef.Values.Rad*Y-MapRef.Values.Ph2*MapRef.Values.Ph0+
                       MapRef.Values.Rad*MapRef.Values.X0+MapRef.Values.Ph2*X);
   FloatPoint.X := (X*MapRef.Values.Ph1-MapRef.Values.Th0*Y+MapRef.Values.Th0*MapRef.Values.X0-
                   MapRef.Values.Ph0*MapRef.Values.Ph1)/(MapRef.Values.Rad*MapRef.Values.Ph1-MapRef.Values.Ph2*MapRef.Values.Th0);
   ErrorLog(0);
end;
end;
//------------------------------------------------------------------------------
////////////////////   Orthographic Projection  ////////////////////////////////
function TEZProjectionL.OrthRange(Theta,Phi: Double):boolean;
var
R1,x,y,z,x0,y0,z0 : Double;
begin
// Calculate relative distance. Phi0 and Theta0 to Theta and Phi on sphere ->
// If distance for phi and Theta is more than 1, point is not visible (backside of view).
 x  := Cos(Theta)*Cos(Phi);
 y  := Sin(Theta)*Cos(Phi);
 z  := Sin(Phi);
 x0 := Cos(MapRef.Values.Ph0)*Cos(MapRef.Values.Th0);
 y0 := Cos(MapRef.Values.Ph0)*Sin(MapRef.Values.Th0);
 z0 := Sin(MapRef.Values.Ph0);
 R1 := sqrt(sqr(x-x0) + sqr(y-y0)+sqr(z-z0))/(sqrt(2));
if R1 <= 1 then result := true else result := false;
end;
//------------------------------------------------------------------------------
Procedure TEZProjectionL.OrthFor;
var
 Theta,Phi:Double;
begin
if MapRef.Calculated = false then ErrorLog(107) else begin
 Theta := FloatPoint.X;
 Phi   := FloatPoint.Y;
if OrthRange(Theta,Phi) then begin
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*(Cos(MapRef.Values.Ph0)*Sin(Phi) -
               Sin(MapRef.Values.Ph0)*Cos(Phi)*Cos(Theta)*Cos(MapRef.Values.Th0) -
               Sin(MapRef.Values.Ph0)*Cos(Phi)*Sin(Theta)*Sin(MapRef.Values.Th0));
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*Cos(Phi)*(Sin(Theta)*Cos(MapRef.Values.Th0)-
               Cos(Theta)*Sin(MapRef.Values.Th0));
  ErrorLog(0);
end else begin
   FloatPoint.X := -1;
   FloatPoint.Y := -1;
   ErrorLog(106);
end;
end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.OrthInv;
var
H,lat,lon,con,z,TmpP0,X,Y : Double;
Po : TFloatPoint;
begin
X     := FloatPoint.X;
Y     := FloatPoint.Y;
Po.X  := 0;
Po.Y  := 0;
TmpP0 := MapRef.Values.Ph0;

if (MapRef.Calculated = false) or (MapRef.Values.Rad < 0.01) then ErrorLog(111)
else begin
      X := X - MapRef.Values.X0;
      Y := MapRef.Values.Y0 - Y;
      H := sqrt(sqr(X) + sqr(Y));
      z := ASinZ(H/MapRef.Values.Rad);
      con := abs(MapRef.Values.Ph0) - Pi/2;
      lat := MapRef.Values.Ph0;
      lon := MapRef.Values.Th0;
  if H > MapRef.Values.Rad then begin
      ErrorLog(110);
  end else if H < 0.1 then begin
      Po.X := lon;
      Po.Y := lat;
      ErrorLog(0);
  end else if (abs(con) <= 0.000000001) then begin
   if H < 1 then lat := MapRef.Values.Ph0 else lat := ASinZ(cos(z)*sin(MapRef.Values.Ph0) -
      (y*sin(z)*cos(MapRef.Values.Ph0))/H);
   if (MapRef.Values.Th0 >= 0) then begin
      lon := MapRef.Values.Th0 - ArcTan2(x,-y);
      if (Sin(TmpP0) = 1) then lon := MapRef.Values.Th0 + ArcTan2(x,-y);
      if (Sin(TmpP0) = -1) then lon := MapRef.Values.Th0 - ArcTan2(-x,y);
      Po.X := lon;
      Po.Y := lat;
      ErrorLog(0);
   end else begin
      lon := MapRef.Values.Th0 - ArcTan2(-x,y);
      Po.X := lon;
      Po.Y := lat;
      ErrorLog(0);
   end;
  end else begin
   if H <= 0.5 then lat := MapRef.Values.Ph0
   else lat := ASinZ(cos(z)*sin(MapRef.Values.Ph0) + (y*sin(z)*cos(MapRef.Values.Ph0))/H);
   con := cos(z) - sin(MapRef.Values.Ph0)*sin(lat);
   lon := MapRef.Values.Th0 + ArcTan2((x*sin(z)*cos(MapRef.Values.Ph0)),(con*H));
   if (cos(TmpP0) <= 0) Then Lon := lon + Pi;
   Po.X := lon;
   Po.Y := lat;
   ErrorLog(0);
  end;
end;
FloatPoint := Po;
end;
//------------------------------------------------------------------------------
///////////////////////////  EZMap subroutines  ////////////////////////////////
function TEZProjectionL.RMinor:Double;
begin
 if (Assigned(FEZConvertL)) and (MapRef.MapIndex > MaxEllipsoid -1) and (MapRef.MapIndex <> 220) then
    result := FEZConvertL.UserDat.Ellips[MapRef.MapIndex].Radius -
              FEZConvertL.UserDat.Ellips[MapRef.MapIndex].Radius/
              FEZConvertL.UserDat.Ellips[MapRef.MapIndex].Flattening
 else if (MapRef.MapIndex > MaxEllipsoid -1) then
    result := Ellipsoid[30,0] - Ellipsoid[30,0]/Ellipsoid[30,1]
 else result := Ellipsoid[MapRef.MapIndex,0] - Ellipsoid[MapRef.MapIndex,0]/Ellipsoid[MapRef.MapIndex,1];
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.phi1z(eccent,qs:Double);
var
eccnts,dphi,con,com,phi:Double;
i:integer;
begin
Flag := 100000;
   phi := asinz(0.5 * qs);
   if (abs(Phi) > Pi/2 - 0.00000001) then

   else if (eccent < 0.000000001) then begin
      TempVal1 := phi;
      Flag := 0;
   end else begin
     eccnts := eccent * eccent;
     for i := 0 to 25 do begin
       con  := eccent * sin(phi);
       com  := 1.0 - con * con;
       dphi := 0.5*com*com/cos(phi)*(qs/(1-eccnts)-sin(phi)/com+
               0.5/eccent*ln((1-con)/(1+con)));
       phi  := phi + dphi;
       if (abs(dphi) <= 0.000000001) then begin
         TempVal1 := phi;
         Flag := 0;
       end;
     end;
   end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.Phi2z(eccent,ts:Double);
var
eccnth,Ph,con,dphi,sinpi:Double;
i:integer;
begin
  TempVal1 := 0;
  Flag  := 10000;
  eccnth  := eccent/2;
  ph      := Pi/2 - 2*ArcTan(ts);
  for i := 0 to 15 do begin
    sinpi := sin(ph);
    con   := eccent*sinpi;
    dphi  := Pi/2 - 2*ArcTan(ts*(power( ((1-con)/(1+con)) ,eccnth)));
    ph    := dphi;
      if (abs(ph-dPhi) <= 0.000000001) then begin TempVal1 := Ph; Flag := 0; Break; end;
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.phi3z(ml,e0,e1,e2,e3:Double);
var
phi,dphi:Double;
i : integer;
begin
Flag := 100000;
phi := ml;
for i := 0 to 15 do begin
  dphi := (ml+e1*sin(2*phi)-e2*sin(4*phi)+e3*sin(6*phi))/e0-phi;
  phi  := phi + dphi;
  if (abs(dphi) <= 0.0000000001) then begin
     Flag := 0;
     break;
  end;
end;
if Flag = 0 then TempVal1 := phi;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.Phi4z(eccent,e0,e1,e2,e3,a,b:Double);
var
sinphi,sin2ph,tanphi,ml,mlp,con1,con2,con3,dphi,Phi,c : Double;
i:integer;
begin
phi := a;
Flag := 100000;
   for i := 1 to 15 do begin
     sinphi := sin(phi);
     tanphi := tan(phi);
     c      := tanphi * sqrt(1-eccent*sinphi*sinphi);
     sin2ph := sin (2*phi);
     ml     := e0*phi-e1*sin2ph+e2*sin(4*phi)-e3*sin(6*phi);
     mlp    := e0-2*e1*cos(2*phi)+4*e2*cos(4*phi)-6*e3*cos(6*phi);
     con1   := 2*ml+c*(ml*ml+b)-2*a*(c*ml+1);
     con2   := eccent*sin2ph*(ml*ml+b-2*a*ml)/(2*c);
     con3   := 2*(a-ml)*(c*mlp-2/sin2ph)-2*mlp;
     dphi   := con1/(con2+con3);
     phi    := phi + dphi;
     if (abs(dphi) < 0.0000001) then begin
       Flag := 0;
       TempVal1 := Phi;
       TempVal2 := c;
     end;
   end;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.E0fn(x:Double):Double;
begin
result := 1-0.25*x*(1+x/16*(3+1.25*x));
end;
//------------------------------------------------------------------------------
function TEZProjectionL.E1fn(x:Double):Double;
begin
result := 0.375*x*(1+0.25*x*(1+0.46875*x));
end;
//------------------------------------------------------------------------------
function TEZProjectionL.E2fn(x:Double):Double;
begin
result := 0.05859375*x*x*(1+0.75*x);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.E3fn(x:Double):Double;
begin
result := x*x*x*(35/3072);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.mlfn(e0,e1,e2,e3,phi:Double):Double;
begin
result := e0*phi-e1*sin(2.0*phi)+e2*sin(4.0*phi)-e3*sin(6.0*phi);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.Msfnz(eccent,sinphi,cosphi:Double):Double;
begin
result := cosphi/sqrt(1-sqr(eccent)*sqr(sinphi));
end;
//------------------------------------------------------------------------------
function TEZProjectionL.ASinZ(con : Double): Double;
begin
 if (abs(con) > 1) then begin
   if (con > 1) then con := 1
   else con := -1;
 end;
 result:= ArcSin(con);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.Tsfnz(eccent,phi,sinphi:Double):Double;
var
Con,Com : Double;
begin
  con := eccent*sinphi;
  com := eccent/2;
  con := power(((1-con)/(1+con)),com);
  result := (tan((Pi/2-phi)/2)/con);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.qsfnz(eccent,sinphi,cosphi:Double):Double;
var
con,R : Double;
begin
  if (eccent > 0.0000000001) then begin
     con := eccent * sinphi;
     R   := (( 1.0- eccent * eccent) * (sinphi /(1.0 - con * con) - (0.5/eccent)*
             ln((1.0 - con)/(1.0 + con))));
  end else R := (2.0 * sinphi);
result := R;
end;
//------------------------------------------------------------------------------
{function TEZProjectionL.MlfnInv(arg,es:double):double;
  //--------
  function MlfnOr(Phi:double):double;
  begin
    Result := phi-0.25*phi*es-0.46875E-1*phi*sqr(es)-0.1953125E-1*phi*Power(es,3)-
      0.375*es*sin(2*phi)-0.9375E-1*sqr(es)*sin(2*phi)-0.439453125E-1*Power(es,3)*
      sin(2*phi)+0.5859375E-1*sqr(es)*sin(4*phi)+0.439453125E-1*Power(es,3)*
      sin(4*phi)-0.1139322917e-1*Power(es,3)*sin(6*phi);
  end;
  //--------
var
  s,t,phi,k:double;
  I : integer;
begin
  k := 1/(1-es);
  phi := arg;
  Result := Phi;
  for I := 0 to 15 do begin
    s := Sin(Phi);
    t := 1-es*s*s;
    t := (MlfnOr(phi)-arg)*(t*sqrt(t))*k;
    phi := Phi - t;
    if (abs(t) < 1e-11) then Result := phi;
  end;
end;}
//------------------------------------------------------------------------------
procedure TEZProjectionL.CylEA(S:double);
begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*(FloatPoint.X-MapRef.Values.Th0);
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*Sin(FloatPoint.Y)*Sec(S);
end;
//------------------------------------------------------------------------------
function TEZProjectionL.CylEAInv(S:double):byte;
var
 ST : double;
begin
  FloatPoint.X := (FloatPoint.X-MapRef.Values.X0+MapRef.Values.Rad*MapRef.Values.Th0)/MapRef.Values.Rad;
  ST := -(FloatPoint.Y-MapRef.Values.Y0)/(MapRef.Values.Rad*Sec(S));
  if (St <= 1) and (St >= -1) then FloatPoint.Y := ArcSin(ST) else begin
    Result := 1;
    Exit;
  end;
  if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) or (FloatPoint.X > MapRef.Values.Th0 + Pi) or
     (FloatPoint.X < MapRef.Values.Th0 - Pi) then begin
     Result := 2;
  end else begin
    if FloatPoint.X < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
    if FloatPoint.X > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
    Result := 0;
  end;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.Asqrt(D:double):Double;
begin
  if (D < 0) then Result := 0 else Result := sqrt(D);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EcT(cx,cy,a,b:double);
begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := FloatPoint.X-MapRef.Values.Th0;
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*FloatPoint.X*cx*(a+asqrt(1-b*sqr(FloatPoint.Y)));
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*cy*FloatPoint.Y;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.EcTI(cx,cy,a,b:double):byte;
begin
   Result := 0;
   FloatPoint.Y := (MapRef.Values.Y0-FloatPoint.Y)/(MapRef.Values.Rad*cy);
   if abs(a+asqrt(1-b*sqr(FloatPoint.Y))) < 1E-9 then Result := 1 else
   FloatPoint.X := (FloatPoint.X-MapRef.Values.X0)/(MapRef.Values.Rad*cx*(a+asqrt(1-b*sqr(FloatPoint.Y))));
   if (FloatPoint.X > Pi) or (FloatPoint.X < -Pi) then Result := 2;
   if (FloatPoint.Y > Pi/2) or (FloatPoint.Y < -Pi/2) then Result := 3;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.EcTV(n,m:double);
var
  K,V,cx,cy : double;
  I : integer;
begin
  K := n*Sin(FloatPoint.Y);
  for I := 0 to 10 do begin
    V := (m*FloatPoint.Y + Sin(FloatPoint.Y)-K)/(m+Cos(FloatPoint.Y));
    FloatPoint.Y := FloatPoint.Y-V;
    if abs(V) < 1E-9 then break;
  end;
  cy := sqrt((m+1)/n);
  cx := cy/(m+1);
  FloatPoint.X := cx*FloatPoint.X*(m+Cos(FloatPoint.Y));
  FloatPoint.Y := cy*FloatPoint.Y;
end;
//------------------------------------------------------------------------------
function TEZProjectionL.EcTVI(n,m:double):byte;
var
  cx,cy : double;
begin
  if (abs(FloatPoint.Y) < n) then begin
    cy := sqrt((m+1)/n);
    cx := cy/(m+1);
    FloatPoint.Y := FloatPoint.Y/cy;
    FloatPoint.X := FloatPoint.X/(cx*(m+Cos(FloatPoint.Y)));
    FloatPoint.Y := AsinZ((m*FloatPoint.Y+Sin(FloatPoint.Y))/n);
    Result := 0;
  end else begin
    FloatPoint.Y := 0;
    FloatPoint.X := 0;
    Result := 1;
  end;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.Fouc(p,q:double;m:byte);
var
  cx,cy,cp,X,Y,C : double;
begin
  cx := q/p;
  cy := p;
  cp := 1/q;

  X := cx*FloatPoint.X*Cos(FloatPoint.Y);
  Y := cy;
  FloatPoint.Y := FloatPoint.Y*cp;
  C := Cos(FloatPoint.Y);
  if (m = 1) then begin
      FloatPoint.X := X*C*C;
      FloatPoint.Y := Y*Tan(FloatPoint.Y);
  end else begin
      FloatPoint.X := X/C;
      FloatPoint.Y := Y*Sin(FloatPoint.Y);
  end;


end;
//------------------------------------------------------------------------------
function  TEZProjectionL.FoucI(p,q:double;m:byte):byte;
var
  cx,cy,cp,Pi,T,C : double;
begin
  cx := q/p;
  cy := p;
  cp := 1/q;
  Result := 0;

  FloatPoint.Y := FloatPoint.Y/cy;
  if (m = 1) then Pi := ArcTan(FloatPoint.Y)
  else Pi := ASinZ(FloatPoint.Y);
  C := Cos(Pi);
  Pi := Pi/cp;
  T := FloatPoint.X/(cx*Cos(Pi));
  if (m = 1) then T := T/sqr(C)
  else T := T*C;
  FloatPoint := SetFloatPoint(T,Pi);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PrepFor1;
begin
  if FloatPoint.X-MapRef.Values.Th0 < -Pi then FloatPoint.X := 2*Pi + FloatPoint.X;
  if FloatPoint.X-MapRef.Values.Th0 > Pi  then FloatPoint.X := FloatPoint.X - 2*Pi;
  FloatPoint.X := FloatPoint.X-MapRef.Values.Th0;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PrepFor2;
begin
  FloatPoint.X := MapRef.Values.X0 + MapRef.Values.Rad*FloatPoint.X;
  FloatPoint.Y := MapRef.Values.Y0 - MapRef.Values.Rad*FloatPoint.Y;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PrepInv1;
begin
  FloatPoint.X := (FloatPoint.X-MapRef.Values.X0)/MapRef.Values.Rad;
  FloatPoint.Y := (MapRef.Values.Y0-FloatPoint.Y)/MapRef.Values.Rad;
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.PrepInv2(E,F:integer);
begin
  if (FloatPoint.X > Pi) or (FloatPoint.X < -Pi) then F := 1;
  if (FloatPoint.Y >= Pi/2-1E-6) or (FloatPoint.Y <= -Pi/2+1E-6) then F := 2;
  if (F = 0) then begin
    ErrorLog(0);
    FloatPoint.X := FloatPoint.X+MapRef.Values.Th0;
  end else ErrorLog(E+F);
end;
//------------------------------------------------------------------------------
procedure TEZProjectionL.Rotate(Theta,Phi:Double;Forw:boolean);
var
px,py,pz,rx,ry,rz,qx,qy,qz,costheta,sintheta,rTheta,rPhi,sign,Sign2:Double;
begin
if (abs(MapRef.Values.Oa) > 0.00001) then begin

  rTheta := MapRef.Values.Ot;
  rPhi   := MapRef.Values.Op;

  rX := cos(rTheta)*cos(rPhi);                                                  // X,Y,Z for rotational point. Rotation occures around line X,Y,Z to centre of sphere (world)
  rY := sin(rTheta)*cos(rPhi);
  rZ := sin(rPhi);

  pX := cos(Theta)*cos(Phi);                                                    // X,Y,Z of cordinat to rotate
  pY := sin(Theta)*cos(Phi);
  pZ := sin(Phi);

  if Forw then begin
    costheta := cos(MapRef.Values.Oa);
    sintheta := sin(MapRef.Values.Oa);
  end else begin
    costheta := cos(-MapRef.Values.Oa);
    sintheta := sin(-MapRef.Values.Oa);
  end;

   qx := (costheta + (1 - costheta) * rx * rx) * px
         +((1 - costheta) * rx * ry - rz * sintheta) * py
         +((1 - costheta) * rx * rz + ry * sintheta) * pz;

   qy := ((1 - costheta) * rx * ry + rz * sintheta) * px
         +(costheta + (1 - costheta) * ry * ry) * py
         +((1 - costheta) * ry * rz - rx * sintheta) * pz;

   qz := ((1 - costheta) * rx * rz - ry * sintheta) * px
         +((1 - costheta) * ry * rz + rx * sintheta) * py
         +(costheta + (1 - costheta) * rz * rz) * pz;


   if (qx >= 0) then sign  := -1 else sign := 1;
   if (qy >= 0) then sign2 := -1 else sign2 := 1;

   if (abs(qx) > 0.00000000001) then begin
      if Sign < 0  then FloatPoint.X := ArcTan(qy/qx);
      if Sign >= 0 then FloatPoint.X := ArcTan(qy/qx)+Pi;
   end else FloatPoint.X := Sign2*Pi/2;

   FloatPoint.Y := ArcSin(qz);
End else begin
  FloatPoint.X := Theta;
  FloatPoint.Y := Phi;
end;
end;
//------------------------------------------------------------------------------
////////////////////// Reference  calculations /////////////////////////////////
procedure TEZProjectionL.OrthRef;
var
a1,a2,a3,a4,b1,b2,b3,b4,c1,c2,c3,c4,d1,d2,d3,d4,x1,x2,x3,
x4,y1,y2,y3,y4,Theta1,Theta2,Theta3,Phi1,Phi2,Phi3 : Double;
R1,R2,CP01,CP02,CP0,X01,X02,Y01,Y02,Y03,Y04,ST01,ST02,ST0,CT0,SP0 : Double;
X0,Y0,Phi0,Theta0,R : Double;
X, Y, i, j, k : integer;
Theta,Phi : Double;
AA : array of integer;
begin
//if MapRef.Input <> ifGeographic then ErrorLog(108) else
if MapRef.CountRef > 2 then begin           // Check for minimal points

SetLength(AA,MapRef.CountRef*2 + 1);        // setup array to iterate through points
for i := 1 to MapRef.CountRef do begin
    AA[i]  := i;
    AA[i+MapRef.CountRef]:= i;
end;
                                   //Lat = Theta, Lon = Phi

//                     ---- Calc X0 ----
X01 := 0;
X0 := 0;
j := 0;
for i := 0 to MapRef.CountRef-1 do begin
  X1 := MapRef.Points[AA[i+0]].X;  a1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Theta));
  X2 := MapRef.Points[AA[i+1]].X;  d2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Phi));
  X3 := MapRef.Points[AA[i+2]].X;  c1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Theta));
  d1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Phi));  c2 := Sin(DegToRad(MapRef.Points[AA[i+1]].Theta));
  a2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Theta));  c3 := Sin(DegToRad(MapRef.Points[AA[i+2]].Theta));
  d3 := Cos(DegToRad(MapRef.Points[AA[i+2]].Phi));  a3 := Cos(DegToRad(MapRef.Points[AA[i+2]].Theta));
  if (d2*d1*a1*c2-d3*a3*d2*c2+d3*a3*d1*c1-d3*c3*d1*a1+d3*c3*a2*d2-d1*c1*a2*d2) <> 0 then begin
  X02   :=   (-d3*a3*X1*d2*c2+d3*a3*d1*c1*X2-d3*c3*X2*d1*a1+d3*c3*a2*X1*d2-
             X3*d2*d1*c1*a2+X3*d2*d1*a1*c2)/(d2*d1*a1*c2-d3*a3*d2*c2+
             d3*a3*d1*c1-d3*c3*d1*a1+d3*c3*a2*d2-d1*c1*a2*d2);
  X01 := X01 + X02;
  j := j+1;
  end;
end;
if j <> 0 then X0 := X01/j        // avarage
else ErrorLog(101);

//                         ---- Calc R ----
j := 0;
R := 0;
for i := 0 to MapRef.CountRef-1 do begin
  X1 := MapRef.Points[AA[i+0]].X;  a1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Theta));  c2 := Sin(DegToRad(MapRef.Points[AA[i+1]].Theta));
  X2 := MapRef.Points[AA[i+1]].X;  d2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Phi));
  X3 := MapRef.Points[AA[i+2]].X;  c1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Theta));  a3 := Cos(DegToRad(MapRef.Points[AA[i+2]].Theta));
  d1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Phi));
  a2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Theta));  c3 := Sin(DegToRad(MapRef.Points[AA[i+2]].Theta));
  d3 := Cos(DegToRad(MapRef.Points[AA[i+2]].Phi));
  R1 :=   2*X3*d1*a1*a2*X1*d2-2*d3*a3*power(X1,2)*a2*d2+
          2*X3*d1*a1*X2*a3*d3-2*X3*d1*a1*d3*a3*X1-
          2*power(X3,2)*d1*a1*d2*a2-2*X2*d1*a1*a2*X1*d2+
          2*X2*d1*a1*d2*a2*X3+power(X3,2)*power(d1,2)*power(a1,2)+
          power(X2,2)*power(a3,2)*power(d3,2)-2*power(X2,2)*d1*a1*a3*d3+
          2*X2*d1*a1*d3*a3*X1+power(X2,2)*power(d1,2)*power(a1,2)+
          power(a2,2)*power(X1,2)*power(d2,2)+
          power(d2,2)*power(a2,2)*power(X3,2)+2*d3*a3*X1*d2*a2*X3+
          2*X2*a3*d3*a2*X1*d2-2*X2*a3*d3*d2*a2*X3+
          power(d3,2)*power(a3,2)*power(X1,2)-2*X2*power(d1,2)*power(a1,2)*X3-
          2*power(a2,2)*X1*power(d2,2)*X3-2*X2*power(a3,2)*power(d3,2)*X1-
          2*power(d1,2)*power(c1,2)*X2*X3-2*X1*power(d3,2)*power(c3,2)*X2
          +power(d1,2)*power(c1,2)*power(X2,2)+power(X1,2)*power(d3,2)*power(c3,2)+
          power(X2,2)*power(c3,2)*power(d3,2)+
          power(d1,2)*power(c1,2)*power(X3,2)+power(c2,2)*power(X3,2)*power(d2,2)+
          2*X2*c3*d3*d1*c1*X3+2*X1*d2*c2*d1*c1*X3+2*d1*c1*X2*X1*d3*c3-
          2*X1*d3*c3*d1*c1*X3-2*c2*power(X3,2)*d2*d1*c1-
          2*power(X1,2)*d2*c2*d3*c3+2*c2*X3*d2*X1*d3*c3-
          2*X1*d2*c2*d1*c1*X2+2*X1*d2*c2*X2*c3*d3-
          2*d1*c1*power(X2,2)*c3*d3+2*c2*X3*d2*d1*c1*X2-
          2*c2*X3*d2*X2*c3*d3+power(X1,2)*power(d2,2)*power(c2,2)-
          2*power(c2,2)*X3*power(d2,2)*X1;
  R2 :=   -d3*a3*d2*c2+d2*c2*d1*a1+d3*a3*d1*c1-d2*d1*c1*a2+d3*c3*d2*a2-d3*c3*d1*a1;
  if (R1 > 0) and (abs(R2) > 1*10E-8) then begin
     R    := R + abs(sqrt(R1)/R2);
     j := j+1;
  end else ErrorLog(102);
end;
if j <> 0 then R := R/j;         // avarage
//else ErrorLog(102);

//                        ---- Calc Theta0 ----
// needs two0 calculations. ArcCos and ArcSin only give positive angles while
// negative might also occure.
j   := 0;
k   := 0;
CP0 := 0;
SP0 := 0;
for i := 0 to MapRef.CountRef-1 do begin
  X1 := MapRef.Points[AA[i+0]].X;  a1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Theta));  a2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Theta));
  X2 := MapRef.Points[AA[i+1]].X;  c1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Theta));  c2 := Sin(DegToRad(MapRef.Points[AA[i+1]].Theta));
                            d1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Phi));  d2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Phi));
  if (abs(d1*R*d2*(-c1*a2+a1*c2)) > 1*10E-6) and (abs(-(-d1*a1*X2+d1*a1*X0+a2*X1*d2-a2*X0*d2)/(d1*(-c1*a2+a1*c2)*R*d2)) < 1) then begin
     CP0  :=  CP0 -(-d1*a1*X2+d1*a1*X0+a2*X1*d2-a2*X0*d2)/(d1*(-c1*a2+a1*c2)*R*d2);
     j := j+1;
  end;
  if (abs(d1*R*d2*(-c1*a2+a1*c2)) > 1*10E-6) and (abs(-(X1*d2*c2-X0*d2*c2-d1*c1*X2+d1*c1*X0)/(d1*R*d2*(-c1*a2+a1*c2))) < 1) then begin
     SP0 := SP0 -(X1*d2*c2-X0*d2*c2-d1*c1*X2+d1*c1*X0)/(d1*R*d2*(-c1*a2+a1*c2));
     k := k+1;
  end;
end;
if (j <> 0) and (k <> 0) then begin
   CP0 := CP0/j;          // avarage
   SP0 := SP0/k;          // avarage
   if ArcSin(SP0) <= 0 then Theta0 := -ArcCos(CP0)
   else Theta0 := ArcCos(CP0);
end else ErrorLog(103);

//                      ---- Calc Phi0 ----
// To calculate Sin(Phi0) Cos(Phi0) needed to be differentiated and
// solved for Cos(Phi0) with result = 0;
j := 0;
ST0 := 0;
Phi0 := 0;
for i := 0 to MapRef.CountRef-1 do begin
  a1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Theta)); a2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Theta));
  b1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Phi));   b2 := Sin(DegToRad(MapRef.Points[AA[i+1]].Phi));
  c1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Theta)); c2 := Sin(DegToRad(MapRef.Points[AA[i+1]].Theta));
  d1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Phi));   d2 := Cos(DegToRad(MapRef.Points[AA[i+1]].Phi));
  // ArcSin -> after diff. from CT0
  ST01:= b2+b1;
  ST02:= sqr(b2)+2*b2*b1+sqr(b1)+sqr(d2)*sqr(a2)*sqr(CP0)+2*sqr(d2)*a2*CP0*c2*SP0+2*d2*a2*sqr(CP0)*d1*a1+
         2*d2*a2*CP0*d1*c1*SP0+sqr(d2)*sqr(c2)*sqr(SP0)+2*d2*c2*SP0*d1*a1*CP0+2*d2*c2*sqr(SP0)*d1*c1+
         sqr(d1)*sqr(a1)*sqr(CP0)+2*sqr(d1)*a1*CP0*c1*SP0+sqr(d1)*sqr(c1)*sqr(SP0);
  if ST02 > 0 then begin
     ST0 := ST0 + ST01/sqrt(ST02);
     j := j+1;
  end;
end;
if j <> 0 then Phi0 := ArcSin(ST0/j)   // avarage
else ErrorLog(104);

//                        ---- Calc Y0 ----
// Remaining constant.
j := 0;
Y0 := 0;
for i := 0 to MapRef.CountRef-1 do begin
  a1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Theta));
  b1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Phi));
  c1 := Sin(DegToRad(MapRef.Points[AA[i+0]].Theta));
  d1 := Cos(DegToRad(MapRef.Points[AA[i+0]].Phi));
  Y1 := MapRef.Points[AA[i+0]].Y;
  Y0 := Y0 + (Y1 - R*(Cos(Phi0)*b1 - Sin(Phi0)*d1*a1*CP0 - Sin(Phi0)*d1*c1*SP0));
  j := j+1;
end;
if j <> 0 then Y0 := Y0/j               // avarage
else ErrorLog(105);


MapRef.Values.Rad := R;
MapRef.Values.X0 := X0;
MapRef.Values.Y0 := Y0;
MapRef.Values.Th0 := Theta0;
MapRef.Values.Ph0 := Phi0;
MapRef.Calculated := true;
ErrorLog(0);
end else begin
ErrorLog(100);
MapRef.Values.Rad := 1000;
MapRef.Values.X0 := 0;
MapRef.Values.Y0 := 0;
MapRef.Values.Th0 := 0;
MapRef.Values.Ph0 := 0;
end;
end;
//------------------------------------------------------------------------------
Procedure TEZProjectionL.UTMRef;
var
AA : array of integer;
i, j : integer;
A,B,C,D,E,F,Theta1,Theta2,Theta3,Phi1,Phi2,Phi3,X1,X2,X3,Y1,Y2,Y3 : Double;
begin
if MapRef.CountRef > 2 then begin           // Check for minimal points

SetLength(AA,MapRef.CountRef*2 + 1);        // setup array to iterate through points
for i := 0 to MapRef.CountRef-1 do begin
    AA[i]  := i;
    AA[i+MapRef.CountRef-1]:= i;
end;

A := 0;
B := 0;
C := 0;
D := 0;
E := 0;
F := 0;

j := 0;
for i := 0 to MapRef.CountRef-1 do begin
  X1 := MapRef.Points[AA[i+0]].X;  Y1 := MapRef.Points[AA[i+0]].Y;
  X2 := MapRef.Points[AA[i+1]].X;  Y2 := MapRef.Points[AA[i+1]].Y;
  X3 := MapRef.Points[AA[i+2]].X;  Y3 := MapRef.Points[AA[i+2]].Y;
  Theta1 := MapRef.Points[AA[i+0]].Theta;  Phi1 := MapRef.Points[AA[i+0]].Phi;
  Theta2 := MapRef.Points[AA[i+1]].Theta;  Phi2 := MapRef.Points[AA[i+1]].Phi;
  Theta3 := MapRef.Points[AA[i+2]].Theta;  Phi3 := MapRef.Points[AA[i+2]].Phi;

  if (abs(-Phi2*Theta3+Phi2*Theta1+Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2) > 0.001) or
     (abs(-Phi2*Theta3+Phi2*Theta1+Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2) > 0.001) then begin
    A := A + (-Phi2*X3-Phi3*X1+Phi2*X1-Phi1*X2+Phi1*X3+Phi3*X2)/(-Phi2*Theta3+Phi2*Theta1+
              Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2);
    B := B + (-Phi2*Y3-Phi3*Y1+Phi2*Y1+Phi1*Y3+Phi3*Y2-Phi1*Y2)/(-Phi2*Theta3+Phi2*Theta1+
              Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2);
    C := C + (-X1*Phi2*Theta3+X1*Phi3*Theta2+Theta1*Phi2*X3-Theta1*Phi3*X2-
                  Phi1*Theta2*X3+Phi1*X2*Theta3)/(-Phi2*Theta3+Phi2*Theta1+Phi1*Theta3-
                  Phi3*Theta1+Phi3*Theta2-Phi1*Theta2);
    D := D + (-Theta1*Y3+Theta2*Y3-Theta2*Y1-Y2*Theta3+Y2*Theta1+Y1*Theta3)/(-Phi2*Theta3+
              Phi2*Theta1+Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2);
    E := E + (-Phi2*Y3-Phi3*Y1+Phi2*Y1+Phi1*Y3+Phi3*Y2-Phi1*Y2)/(-Phi2*Theta3+Phi2*Theta1+
              Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-Phi1*Theta2);
    F := F + (-Y1*Phi2*Theta3+Theta2*Phi3*Y1-Theta2*Phi1*Y3+Y2*Phi1*Theta3+Theta1*Phi2*Y3-
              Theta1*Phi3*Y2)/(-Phi2*Theta3+Phi2*Theta1+Phi1*Theta3-Phi3*Theta1+Phi3*Theta2-
              Phi1*Theta2);
  end;
  j := j + 1;
end;
if j <> 0 then begin
  MapRef.Values.Rad := A/j;
  MapRef.Values.Th0 := B/j;
  MapRef.Values.Ph0 := C/j;
  MapRef.Values.Ph1 := D/j;
  MapRef.Values.Ph2 := E/j;
  MapRef.Values.X0  := F/j;
  MapRef.Calculated := true;
  ErrorLog(0);
end else ErrorLog(151);
end else ErrorLog(150);
end;
////////////////////////////////////////////////////////////////////////////////
end.
